// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Aff_transformation_2.h,v $
// $Revision: 1.13 $ $Date: 2003/10/21 12:18:00 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
//                 Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_AFF_TRANSFORMATION_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_AFF_TRANSFORMATION_T2_H

#include <cmath>

CGAL_BEGIN_NAMESPACE

class Identity_transformation;
template < class R > class Aff_transformation_rep_baseT2;
template < class R > class Aff_transformation_repT2;
template < class R > class Translation_repT2;
template < class R > class Rotation_repT2;
template < class R > class Scaling_repT2;

CGAL_END_NAMESPACE

#include <CGAL/Aff_transformation_rep_T2.h>
#include <CGAL/Translation_rep_T2.h>
#include <CGAL/Rotation_rep_T2.h>
#include <CGAL/Scaling_rep_T2.h>

CGAL_BEGIN_NAMESPACE

template < class R_ >
class Aff_transformation_T2
  : public Handle_for_virtual<Aff_transformation_rep_baseT2<R_> >
{
CGAL_VC7_BUG_PROTECTED
  typedef typename R_::FT                   FT;
  typedef Aff_transformation_rep_baseT2<R_> Aff_t_base;

  typedef typename R_::Point_T2              TPoint_T2;
  typedef typename R_::Vector_T2             TVector_T2;
  typedef typename R_::Direction_T2          TDirection_T2;
  typedef typename R_::Line_T2               TLine_T2;
  typedef typename R_::Aff_transformation_T2 TAff_transformation_T2;

public:
  typedef R_                                R;
   
  Aff_transformation_T2()
  {
    initialize_with(Aff_transformation_repT2<R>(FT(1), FT(0), FT(0), FT(1)));
  }

  Aff_transformation_T2(const Identity_transformation)
  {
    initialize_with(Aff_transformation_repT2<R>(FT(1), FT(0), FT(0), FT(1)));
  }

  Aff_transformation_T2(const Translation, const TVector_T2 &v)
  {
    initialize_with(Translation_repT2<R>(v));
  }

  // Rational Rotation:
  Aff_transformation_T2(const Rotation,
                       const TDirection_T2 &d,
                       const FT &num,
                       const FT &den = FT(1))
  {
    initialize_with(Rotation_repT2<R>(d, num, den));
  }

  Aff_transformation_T2(const Rotation,
                       const FT &sine,
                       const FT &cosine)
  {
    initialize_with(Rotation_repT2<R>(sine, cosine));
  }

  Aff_transformation_T2(const Scaling, const FT &s)
  {
    initialize_with(Scaling_repT2<R>(s));
  }

  // The general case:
  // a 3x2 matrix for the operations combining rotation, scaling, translation
  Aff_transformation_T2(const FT & m11, const FT & m12, const FT & m13,
                       const FT & m22, const FT & m23, const FT & m32,
                       const FT & m33)
  {
    initialize_with(Aff_transformation_repT2<R>(m11, m12, m13,
                                                m22, m23, m32, m33));
  }

  Aff_transformation_T2(const FT & m22,
                       const FT & m23, const FT & m32,
                       const FT & m33)
  {
    initialize_with(Aff_transformation_repT2<R>(m22, m23, m32, m33));
  }

  TPoint_T2
  transform(const TPoint_T2 &p) const 
  { return Ptr()->transform(p); } 

  TPoint_T2
  operator()(const TPoint_T2 &p) const
  { return transform(p); }

  TVector_T2
  transform(const TVector_T2 &v) const 
  { return Ptr()->transform(v); }

  TVector_T2
  operator()(const TVector_T2 &v) const
  { return transform(v); } // FIXME : not compiled by the test-suite.

  TDirection_T2
  transform(const TDirection_T2 &d) const
  { return Ptr()->transform(d); }

  TDirection_T2
  operator()(const TDirection_T2 &d) const
  { return transform(d); }

  TLine_T2
  transform(const TLine_T2 &l) const
  { return l.transform(*this); }

  TLine_T2
  operator()(const TLine_T2 &l) const
  { return transform(l); }

  TAff_transformation_T2 inverse() const { return Ptr()->inverse(); }

  bool is_even() const { return Ptr()->is_even(); }
  bool is_odd() const { return ! (Ptr()->is_even()); }

  FT cartesian(int i, int j) const { return Ptr()->cartesian(i,j); }
  FT homogeneous(int i, int j) const { return cartesian(i,j); }
  FT m(int i, int j) const { return cartesian(i,j); }
  FT hm(int i, int j) const { return cartesian(i,j); }

  TAff_transformation_T2 operator*(const TAff_transformation_T2 &t) const
  {
    return (*Ptr()) * (*t.Ptr());
  }

  std::ostream &
  print(std::ostream &os) const;
};

template < class R >
std::ostream&
Aff_transformation_T2<R>::print(std::ostream &os) const
{
  Ptr()->print(os);
  return os;
}

#ifndef CGAL_NO_OSTREAM_INSERT_AFF_TRANSFORMATIONC2
template < class R >
std::ostream&
operator<<(std::ostream& os, const Aff_transformation_T2<R>& t)
{
  t.print(os);
  return os;
}
#endif // CGAL_NO_OSTREAM_INSERT_AFF_TRANSFORMATIONC2

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_AFF_TRANSFORMATION_T2_H
