// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/H2/include/CGAL/Homogeneous/Aff_transformationH2.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:16:07 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_AFF_TRANSFORMATION_REP_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_AFF_TRANSFORMATION_REP_T2_H

#include <CGAL/determinant.h>

CGAL_BEGIN_NAMESPACE

template < class R >
class Aff_transformation_rep_baseT2
  : public Ref_counted_virtual
{
public:
  typedef typename R::FT                    FT;
  typedef typename R::Point_T2              TPoint_T2;
  typedef typename R::Vector_T2             TVector_T2;
  typedef typename R::Direction_T2          TDirection_T2;
  typedef typename R::Aff_transformation_T2 TAff_transformation_T2;

  virtual ~Aff_transformation_rep_baseT2() {}

  virtual TPoint_T2     transform(const TPoint_T2 &p) const  = 0;
  virtual TVector_T2    transform(const TVector_T2 &v) const = 0;
  virtual TDirection_T2 transform(const TDirection_T2 &d) const=0;

  virtual TAff_transformation_T2 operator*(
                       const Aff_transformation_rep_baseT2 &t) const = 0;

  virtual TAff_transformation_T2 compose(
                       const Aff_transformation_repT2<R> &t) const  = 0;

  virtual TAff_transformation_T2 compose(
                       const Translation_repT2<R> &t) const  = 0;

  virtual TAff_transformation_T2 compose(
                       const Rotation_repT2<R> &t) const  = 0;

  virtual TAff_transformation_T2 compose(
                       const Scaling_repT2<R> &t) const  = 0;

  virtual TAff_transformation_T2 inverse() const  = 0;
  virtual bool                 is_even() const  = 0;
  virtual FT                   cartesian(int i, int j) const = 0;
  virtual std::ostream         &print(std::ostream &os) const = 0;
};

template < class R >
class Aff_transformation_repT2
  : public Aff_transformation_rep_baseT2<R>
{
public:
  typedef typename R::FT                              FT;
  typedef Aff_transformation_repT2<R>                 Self;
  typedef Aff_transformation_rep_baseT2<R>            Aff_t_base;
  typedef typename Aff_t_base::TPoint_T2              TPoint_T2;
  typedef typename Aff_t_base::TVector_T2             TVector_T2;
  typedef typename Aff_t_base::TDirection_T2          TDirection_T2;
  typedef typename Aff_t_base::TAff_transformation_T2  TAff_transformation_T2;

friend class Translation_repT2<R>;
friend class Rotation_repT2<R>;
friend class Scaling_repT2<R>;

  Aff_transformation_repT2()
  {}

  Aff_transformation_repT2( const FT& m22, const FT& m23,
                            const FT& m32, const FT& m33)
  : t11(FT(1)), t12(FT(0)), t13(FT(0)),
    t22(m22), t23(m23), t32(m32), t33(m33)
  {
     
  }

  Aff_transformation_repT2( const FT& m11, const FT& m12, const FT& m13,
                            const FT& m22, const FT& m23, const FT& m32,
                            const FT& m33)
  : t11(m11), t12(m12), t13(m13),
    t22(m22), t23(m23), t32(m32), t33(m33)
  
  {
  }

  TPoint_T2 transform(const TPoint_T2& p) const // FIXME : construction
  {
    return TPoint_T2(t11 * p.w(), 
                     t12 * p.w() + t22 * p.x() + t32 * p.y(),
                     t13 * p.w() + t23 * p.x() + t33 * p.y());
  }

  // note that a vector is not translated
  TVector_T2 transform(const TVector_T2& v) const // FIXME : construction
  {
    return TVector_T2(v.w(), 
                      t12 * v.w() + t22 * v.x() + t32 * v.y(),
                      t13 * v.w() + t23 * v.x() + t33 * v.y());
  }

  // note that a direction is not translated
  TDirection_T2 transform(const TDirection_T2& dir) const // FIXME : construction
  {
    TVector_T2 v = dir.to_vector();
    return TDirection_T2(v.w(), 
                       t12 * v.w() + t22 * v.x() + t32 * v.y(),
                       t13 * v.w() + t23 * v.x() + t33 * v.y());
  }

  // Note that Aff_transformation is not defined yet,
  // so the following 6 functions have to be implemented later...
  TAff_transformation_T2 inverse() const;
  TAff_transformation_T2 operator*(const Aff_t_base &t) const;
  TAff_transformation_T2 compose(const Self &t) const;
  TAff_transformation_T2 compose(const Translation_repT2<R> &t) const;
  TAff_transformation_T2 compose(const Rotation_repT2<R> &t) const;
  TAff_transformation_T2 compose(const Scaling_repT2<R> &t) const;

  bool is_even() const
  {
    return sign_of_determinant2x2(t22, t23, t32, t33) == POSITIVE;
  }

  FT cartesian(int i, int j) const
  {
    switch (i)
    {
    case 0: switch (j)
            {
              case 0: return t11;
              case 1: return t12;
              case 2: return t13;
            }
    case 1: switch (j)
            {
              case 0: return FT(0);
              case 1: return t22;
              case 2: return t23;
            }
    case 2: switch (j)
            {
              case 0: return FT(0);
              case 1: return t32;
              case 2: return t33;
            }
    }
    return FT(0);
  }

  std::ostream &print(std::ostream &os) const
  {
    os <<"Aff_transformationT2(" <<t11<<" "<<t12<<" "<<t13<<std::endl;
    os <<"                     " <<t22<<" "<<t23<<" "<<t32<<" "<<t33<<")";
    return os;
  }

private:
    FT t11,t12,t13,t22,t23,t32,t33;
};

#ifdef CGAL_CFG_TYPENAME_BUG
#define typename
#endif

template < class R >
CGAL_KERNEL_LARGE_INLINE
typename Aff_transformation_repT2<R>::TAff_transformation_T2
Aff_transformation_repT2<R>::
inverse() const
{
  FT det = FT(1)/(t23*t32-t22*t33);
  FT a = FT(1)/t11;
  
  return TAff_transformation_T2(a, (t33*t12-t13*t32)*det*a, 
                                  (-t12*t23+t13*t22)*det*a, 
                                                  -t33*det,
                                                   t23*det,
                                                   t32*det,
                                                  -t22*det);
}


template < class R >
CGAL_KERNEL_INLINE
typename Aff_transformation_repT2<R>::TAff_transformation_T2
Aff_transformation_repT2<R>::
operator*(const Aff_transformation_rep_baseT2<R> &t) const
{
  return t.compose(*this);
}
 
template < class R >
CGAL_KERNEL_LARGE_INLINE
typename Aff_transformation_repT2<R>::TAff_transformation_T2
Aff_transformation_repT2<R>::
compose(const Aff_transformation_repT2<R> &t) const
{
  return TAff_transformation_T2(t.t11*t11,
                               t.t11*t12 + t.t12*t22 + t.t13*t32,
                               t.t11*t13 + t.t12*t23 + t.t13*t33,
                               t.t22*t22 + t.t23*t32,
                               t.t22*t23 + t.t23*t33,
                               t.t32*t22 + t.t33*t32,
                               t.t32*t23 + t.t33*t33);
}

template < class R >
CGAL_KERNEL_LARGE_INLINE
typename Aff_transformation_repT2<R>::TAff_transformation_T2
Aff_transformation_repT2<R>::
compose(const Translation_repT2<R> &t) const
{
  return TAff_transformation_T2(t11 * t.translationvector_.w(),
                              t12 * t.translationvector_.w() + 
                              t22 * t.translationvector_.x() +
                              t32 * t.translationvector_.y(),
                              t13 * t.translationvector_.w() +
                              t23 * t.translationvector_.x() +
                              t33 * t.translationvector_.y(),
                              t22 * t.translationvector_.w(),
                              t23 * t.translationvector_.w(),
                              t32 * t.translationvector_.w(),
                              t33 * t.translationvector_.w());
}


template < class R >
CGAL_KERNEL_LARGE_INLINE
typename Aff_transformation_repT2<R>::TAff_transformation_T2
Aff_transformation_repT2<R>::
compose(const Rotation_repT2<R> &t) const
{
  return TAff_transformation_T2(t11, t12, t13,
                              t.cosinus_*t22 + t.sinus_*t32,
                              t.cosinus_*t23 + t.sinus_*t33,
                              -t.sinus_*t22 + t.cosinus_*t32,
                              -t.sinus_*t23 + t.cosinus_*t33);
}

template < class R >
CGAL_KERNEL_LARGE_INLINE
typename Aff_transformation_repT2<R>::TAff_transformation_T2
Aff_transformation_repT2<R>::
compose(const Scaling_repT2<R> &t) const
{
   return TAff_transformation_T2(t11,
                               t12,
                               t13,
                               t.scalefactor_ * t22,
                               t.scalefactor_ * t23,
                               t.scalefactor_ * t32,
                               t.scalefactor_ * t33);
}

#ifdef CGAL_CFG_TYPENAME_BUG
#undef typename
#endif

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_AFF_TRANSFORMATION_REP_T2_H
