// Copyright (c) 1998-2003  ETH Zurich (Switzerland).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Matrix_search/include/CGAL/all_furthest_neighbors_2.h,v $
// $Revision: 1.58 $ $Date: 2003/09/29 08:41:46 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Michael Hoffmann <hoffmann@inf.ethz.ch>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_ALL_NEAREST_NEIGHBORS_T2_H
#define CGAL_ALL_NEAREST_NEIGHBORS_T2_H


#include <iostream>

#include <CGAL/Delaunay_triangulation_T2.h>

CGAL_BEGIN_NAMESPACE

template < class Gt,
           class Tds = Triangulation_data_structure_T2 <
                             Triangulation_vertex_base_2<Gt>,
                             Triangulation_face_base_2<Gt, Triangulation_ds_face_base_T2<> > > >
class All_nearest_neighbors_T2
  : private Delaunay_triangulation_T2<Gt,Tds>
{
  public :
  typedef Delaunay_triangulation_T2<Gt,Tds>  Delaunay_triangulation;
  typedef Tds                                Triangulation_data_structure;
  typedef Gt                                 Geom_traits;
  typedef typename Gt::Point_T1		     Point_T1;
  typedef typename Delaunay_triangulation::Edge_circulator Edge_circulator;
  typedef typename Delaunay_triangulation::Face_circulator Face_circulator;
  typedef typename Delaunay_triangulation::Face_handle Face_handle;
  typedef typename Delaunay_triangulation::Point Point;
  typedef typename Delaunay_triangulation::Finite_vertices_iterator Finite_vertices_iterator;

  private :
  struct Graph_edge {
    private :
      Point vertex;
      Point vertex_neighbor;

    public :

    Graph_edge() {}
    Graph_edge(const Point& _vertex, const Point& _vertex_neighbor) :
    vertex(_vertex), vertex_neighbor(_vertex_neighbor)
    {}

    Point get_vertex() const {
      return vertex;
    }

    Point get_vertex_neighbor() const {
      return vertex_neighbor;
    }

    void set_vertex(const Point& _vertex) {
      vertex = _vertex;
    }

    void set_vertex_neighbor(const Point& _vertex_neighbor) {
      _vertex_neighbor = vertex_neighbor;
    }
  };

  Geom_traits gt;
  std::vector<Graph_edge> all_nearest_neighbors;

  public:
  typedef Graph_edge			all_nearest_neighbor_node;
  typedef std::vector<Graph_edge>	all_nearest_neighbor_graph;
  typedef typename std::vector<Graph_edge>::iterator	all_nearest_neighbor_iterator;


  All_nearest_neighbors_T2(const Geom_traits& geom_traits=Geom_traits()) {
    gt = geom_traits;
  }

  all_nearest_neighbor_iterator  all_nearest_neighbor_begin() {
    return all_nearest_neighbors.begin();
  }

  all_nearest_neighbor_iterator  all_nearest_neighbor_end() {
    return all_nearest_neighbors.end();
  }

  template <class InputIterator>
  void compute(InputIterator first, InputIterator last)
  {
    insert(first,last);
    Finite_vertices_iterator vertices = finite_vertices_begin();
    if (number_of_vertices() == 1)
     all_nearest_neighbors.push_back(Graph_edge(vertices->point(),vertices->point()));
    else if (number_of_vertices() == 2) {
      Point q = vertices->point();
      vertices++;
      all_nearest_neighbors.push_back(Graph_edge(q,vertices->point()));
    }
    else {
      while (vertices !=finite_vertices_end()) {
        while ((void*)(&(*(vertices->face()))) == 0) vertices++;
        if (vertices == finite_vertices_end())
          break;

        Face_circulator fc = incident_faces(vertices), done(fc);
        int i = fc->index(vertices);
        while (is_infinite(fc)) fc++;
        Point nearest_neighbor = fc->vertex(ccw(i))->point();
        fc++;

        while (done != fc){

          i = fc->index(vertices);
	  if (!is_infinite(fc)) {
            if (gt.compare_distance(vertices->point(), fc->vertex(ccw(i))->point(),
                                    vertices->point(), nearest_neighbor) == SMALLER) {
              nearest_neighbor = fc->vertex(ccw(i))->point();
            }
	  }

          fc++;
        }
       all_nearest_neighbors.push_back(Graph_edge(vertices->point(),                                                                     nearest_neighbor));
        vertices++;

      }
    }
  }

};

template < class Gt, class Tds > std::istream& operator>>
    (std::istream& is, All_nearest_neighbors_T2<Gt,Tds> &ann);
template < class Gt, class Tds >  std::ostream& operator<<
  (std::ostream& os, All_nearest_neighbors_T2<Gt,Tds> &ann);


template <class Gt, class Tds >
std::ostream&
operator<<(std::ostream& os, All_nearest_neighbors_T2<Gt,Tds> &ann)
{
  typedef typename All_nearest_neighbors_T2<Gt,Tds>::all_nearest_neighbor_iterator
					    all_nearest_neighbor_iterator;
  all_nearest_neighbor_iterator it = ann.all_nearest_neighbor_begin();
  for (; it != ann.all_nearest_neighbor_end(); it++) {
    os << it->get_vertex();
    os << it->get_vertex_neighbor();
  }
  return os;
}



template < class Gt, class Tds >
std::istream&
operator>>(std::istream& is, All_nearest_neighbors_T2<Gt,Tds> &ann)
{
  /*int num_vertices;
  is >> num_vertices;
  int i = 0;
  it 
  for (i; i < num_vertices; i++) {
    is >> _vertex;
    is >> _vertex_neighbor;
    	
  }*/
  return is;
}

  
CGAL_END_NAMESPACE  
  
#endif
  
