// Copyright (c) 2003  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Arrangement/include/CGAL/Arr_polyline_traits_2.h,v $
// $Revision: 1.12 $ $Date: 2003/09/18 10:19:40 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Ron Wein <wein@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_ARR_POLYLINE_TRAITS_T2_H
#define CGAL_ARR_POLYLINE_TRAITS_T2_H

#include <CGAL/basic.h>
#include <CGAL/tags.h>
#include <CGAL/intersections.h>

#include <list>
#include <fstream>
#include <vector>

CGAL_BEGIN_NAMESPACE

template <class Segment_traits_> class Polyline_T2;

template <class Segment_traits_>
class Arr_polyline_traits_T2
{
public:

  typedef Tag_true                                                Has_left_category;
    
  typedef Segment_traits_                                         Segment_traits_T2;
  typedef typename Segment_traits_T2::Kernel                      Kernel;
  typedef Arr_polyline_traits_T2<Segment_traits_T2>               Self;
  typedef typename Segment_traits_T2::Point_T2                    Point_T2;
  typedef typename Segment_traits_T2::Curve_T2                    Segment_T2;
  typedef typename Segment_traits_T2::Object_T2                   Object_T2;

  typedef Polyline_T2<Segment_traits_T2>                          Curve_T2;
  typedef Polyline_T2<Segment_traits_T2>                          X_monotone_curve_T2;

  // Obsolete, for backward compatibility
  typedef Point_T2                                                Point;
  typedef X_monotone_curve_T2                                     X_curve;
  typedef Curve_T2                                                Curve;
  
  typedef typename Kernel::Collinear_has_on_T2                    Collinear_has_on_T2;
  typedef typename Kernel::FT                                     FT;
  
protected:

  typedef typename Kernel::rep                                    rep;
  typedef typename rep::Point_3                                   Point_R3;

  
  Segment_traits_T2   seg_traits;   // A traits class for segments.

public:

  /*!
   * Default constructor.
   */
  Arr_polyline_traits_T2 () :
    seg_traits()
  {}

  /*!
   * Check whether the given curve is a vertical segment.
   * \param cv The curve.
   * \return (true) if the curve is vertical.
   */

  bool curve_overlapSweepLine(const X_monotone_curve_T2& cv) const
  {
    // An x-monotone polyline can represent a overlap sweep line
    //segment only if it is comprised of a single segment.
    return (cv._size() == 1 && seg_traits.curve_overlapSweepLine(cv[0]));
  } 



  Comparison_result curves_compare_forward(const X_monotone_curve_T2 & cv1,
                                           const X_monotone_curve_T2 & cv2,
                                           const Point_T2 & q) const
  {
    int i1 = _locate_point(cv1,q);
    std::cout << "cv1 = " << cv1 << std::endl;
    int i2 = _locate_point(cv2,q);
    std::cout << "cv2 = " << cv2 << std::endl;
    std::cout << "q = " << q << std::endl;
    CGAL_kernel_precondition(i1 != -1 && i2 != -1);
    std::cout << "cv1[" << i1 << "] = " << cv1[i1] << std::endl;
    std::cout << "cv2[" << i2 << "] = " << cv2[i2] << std::endl;
    return seg_traits.curves_compare_forward(cv1[i1],cv2[i2],q);
  }


  Comparison_result compare_distance(const Point_T2 & p1, const Point_T2 & p2,
                                                          const Point_T2 & p3) const
  {
    return seg_traits.compare_distance(p1,p2,p3);
  }
  
  Point antipode (const Point p) const
  {
    return seg_traits.antipode(p);
  }
 
  bool compare_around_origin(const Point p,const Point q) const
  {
    return seg_traits.compare_around_origin(p,q);
  }

  Comparison_result curve_compare_dirSweepLine_at_perpendicularDir(
                                            const Point_T2 & q,
                                            const X_monotone_curve_T2 & cv1) const
  {
    int i=0;
    bool in_range = false;
    while (i < cv1._size() && !in_range)
    {
       std::cout << "Perpendicular 1 "<< std::endl;
       if (seg_traits.point_in_PerpendicularDirSweep_range(cv1[i], q))
         in_range = true;
       else
         i++;
    }
    CGAL_kernel_precondition(in_range);
    return seg_traits.curve_compare_dirSweepLine_at_perpendicularDir(q,cv1[i]);
  }
 
  Comparison_result curves_compare_dirSweepLine_at_perpendicularDir(
                                            const X_monotone_curve_T2 & cv1,
                                            const X_monotone_curve_T2 & cv2,
                                            const Point_T2 & q) const
  {
    std::cout << "cv1 = "<< cv1 << std::endl;
    std::cout << "cv2 = "<< cv2 << std::endl;
    std::cout << "q = "<< q << std::endl;
    int i=0;
    bool in_range = false;
    while (i < cv1._size() && !in_range)
    {
      std::cout << "Perpendicular 2 "<< std::endl;
      if (seg_traits.point_in_PerpendicularDirSweep_range(cv1[i], q))
        in_range = true;
      else
        i++;
    }
    CGAL_kernel_precondition(in_range); //q is in range of cv1
    in_range = false;
    int j=0;
    while (j < cv2._size() && !in_range)
    {
      std::cout << "Perpendicular 3 "<< std::endl;
      if (seg_traits.point_in_PerpendicularDirSweep_range(cv2[j], q)) {
        in_range = true;
        std::cout << "Perpendicular 4 "<< std::endl;
      }  
      else
        j++;
      std::cout << "Perpendicular 5 "<< std::endl; 
    }
    CGAL_kernel_precondition(in_range); //q is in range of cv2
    std::cout << "Perpendicular 6 "<< std::endl;
    return seg_traits.curves_compare_dirSweepLine_at_perpendicularDir(cv1[i],cv2[j],q);
  }


  

  bool subdivide_curve(X_monotone_curve_T2 & curve_base,
                       X_monotone_curve_T2 &subcurve1,
                       X_monotone_curve_T2 &subcurve2)
  {
    //std::cout << "subdivide 1 "<< std::endl;
    std::vector<Point> polyline1;
    //std::cout << "subdivide 2 "<< std::endl;
    std::vector<Point> polyline2;
    //std::cout << "subdivide 3 "<< std::endl;
    Point origin(FT(1),FT(0),FT(0));
    //std::cout << "subdivide 4 "<< std::endl;
    Point p(FT(0),FT(1),FT(0));
    //std::cout << "subdivide 5 "<< std::endl;
    bool intersect = false;
    //std::cout << "subdivide 6 "<< std::endl;
    Point_T2 point;
    //std::cout << "subdivide 7 "<< std::endl;
    Collinear_has_on_T2 collinear_has_on = seg_traits.collinear_has_on_T2_object();
    //std::cout << "subdivide 8 "<< std::endl;
    int i = 0;
    //std::cout << "subdivide 9 "<< std::endl; 
    while (i < curve_base._size() && !intersect) {
      //std::cout << "subdivide 10 "<< std::endl;
      if (i < curve_base._size())  
        polyline1.push_back(seg_traits.curve_source(curve_base[i]));     
      if (orientation(origin,seg_traits.curve_source(curve_base[i]),
                        seg_traits.curve_target(curve_base[i]))==0) {
        //std::cout << "subdivide 11 "<< std::endl;
        if (collinear_has_on(origin,seg_traits.curve_source(curve_base[i]),
                                    seg_traits.curve_target(curve_base[i]))) {
          intersect = true;
          //std::cout << "subdivide 12 "<< std::endl;
          point = origin;
        }
        else if (collinear_has_on(seg_traits.antipode(origin),
                                  seg_traits.curve_source(curve_base[i]),
                                  seg_traits.curve_target(curve_base[i]))) {
          intersect = true;
          //std::cout << "subdivide 13 "<< std::endl;
          point = antipode(origin);
        }
        else {
          //std::cout << "subdivide 14 "<< std::endl;
          i++;
        }
      }
      else {
        //std::cout << "subdivide 15 "<< std::endl;
        Object_T2 o = intersection(curve_base[i],Segment_T2(origin,p));
        //std::cout << "subdivide 16 "<< std::endl;
        if (assign(point,o))
          intersect = true;
        else {
          //std::cout << "subdivide 17 "<< std::endl;
          Object_T2 o = intersection(curve_base[i],Segment_T2(antipode(origin),p));
          //std::cout << "subdivide 18 "<< std::endl;
          if (assign(point,o))
            intersect = true;
          else
            i++;
          //std::cout << "subdivide 19 "<< std::endl; 
        }
      }
      //std::cout << "subdivide 20 "<< std::endl;
      
    }
    if (intersect) {
      //std::cout << "subdivide 21 "<< std::endl;
      if (point != seg_traits.curve_source(curve_base[i]) || 
          polyline1.size() == 1)
        polyline1.push_back(point); 
      //std::cout << "subdivide 22 "<< std::endl;
      if (point != seg_traits.curve_target(curve_base[i]) || 
          i == curve_base._size()-1)
        polyline2.push_back(point);
      //std::cout << "subdivide 23 "<< std::endl;
      polyline2.push_back(seg_traits.curve_target(curve_base[i]));
      //std::cout << "subdivide 24 "<< std::endl;
      i++;    
      //std::cout << "subdivide 25 "<< std::endl;
      while(i < curve_base._size()) {
        //std::cout << "subdivide 26 "<< std::endl; 
        polyline2.push_back(seg_traits.curve_target(curve_base[i]));
        //std::cout << "subdivide 27 "<< std::endl;
        i++;
      } 
      std::cout << "curve_base = " << curve_base << std::endl;
      subcurve1 = Curve_T2(polyline1.begin(),polyline1.end());
      std::cout << "subcurve1 = " << subcurve1 << std::endl;
      subcurve2 = Curve_T2(polyline2.begin(),polyline2.end());
      std::cout << "subcurve2 = " << subcurve2 << std::endl;
      return true;
    }
    else {
      std::cout << "Nao ha interseccao"<< std::endl;
      //std::cout << "subdivide 31 "<< std::endl;
      return false;  
    }
  }

  Orientation orientation(const Point p,const Point q,const Point r) const
  {
    return seg_traits.orientation(p,q,r);
  }
 
  //returns true if the curves cv0 and cv1 are equal
  bool curves_equal(const X_monotone_curve_T2& cv0,
                    const X_monotone_curve_T2& cv1) const
  {
    if (cv0._size() != cv1._size())
      return false;
    int i = 0;
    bool equal_curves = true;
    do {
      if (!seg_traits.curves_equal(cv0[i],cv1[i]))
        equal_curves = false;
      i++;
    } while (i < cv0._size() && equal_curves);
    return equal_curves;
  }

  bool point_equal(const Point p,const Point q) const
  {
    return seg_traits.point_equal(p,q);
  }

  bool has_on(const X_monotone_curve_T2& c,const Point q)
  {
    int i = 0;
    bool pt_has_on = false;
    do {
      if (seg_traits.has_on(c[i],q))
        pt_has_on = true;
      i++;
    } while (i < c._size() && !pt_has_on);
    return pt_has_on;
  }

  Comparison_result cmp_dist_collinear_origin(const Point p,const Point q)
  {
    return seg_traits.cmp_dist_collinear_origin(p,q);
  }

  bool point_in_PerpendicularDirSweep_range(const X_monotone_curve_T2& cv,
                                            const Point & p) const
  {
    bool in_range = false;
    int i = 0;
    do {
      std::cout << "Perpendicular 4 "<< std::endl;
      if (seg_traits.point_in_PerpendicularDirSweep_range(cv[i],p))
        in_range = true;
      i++;
    } while (!in_range && i < cv._size());
    return in_range;
  }  

  
 

 

  /*!
   * Get the curve source (the source of the first segment in the polyline).
   * \param cv The curve.
   * \return The source point.
   */
  Point curve_source (const X_monotone_curve_T2& cv) const
  {
    CGAL_assertion(cv._size() > 0);
    return (seg_traits.curve_source(cv[0]));
  }

  /*!
   * Get the curve target (the target of the last segment in the polyline).
   * \param cv The curve.
   * \return The target point.
   */
  Point curve_target (const X_monotone_curve_T2 & cv) const
  {
    CGAL_assertion(cv._size() > 0);
    return (seg_traits.curve_target(cv[cv._size() - 1]));
  }

  /*!
   * Flip a given curve.
   * \param cv The input curve.
   * \return The flipped curve.
   */
  X_monotone_curve_T2 curve_opposite (
                            const X_monotone_curve_T2& cv) const
  {
    // Copy the segments in revered order while 
    //flipping each one individually.
    int n = cv._size();
    int i;
    X_monotone_curve_T2 flip_cv;

    for (i = 0; i < n; i++)
    {
      flip_cv._push_back(seg_traits.curve_opposite(cv[n - i - 1]));
    }

    return (flip_cv); 
  }

  /*!
   * Check whether the curve is DirSweepLine-monotone.
   * \param curve The curve.
   * \return (true) if the curve is perpendicularDirSweepLine-monotone.
   */
  bool is_dirSweepLine_monotone(const Curve_T2& curve) const
  {
    bool monotone = true;

    int size = curve._size();
    int i = 1;
    
    Point_T2 ori(1,0,0);
 
    if (orientation(ori,seg_traits.curve_target(curve[0]),seg_traits.curve_source(curve[0])) == 0) {
      return false;
    }
    while (i <= size-1 && monotone) {
      if (int(orientation(ori,seg_traits.curve_target(curve[i-1]),seg_traits.curve_source(curve[i-1])))*
          int(orientation(ori,seg_traits.curve_target(curve[i-1]),seg_traits.curve_target(curve[i]))) != -1) {
        monotone = false;
      } 
      i++;
    }
    return monotone;
  }
  
  /*! 
   * Cut the given curve into perpendicularDirSweepLine-monotone
   * subcurves and insert them to the
   * given output iterator.
   * \param cv The curve.
   * \param o The output iterator
   * \return The past-the-end iterator
   */
  template<class OutputIterator>
  OutputIterator curve_make_dirSweepLine_monotone 
                                        (const Curve_T2& curve,
                                         OutputIterator o)
  {
    // Go over all curve segments.
    X_monotone_curve_T2 x_cv;
    int n = curve._size();
    //std::cout << "curve_make_dirSweepLine_monotone curve size =  "<< n <<  std::endl; 
    Point_T2 ori(1,0,0);
    
    int i = 1;
    x_cv._push_back(curve[0]);
    if (orientation(ori,seg_traits.curve_target(curve[0]),
                        seg_traits.curve_source(curve[0])) == 0) {
      *o++ = x_cv;
      x_cv._clear();   
    }
    while (i <= n-1) {
      std::cout << "seg_traits.curve_target(curve[i-1]) = "
                << seg_traits.curve_target(curve[i-1]) << std::endl;
      std::cout << "seg_traits.curve_source(curve[i-1]) = "
                << seg_traits.curve_source(curve[i-1]) << std::endl;
      std::cout << "seg_traits.curve_source(curve[i]) = "
                << seg_traits.curve_source(curve[i]) << std::endl;
      if (int(orientation(ori,seg_traits.curve_target(curve[i-1]),
                              seg_traits.curve_source(curve[i-1])))*
          int(orientation(ori,seg_traits.curve_target(curve[i-1]),
                              seg_traits.curve_target(curve[i]))) == 1) {
        std::cout << "entra aqui i =  "<< i <<  std::endl; 
        *o++ = x_cv;
        x_cv._clear(); 
        //i++;   
        if (i <= n-1)
          x_cv._push_back(curve[i]);      
        i++; 
      } 
      else if (orientation(ori,seg_traits.curve_target(curve[i]),
                               seg_traits.curve_source(curve[i])) == 0) {
        if (x_cv._size() != 0)
        {
          *o++ = x_cv;
          x_cv._clear();  
        }
        x_cv._push_back(curve[i]);  
        *o++ = x_cv;
        x_cv._clear();
        i++;
        if (i <= n-1)
          x_cv._push_back(curve[i]);  
        //i++;
      }
      else {
        x_cv._push_back(curve[i]);
      } 
      i++;
    }
    // Append the last x-monotone chain.
    if (x_cv._size() != 0)
    {
      // Cut the previous chain.
      *o++ = x_cv;
      x_cv._clear();
    }
    
    return (o);

  }

  /*!
   * Split a given curve at a given split point into two sub-curves.
   * \param cv the curve to split
   * \param c1 the output first part of the split curve. Its source is the
   * source of the original curve.
   * \param c2 the output second part of the split curve. Its target is the
   * target of the original curve.
   * \param p the split point.
   * \pre p lies on cv but is not one of its end-points.
   */
  /*void curve_split (const X_monotone_curve_2& cv, 
		    X_monotone_curve_2& c1, X_monotone_curve_2& c2, 
		    const Point_2& p) const
  {
    // Locate the segment on the polyline cv that contains p.
    int    i = _locate_point (cv, p);

    // Check preconditions.
    CGAL_precondition(i >= 0);
    CGAL_precondition(! seg_traits.point_equal(curve_source(cv), p));
    CGAL_precondition(! seg_traits.point_equal(curve_target(cv), p));

    // Clear the output curves.
    c1._clear(); 
    c2._clear();

    // Push all segments labeled (0, 1, ... , i-1) into c1.
    int    j;

    for (j = 0; j < i; j++)
      c1._push_back (cv[j]);

    // Check whether the split point is cv[i]'s source of target.
    if (seg_traits.point_equal (seg_traits.curve_source(cv[i]), p))
    {
      // cv[i] should be the first segment in c2.
      c2._push_back (cv[i]);
    }
    else if (seg_traits.point_equal (seg_traits.curve_target(cv[i]), p))
    {
      // cv[i] should be the last segment in c1.
      c1._push_back (cv[i]);
    }
    else
    {
      // cv[i] should be split.
      Segment_2    cvi_1, cvi_2;

      seg_traits.curve_split (cv[i],
			      cvi_1, cvi_2,
			      p);

      // The first part should go into c1 and the second into c2.
      c1._push_back (cvi_1);
      c2._push_back (cvi_2);
    }

    // Push all segments labeled (i+1, i+2, ... , n-1) into cv1.
    int    n = cv._size();

    for (j = i+1; j < n; j++)
      c2._push_back (cv[j]);

    return;
  }*/

  
  /*!
   * Find the nearest intersection point (or points) of two given curves to
   * the left lexicographically of a given point not including the point
   * itself:
   *  - If the intersection of the two curves is a point to the left of the 
   *    given point, it is returned through both point references p1 and p2.
   *  - If the intersection of the two curves is an X_monotone_curve_2, 
   *    that is, they overlap at infinitely many points, then the leftmost
   *    segment of the intersection is returned.
   * NOTE: When there is an overlap we will always return a SEGMENT (i.e.,
   *       p1 and p2 will be on a segment) even if the overlap is a polyline,
   *       but this is still sufficient for the arrangement.
   * \param cv1 The first curve.
   * \param cv2 The second curve.
   * \param p The refernece point.
   * \param p1 The first output point.
   * \param p2 The second output point.
   * \return (true) if c1 and c2 do intersect to the right of p, or (false)
   * if no such intersection exists.
   */
/*  bool nearest_intersection_to_left (const X_monotone_curve_2& cv1,
				     const X_monotone_curve_2& cv2,
				     const Point_2& p,
				     Point_2& p1, Point_2& p2) const
  {
    return (_nearest_intersection_to_side (cv1, cv2,
					   p,
					   false,     // To the left of p.
					   p1, p2));
  }*/

  /*!
   * Check whether the two given curves overlap.
   * \patam cv1 The first curve.
   * \patam cv2 The second curve.
   * \return (true) if the two curves overlap in a one-dimensional subcurve
   * (i.e., not in a finite number of points). Otherwise, if they have a finite
   * number of intersection points, or none at all, return (false).
   */
  

  /*!
   * Check whether the two given curves overlap.
   * \patam cv1 The first curve.
   * \patam cv2 The second curve.
   * \return (true) if the two curves overlap in a one-dimensional subcurve
   * (i.e., not in a finite number of points). Otherwise, if they have a finite
   * number of intersection points, or none at all, return (false).
   */
  bool curves_overlap(const X_monotone_curve_T2& cv1,
                      const X_monotone_curve_T2& cv2) const
  {
    // Get the leftmost point of cv1 and of cv2.
    const Point_T2& left1 = _is_curve_to_right(cv1) ? curve_source(cv1) : 
                                                     curve_target(cv1);
    const Point_T2& left2 = _is_curve_to_right(cv2) ? curve_source(cv2) : 
                                                     curve_target(cv2);

    typedef typename Point_T2::FT FT;

    Point_T2 o(FT(1),FT(0),FT(0));

    // Pick the righmost point of the two to start with.
    Point_T2 p = (orientation(o,left1, left2) == CLOCKWISE) ? left1 : left2;

    // Try to find an intersection to the right of p.
    Point_T2 p1, p2;

    while (nearest_intersection_to_right(cv1, cv2,
                                          p,
                                          p1, p2))
    {
      // Check if an overlap has been detected:
      if (! seg_traits.point_equal (p1, p2))
        return (true);

      // Otherwise, p1==p2 is the next intersection point to the right.
      // Continue checking the intersections to its right.
      p = p1;
    }

    // If we reached here, no overlap has been detected.
    return (false);
  }

  Orientation curve_orientation(const X_monotone_curve_T2& cv) const 
  {
    int n = cv._size();
    int i=0;
    Orientation ori,get_orientation = COLLINEAR;
    Point o(FT(1),FT(0),FT(0));
    while (i < n && (get_orientation==COLLINEAR))
    {
      if ((ori=orientation(o,seg_traits.curve_source(cv[i]),
                             seg_traits.curve_target(cv[i])))!=COLLINEAR)
        get_orientation = ori;
        i++;
    }
    return get_orientation;
  }


  /*!
   * Return the index of the segment in the polyline cv that contains the
   * point q in its perpendicular-Direction-to-SweepLine-range. 
   *The function examines each segment, so if the
   * point q is in the x-range of the polyline with n segments, 
   * the segment
   * containing it can be located in O(n) operations.
   * \param cv The polyline curve.
   * \param q The point.
   * \return An index i such that q is in 
   * the perpendicular-Direction-to-SweepLine-range of cv[i].
   * If q is not in the perpendicular-Direction-to-SweepLine-
   * range of cv, returns (-1).
   */
  /*int _locate_point (X_monotone_curve_T2& cv, 
                     const Point& q) const
  {
    // First check whether the polyline curve 
    //really contains q in its x-range.
    std::cout << "locate 1 "<< std::endl;
    Point o(FT(1),FT(0),FT(0));
    Point ao(FT(-1),FT(0),FT(0));
    Point s0,s1;
    std::cout << "locate 2 "<< std::endl;
    int n = cv._size();
    std::cout << "locate 3 "<< std::endl;
    if (curve_orientation(cv) != COUNTERCLOCKWISE) 
      cv = curve_opposite(cv);
    if (seg_traits.curve_source(cv[0]) == o || seg_traits.curve_source(cv[0]) == ao)
       s0 = seg_traits.curve_target(cv[0]);
    else
       s0 = seg_traits.curve_source(cv[0]);
    if (seg_traits.curve_source(cv[n-1]) == o || seg_traits.curve_source(cv[n-1]) == ao)
       s1 = seg_traits.curve_target(cv[n-1]);
    else
       s1 = seg_traits.curve_source(cv[n-1]);
    if (!(orientation(o,s0,q)!=-1 ||
          orientation(o,s1,q)!=1)) {
      std::cout << "ERRADO "<< std::endl;
      return -1;
    }
    std::cout << "locate 4 "<< std::endl;
    int i=0;
    int segment_range = -1;
    std::cout << "locate 5 "<< std::endl;
    Point_R3 o_R3(FT(1),FT(0),FT(0));
    Point_R3 oa_R3(FT(-1),FT(0),FT(0));
    Point_R3 q_R3(q.w(),q.x(),q.y());
    //std::cout << "locate 6 "<< std::endl;
    while (i < n && (segment_range==-1)) {
      //std::cout << "locate 7 "<< std::endl;
      Point_T2 source=seg_traits.curve_source(cv[i]);
      //std::cout << "locate 8 "<< std::endl;
      Point_T2 target=seg_traits.curve_target(cv[i]);
      //std::cout << "locate 9 "<< std::endl;
      Point_R3 source_R3(source.w(),source.x(),source.y());
      //std::cout << "locate 10 "<< std::endl;
      Point_R3 target_R3(target.w(),target.x(),target.y());
      //std::cout << "locate 11 o = " << o << std::endl;
      //std::cout << "source = " << source << std::endl;
      //std::cout << "q = " << q << std::endl;
      //std::cout << "orientation(o,source,q) = "<< orientation(o,source,q) << std::endl;
      //std::cout << "orientation(o,target,q) = "<< orientation(o,target,q) << std::endl;
      //std::cout << orientation(o,seg_traits.curve_target(cv[i]),q) << std::endl;
      if (orientation(o,source,q) == 1 && orientation(o,target,q) == -1) {
        //std::cout << "locate 12 "<< std::endl;
        segment_range = i;
      }
      else if (orientation(o,source,q) == 0) {
        //std::cout << "locate 13 "<< std::endl;
        if (int(coplanar_orientation(o_R3,oa_R3,q_R3)) *
            int(coplanar_orientation(o_R3,oa_R3,source_R3)) != -1) {
          //std::cout << "locate 14 "<< std::endl;
          segment_range = i;
        }
      }
      else if (orientation(o,seg_traits.curve_target(cv[i]),q)==0) {
        //std::cout << "locate 15 "<< std::endl;
        if (int(coplanar_orientation(o_R3,oa_R3,q_R3)) *
            int(coplanar_orientation(o_R3,oa_R3,target_R3)) != -1) {
          //std::cout << "locate 16 "<< std::endl;
          segment_range = i;
        } 
      } 
      //std::cout << "locate 17 "<< std::endl;
      i++;
    }
    //std::cout << "locate 18 "<< std::endl;
    return segment_range;
  }*/

  /*!
   * Find the index of the segment in the polyline that is defined to the
   * left (or to the right) of a the point q.
   * \param cv The polyline curve.
   * \param q The point.
   * \param to_right (true) if we wish to locate a segment to the right of q,
   *                 (false) if we wish to locate a segment to its right.
   * \return An index i such that cv[i] is defined to the left (or to the
   *         right) of q, or (-1) if no such segment exists.
   */
/*  int _locate_point_side (const X_monotone_curve_2& cv, const Point_2& q,
			  const bool& to_right) const
  {
    // First locate a segment cv[i] that contains q in its x-range.
    int    i = _locate_point (cv, q);

    if (i < 0)
      return (-1);

    // If we seek an end-point to the right of q, q must be smaller than it.
    // If we seek an end-point to its left, q must be larger.
    const Comparison_result  cres = (to_right) ? SMALLER : LARGER;

    // Check whether x(q) is the x coordinate of an end-point of cv.
    if (seg_traits.compare_x (q, seg_traits.curve_source(cv[i])) == EQUAL)
    {
      // If x(q) is at cv[i]'s source and its target is to the right (or left) 
      // of q, then cv[i] is defined to the right (or left) of q. 
      if (seg_traits.compare_x (q, 
				seg_traits.curve_target(cv[i])) == cres)
      {
	return (i);
      }

      // Otherwise, since cv[i]'s source is cv[i-1]'s target, we should check
      // if the source of the previous curve is to the right (or left) of q.
      if (i > 0 &&
	  seg_traits.compare_x (q, 
			        seg_traits.curve_source(cv[i-1])) == cres)
      {
	return (i-1);
      }
      else
      {
	return (-1);
      }
    }
    else if (seg_traits.compare_x (q, seg_traits.curve_target(cv[i])) == EQUAL)
    {
      // If x(q) is at cv[i]'s target and its source is to the right (or left) 
      // of q, then cv[i] is defined to the right (or left) of q. 
      if (seg_traits.compare_x (q, 
				seg_traits.curve_source(cv[i])) == cres)
      {
	return (i);
      }

      // Otherwise, since cv[i]'s target is cv[i+1]'s source, we should check
      // if the target of the next curve is to the right (or left) of q.
      if (i < (cv._size() - 1) &&
	  seg_traits.compare_x (q, 
			        seg_traits.curve_target(cv[i+1])) == cres)
      {
	return (i+1);
      }
      else
      {
	return (-1);
      }
    }

    // In case q is in cv[i]'s interior:
    return (i);
  }*/

  /*!
   * Check whether the cuve is defined from the left to the right.
   * \param cv The polyline curve.
   * \return (true) if the curve target is lexicographically larger than its
   *         source; (false) otherwise.
   */
  bool _is_curve_to_right (const X_monotone_curve_T2& cv) const
  {
    const Point&    ps = seg_traits.curve_source(cv[0]);
    const Point&    pt = seg_traits.curve_target(cv[0]);
    const Point&    o = Point(FT(1),FT(0),FT(0));
    return (orientation(ps,pt,o) == 1);
  }

  /*!
   * Find the nearest intersection point (or points) of two given curves to
   * the right lexicographically of a given point not including the point
   * itself:
   *  - If the intersection of the two curves is a point to the right of the 
   *    given point, it is returned through both point references p1 and p2.
   *  - If the intersection of the two curves is an X_monotone_curve_2, 
   *    that is, they overlap at infinitely many points, then the rightmost
   *    segment of the intersection is returned.
   * NOTE: When there is an overlap we will always return a SEGMENT (i.e.,
   *       p1 and p2 will be on a segment) even if the overlap is a polyline,
   *       but this is still sufficient for the arrangement.
   * \param cv1 The first curve.
   * \param cv2 The second curve.
   * \param p The refernece point.
   * \param p1 The first output point.
   * \param p2 The second output point.
   * \return (true) if c1 and c2 do intersect to the right of p, or (false)
   * if no such intersection exists.
   */
  bool nearest_intersection_to_right (const X_monotone_curve_T2& cv1,
                                      const X_monotone_curve_T2& cv2,
                                      const Point& p,
                                      Point& p1, Point& p2) const
  {
    // Get the indices of the segments in cv1 and cv2 containing p.
    std::cout << "nearest_intersection_to_right 0 "<< std::endl;
    std::cout << "cv1 = "<< cv1 << std::endl;
    std::cout << "cv2 = "<< cv2 << std::endl;
    std::cout << "p = "<< p << std::endl;
    X_monotone_curve_T2 cv1_1 = cv1;
    X_monotone_curve_T2 cv2_1 = cv2;
    std::cout << "CHAMANDO LOCATE PARA CURVA1 "<< std::endl;
    int    i1 = _locate_point (cv1_1, p);
    std::cout << "nearest_intersection_to_right 1 "<< std::endl;
    std::cout << "CHAMANDO LOCATE PARA CURVA2 "<< std::endl;
    int    i2 = _locate_point (cv2_1, p);
    std::cout << "nearest_intersection_to_right 2 "<< std::endl;
    std::cout << "i1 = "<< i1 <<  std::endl;
    std::cout << "i2 = "<< i2 <<  std::endl;
    std::cout << "NOVO nearest_intersection_to_right 2.1 "<< std::endl;
    typedef typename Point_T2::FT FT;

    Point o(FT(1),FT(0),FT(0));

    // Check if cv1 and cv2 are defined from left to right, and also
    // determine the desired comparison result (and its inverse).
    std::cout << "nearest_intersection_to_right 3 "<< std::endl;
    Orientation d_res;
    Orientation i_res;
    int    n1 = cv1._size();
    int    n2 = cv2._size();
    int    inc1;
    int    inc2;

    d_res = CLOCKWISE;
    i_res = COUNTERCLOCKWISE;
    inc1 = _is_curve_to_right(cv1) ? 1 : -1;
    inc2 = _is_curve_to_right(cv2) ? 1 : -1;
    std::cout << "nearest_intersection_to_right 4 "<< std::endl;
    // Check if the entire curve cv1 is defined to the right of p. 
    if (i1 < 0)
    {
      if (inc1 == 1)
      {
        // The source is the leftmost (rightmost) point in cv1, so p must be 
        // to its left (right).
        // Otherwise, p is to the right (left) of the x-range of cv1. 
        if (orientation(o,seg_traits.curve_source(cv1[0]),p) == d_res)
          i1 = 0;
        else
          return (false);
      }
      else // if (inc1 == -1)
      {
        // The target is the leftmost (rightmost) point in cv1, so p must be 
        // to its left (right).
        // Otherwise, p is to the right (left) of the x-range of cv1. 
        if (orientation(o,seg_traits.curve_target(cv1[n1-1]),p) == d_res)
          i1 = n1-1;
        else
         return (false);
      }
    }    

    // Check if the entire curve cv2 is defined to the right (left) of p. 
    if (i2 < 0)
    {
      std::cout << "nearest_intersection_to_right 5 "<< std::endl; 
      if (inc2 == 1)
      {
        // The source is the leftmost (rightmost) point in cv2, so p must be 
        // to its left (right).
        // Otherwise, p is to the right (left) of the x-range of cv2. 
        std::cout << "nearest_intersection_to_right 6"<< std::endl;  
        if (orientation(o,seg_traits.curve_source(cv2[0]),p) == d_res)
          i2 = 0;
        else {
          std::cout << "nearest_intersection_to_right 7 "<< std::endl;
          return (false);
        }  
      }
      else // if (inc2 == -1)
      {
        // The target is the leftmost (rightmost) point in cv2, so p must be 
        // to its left (right).
        // Otherwise, p is to the right (left) of the x-range of cv2.
        std::cout << "nearest_intersection_to_right 8"<< std::endl;   
        if (orientation(o,seg_traits.curve_target(cv2[n2-1]),p) == d_res)
          i2 = n2-1;
        else {
          std::cout << "nearest_intersection_to_right 9"<< std::endl;   
          return (false);
        } 
      }
    }

    // Try to locate the intersection point.
    bool               found;
    Orientation        res;

    while (i1 >= 0 && i1 < n1 && i2 >= 0 && i2 < n2)
    {
      // Check if the two current segment intersect to the right (left) of p.
  
      found = seg_traits.nearest_intersection_to_right (cv1[i1], cv2[i2],
                                                          p,
                                                          p1, p2);
      std::cout << "nearest_intersection_to_right 10"<< std::endl;   
      std::cout << "i1 = "<< i1 << std::endl;    
      std::cout << "i2 = "<< i2 << std::endl;    
      if (found)
      {
        std::cout << "nearest_intersection_to_right 11"<< std::endl;   
        // In case an overlap was detected, stop here:
        if (! seg_traits.point_equal (p1, p2))
          return (true);

        // In case we found a single intersection point, check whether it
        // is the next end-point of cv1[i1] or of cv2[i2].
        bool    eq1, eq2;

        eq1 = seg_traits.point_equal (p1, 
                                      (inc1 == 1) ? 
                                      seg_traits.curve_target(cv1[i1]) :
                                      seg_traits.curve_source(cv1[i1]));

        eq2 = seg_traits.point_equal (p1,
                                     (inc2 == 1) ? 
                                      seg_traits.curve_target(cv2[i2]) :
                                      seg_traits.curve_source(cv2[i2]));

        if (!eq1 && !eq2)
          return (true);

        // Proceed to the next curves.
        if (eq1)
          i1 += inc1;
        if (eq2)
          i2 += inc2;

        if (i1 < 0 || i1 >= n1 || i2 < 0 || i2 >= n2)
          return (true);

        // Check if the next curves overlap, and the nearest overlap endpoint
        // equals p1.
        Point_T2  q1, q2;

        found = seg_traits.nearest_intersection_to_right (cv1[i1], cv2[i2],
                                                          p,
                                                          q1, q2);

        if (found &&
            !seg_traits.point_equal (q1, q2) &&
            seg_traits.point_equal (p1, q1))
        {
          // Now p1 (== q1) --> q2 is an overlapping segment.
          p2 = q2;
        }

        return (true);
      }

      // Find the segment whose end-point is the leftmost (rightmost) and move
      //  forward (or backward) on its polyline.
      if (!compare_around_origin(seg_traits.curve_target(cv1[i1]),
                         seg_traits.curve_target(cv2[i2])) &&
          !compare_around_origin(seg_traits.curve_target(cv2[i2]), 
                         seg_traits.curve_target(cv1[i1])))
        res = COLLINEAR;
      else if (compare_around_origin ((inc1 == 1) ? 
                            seg_traits.curve_target(cv1[i1]) :
                            seg_traits.curve_source(cv1[i1]),
                            (inc2 == 1) ? 
                            seg_traits.curve_target(cv2[i2]) :
                            seg_traits.curve_source(cv2[i2])))
        res = CLOCKWISE;
      else
        res = COUNTERCLOCKWISE;

   //   res = (orientation (o, (inc2 == 1) ? 
                   //        seg_traits.curve_target(cv2[i2]) :
                   //        seg_traits.curve_source(cv2[i2]),
                    //      (inc1 == 1) ? 
                       //     seg_traits.curve_target(cv1[i1]) :
                       //     seg_traits.curve_source(cv1[i1])));
      
      std::cout << "inc1 = "<< inc1 << std::endl;    
      std::cout << "inc2 = "<< inc2 << std::endl;    
      
      if (res == d_res)
      {
        i1 += inc1;
      }
      else if (res == i_res)
      {
        i2 += inc2;
      }
      else
      {
        i1 += inc1;
        i2 += inc2;
      }
      std::cout << "nearest_intersection_to_right 12"<< std::endl;   
      std::cout << "i1 = "<< i1 << std::endl;    
      std::cout << "i2 = "<< i2 << std::endl;    
      
    }

    // No intersection found:
    std::cout << "nearest_intersection_to_right - VAI SAIR"<< std::endl;   
    return (false);
  }

  bool is_defined_left(const Point_T2 & p, const X_monotone_curve_T2 &curve) const
  {
     Point_T2 source = curve_point(curve,0);
     const Point_T2 o(FT(1),FT(0),FT(0));
     if (p == source) {
       const Point_T2 &target = curve_point(curve,1);
       return (orientation(o,p,target) < 0);
     }
     else { 
       Point_T2 source = curve_point(curve,curve._size()-1);
       const Point_T2 &target = curve_point(curve,curve._size());
       return (orientation(o,target,source) < 0);
     }
  }


  Point_T2 curve_point(const X_monotone_curve_T2 & cv, const int i) const
  {
    CGAL_precondition(i >= 0 && i <= cv._size());
    if (i == cv._size())
      return cv[i-1].target();
    return cv[i].source();    
  }

  bool is_SourceFirstProcessed(const X_monotone_curve_T2 &  curve,
                               Point_T2 & m_lastPoint) {

    Point_T2 o(FT(1),FT(0),FT(0));
    bool m_isSourceFirstProcessed;
    const Point_T2 & m_source = curve_point(curve,0);
    const Point_T2 & m_target = curve_point(curve,1);
    Orientation ori = orientation(o, m_source, m_target);
    std::cout << "ORI = "<< ori << std::endl;
    std::cout << "M_SOURCE = "<< m_source << std::endl;
    std::cout << "M_TARGET = "<< m_target << std::endl;
    if ( ori  == -1 )
    {
      m_lastPoint = curve_target(curve); 
      std::cout << "m_lastPoint = " << m_lastPoint << std::endl;
      m_isSourceFirstProcessed = false;
    }
    else if ( ori == 1)
    {
      m_lastPoint = m_source; 
      std::cout << "m_lastPoint = " << m_lastPoint << std::endl;
      m_isSourceFirstProcessed = true;
    }
   else {
      std::cout << "cmp_dist_collinear_origin\n\n";
      std::cout << "source = " << m_source << std::endl;
      std::cout << "target = " << m_target << std::endl;
      std::cout << cmp_dist_collinear_origin(m_source,m_target) << std::endl;
      if (m_source == antipode(o)) {
        if (!has_on_segment(m_target,o,Point_T2(0,1,0)) && 
                !has_on_segment(m_target,antipode(o),Point_T2(0,1,0))) {
          m_isSourceFirstProcessed = true;
          m_lastPoint = m_source;
        } 
        else {
          m_isSourceFirstProcessed = false;
          m_lastPoint = m_target;
        }
      } 
      else if (m_target == antipode(o)) {
        if (has_on_segment(m_source,o,Point_T2(0,1,0)) ||
                has_on_segment(m_source,antipode(o),Point_T2(0,1,0))){  
          m_isSourceFirstProcessed = true;
          m_lastPoint = m_source;
        } 
        else {
          m_isSourceFirstProcessed = false;
          m_lastPoint = m_target;
        }
      } 
      else if (cmp_dist_collinear_origin(m_source,m_target)  == SMALLER) 
      {
        m_isSourceFirstProcessed = true;
        m_lastPoint = m_source;
      }
      else {
        m_isSourceFirstProcessed = false;
        m_lastPoint = curve_target(curve);
      }
    }
    return m_isSourceFirstProcessed;
  }
  

private:

  int _locate_point (const X_monotone_curve_T2& cv, const Point_T2& q) const
  {
    // First check whether the polyline curve really contains q in its dirSweepLine-range.
    int          from = 0;
    Orientation  res_from;
    int          to = cv._size() - 1;
    Orientation  res_to;

    typedef typename Point_T2::FT  FT;

    const Point_T2 o(FT(1),FT(0),FT(0));

    res_from = orientation(o,seg_traits.curve_source(cv[from]), q);
    //bool r_from = compare_around_origin(seg_traits.curve_source(cv[to]), q);
    std::cout << "seg_traits.curve_source(cv[from]) =  "
              << seg_traits.curve_source(cv[from]) << std::endl; 
    std::cout << "q =  " << q << std::endl; 
    std::cout << "res_from =  "<< res_from << std::endl;
    if (res_from == COLLINEAR) {
      Point_T2 source = seg_traits.curve_source(cv[from]);
      Point_R3 oR3(FT(1),FT(0),FT(0));
      Point_R3 oaR3(FT(-1),FT(0),FT(0));
      Point_R3 sourceR3(source.w(),source.x(),source.y());
      Point_R3 qR3(q.w(),q.x(),q.y());
      if (coplanar_orientation(oR3,oaR3,sourceR3)*
          coplanar_orientation(oR3,oaR3,qR3) == 1)
        return (from);
    }
    res_to = orientation(o, seg_traits.curve_target(cv[to]), q);
    //bool r_to = compare_around_origin(seg_traits.curve_target(cv[to]), q);
    
    std::cout << "res_to =  "<< res_to << std::endl;
    if (res_to == COLLINEAR)
      return (to);
    
    if (int(orientation(o,curve_source(cv),curve_point(cv,1))*
        orientation(o,curve_source(cv),curve_target(cv)))!=-1) {
      std::cout << "ENTROU CURVA MENOR QUE 180 "<< std::endl;
      if (res_from == res_to)
        return (-1);
    }
    else {
      std::cout << "ENTROU CURVA MAIOR QUE 180 "<< std::endl;
      if (int(orientation(o,curve_source(cv),q)*
          orientation(o,curve_source(cv),curve_point(cv,1)))==-1 &&
          int(orientation(o,curve_target(cv),q)*
          orientation(o,curve_target(cv),curve_point(cv,cv._size()-1)))==-1)
        return (-1);
      
    }
   
    // Perform a binary search and locate the segment that contains q in its
    // x-range.
    int i=0;
    Orientation  o1;
    bool found = false;

    while (i < to && !found) {
      o1 = orientation(o,cv[i].source(),cv[i].target());
      std::cout << "o1 = "<< o1 <<  std::endl;
      if ((orientation(o,cv[i].source(),q) == o1 || orientation(o,cv[i].source(),q) == 0) &&
          (orientation(o,cv[i].target(),q) == -o1 || orientation(o,cv[i].target(),q) == 0))
      {
        found = true;
        std::cout << "orientation(o,cv[i].source(),q) = "
                  << orientation(o,cv[i].source(),q) <<  std::endl;
        std::cout << "orientation(o,cv[i].target(),q) = "
                  << orientation(o,cv[i].target(),q) <<  std::endl;
      }
      else
        i++;
    } 
    return i;
    /*while (to > from)
    {
      mid = (from + to) / 2;

      if (mid > from)
      {
        if (has_on_segment(q,o,seg_traits.curve_source(cv[mid])) ||
            has_on_segment(q,antipode(o),seg_traits.curve_source(cv[mid])))
          return (mid);
           
        bool r_mid_s = compare_around_origin(seg_traits.curve_source(cv[mid]), q);
        
        if (!r_mid_s && !r_from)
          from = mid; 
        else
          to = mid - 1; 
      }
      else
      {
        CGAL_assertion (mid < to);
        if (has_on_segment(q,o,seg_traits.curve_target(cv[mid])) ||
            has_on_segment(q,antipode(o),seg_traits.curve_target(cv[mid])))
          return (mid);
          
        bool r_mid_t = compare_around_origin(seg_traits.curve_target(cv[mid]), q);

        if (!r_mid_t && !r_to)
          from = mid + 1;
        else
          to = mid;
          
      }
    }

    // In case (from == to), and we know that the polyline contains the q:
    CGAL_assertion (from == to);
    return (from);*/
    
  }

    
};

/*!
 * A representation of a polyline, as used by the Arr_polyline_traits_2
 * traits class.
 */
template <class Segment_traits_>
class Polyline_T2
{
  friend class Arr_polyline_traits_T2<Segment_traits_>;

public:

  typedef Segment_traits_                         Segment_traits_T2;
  typedef typename Segment_traits_T2::Point_T2    Point_T2;
  typedef typename Segment_traits_T2::Curve_T2    Segment_T2;
  typedef ::std::vector<Segment_T2>               Base;

private:
  
  // The segments that comprise the poyline:
  ::std::vector<Segment_T2>                          segments;

public:

  /*!
   * Default constructor.
   */
  Polyline_T2 () :
    segments()
  {}

  /*!
   * Constructor from a range of points, defining the endpoints of the
   * polyline segments.
   * \param pts_begin An iterator pointing to the first point in the range.
   * \param pts_end An iterator pointing after the last point in the range.
   * \pre The are at least 2 points in the range.
   *      In other cases, an empty polyline will be created.
   */
  template <class InputIterator>
  Polyline_T2 (const InputIterator& pts_begin,
               const InputIterator& pts_end) :
    segments()
  {
    // Check if there are no points in the range:
    InputIterator  ps = pts_begin;

    if (ps == pts_end)
      return;

    // Construct a segment from each to adjacent points.
    InputIterator  pt = ps;
    pt++;

    while (pt != pts_end)
    {
      segments.push_back (Segment_T2 (*ps, *pt));
      ps++; pt++;
    }
  }

  /*!
   * Create a bounding-box for the polyline.
   * \return The bounding-box.
   */
  /*Bbox_2 bbox() const
  {
    // Compute the union of the bounding boxes of all segments.
    Bbox_2 bbox;
    int    n = _size();
    int    i;

    for (i = 0; i < n; i++)
    {
      if (i == 0)
	bbox = (*this)[i].bbox();
      else
	bbox = bbox + (*this)[i].bbox();
    }

    return (bbox);
  }*/

  class const_iterator;
  friend class const_iterator;

  /*!
   * An iterator for the polyline points.
   */
  class const_iterator
  {
    private:
    // The polyline curve.
    const Polyline_T2<Segment_traits_> *cvP;       

    // Its number of points. 
    int  n_pts;       

    // Auxiliary variable.
    Segment_traits_                   seg_traits;  

    // The current point.
    int                               i;           

    /*!
     * Private constructor.
     * \param cv The scanned curve.
     * \param index The index of the segment.
     */
    const_iterator (const Polyline_T2<Segment_traits_>* _cvP,
                    const int& index) :
      cvP(_cvP),
      i(index)
    {
      if (cvP == NULL)
        n_pts = 0;
      else
        n_pts = (cvP->_size() == 0) ? 0 : (cvP->_size() + 1);
    }

  public:
    
    /*!
     * Default constructor.
     */
    const_iterator () :
      cvP(NULL),
      n_pts(0),
      i(-1)
    {}

    /*!
     * Dereference operator.
     * \return The current point.
     */
    Point_T2 operator* () const
    {
      CGAL_assertion(cvP != NULL);
      CGAL_assertion(i >= 0 && i < n_pts);

      if (i == 0)
        // First point is the source of the first segment.
        return (seg_traits.curve_source ((*cvP)[0]));
      else
        // Return the target of the (i-1)'st segment
        return (seg_traits.curve_target ((*cvP)[i-1]));
    }

    /*!
     * Increment operators.
     */
    void operator++ () 
    {
      if (cvP != NULL && i < n_pts)
        i++;
    }

    void operator++ (int)
    {
      if (cvP != NULL && i < n_pts)
        i++;
    }

    /*!
     * Decrement operators.
     */
    void operator-- ()
    {
      if (cvP != NULL && i >= 0)
        i--;
    }

    void operator-- (int)
    {
      if (cvP != NULL && i >= 0)
        i--;
    }

    /*!
     * Equality operators.
     */
    bool operator== (const const_iterator& it) const
    {
      return (cvP == it.cvP && i == it.i);
    }

    bool operator!= (const const_iterator& it) const
    {
      return (cvP != it.cvP || i != it.i);
    }

    friend class Polyline_T2<Segment_traits_>;
  };

  /*!
   * Get an iterator for the polyline points.
   * \return An iterator that points on the first point.
   */
  const_iterator begin () const
  {
    if (_size() == 0)
      return (const_iterator (NULL, -1));
    else
      return (const_iterator (this, 0));
  }

  /*!
   * Get a past-the-end iterator for the polyline points.
   * \return A past-the-end iterator.
   */
  const_iterator end () const
  {
    if (_size() == 0)
      return (const_iterator (NULL, -1));
    else
      return (const_iterator (this, _size() + 1));
  }

  /*!
   * Get an reverse iterator for the polyline points.
   * \return An iterator that points on the last point.
   */
  const_iterator rbegin () const
  {
    if (_size() == 0)
      return (const_iterator (NULL, -1));
    else
      return (const_iterator (this, _size()));
  }

  /*!
   * Get a reverse past-the-end iterator for the polyline points.
   * \return A reverse past-the-end iterator.
   */
  const_iterator rend () const
  {
    if (_size() == 0)
      return (const_iterator (NULL, -1));
    else
      return (const_iterator (this, -1));
  }

  /*!
   * Get the number of points contained in the polyline.
   * \return The number of points.
   */
  unsigned int points () const
  {
    if (_size() == 0)
      return (0);
    else
      return (_size() + 1);
  }

private:

  /*!
   * Append a segment to the polyline.
   * \param seg The new segment to be appended to the polyline.
   * \pre If the polyline is not empty, the segment source must be the
   *      same as the target point of the last segment in the polyline.
   */
  inline void _push_back (const Segment_T2& seg)
  {
    segments.push_back(seg);
    return;
  }

  /*!
   * Get the number of segments that comprise the poyline.
   * \return The number of segments.
   */
  inline int _size () const
  {
    return (static_cast<int>(segments.size()));
  }

  /*!
   * Get the i'th segment of the polyline.
   * \param i The segment index (from 0 to size()-1).
   * \return A const reference to the segment.
   */
  inline const Segment_T2& operator[] (const int& i) const
  {
    return (segments[i]);
  }

  /*!
   * Clear the polyline.
   */
  inline void _clear ()
  {
    segments.clear();
    return;
  }
};

/*!
 * Output operator for a polyline.
 */
template <class Segment_traits_, class Stream_>
Stream_ & operator<< (Stream_ & os, const Polyline_T2<Segment_traits_> & cv)
{
  typename Polyline_T2<Segment_traits_>::const_iterator ps = cv.begin();
  typename Polyline_T2<Segment_traits_>::const_iterator pt = ps; pt++;

  while (pt != cv.end()) {
    typename Segment_traits_::Curve_T2 seg(*ps, *pt);
    os << seg;
    ps++; pt++;
  }
  return (os);
}

CGAL_END_NAMESPACE

#endif
