// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Circle_2_Circle_2_intersection.h,v $
// $Revision: 1.7 $ $Date: 2003/10/21 12:16:43 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CIRCLE_T2_CIRCLE_T2_INTERSECTION_H
#define CGAL_CIRCLE_T2_CIRCLE_T2_INTERSECTION_H

#include <CGAL/Circle_T2.h>
//#include <CGAL/squared_distance_2_1.h>
CGAL_BEGIN_NAMESPACE

template <class R>
bool
do_intersect(const Circle_T2<R> & circ1, const Circle_T2<R>& circ2)
{
  if (circ1.shape_circle()*circ2.shape_circle())
    return less_equalT1p(squared_distance(circ1.center(),circ2.center()),
                   circ1.squared_radius() + circ2.squared_radius());
  else if (circ1.shape_circle() == LINE && circ2.shape_circle() == CIRCLE) {
    Line_T2<R> l(circ1.point0(),circ1.point1());
    Line_T2<R> s(circ2.center(),perpendicular_direction(l));
    Line_T2_Line_T2_pair<R> llpair(&s,&l);
    Point_T2<R> pt;
    llpair.intersection(pt);
    if (!circ2.has_on_unbounded_side(pt))
      return true;
    return (!circ2.has_on_unbounded_side(antipode(pt)));
  }
  else if (circ2.shape_circle() == LINE && circ1.shape_circle() == CIRCLE) {
    Line_T2<R> l(circ2.point0(),circ2.point1());
    Line_T2<R> s(circ1.center(),perpendicular_direction(l));
    Line_T2_Line_T2_pair<R> llpair(&s,&l);
    Point_T2<R> pt;
    llpair.intersection(pt);
    if (!circ1.has_on_unbounded_side(pt))
      return true;
    return (!circ1.has_on_unbounded_side(antipode(pt)));
  }
  
  return true;
}

CGAL_END_NAMESPACE

#endif
