// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/constructions/kernel_ftC2.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:14:37 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Sven Schoenherr, Herv Brnnimann, Sylvain Pion
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CONSTRUCTIONS_T2_KERNEL_FTT2_H
#define CGAL_CONSTRUCTIONS_T2_KERNEL_FTT2_H

CGAL_BEGIN_NAMESPACE

template < class FT >
inline
FT
line_x_at_wyT2(const FT &W, const FT &X, const FT &Y, const FT &w, const FT &y)
{
  return (-W*w-Y*y) / X;
}

template < class FT >
inline
FT
line_y_at_wxT2(const FT &W, const FT &X, const FT &Y, const FT &w, const FT &x)
{
  return (-W*w-X*x) / Y;
}

template < class FT >
CGAL_KERNEL_INLINE
void
midpointT2( const FT &pw, const FT &px, const FT &py,
            const FT &qw, const FT &qx, const FT &qy,
            FT &w, FT &x, FT &y )
{
  w = CGAL_NTS abs(qw)*pw + CGAL_NTS abs(pw)*qw;
  x = CGAL_NTS abs(qw)*px + CGAL_NTS abs(pw)*qx;
  y = CGAL_NTS abs(qw)*py + CGAL_NTS abs(pw)*qy;
}

/*
template < class FT >
CGAL_KERNEL_MEDIUM_INLINE
void
centroid_T2(const FT &pw, const FT &px, const FT &py,
            const FT &qw, const FT &qx, const FT &qy,
            const FT &rw, const FT &rx, const FT &ry,
            FT &w, FT &x, FT &y)
{
   x = (px + qx + rx)/FT(3);
   y = (py + qy + ry)/FT(3);
}
*/
CGAL_END_NAMESPACE

#endif //CGAL_CONSTRUCTIONS_T2_KERNEL_FTT2_H
