// Copyright (c) 1997  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Delaunay_triangulation_2.h,v $
// $Revision: 1.66 $ $Date: 2003/09/18 10:26:05 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Mariette Yvinec
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_DELAUNAY_TRIANGULATION_T2_H
#define CGAL_DELAUNAY_TRIANGULATION_T2_H

#include <CGAL/Triangulation_T2.h>
#include <CGAL/Triangulation_not_Delaunay_T2.h>
#include <CGAL/iterator.h>

CGAL_BEGIN_NAMESPACE

template < class Gt,
           class Tds = Triangulation_data_structure_T2 <
                           Triangulation_vertex_base_2<Gt> ,
Triangulation_face_base_2<Gt, Triangulation_ds_face_base_T2<> > > >
class Delaunay_triangulation_T2 : public Triangulation_T2<Gt,Tds>
{
public:
  typedef Gt                                                    Geom_traits;
  typedef typename Geom_traits::K                               K;
  typedef typename K::FT                                        FT;
  typedef typename Geom_traits::Point_2                         Point;
  typedef typename Geom_traits::Point_T1                        Point_T1;
  typedef typename Geom_traits::Segment_2                       Segment;
  typedef typename Geom_traits::Triangle_2                      Triangle;
  typedef typename Geom_traits::Line_2                          Line;
  typedef typename Geom_traits::Direction_T2                    Direction;
  typedef typename Geom_traits::Orientation_2                   Orientation_2;


  typedef Triangulation_T2<Gt,Tds>                              Triangulation;
  typedef Triangulation_not_Delaunay_T2<Gt,Tds>
                                                   Triangulation_not_Delaunay;
  typedef typename Triangulation::Locate_type                   Locate_type;
  typedef typename Triangulation::Face_handle                   Face_handle;
  typedef typename Triangulation::Face_circulator               Face_circulator;
  typedef typename Triangulation::Face_iterator                 Face_iterator;
  typedef typename Triangulation::Vertex_handle                 Vertex_handle;
  typedef typename Triangulation::Edge                          Edge;
  typedef typename Triangulation::Edge_circulator               Edge_circulator;
  typedef typename Triangulation::Finite_edges_iterator         Finite_edges_iterator;
  typedef typename Triangulation::Finite_faces_iterator         Finite_faces_iterator;
  typedef typename Triangulation::Finite_vertices_iterator      Finite_vertices_iterator;
  typedef typename std::vector<Segment>::iterator               MDST_iterator;


  Delaunay_triangulation_T2(const Gt& gt = Gt())
  : Triangulation_T2<Gt,Tds>(gt) {}

  Delaunay_triangulation_T2(
        const Delaunay_triangulation_T2<Gt,Tds> &tr)
      : Triangulation_T2<Gt,Tds>(tr)
  {   CGAL_triangulation_postcondition( is_valid() );
         }

 bool is_delaunay_aux(const Point & a, const Point & b,
                const Point & c, const Point & d) const;

 bool is_delaunay(Face_handle& f, Face_handle& n, int index);

  // CHECK -QUERY
  bool is_valid(bool verbose = false, int level = 0) const;

  Vertex_handle
  nearest_vertex(const Point& p, Face_handle f= Face_handle(NULL)) const;

  bool does_conflict(const Point  &p, Face_handle fh) const;// deprecated
  bool test_conflict(const Point  &p, Face_handle fh) const;
  bool find_conflicts(const Point  &p,                //deprecated
		      std::list<Face_handle>& conflicts,
		      Face_handle start= Face_handle(NULL) ) const;



   struct MDST_edge {
     Edge edge;

     MDST_edge(){}
     MDST_edge(Edge _edge) :
     edge(_edge)
     {}

     Edge get_edge() const {
       return edge;
     }

  };

  template <typename Segment_iterator>
     Segment_iterator dual(const Segment_iterator &out) const      {
       Segment_iterator out2(out);
       bool in_both_range = false;
       if (check_ch_defined()) {
       Finite_vertices_iterator fvi = finite_vertices_begin();
       Sign sg = sign(fvi->point().w());
       fvi++;
       while (fvi != finite_vertices_end() && !in_both_range) {
         if (sg != sign(fvi->point().w())) 
	   in_both_range = true;
         fvi++;
       }
       }
       if (!(check_ch_defined() && !in_both_range) ) {
          Finite_edges_iterator eit= finite_edges_begin();
          if (dimension() == 2) {
          for (; eit != finite_edges_end(); ++eit) {
            Point other_point;
            
            if (!is_infinite(eit->first) && 
                !is_infinite(eit->first->neighbor(eit->second))) {
              if (circumcenter(eit->first).w() == 0 && 
                  circumcenter(eit->first) == 
                  circumcenter(eit->first->neighbor(eit->second))) {  
                if (eit->first->vertex(cw(eit->second))->point() == 
                    circumcenter(eit->first))
                  other_point = midpoint(eit->first->vertex(eit->second)->point(),
                           eit->first->vertex(ccw(eit->second))->point());
                else
                  other_point = midpoint(eit->first->vertex(eit->second)->point(),
                           eit->first->vertex(cw(eit->second))->point());
                Segment s = Segment(circumcenter(eit->first),other_point); 
                *out2 = s;
                out2++;
              }
              else  {
                Point pt1 = circumcenter(eit->first);
                Point pt2 = circumcenter(eit->first->neighbor(eit->second)); 
                Segment s;
                if (!antipodal(pt1,pt2)) {
                  s = Segment(pt1,pt2);
                  *out2 = s;
	          out2++; 
                }
                
              }
            }
            else if (is_infinite(eit->first))  {
              int ind = eit->first->index(infinite_vertex());
              Point pt1 = circumcenter(eit->first->neighbor(eit->second));
              if (eit->first->vertex(ccw(ind))->point().w() != 0 ||
                  eit->first->vertex(cw(ind))->point().w() != 0) {
                Direction dr = perpendicular_direction(
                                   eit->first->vertex(ccw(ind))->point(),
                                   eit->first->vertex(cw(ind))->point());
                if (!antipodal(pt1,Point(0,dr.dx(),dr.dy()))) {  
                  Segment s = Segment(pt1,Point(0,dr.dx(),dr.dy()));
                  *out2 = s;
                  out2++;
                }
                else {
                  Point q = midpoint(eit->first->vertex(ccw(eit->second))->point(),
                                     eit->first->vertex(cw(eit->second))->point());
                  Segment s = Segment(pt1,q);
                  *out2 = s;
                  out2++;
                  s = Segment(q,Point(0,dr.dx(),dr.dy()));
                  *out2 = s;
                  out2++;
                }
              }
            }  
            else if (is_infinite(eit->first->neighbor(eit->second)))  {
              int ind = eit->first->neighbor(eit->second)->index((infinite_vertex()));
              Point pt1 = circumcenter(eit->first);
              if (eit->first->neighbor(eit->second)->vertex(ccw(ind))->point().w() != 0 ||
                  eit->first->neighbor(eit->second)->vertex(cw(ind))->point().w() != 0) {
                Direction dr = perpendicular_direction(
                               eit->first->neighbor(eit->second)->vertex(ccw(ind))->point(),
                               eit->first->neighbor(eit->second)->vertex(cw(ind))->point());
                if (!antipodal(pt1,Point(0,dr.dx(),dr.dy()))) {  
                  Segment s = Segment(pt1,Point(0,dr.dx(),dr.dy()));
                  *out2 = s;
                  out2++;
                }
                else {
                  int iv = eit->first->neighbor(eit->second)->index(infinite_vertex());
                  Point q = midpoint(
                    eit->first->neighbor(eit->second)->vertex(ccw(iv))->point(),
                    eit->first->neighbor(eit->second)->vertex(cw(iv))->point());
                  Segment s = Segment(pt1,q);
                  *out2 = s;
                  out2++;
                  s = Segment(q,Point(0,dr.dx(),dr.dy()));
                  *out2 = s;
                  out2++;
                }
              }
            } 
            }
            *out2 = Segment(Point(0,1,0),Point(0,0,1));
            out2++;
            *out2 = Segment(Point(0,-1,0),Point(0,0,1));
            out2++; 
            *out2 = Segment(Point(0,-1,0),Point(0,0,-1));
            out2++;
            *out2 = Segment(Point(0,1,0),Point(0,0,-1)); 
            out2++;
          
          
          std::vector<Vertex_handle> v_front, v_back;
          Finite_vertices_iterator fv;
          for (fv = finite_vertices_begin(); fv != finite_vertices_end(); fv++)
          {
            if ((void*)(&(*(fv->face()))) != NULL) {
              if (fv->point().w() > 0) {
                v_front.push_back(fv);
              }
              else if (fv->point().w() < 0) {
                v_back.push_back(fv);
              }
            }
          }
          if (v_front.size() == 2) {
            Direction dr = perpendicular_direction(v_front[0]->point(),v_front[1]->point());
            *out2 = Segment(midpoint(v_front[0]->point(),v_front[1]->point()),
                            Point(0,dr.dx(),dr.dy())); 
            out2++;
            *out2 = Segment(midpoint(v_front[0]->point(),v_front[1]->point()),
                            Point(0,-dr.dx(),-dr.dy())); 
            out2++;
           }
           else { //collinear
             unsigned i = 0;
             Orientation c = COUNTERCLOCKWISE;
             if (v_front.size() >= 3) {
               do {
                 c = orientation(v_front[i]->point(),v_front[i+1]->point(),v_front[i+2]->point());
                 i++;
               } while (i < v_front.size()-2 && c==0);
             }
             if (c == 0) { 
               for (eit = finite_edges_begin(); eit != finite_edges_end(); eit++)
               {
                 if (eit->first->vertex(ccw(
                             eit->second))->point().w() > 0 &&
                     eit->first->vertex(ccw(
                             eit->second))->point().w() *
                     eit->first->vertex(cw(
                             eit->second))->point().w() > 0) {
                     Direction dr = perpendicular_direction(
                                          eit->first->vertex(ccw(
                             eit->second))->point(),
                                          eit->first->vertex(cw(
                             eit->second))->point());
                     *out2 = Segment(midpoint(eit->first->vertex(ccw(
                             eit->second))->point(),
                             eit->first->vertex(cw(
                             eit->second))->point()),
                            Point(0,-dr.dx(),-dr.dy())); 
                     out2++; 
                     *out2 = Segment(midpoint(eit->first->vertex(ccw(
                             eit->second))->point(),
                             eit->first->vertex(cw(
                             eit->second))->point()),
                            Point(0,dr.dx(),dr.dy())); 
                     out2++; 
                  }
                }
              }
           } 
           if (v_back.size() == 2) {
             Direction dr = perpendicular_direction(v_back[0]->point(),v_back[1]->point());
             *out2 = Segment(midpoint(v_back[0]->point(),v_back[1]->point()),
                           Point(0,dr.dx(),dr.dy())); 
             out2++;
             *out2 = Segment(midpoint(v_back[0]->point(),v_back[1]->point()),
                           Point(0,-dr.dx(),-dr.dy())); 
             out2++;
           }
           else { //collinear
             unsigned i = 0;
             Orientation c = COUNTERCLOCKWISE;
             if (v_back.size() >= 3) {
               do {
                 c = orientation(v_back[i]->point(),
                                 v_back[i+1]->point(),
                                 v_back[i+2]->point());
                 i++;
               } while (i < v_back.size()-2 && c==0);
             }
             if (c == 0) { 
               for (eit = finite_edges_begin(); eit != finite_edges_end(); eit++)
               {
                 if (eit->first->vertex(ccw(
                             eit->second))->point().w() < 0 &&
                     eit->first->vertex(ccw(
                             eit->second))->point().w() *
                     eit->first->vertex(cw(
                             eit->second))->point().w() > 0) {
                    Direction dr = perpendicular_direction(
                                          eit->first->vertex(ccw(
                             eit->second))->point(),
                                          eit->first->vertex(cw(
                             eit->second))->point());
                    *out2 = Segment(midpoint(eit->first->vertex(ccw(
                             eit->second))->point(),
                             eit->first->vertex(cw(
                             eit->second))->point()),
                            Point(0,-dr.dx(),-dr.dy())); 
                     out2++; 
                     *out2 = Segment(midpoint(eit->first->vertex(ccw(
                             eit->second))->point(),
                             eit->first->vertex(cw(
                             eit->second))->point()),
                            Point(0,dr.dx(),dr.dy())); 
                      out2++; 
                  }
                }
              }
           }  
  
       }
       else {
         Finite_edges_iterator eit= finite_edges_begin();
         for (; eit != finite_edges_end(); ++eit) {
           if (eit->first->vertex(0)->point().w()!=0 ||
               eit->first->vertex(1)->point().w()!=0) {
             Direction dr = perpendicular_direction(eit->first->vertex(0)->point(),
                                  eit->first->vertex(1)->point());
             Point m = midpoint(eit->first->vertex(0)->point(),
                                eit->first->vertex(1)->point());
             *out2 = Segment(m,Point(0,dr.dx(),dr.dy())); 
             out2++; 
             *out2 = Segment(m,Point(0,-dr.dx(),-dr.dy())); 
             out2++;  
           }
         }
         if (check_ch_defined()) {
           Face_handle ff = infinite_face();
           int iv = ff->index(infinite_vertex());
           Face_handle f = ff->neighbor(1-iv);
           Vertex_handle vh0= ff->vertex(1-iv);
           iv = f->index(infinite_vertex());
           Vertex_handle vh1= f->vertex(1-iv);
    
           if (vh0->point().w()!=0 ||
               vh1->point().w()!=0) {
             Direction dr = perpendicular_direction(vh0->point(),
                                                    vh1->point());
             Point m = midpoint(vh0->point(),
                                vh1->point());
             *out2 = Segment(antipode(m),Point(0,dr.dx(),dr.dy())); 
             out2++; 
             *out2 = Segment(antipode(m),Point(0,-dr.dx(),-dr.dy())); 
             out2++;  
           }
         }
       }
       }
       else {
         if (dimension() == 2) {
           int i;
           std::vector<Point> v;
           Triangulation_not_Delaunay tri;
           Vertex_handle vh = infinite_vertex();

	   Face_circulator fc = incident_faces(vh), done(fc);
           do {
             i = fc->index(vh);
             v.push_back(fc->vertex(ccw(i))->point());
             fc++;
           } while(fc!=done);
           Finite_vertices_iterator vit= finite_vertices_begin();
           for(; vit != finite_vertices_end(); vit++) {
             fc = incident_faces(vh);
             done = fc;
             if ((void*)(&(*(vit->face()))) != NULL) {
               do {
                 i = fc->index(vh);
                 Segment seg = Segment(fc->vertex(ccw(i))->point(),
                                       fc->vertex(cw(i))->point()); 
                 if (seg.has_on(vit->point()))
                   v.push_back(vit->point());
                 fc++;
               } while(fc!=done);
             }
           }
           tri.insert(v.begin(),v.end());
           v.erase(v.begin(),v.end()); 
           Finite_edges_iterator eit= finite_edges_begin();
           for (; eit != finite_edges_end(); ++eit) {
             if ((!is_infinite(eit->first)) &&
                 (!is_infinite(eit->first->neighbor(eit->second))) ) {
               Point p = dual(eit->first);
               Point q = dual(eit->first->neighbor(eit->second));
               if (p.w() == 0) {
                 if (orientation(eit->first->vertex(0)->point()
                    ,eit->first->vertex(2)->point(),p)!=1) 
                    p = antipode(p);
               }
               if (q.w() == 0) {
                 if (orientation(
                      eit->first->neighbor(eit->second)->vertex(0)->point(),
                      eit->first->neighbor(eit->second)->vertex(2)->point(),p)!=1) 
                    q = antipode(q);
               }
               Segment s = geom_traits().construct_segment
                 (p,q);
               *out2 = s;
               out2++;
             }
             else if (is_infinite(eit->first)) {
               Point q;
               Point p = dual(eit->first->neighbor(eit->second));
               std::vector<Face_handle> v1;
	       int index_iv = eit->first->index(infinite_vertex());
               fc = tri.incident_faces(tri.infinite_vertex());
	       done = fc;
	       Vertex_handle vh0,vh1;
	       bool found1 = false;
	       bool found2 = false;
	       int index_v2;
	       do {
                 index_v2 = fc->index(tri.infinite_vertex());
	         if (fc->vertex(cw(index_v2))->point() ==
	             eit->first->vertex(cw(index_iv))->point()) {
                   vh0 = fc->vertex(cw(index_v2));
		   found1 = true;
	         }
                 else if (fc->vertex(ccw(index_v2))->point() ==
	                  eit->first->vertex(cw(index_iv))->point()) {
                   vh0 = fc->vertex(ccw(index_v2));
		   found1 = true;
	         }
                 if (fc->vertex(cw(index_v2))->point() ==
	             eit->first->vertex(ccw(index_iv))->point()) {
                   vh1 = fc->vertex(cw(index_v2));
		   found2 = true;
	         }
                 else if (fc->vertex(ccw(index_v2))->point() ==
	                  eit->first->vertex(ccw(index_iv))->point()) {
                   vh1 = fc->vertex(ccw(index_v2));
		   found2 = true;
	         }
	         fc++;
	       }while(done != fc && (!found1 || !found2));
               v1 =tri.incident_faces_not_Delaunay(
	           vh0,vh1);
	       int j = 0;
               while (tri.is_infinite(v1[j]))
	         j++;
               q = circumcenter(v1[j]->vertex(0)->point(),
                                v1[j]->vertex(1)->point(),
                                v1[j]->vertex(2)->point());
               if (p.w() == 0) {
                 if (orientation(eit->first->neighbor(eit->second)->vertex(0)->point(),
                                 eit->first->neighbor(eit->second)->vertex(2)->point(),q)!=1)
                   p = antipode(p);
               }
               if (q.w() == 0) {
                 if (orientation(v1[j]->vertex(0)->point(),
                                 v1[j]->vertex(2)->point(),q)!=1)
                   q = antipode(q);
               }
               else if (sign(q.w()) == sign(v1[j]->vertex(0)->point().w()))
	         q = antipode(q);
               if (!antipodal(p,q)) {
               *out2 = Segment(p,q);
               out2++;}
               else {
                 *out2 = Segment(p,midpoint(vh0->point(),vh1->point()));
                 out2++;
                 *out2 = Segment(q,midpoint(vh0->point(),vh1->point()));
                 out2++;
               } 
	       v1.erase(v1.begin(),v1.end());
             }
             else {
	       Point q;
               Point p = dual(eit->first);
               std::vector<Face_handle> v1;
	       int index_iv = eit->first->neighbor(eit->second)->index(
	                                           infinite_vertex());
               fc = tri.incident_faces(tri.infinite_vertex());
               done = fc;
	       Vertex_handle vh0,vh1;
	       bool found1 = false;
	       bool found2 = false;
	       int index_v2;
	       do {
                 index_v2 = fc->index(tri.infinite_vertex());
	         if (fc->vertex(cw(index_v2))->point() ==
	             eit->first->neighbor(eit->second)->vertex(
		                        cw(index_iv))->point()) {
                   vh0 = fc->vertex(cw(index_v2));
		   found1 = true;
	         }
                 else if (fc->vertex(ccw(index_v2))->point() ==
	                  eit->first->neighbor(eit->second)->vertex(
		 	           cw(index_iv))->point()) {
                   vh0 = fc->vertex(ccw(index_v2));
		   found1 = true;
	         }
                 if (fc->vertex(cw(index_v2))->point() ==
	             eit->first->neighbor(eit->second)->vertex(
		             ccw(index_iv))->point()) {
                   vh1 = fc->vertex(cw(index_v2));
		   found2 = true;
	         }
                 else if (fc->vertex(ccw(index_v2))->point() ==
	                  eit->first->neighbor(eit->second)->vertex(
			           ccw(index_iv))->point()) {
                   vh1 = fc->vertex(ccw(index_v2));
		   found2 = true;
	         }
	         fc++;
	       }while(done != fc && (!found1 || !found2));
               v1 = tri.incident_faces_not_Delaunay(vh0,vh1);
	       int j=0;
               while (tri.is_infinite(v1[j]))
	         j++;
	       q = circumcenter(v1[j]->vertex(0)->point(),
                                v1[j]->vertex(1)->point(),
                                v1[j]->vertex(2)->point());
               if (p.w() == 0) {
                 if (orientation(eit->first->neighbor(eit->second)->vertex(0)->point(),
                  eit->first->neighbor(eit->second)->vertex(2)->point(),p) != 1)
                   p = antipode(p);
               } 
               if (q.w() == 0) {
                 if (orientation(v1[j]->vertex(0)->point(),
                                v1[j]->vertex(2)->point(),q) != 1)
                   q = antipode(q);
               }
               else if (sign(q.w()) == sign(v1[j]->vertex(0)->point().w()))
                 q = antipode(q);
               if (!antipodal(p,q)) {
	         *out2 = Segment(p,q);
                 out2++;
               }
               else {
                 *out2 = Segment(p,midpoint(vh0->point(),vh1->point()));
                 out2++;
                 *out2 = Segment(q,midpoint(vh0->point(),vh1->point()));
                 out2++;
               } 
	       v1.erase(v1.begin(),v1.end());
             }
           }
	   eit= tri.finite_edges_begin();
           for (; eit != tri.finite_edges_end(); ++eit) {
             if ((!tri.is_infinite(eit->first)) &&
                (!tri.is_infinite(eit->first->neighbor(eit->second))) ) {
               Point p = antipode(dual(eit->first));
               Point q = antipode(dual(eit->first->neighbor(eit->second)));
               if (p.w() == 0) {
                 if (orientation(eit->first->vertex(0)->point(),
                                 eit->first->vertex(2)->point(),p) != 1)
                   p = antipode(p);
               } 
               if (q.w() == 0) {
                 if (orientation(eit->first->neighbor(eit->second)->vertex(0)->point(),
                                eit->first->neighbor(eit->second)->vertex(2)->point(),q) != 1)
                   q = antipode(q);
               }
               Segment s = geom_traits().construct_segment(p,q);
               *out2 = s;
               out2++;
             }
           }
           Finite_faces_iterator ite = tri.finite_faces_begin();
           for(; ite != tri.finite_faces_end(); ite++) {
           }
         }
         else {
           Finite_edges_iterator eit= finite_edges_begin();
           for (; eit != finite_edges_end(); ++eit) {
             if (eit->first->vertex(0)->point().w()!=0 ||
               eit->first->vertex(1)->point().w()!=0) {
               Direction dr = perpendicular_direction(eit->first->vertex(0)->point(),
                                eit->first->vertex(1)->point());
               Point m = midpoint(eit->first->vertex(0)->point(),
                                eit->first->vertex(1)->point());
               *out2 = Segment(m,Point(0,dr.dx(),dr.dy())); 
               out2++; 
               *out2 = Segment(m,Point(0,-dr.dx(),-dr.dy())); 
               out2++;  
             } 
           }
           Face_handle ff = infinite_face();
           int iv = ff->index(infinite_vertex());
           Face_handle f = ff->neighbor(1-iv);
           Vertex_handle vh0= ff->vertex(1-iv);
           iv = f->index(infinite_vertex());
           Vertex_handle vh1= f->vertex(1-iv);
    
           if (vh0->point().w()!=0 ||
               vh1->point().w()!=0) {
             Direction dr = perpendicular_direction(vh0->point(),
                                                    vh1->point());
             Point m = midpoint(vh0->point(),
                                vh1->point());
             *out2 = Segment(antipode(m),Point(0,dr.dx(),dr.dy())); 
             out2++; 
             *out2 = Segment(antipode(m),Point(0,-dr.dx(),-dr.dy())); 
             out2++;    
           }
         }
       }
     return out2;
   }


  class Smaller_distance{
    private:
      Geom_traits _gt;

    public:
      Smaller_distance(const Geom_traits &gt) : _gt(gt) {}
      typedef bool result_type;

      bool
      operator()(const MDST_edge e1, const MDST_edge e2) const
      {
        return (_gt.compare_distance(e1.get_edge().first->vertex(
                                        ccw(e1.get_edge().second))->point(),
                                     e1.get_edge().first->vertex(
                                        cw(e1.get_edge().second))->point(),
                                     e2.get_edge().first->vertex(
                                        ccw(e2.get_edge().second))->point(),
                                     e2.get_edge().first->vertex(
                                        cw(e2.get_edge().second))->point())
                                               == SMALLER);
      }
  };


  MDST_iterator MDST_begin() {
    typename std::vector< std::vector<Segment> >::iterator it = MDST_vector.begin();
    return (*it).begin();
  }

  MDST_iterator MDST_end() {
    typename std::vector< std::vector<Segment> >::iterator it = MDST_vector.begin();
    return (*it).end();
  }

  void MDST() {
    std::vector<MDST_edge> EdgeList;
    initialize_EdgeList(finite_edges_begin(),finite_edges_end(),EdgeList);
    Smaller_distance smaller_distance(_gt);
    std::sort(EdgeList.begin(), EdgeList.end(),smaller_distance);
    typename std::vector<MDST_edge>::iterator ite = EdgeList.begin();

    std::vector<Segment> vseg,vseg_aux,v_edges;
    vseg_aux = segment(ite->get_edge());
    typename std::vector<Segment>::iterator ite1 = vseg_aux.begin();
    do {
      vseg.push_back(*ite1);
      ite1++;
    } while (ite1 != vseg_aux.end());
    ite++;
    MDST_vector.push_back(vseg);
    bool found_source, found_target;
    typename std::vector< std::vector<Segment> >::iterator it, ccomp1, ccomp2;
    typename std::vector<Segment>::iterator vseg_it;
    int number_edges = 0;
    while(ite != EdgeList.end()) {
      it = MDST_vector.begin();
      found_source = false;
      found_target = false;
      //for each connected component
      while (it != MDST_vector.end() && (!found_source || !found_target)) {
        vseg_it = (*it).begin();
        while ((vseg_it != (*it).end()) && (!found_source || !found_target)) {
          if (ite->edge.first->vertex(ccw(ite->edge.second))->point() ==
              vseg_it->source()) {
            ccomp1 = it;
            found_source = true;
          }
          else if (ite->edge.first->vertex(ccw(ite->edge.second))->point() ==
                   vseg_it->target()) {
            ccomp1 = it;
            found_source = true;
          }
          if (ite->edge.first->vertex(cw(ite->edge.second))->point() ==
              vseg_it->source()) {
            ccomp2 = it;
            found_target = true;
          }
          else if (ite->edge.first->vertex(cw(ite->edge.second))->point() ==
                   vseg_it->target()) {
            ccomp2 = it;
            found_target = true;
          }
          vseg_it++;
        }
        it++;
      }
      if (found_source && found_target && ccomp2 != ccomp1) {
        typename std::vector<Segment>::iterator iterator = (*ccomp2).begin();
        for (; iterator != (*ccomp2).end(); iterator++)
          (*ccomp1).push_back(*iterator);
        v_edges=segment(ite->get_edge());
        for (unsigned int k=0; k < v_edges.size(); k++)
          (*ccomp1).push_back(v_edges[k]);
        MDST_vector.erase(ccomp2);
        number_edges++;
      }
      else if (found_source && !found_target) {
        v_edges=segment(ite->get_edge());
        for (unsigned int k=0; k < v_edges.size(); k++)
          (*ccomp1).push_back(v_edges[k]);
        number_edges++;
      }
      else if (found_target && !found_source) {
        v_edges=segment(ite->get_edge());
        for (unsigned int k=0; k < v_edges.size(); k++)
          (*ccomp2).push_back(v_edges[k]);
        number_edges++;
      }
      else if (!found_source || !found_target) {
        vseg.erase(vseg.begin(),vseg.end());
        vseg = segment(ite->get_edge());
        number_edges++;
        MDST_vector.push_back(vseg);
      }
      ite++;
    }

   }


  // DUAL
  Point dual (Face_handle f) const;
  Object dual(const Edge &e) const ;
  Object dual(const Edge_circulator& ec) const;
  Object dual(const Finite_edges_iterator& ei) const;

  //INSERTION-REMOVAL
  Vertex_handle insert(const Point  &p,
                       Face_handle start = Face_handle(NULL) );
  Vertex_handle insert(const Point& p,
                       Locate_type lt,
                       Face_handle loc, int li );
  Vertex_handle push_back(const Point &p);

  void  remove(Vertex_handle v );


private:
  void restore_Delaunay(Vertex_handle v);
  void propagating_flip(Face_handle& f,int i);
  void remove_2D(Vertex_handle v );

  Vertex_handle nearest_vertex_2D(const Point& p, Face_handle f) const;
  Vertex_handle nearest_vertex_1D(const Point& p) const;

  void  look_nearest_neighbor(const Point& p,
                              Face_handle f,
                              int i,
                              Vertex_handle& nn) const;

public:
  template < class Stream>
  Stream& draw_dual(Stream & ps)
  {
    Finite_edges_iterator eit= finite_edges_begin();
    for (; eit != finite_edges_end(); ++eit) {
      Object o = dual(eit);
      Segment s;
      if (CGAL::assign(s,o)) ps << s;
    }
    return ps;
  }

  template < class InputIterator >
  int
  insert(InputIterator first, InputIterator last)
  {
    int n = number_of_vertices();
    std::vector<Point> v1,v2;
    while(first != last) {
      if (first->w() == 0)
        v2.push_back(*first);
      else
        v1.push_back(*first);
      first++;
    }
    unsigned int i=0;
    while(i < v1.size()){
      insert(v1[i]);
      ++i;
    }
    i=0;
    while(i < v2.size()){
      insert(v2[i]);
      ++i;
    }
    return number_of_vertices() - n;
  }


  template <class OutputItFaces, class OutputItBoundaryEdges>
  std::pair<OutputItFaces,OutputItBoundaryEdges>
  get_conflicts_and_boundary(const Point  &p,
                             OutputItFaces fit,
                             OutputItBoundaryEdges eit,
                             Face_handle start = Face_handle(NULL)) const {
    CGAL_triangulation_precondition( dimension() == 2);
    int li;
    Locate_type lt;
    Face_handle fh = locate(p,lt,li, start);
    switch(lt) {
    case Triangulation::OUTSIDE_AFFINE_HULL:
    case Triangulation::VERTEX:
      return std::make_pair(fit,eit);
    case Triangulation::FACE:
    case Triangulation::EDGE:
    case Triangulation::OUTSIDE_CONVEX_HULL:
      *fit++ = fh; //put fh in OutputItFaces
      std::pair<OutputItFaces,OutputItBoundaryEdges>
      pit = std::make_pair(fit,eit);
      pit = propagate_conflicts(p,fh,0,pit);
      pit = propagate_conflicts(p,fh,1,pit);
      pit = propagate_conflicts(p,fh,2,pit);
      return std::make_pair(fit,eit);
    }
    CGAL_triangulation_assertion(false);
    return std::make_pair(fit,eit);
  }

  template <class OutputItFaces>
  OutputItFaces
  get_conflicts (const Point  &p,
		 OutputItFaces fit,
		 Face_handle start= Face_handle(NULL)) const {
    std::pair<OutputItFaces,Emptyset_iterator> pp =
      get_conflicts_and_boundary(p,fit,Emptyset_iterator(),start);
    return pp.first;
  }

  template <class OutputItBoundaryEdges>
  OutputItBoundaryEdges
  get_boundary_of_conflicts(const Point  &p,
			    OutputItBoundaryEdges eit,
			    Face_handle start= Face_handle(NULL)) const {
    std::pair<Emptyset_iterator, OutputItBoundaryEdges> pp =
      get_conflicts_and_boundary(p,Emptyset_iterator(),eit,start);
    return pp.second;
  }

private:
 template <class OutputItFaces, class OutputItBoundaryEdges>
 std::pair<OutputItFaces,OutputItBoundaryEdges>
 propagate_conflicts (const Point  &p,
		      Face_handle fh,
		      int i,
		      std::pair<OutputItFaces,OutputItBoundaryEdges>
		      pit)  const {
   Face_handle fn = fh->neighbor(i);
   OutputItFaces fit = pit.first;
   OutputItBoundaryEdges eit = pit.second;
   if (! test_conflict(p,fn)) {
     *eit++ = Edge(fn, fn->index(fh));
     return std::make_pair(fit,eit);
   }
   *fit++ = fn;
   int j = fn->index(fh);
   pit = propagate_conflicts(p,fn,ccw(j),pit);
   pit = propagate_conflicts(p,fn,cw(j), pit);
   return pit;
 }

 protected:
   std::vector< std::vector <Segment> > MDST_vector;

   void initialize_EdgeList(const Finite_edges_iterator begin,
                            const Finite_edges_iterator end,
                            std::vector<MDST_edge> &Edgelist) const;
};

template < class Gt, class Tds >
inline bool
Delaunay_triangulation_T2<Gt,Tds>::
test_conflict(const Point  &p, Face_handle fh) const
{
  return (side_of_oriented_circle(fh,p) == ON_POSITIVE_SIDE);
}

template < class Gt, class Tds >
inline bool
Delaunay_triangulation_T2<Gt,Tds>::
does_conflict(const Point  &p, Face_handle fh) const
{
  return test_conflict(p,fh);
}

template < class Gt, class Tds >
inline bool
Delaunay_triangulation_T2<Gt,Tds>::
find_conflicts(const Point  &p,
	       std::list<Face_handle>& conflicts,
	       Face_handle start ) const
{
  get_conflicts(p, std::back_inserter(conflicts), start);
  return (! conflicts.empty());
}

template < class Gt, class Tds >
bool
Delaunay_triangulation_T2<Gt,Tds>::
is_valid(bool verbose, int level) const
{
  bool result = Triangulation_T2<Gt,Tds>::is_valid(verbose, level);

  for( Finite_faces_iterator it = finite_faces_begin();
       it != finite_faces_end() ; it++) {
    for(int i=0; i<3; i++) {
      if ( ! is_infinite( it->mirror_vertex(i))) {
	result = result &&  ON_POSITIVE_SIDE !=
	  side_of_oriented_circle( it, it->mirror_vertex(i)->point());
      }
      CGAL_triangulation_assertion( result );
    }
  }
  return result;
}

template < class Gt, class Tds >
typename Delaunay_triangulation_T2<Gt,Tds>::Vertex_handle
Delaunay_triangulation_T2<Gt,Tds>::
nearest_vertex(const Point  &p, Face_handle f) const
{
  switch (dimension()) {
  case 0:
    if (number_of_vertices() == 0) return NULL;
    if (number_of_vertices() == 1) return finite_vertex();
    //break;
  case 1:
    return nearest_vertex_1D(p);
    //break;
  case 2:
    return nearest_vertex_2D(p,f);
    //break;
  }
  return NULL;
}

template < class Gt, class Tds >
typename Delaunay_triangulation_T2<Gt,Tds>::Vertex_handle
Delaunay_triangulation_T2<Gt,Tds>::
nearest_vertex_2D(const Point& p, Face_handle f) const
{
  CGAL_triangulation_precondition(dimension() == 2);
  if (f == Face_handle(NULL)) f = locate(p);
  else
    CGAL_triangulation_precondition(oriented_side(f,p)!=ON_NEGATIVE_SIDE);

  typename Geom_traits::Compare_distance_2
    compare_distance =  geom_traits().compare_distance_2_object();
  Vertex_handle nn =  !is_infinite(f->vertex(0)) ? f->vertex(0):f->vertex(1);
  if ( !is_infinite(f->vertex(1)) && compare_distance(p,
					    f->vertex(1)->point(),
					    nn->point()) == SMALLER)
    nn=f->vertex(1);
  if ( !is_infinite(f->vertex(2)) && compare_distance(p,
					    f->vertex(2)->point(),
					    nn->point()) == SMALLER)
    nn=f->vertex(2);

  look_nearest_neighbor(p,f,0,nn);
  look_nearest_neighbor(p,f,1,nn);
  look_nearest_neighbor(p,f,2,nn);
  return nn;
}

template < class Gt, class Tds >
typename Delaunay_triangulation_T2<Gt,Tds>::Vertex_handle
Delaunay_triangulation_T2<Gt,Tds>::
nearest_vertex_1D(const Point& p) const
{
  typename Geom_traits::Compare_distance_2
    compare_distance =  geom_traits().compare_distance_2_object();
  Vertex_handle nn;

  Finite_vertices_iterator vit=finite_vertices_begin();
  nn = vit;
  for ( ; vit != finite_vertices_end(); ++vit){
    if (compare_distance(p, vit->point(), nn->point()) == SMALLER)
      nn = vit;
  }
  return nn;
}

template < class Gt, class Tds >
void
Delaunay_triangulation_T2<Gt,Tds>::
look_nearest_neighbor(const Point& p,
                      Face_handle f,
		      int i,
		      Vertex_handle& nn) const
{
  Face_handle  ni=f->neighbor(i);
  if ( ON_POSITIVE_SIDE != side_of_oriented_circle(ni,p) ) return;

  typename Geom_traits::Compare_distance_2
    compare_distance =  geom_traits().compare_distance_2_object();
  i = ni->index(f);
  if ( !is_infinite(ni->vertex(i)) &&
       compare_distance(p,
	      ni->vertex(i)->point(),
	      nn->point())  == SMALLER)  nn=ni->vertex(i);

  // recursive exploration of triangles whose circumcircle contains p
  look_nearest_neighbor(p, ni, ccw(i), nn);
  look_nearest_neighbor(p, ni, cw(i), nn);
}

//DUALITY




template<class Gt, class Tds>
inline
typename Delaunay_triangulation_T2<Gt,Tds>::Point
Delaunay_triangulation_T2<Gt,Tds>::
dual (Face_handle f) const
{
  CGAL_triangulation_precondition (dimension()==2);
  return circumcenter(f);
}


template < class Gt, class Tds >
Object
Delaunay_triangulation_T2<Gt,Tds>::
dual(const Edge &e) const
{
  CGAL_triangulation_precondition (!is_infinite(e));
    if( (!is_infinite(e.first)) &&
        (!is_infinite(e.first->neighbor(e.second))) ) {
      Segment s = geom_traits().construct_segment
                  (dual(e.first),dual(e.first->neighbor(e.second)));
      return make_object(s);
    }
    Face_handle f; int i;
    if (is_infinite(e.first)) {
      f=e.first->neighbor(e.second); f->has_neighbor(e.first,i);
    }
    else {
      f=e.first; i=e.second;
    }
}

template < class Gt, class Tds >
inline Object
Delaunay_triangulation_T2<Gt,Tds>::
dual(const Edge_circulator& ec) const
{
  return dual(*ec);
}

template < class Gt, class Tds >
inline Object
Delaunay_triangulation_T2<Gt,Tds>::
dual(const Finite_edges_iterator& ei) const
{
  return dual(*ei);
}

template < class Gt, class Tds >
inline
typename Delaunay_triangulation_T2<Gt,Tds>::Vertex_handle
Delaunay_triangulation_T2<Gt,Tds>::
insert(const Point  &p,  Face_handle start)
{
  Locate_type lt;
  int li;
  Face_handle loc = locate (p, lt, li, start);
  return insert(p, lt, loc, li);
}

template < class Gt, class Tds >
inline
typename Delaunay_triangulation_T2<Gt,Tds>::Vertex_handle
Delaunay_triangulation_T2<Gt,Tds>::
push_back(const Point &p)
{
  return insert(p);
}

template < class Gt, class Tds >
inline
typename Delaunay_triangulation_T2<Gt,Tds>::Vertex_handle
Delaunay_triangulation_T2<Gt,Tds>::
insert(const Point  &p, Locate_type lt, Face_handle loc, int li)
{
  Vertex_handle v = Triangulation_T2<Gt,Tds>::insert(p,lt,loc,li);
  restore_Delaunay(v);
  return(v);
}


template < class Gt, class Tds >
void
Delaunay_triangulation_T2<Gt,Tds>::
restore_Delaunay(Vertex_handle v)
{
  if(dimension() <= 1) return;
  if ((void*)(&(*(v->face()))) != NULL) {
  Face_handle f=v->face();
  Face_handle next;
  
  int i;
  Face_handle start(f);
  do {
    i = f->index(v);
    
    next = f->neighbor(ccw(i));
    
    propagating_flip(f,i);
    
    f=next;
    
  } while(next != start);

  } 
  return;
}


template < class Gt, class Tds >
bool
Delaunay_triangulation_T2<Gt,Tds>::
is_delaunay(Face_handle& f, Face_handle& n, int index)
{
  int i = f->index(n);
  int in = n->index(f);

  if (is_infinite(f))
    return true;
  Point a = f->vertex(i)->point();
  Point b = f->vertex(cw(i))->point();
  Point d = f->vertex(ccw(i))->point();
  Point c = n->vertex(in)->point();
  

  if (Triangulation().antipodal(a,b))
    return true;
  if (Triangulation().antipodal(a,c))
    return true;
  if (Triangulation().antipodal(a,d))
    return true;
  if (Triangulation().antipodal(b,c))
    return true;
  if (Triangulation().antipodal(b,d))
    return true;
  if (Triangulation().antipodal(c,d))
    return true;
  if (orientation(a,b,c)==COLLINEAR || orientation(a,d,c)==COLLINEAR) {
    return true;
  }
  if (orientation(b,a,d)==COLLINEAR || orientation(b,c,d)==COLLINEAR) {
    return false;
  }
  if (orientation(a,b,c)==COUNTERCLOCKWISE || orientation(c,d,a)==COUNTERCLOCKWISE) {
    return true;
  }
  if (orientation(d,a,b)==COUNTERCLOCKWISE || orientation(b,c,d)==COUNTERCLOCKWISE) {
    return false;
  }
  
  
  if (sign(a.w()) == sign(b.w()) && sign(b.w()) == sign(c.w())  && sign(c.w()) == sign(d.w()) 
      && a.w()!=0) {
    Point c1 = circumcenter(a,b,d);
    if (sign(c1.w()) != sign(a.w()))
      c1 = Point(-c1.w(),-c1.x(),-c1.y());
    if (geom_traits().compare_T1(squared_distance(c1,a),squared_distance(c1,c)) == LARGER)
      return false;
    Point c2 = circumcenter(c,d,b);
    if (sign(c2.w()) != sign(c.w()))
      c2 = Point(-c2.w(),-c2.x(),-c2.y());
    if (geom_traits().compare_T1(squared_distance(c2,c),squared_distance(c2,a)) == LARGER)
      return false;
     return true;
  }    
  //if the four points are proper
  else if (a.w()*b.w()*c.w()*d.w() != 0) {
      Point c1;
      if (a.w()*b.w() > 0 && sign(b.w())==sign(d.w()) && sign(b.w())!=sign(c.w())) {
        c1 = circumcenter(b,c,d);
        return true;
      }
      if (a.w()*c.w() > 0 && sign(c.w())==sign(d.w()) && sign(d.w())!=sign(b.w())) {
        c1 = circumcenter(a,b,c);
        return false; 
      }
      if (b.w()*c.w() > 0 && sign(b.w())==sign(d.w()) && sign(d.w())!=sign(a.w())) {
        c1 = circumcenter(a,b,d);
        return true;
      }
      if (a.w()*b.w() > 0 && sign(b.w())==sign(c.w()) && sign(c.w())!=sign(d.w())) {
        c1 = circumcenter(a,c,d);
        return false; 
      } }
     return (geom_traits().compare_distance(b,d,a,c)!=LARGER); 

}


template <class Gt, class Tds >
void
Delaunay_triangulation_T2<Gt, Tds>::
initialize_EdgeList(const Finite_edges_iterator begin, const Finite_edges_iterator end,
                        std::vector<MDST_edge> &Edgelist) const
{
  Finite_edges_iterator eit(begin);
  while (eit != end) {
    Edgelist.push_back(MDST_edge(*eit));
    eit++;
  }
}


template < class Gt, class Tds >
void
Delaunay_triangulation_T2<Gt,Tds>::
propagating_flip(Face_handle& f,int i)
{
  Face_handle n = f->neighbor(i);


  if (is_delaunay(n, f, i))
    return;
  if (!(orientation(f->vertex(i)->point(),f->vertex(cw(i))->point(),
                    f->mirror_vertex(i)->point()) == RIGHT_TURN &&
        orientation(f->vertex(i)->point(),f->vertex(ccw(i))->point(),
                    f->mirror_vertex(i)->point()) == LEFT_TURN))
  return;
  flip(f, i);
  propagating_flip(f,i);
  i = n->index(f->vertex(i));
  propagating_flip(n,i);
}


CGAL_END_NAMESPACE

#endif // CGAL_DELAUNAY_TRIANGULATION_T2_H
