// Copyright (c) 1997  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Delaunay_triangulation_2.h,v $
// $Revision: 1.66 $ $Date: 2003/09/18 10:26:05 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Mariette Yvinec
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_DELAUNAY_TRIANGULATION_TRAITS_T2_H
#define CGAL_DELAUNAY_TRIANGULATION_TRAITS_T2_H

#include <CGAL/direction_constructions_T2.h>

CGAL_BEGIN_NAMESPACE

template <class kernel>
class Delaunay_triangulation_traits_T2 :
         public Triangulation_traits_T2<kernel>{
public:
  typedef kernel K;
  typedef typename kernel::FT                                FT;
  typedef typename kernel::Direction_T2                      Direction_T2;
  typedef typename kernel::Circle_T2                         Circle_T2;
  typedef typename kernel::Point_T1                          Point_T1;
  typedef typename kernel::Point_T2                          Point_T2;
  typedef typename kernel::Line_T2                           Line_T2;
  typedef typename kernel::Compare_angle_with_x_axis_T2      Compare_angle_T2;
  typedef typename kernel::Compare_T1p                       Compare_T1p;

  Comparison_result compare_T1(const Point_T1 &a, const Point_T1 &b) const
  {
    return K().compare_T1p_object()(a,b);
  }

  //finds the minimum_angle of the triangle abd
  Point_T2 find_minimum_angle(const Point_T2 &a, const Point_T2 &b, const Point_T2 &d) const
  {
    Direction_T2 dir_ab(a,b);
    Point_T2 d_ab(FT(0),dir_ab.dx(),dir_ab.dy());
    Direction_T2 dir_ad(a,d);
    Point_T2 d_ad(FT(0),dir_ad.dx(),dir_ad.dy());
    Point_T2 angle_a,angle_b,angle_d;
    bool got_angle_a = false;
    bool got_angle_b = false;
    bool got_angle_d = false;
    if (a.w() == 0) { //a is infinite
      if (b.w()*d.w() < 0)
        angle_a = Point_T2(FT(0),FT(-1),FT(0));
      else if (b.w()*d.w() > 0) {
        angle_a = Point_T2(FT(0),FT(-1),FT(0));
      }
      else {
        angle_a = Point_T2(FT(0),FT(0),FT(1));
        if (d.w() == 0) {
          angle_d = Point_T2(FT(0),FT(0),FT(1));
          got_angle_d = true;
        }
        else if (b.w() == 0) {
          angle_b = Point_T2(FT(0),FT(0),FT(1));
          got_angle_b = true;
        }
      }
    }
    else
      angle_a = Point_T2(FT(0),
                         d_ab.x()*d_ad.x() + d_ab.y()*d_ad.y(),
                         abs(d_ab.y()*d_ad.x() - d_ab.x()*d_ad.y()));
    got_angle_a = true;
    Direction_T2 dir_bd(b,d);
    Point_T2 d_bd(FT(0),dir_bd.dx(),dir_bd.dy());
    if (!got_angle_b) {
      if (b.w() == 0) { //b is infinite
        if (a.w()*d.w() < 0)
          angle_b = Point_T2(FT(0),FT(-1),FT(0));
        else if (a.w()*d.w() > 0)
          angle_b = Point_T2(FT(0),FT(-1),FT(0));
        else {
          angle_b = Point_T2(FT(0),FT(0),FT(1));
          if (d.w() == 0 && !got_angle_d) {
            angle_d = Point_T2(FT(0),FT(0),FT(1));
            got_angle_d = true;
          }
          else if (a.w() == 0 && !got_angle_a) {
            angle_a = Point_T2(FT(0),FT(0),FT(1));
            got_angle_a = true;
          }
        }
      }
      else
        angle_b = Point_T2(FT(0),
                           d_bd.x()*-d_ab.x() + d_bd.y()*-d_ab.y(),
                           abs(d_bd.y()*-d_ab.x() - d_bd.x()*-d_ab.y()));
      
    }
    got_angle_b = true;
    if (!got_angle_d) {
      if (d.w() == 0) { //a is infinite
        if (a.w()*b.w() < 0)
          angle_d = Point_T2(FT(0),FT(-1),FT(0));
        else if (a.w()*b.w() > 0) {
          angle_d = Point_T2(FT(0),FT(-1),FT(0));
        }
        else {
          angle_d = Point_T2(FT(0),FT(0),FT(1));
          if (a.w() == 0 && !got_angle_a) {
            angle_a = Point_T2(FT(0),FT(0),FT(1));
            got_angle_a = true;
          }
          else if (b.w() == 0 && got_angle_b) {
            angle_b = Point_T2(FT(0),FT(0),FT(1));
            got_angle_b = true;
          }
        }
      }
      else
        angle_d = Point_T2(FT(0),
                           -d_bd.x()*-d_ad.x() + -d_bd.y()*-d_ad.y(),
                           abs(-d_bd.y()*-d_ad.x() - (-d_bd.x())*-d_ad.y()));
    }
    Compare_angle_T2 cmp_angle = K().compare_angle_with_x_axis_T2_object();
    if (cmp_angle(angle_a,angle_b) == SMALLER) {
      if (cmp_angle(angle_a,angle_d) == SMALLER) {
        return angle_a;
      }
      return angle_d;
    }
    if (cmp_angle(angle_b,angle_d) == SMALLER) {
        return angle_b;
    }
    return angle_d;
  }

  Comparison_result compare_radius(const Circle_T2 &c1, const Circle_T2 &c2) const
  {
     Point_T1 sr1 = c1.squared_radius();
     Point_T1 sr2 = c2.squared_radius();
     if (sr1.w() == 0 && sr2.w() == 0) {
      Point_T2 center1 = c1.center();
      FT norm_center1 = sqrt(center1.w()*center1.w() +
                             center1.x()*center1.x() +
                             center1.y()*center1.y());
      Point_T2 p1 = c1.point();
      FT norm_p1 = sqrt(p1.w()*p1.w()+p1.x()*p1.x()+p1.y()*p1.y());
      Point_T2 center2 = c2.point();
      Point_T2 p2 = c2.point();
      FT norm_center2 = sqrt(center2.w()*center2.w() +
                             center2.x()*center2.x() +
                             center2.y()*center2.y());


      FT norm_p2 = sqrt(p2.w()*p2.w()+p2.x()*p2.x()+p2.y()*p2.y());
      FT length_sr1 = norm_p2*norm_center2*(center1.x()*p1.y()+center1.y()*p1.w()+
                                        center1.w()*p1.x()-center1.x()*p1.w()-
                                        center1.y()*p1.x()-center1.w()*p1.y());
      FT length_sr2 = norm_p1*norm_center1*(center2.x()*p2.y()+center2.y()*p2.w()+
                                        center2.w()*p2.x()-center2.x()*p2.w()-
                                        center2.y()*p2.x()-center2.w()*p2.y());
      if (abs(length_sr1) > abs(length_sr2)) return LARGER;
      if (abs(length_sr1) < abs(length_sr2)) return SMALLER;
      return EQUAL;
    }
    return compare_T1p(sr1,sr2);
  }




  bool exist_circle(const Point_T2 &p,
                    const Point_T2 &q, 
                    const Point_T2 &r,
                    Point_T2 & c) const
  {
    if (sign(p.w()) == sign(q.w()) && sign(q.w()) == sign(r.w()) && r.w() != 0) {
      c=circumcenter(p,q,r);
      return true;
    }
    else
      return false;
    return false;
  }

  bool exist_circle(const Point_T2 &p,
                    const Point_T2 &q,
                    Point_T2 & c) const
  {
    if (sign(p.w()) == sign(q.w()) && p.w() != 0) {
      c=midpoint(p,q);
      return true;
    }
    return false;   
  }

  

 private:
  
 bool has_equal_relative_distance(const Point_T2 &p,const Point_T2 &q,const Point_T2 &r) const
 {
   CGAL_kernel_precondition(is_at_infinity(r) &&
                           !is_at_infinity(p) &&
                           !is_at_infinity(q));
   Line_T2 *l0 = new Line_T2(p,r);
   Line_T2 *l = new Line_T2(p,perpendicular_direction(*l0));

   return (l->has_on(q));
 }	

};

CGAL_END_NAMESPACE

#endif // CGAL_DELAUNAY_TRIANGULATION_TRAITS_T2_H
