// Copyright (c) 1997  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Generator/include/CGAL/point_generators_2.h,v $
// $Revision: 1.12 $ $Date: 2003/10/21 12:15:40 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Lutz Kettner  <kettner@inf.ethz.ch>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#include <CGAL/MP_Float.h>
#include <CGAL/Signed_homogeneous.h>


typedef CGAL::Signed_homogeneous<CGAL::MP_Float> K;
typedef CGAL::Point_T2<K> Point;

Point front_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;

  do {
    w = ((rand()%range));
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0) || (w == 0));
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  int w1;
  do {
    w1  = ((rand()%range));
  }while (w1 == 0);
  w = w1/w;
  p = new Point(w,x,y);
  return *p;
}


Point back_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;

  do {
    w = -((rand()%range));
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0) || (w == 0));
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  int w1;
  do {
    w1  = ((rand()%range));
  }while (w1 == 0);
  w = w1/w;
  p = new Point(w,x,y);
  return *p;
}


Point front_back_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;

  do {
    w = ((rand()%range));
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0) || (w == 0));
  int signal_w = rand()%2;
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_w)
    w = -w;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  int w1;
  do {
    w1  = ((rand()%range));
  }while (w1 == 0);
  w = w1/w;
  p = new Point(w,x,y);
  return *p;
}


Point front_infinite_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;

  do {
    w = ((rand()%range));
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0) && (w == 0));
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  if (w != 0 && (x!=0 || y!=0))
    w = ((rand()%range))/w;
  p = new Point(w,x,y);
  return *p;
}


Point infinite_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;
  w = 0;
  do {
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0));
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  p = new Point(w,x,y);
  return *p;
}


Point back_infinite_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;

  do {
    w = -((rand()%range));
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0) && (w == 0));
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  if (w != 0 && (x!=0 || y!=0))
    w = ((rand()%range))/w;
  p = new Point(w,x,y);
  return *p;
}


Point all_point(int range) {
  Point *p;
  CGAL::MP_Float w,x,y;
  do {
    w = ((rand()%range));
    y = ((rand()%range));
    x = ((rand()%range));
  } while ((x == 0) && (y == 0) && (w == 0));
  int signal_w = rand()%2;
  int signal_x = rand()%2;
  int signal_y = rand()%2;
  if (signal_w)
    w = -w;
  if (signal_x)
    x = -x;
  if (signal_y)
    y = -y;
  if (w != 0 && (x!=0 || y!=0))
    w = ((rand()%range))/w;
  p = new Point(w,x,y);
  return *p;
}
