// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Iso_rectangle_2.h,v $
// $Revision: 1.16 $ $Date: 2003/10/21 12:18:06 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_ISO_RECTANGLE_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_ISO_RECTANGLE_T2_H

#include<CGAL/polygon_operations_T2.h>

CGAL_BEGIN_NAMESPACE

template <class R_>
class Iso_rectangle_T2
  : public Handle_for<Triple< typename R_::Point_T2,
															typename R_::Point_T1, 
															typename R_::Point_T1> >
{
CGAL_VC7_BUG_PROTECTED
  typedef typename R_::FT                           FT;
  typedef typename R_::Point_T2                     TPoint_T2;
  typedef typename R_::Point_T1                     TPoint_T1;
  typedef typename R_::Iso_rectangle_T2             TIso_rectangle_T2;
  typedef typename R_::Aff_transformation_T2        TAff_transformation_T2;
  typedef typename R_::Iso_rectangle_handle_T2      base;
  typedef typename base::element_type               rep;

private:

  void initialize(const TPoint_T2 &p, TPoint_T1 & x, TPoint_T1 & y)
  {
    if (is_at_infinity(p)) 
      CGAL_kernel_precondition((p==TPoint_T2(0,1,0)) || 
                               (p==TPoint_T2(0,0,1)) ||
                               (p==TPoint_T2(0,-1,0)) || 
                               (p==TPoint_T2(0,0,-1)));
    initialize_with(rep(p,x,y));
  }

public:
  typedef R_                                        R;

  Iso_rectangle_T2()
    : base(rep()) {}

  Iso_rectangle_T2(const TPoint_T2 &p, TPoint_T1 & x, TPoint_T1 & y)
  {
    initialize(p,x,y);    
  }

  Iso_rectangle_T2(const FT& w1, const FT& x1, const FT& y1, 
                   const FT& w2, const FT& x2, 
                   const FT& w3, const FT& y3) 
  {
    TPoint_T2 p(w1, x1, y1);
    TPoint_T1 x(w2, x2);
    TPoint_T1 y(w3, y3);
    initialize(p,delta_x,delta_y,orient);    
  }

  bool            operator==(const Iso_rectangle_T2 &s) const;
  bool            operator!=(const Iso_rectangle_T2 &s) const;

  const TPoint_T2 & p1() const
  {
    return Ptr()->first;
  }
  
  const TPoint_T2 &p3() const
  {
    TPoint_T2 *p = new TPoint_T2(vertex(2));
    return *p;
  }

  const TPoint_T1 & x() const
  {
    return Ptr()->second;
  }

  const TPoint_T1  & y() const
  {
    return Ptr()->third; 
  }
 
  const Orientation & orientation() const
  {
    Orientation *o = new Orientation(CGAL::orientation(vertex0(),vertex1(),vertex2()));
    return *o;
  } 

  TPoint_T2 vertex(int i) const;
  TPoint_T2 operator[](int i) const;

  TPoint_T2 vertex0() const;
  TPoint_T2 vertex1() const;
  TPoint_T2 vertex2() const;  
  TPoint_T2 vertex3() const;


  Oriented_side oriented_side(const TPoint_T2 &p) const;
  bool has_on_oriented_boundary(const TPoint_T2 &p) const;
  bool has_on_positive_side(const TPoint_T2 &p) const;
  bool has_on_negative_side(const TPoint_T2 &p) const;
  Bounded_side bounded_side(const TPoint_T2 &p) const;
  bool has_on_boundary(const TPoint_T2 &p) const;
  bool has_on_bounded_side(const TPoint_T2 &p) const;
  bool has_on_unbounded_side(const TPoint_T2 &p) const;

  bool is_degenerate() const;



  const FT & p1w() const;
  const FT & p1x() const;
  const FT & p1y() const;
  const FT & p3w() const;
  const FT & p3x() const;
  const FT & p3y() const;
  const FT & p1_coord(int i) const;
  const FT & p3_coord(int i) const;

  TPoint_T1 area() const;
};

#ifdef CGAL_CFG_TYPENAME_BUG
#define typename
#endif

template < class R >
inline
bool
Iso_rectangle_T2<R>::
operator==(const Iso_rectangle_T2<R> &r) const
{
  if (identical(r))
    return true;
  return p1() == r.p1() && x() == r.x() &&
         y() == r.y();
}

template < class R >
inline
bool
Iso_rectangle_T2<R>::
operator!=(const Iso_rectangle_T2<R> &r) const
{
  return !(*this == r);
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p1w() const
{
  return p1().w();
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p1x() const
{
  return p1().x();
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p1y() const
{
  return p1().y();
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p3w() const
{
  return p3().w();
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p3x() const
{
  return p3().x();
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p3y() const
{
  return p3().y();
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p1_coord(int i) const
{
  CGAL_kernel_precondition( i == 0 || i == 1 || i == 2 );
  switch(i) {
    case 0 : return p1w();
    case 1 : return p1x();
    case 2 : return p1y();
  }
}

template < class R >
inline
const typename Iso_rectangle_T2<R>::FT &
Iso_rectangle_T2<R>::p3_coord(int i) const
{
  CGAL_kernel_precondition( i == 0 || i == 1 || i == 2);
  switch(i) {
    case 0 : return p3w();
    case 1 : return p3x();
    case 2 : return p3y();
  }
}


template < class R >
typename Iso_rectangle_T2<R>::TPoint_T2
Iso_rectangle_T2<R>::vertex0() const
{
  return p1(); 
}

template < class R >
typename Iso_rectangle_T2<R>::TPoint_T2
Iso_rectangle_T2<R>::vertex1() const
{
      return TPoint_T2(p1().w()*x().w(),
                       p1().x()*x().w() +
                       p1().w()*x().x(), 
                       p1().y()*x().w());

}   



template < class R >
typename Iso_rectangle_T2<R>::TPoint_T2
Iso_rectangle_T2<R>::vertex2() const
{
  TPoint_T2 *p;
  if (is_at_infinity_T1(y()) && is_at_infinity_T1(x())) {

    p = new TPoint_T2(0,p1().x()*x().x(),p1().y()*y().x());
  }
  else {
    p = new TPoint_T2(p1().w()*x().w()*y().w(),
                   y().w()*
                   (p1().x()*x().w() +
                   p1().w()*x().x()), 
                   x().w()* 
                   (p1().y()*y().w() + 
                   p1().w()*y().x()));
  }
  return *p;
}

template < class R >
typename Iso_rectangle_T2<R>::TPoint_T2
Iso_rectangle_T2<R>::vertex3() const
{
  return TPoint_T2(p1().w()*y().w(),
                   p1().x()*y().w(), 
                   p1().y()*y().w() + 
                   p1().w()*y().x());
 }


template < class R >
typename Iso_rectangle_T2<R>::TPoint_T2
Iso_rectangle_T2<R>::vertex(int i) const
{
  if ( is_at_infinity(p1()) ) 
    return p1(); 
  switch(i%4) { 
    case 0: return vertex0();
    case 1: return vertex1();
    case 2: return vertex2();
    case 3: return vertex3();
    default: { // should not happen. 
      CGAL_kernel_assertion_msg(false, "Internal CGAL error.");
      return TPoint_T2(FT(1),FT(0),FT(0));
    }
  }
}

template < class R >
inline
typename Iso_rectangle_T2<R>::TPoint_T2
Iso_rectangle_T2<R>::operator[](int i) const
{
  return vertex(i);
}

template < class R >
inline
typename Iso_rectangle_T2<R>::TPoint_T1
Iso_rectangle_T2<R>::area() const
{
  Vector_T2<R> v0(vertex(0),vertex(1));
  Vector_T2<R> v1(vertex(0),vertex(3));
  return TPoint_T1(v0.w()*v1.w(), v0.x()*v1.y()-v0.y()*v1.x());
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
Oriented_side 
Iso_rectangle_T2<R>::
oriented_side(const typename Iso_rectangle_T2<R>::TPoint_T2 &p) const
{
  if (is_degenerate()) {
    if (has_on_segment(p,p1(),vertex(2)))
      return ON_ORIENTED_BOUNDARY;
    if (orientation() == COUNTERCLOCKWISE)
      return ON_NEGATIVE_SIDE;
    return ON_POSITIVE_SIDE;
  }
  std::vector<TPoint_T2> simplex;
  for(int i=0; i < 4; i++)
    simplex.push_back(vertex(i));
  return oriented_side_polygon_T2(simplex.begin(),simplex.end(),p);
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
bool
Iso_rectangle_T2<R>::
has_on_positive_side(const typename Iso_rectangle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_POSITIVE_SIDE;
}           

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
bool
Iso_rectangle_T2<R>::
has_on_negative_side(const typename Iso_rectangle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_NEGATIVE_SIDE;
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
bool
Iso_rectangle_T2<R>::
has_on_oriented_boundary
(const typename Iso_rectangle_T2<R>::TPoint_T2 &p) const
{
  return oriented_side(p) == ON_ORIENTED_BOUNDARY;
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
bool
Iso_rectangle_T2<R>::
has_on_boundary(const typename Iso_rectangle_T2<R>::TPoint_T2 &p) const
{
  return has_on_oriented_boundary(p);
}

template < class R >
CGAL_KERNEL_MEDIUM_INLINE
Bounded_side
Iso_rectangle_T2<R>::
bounded_side(const typename Iso_rectangle_T2<R>::TPoint_T2 &p) const
{ // FIXME : predicate
  if (has_on_oriented_boundary(p))
    return ON_BOUNDARY;
  if (orientation() == COUNTERCLOCKWISE)
  {
    if (has_on_positive_side(p))
      return ON_BOUNDED_SIDE;
    else
      return ON_UNBOUNDED_SIDE;
  }
  else {
    if (has_on_negative_side(p))
      return ON_BOUNDED_SIDE;
    else
      return ON_UNBOUNDED_SIDE;
  }
}

template < class R >
inline
bool
Iso_rectangle_T2<R>::
has_on_bounded_side(const typename Iso_rectangle_T2<R>::TPoint_T2 &p)
    const
{
  return bounded_side(p) == ON_BOUNDED_SIDE;
}

template < class R >
inline
bool
Iso_rectangle_T2<R>::
has_on_unbounded_side(const typename Iso_rectangle_T2<R>::TPoint_T2 &p)
    const
{
  return bounded_side(p) == ON_UNBOUNDED_SIDE;
}

template < class R >
inline
bool
Iso_rectangle_T2<R>::is_degenerate() const
{
  return collinear(p1(),p3(),TPoint_T2(0,1,0)) ||
         collinear(p1(),p3(),TPoint_T2(0,0,1));
}

//#ifndef CGAL_NO_OSTREAM_INSERT_ISO_RECTANGLEC2
template < class R >
std::ostream&
operator<<(std::ostream &os, const Iso_rectangle_T2<R> &r)
{
  switch(os.iword(IO::mode)) {
    case IO::ASCII :
      return os << r[0] << ' ' << r.x() << ' ' << r.y();
    case IO::BINARY : 
      return os << r[0] << ' ' << r.x() << ' ' << r.y();
    default: 
      return os << "Iso_rectangle_T2(" << r[0] << ", " << r.x() << ", " 
                                       << r.y() << ")";

  }
}
//#endif // CGAL_NO_OSTREAM_INSERT_ISO_RECTANGLEC2

//#ifndef CGAL_NO_ISTREAM_EXTRACT_ISO_RECTANGLEC2
template < class R >
CGAL_KERNEL_MEDIUM_INLINE
std::istream &
operator>>(std::istream &is, Iso_rectangle_T2<R> &r)
{
  typename R::Point_T2 p;
  typename R::Point_T1 x, y;

  is >> p >> x >> y;
 
  if (is)
  	r = Iso_rectangle_T2<R>(p, x, y);
  return is;
}
//#endif // CGAL_NO_ISTREAM_EXTRACT_ISO_RECTANGLEC2


CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_ISO_RECTANGLE_T2_H
