// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Iso_rectangle_2_Iso_rectangle_2_intersection.h,v $
// $Revision: 1.6 $ $Date: 2003/10/21 12:16:44 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_ISO_RECTANGLE_T2_ISO_RECTANGLE_T2_INTERSECTION_H
#define CGAL_ISO_RECTANGLE_T2_ISO_RECTANGLE_T2_INTERSECTION_H

#include <CGAL/Segment_T2.h>
#include <CGAL/Iso_rectangle_T2.h>
#include <CGAL/Segment_T2_Segment_T2_intersection.h>
#include <CGAL/Triangle_T2_Triangle_T2_intersection.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Iso_rectangle_T2_Iso_rectangle_T2_pair {
public:
    enum Intersection_results {NO, POINT, SEGMENT, ISO_RECTANGLE};
    Iso_rectangle_T2_Iso_rectangle_T2_pair() ;
    Iso_rectangle_T2_Iso_rectangle_T2_pair(Iso_rectangle_T2<R> const *irect1,
                               Iso_rectangle_T2<R> const *irect2);
    ~Iso_rectangle_T2_Iso_rectangle_T2_pair() {}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
    Intersection_results intersection_type() const;
#else
    Intersection_results intersection_type() const
  {
    if (_known)
    return _result;
    // The non const this pointer is used to cast away const.
    _known = true;

    _init_list(_pointlist, *_irect1);

    if (_irect2->is_degenerate()) {

      CGAL_kernel_assertion(false);
    }else {
      Point_T2<R> v0 =_irect2->vertex(0);
      Point_T2<R> v1 =_irect2->vertex(1);
      Point_T2<R> v2 =_irect2->vertex(2);
      Point_T2<R> v3 =_irect2->vertex(3);
      if  (!_irect2->has_on_unbounded_side(_irect1->vertex(0)) && 
           !_irect2->has_on_unbounded_side(_irect1->vertex(1)) &&
           !_irect2->has_on_unbounded_side(_irect1->vertex(2)) &&
           !_irect2->has_on_unbounded_side(_irect1->vertex(2))){
        _result = ISO_RECTANGLE;
        return _result;
      }
      if  (!_irect1->has_on_unbounded_side(v0) && 
           !_irect1->has_on_unbounded_side(v1) &&
           !_irect1->has_on_unbounded_side(v2) &&
           !_irect1->has_on_unbounded_side(v3)){
        _pointlist._finalize_list();
        _init_list(_pointlist,*_irect2);
        _result = ISO_RECTANGLE;
        return _result;
      }
      if (_irect2->orientation() == COUNTERCLOCKWISE) {
        _cut_off(_pointlist, v0, v1);

        _cut_off(_pointlist, v1, v2);

        _cut_off(_pointlist, v2, v3);

        _cut_off(_pointlist, v3, v0);

          
       } else {
        _cut_off(_pointlist, v1, v0);
        _cut_off(_pointlist, v0, v3);
        _cut_off(_pointlist, v3, v2);
        _cut_off(_pointlist, v2, v1);
       }
     } 
     switch (_pointlist.size) {
       case 0:
         _result = NO;
         break;
       case 1:
         _result = POINT;
         break;
       case 2:
         _result = SEGMENT;
         break;
       default:
         _result = ISO_RECTANGLE;
     }
     return _result;
  }   

#endif // CGAL_CFG_RETURN_TYPE_BUG_2
  bool                 intersection(Point_T2<R> &result) const;
  bool                 intersection(Segment_T2<R> &result) const;
  bool                 intersection(Iso_rectangle_T2<R> &result) const;

protected:
  Iso_rectangle_T2<R> const*   _irect1;
  Iso_rectangle_T2<R> const *  _irect2;
  mutable bool                       _known;
  mutable int                  _iso;
  mutable Intersection_results       _result;
  mutable Point_T2<R>            _intersection_point, _other_point;
  mutable Pointlist_T2_<R>    _pointlist;
};




template <class R>
bool
do_intersect(const Iso_rectangle_T2<R> &irect1, 
             const Iso_rectangle_T2<R> &irect2)
{
  typedef Iso_rectangle_T2_Iso_rectangle_T2_pair<R> pair_t;
  pair_t pair(&irect1, &irect2);
  return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE

CGAL_BEGIN_NAMESPACE



template <class R>
void _init_list(Pointlist_T2_<R> &list,
                const Iso_rectangle_T2<R> &irect)
{
  // check on degeneracies of trian.
 // if (!trian.is_degenerate()) {
    list.size = 4;
    list.first = 0;
    for (int i=0; i<4; i++) {
      Pointlist_T2_rec_<R> *newrec =
                new Pointlist_T2_rec_<R>;
      newrec->next = list.first;
      list.first = newrec;
      newrec->point = irect.vertex(i);
    }

}

template <class R>
Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::
Iso_rectangle_T2_Iso_rectangle_T2_pair()
{
  _irect1 = 0;
  _irect2 = 0;
  _known = false;
}

template <class R>
Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::
Iso_rectangle_T2_Iso_rectangle_T2_pair(
    Iso_rectangle_T2<R> const *irect1, Iso_rectangle_T2<R> const *irect2)
{
  _irect1 = irect1;
  _irect2 = irect2;
  _known = false;
}


#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::Intersection_results
Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
  _known = true;

  _init_list(_pointlist, *_irect1);

  if (_irect2->is_degenerate()) {
    // _not_implemented();
    CGAL_kernel_assertion(false);
  }else {
    Point_T2<R> v0 =_irect2->vertex(0);
    Point_T2<R> v1 =_irect2->vertex(1);
    Point_T2<R> v2 =_irect2->vertex(2);
    Point_T2<R> v3 =_irect2->vertex(3);
    if  (!_irect2->has_on_unbounded_side(_irect1->vertex(0)) && 
         !_irect2->has_on_unbounded_side(_irect1->vertex(1)) &&
         !_irect2->has_on_unbounded_side(_irect1->vertex(2)) &&
         !_irect2->has_on_unbounded_side(_irect1->vertex(2))){
      _result = ISO_RECTANGLE;
      return _result;
    }
    if  (!_irect1->has_on_unbounded_side(v0) && 
         !_irect1->has_on_unbounded_side(v1) &&
         !_irect1->has_on_unbounded_side(v2) &&
         !_irect1->has_on_unbounded_side(v3)){
      _pointlist._finalize_list();
      _init_list(_pointlist,*_irect2);
      _result = ISO_RECTANGLE;
      return _result;
    }
    if (_irect2->orientation() == COUNTERCLOCKWISE) {
      // counterclockwise triangle

      _cut_off(_pointlist, v0, v1);

      _cut_off(_pointlist, v1, v2);

      _cut_off(_pointlist, v2, v3);

      _cut_off(_pointlist, v3, v0);

        
       } else {
        _cut_off(_pointlist, v1, v0);
        _cut_off(_pointlist, v0, v3);
        _cut_off(_pointlist, v3, v2);
        _cut_off(_pointlist, v2, v1);
      }
  }
  switch (_pointlist.size) {
    case 0:
      _result = NO;
      break;
    case 1:
      _result = POINT;
      break;
    case 2:
      _result = SEGMENT;
      break;
    default:
      _result = ISO_RECTANGLE;
   }
   return _result;
}


#endif

template <class R>
bool
Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::
intersection(Point_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != POINT)
    return false;
  result = _pointlist.first->point;;
  return true;
}

template <class R>
bool
Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::
intersection(Segment_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != SEGMENT)
    return false;
  result = Segment_T2<R>(_pointlist.first->point,
                    _pointlist.first->next->point);
  return true;
}

template <class R>
bool
Iso_rectangle_T2_Iso_rectangle_T2_pair<R>::
intersection(Iso_rectangle_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != ISO_RECTANGLE)
    return false;
  if (_pointlist.first->point!=_pointlist.first->next->next->point) {
    result = Iso_rectangle_T2<R>(_pointlist.first->point,
                                 _pointlist.first->next->next->point);
    return true;
  }
  if (_pointlist.first->point!=_pointlist.first->next->point) {
    result = Iso_rectangle_T2<R>(_pointlist.first->point,
                                 _pointlist.first->next->point);
    return true;
  }
  if (_pointlist.first->point!=_pointlist.first->next->next->next->point) {
    result = Iso_rectangle_T2<R>(_pointlist.first->point,
                                 _pointlist.first->next->next->next->point);
    return true;
  }
  if (_pointlist.first->next->point!=_pointlist.first->next->next->point) {
    result = Iso_rectangle_T2<R>(_pointlist.first->next->point,
                                 _pointlist.first->next->next->point);
    return true;
  }

  if (_pointlist.first->next->point!=_pointlist.first->next->next->next->point) {
    result = Iso_rectangle_T2<R>(_pointlist.first->next->point,
                                 _pointlist.first->next->next->next->point);
    return true;
  }

  result = Iso_rectangle_T2<R>(_pointlist.first->next->next->point,
                               _pointlist.first->next->next->next->point);
  return true;
}

CGAL_END_NAMESPACE



#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Iso_rectangle_T2<R> &irect1, 
             const Iso_rectangle_T2<R> &irect2)
{
  typedef Iso_rectangle_T2_Iso_rectangle_T2_pair<R> is_t;
  is_t ispair(&irect1, &irect2);
  switch (ispair.intersection_type()) {
    case is_t::NO:
    default:
      return Object();
    case is_t::POINT: {
      Point_T2<R> pt;
      ispair.intersection(pt);
      return make_object(pt);
    }
    case is_t::SEGMENT: {
      Segment_T2<R> iseg;
      ispair.intersection(iseg);
      return make_object(iseg);
    }
    case is_t::ISO_RECTANGLE: {
      Iso_rectangle_T2<R> irect;
      ispair.intersection(irect);
      return make_object(irect);
    }
      
  }
}

CGAL_END_NAMESPACE

#endif
