// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Line_2_Line_2_intersection.h,v $
// $Revision: 1.9 $ $Date: 2003/10/21 12:16:46 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_LINE_T2_LINE_T2_INTERSECTION_H
#define CGAL_LINE_T2_LINE_T2_INTERSECTION_H

#include <CGAL/Line_T2.h>
#include <CGAL/Point_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Line_T2_Line_T2_pair {
public:
  enum Intersection_results {TWOPOINTS, LINE};
  Line_T2_Line_T2_pair() ;
  Line_T2_Line_T2_pair(Line_T2<R> const *line1,
                            Line_T2<R> const *line2);
  ~Line_T2_Line_T2_pair() {}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
  Intersection_results intersection_type() const;

#else
  Intersection_results intersection_type() const
  {
    if (_known)
      return _result;
    const typename R::FT a = _line1->X()*_line2->Y()-_line2->X()*_line1->Y();
    const typename R::FT b = -_line1->W()*_line2->Y()+_line2->W()*_line1->Y();
    const typename R::FT c = _line1->W()*_line2->X()-_line2->W()*_line1->X();
    if ((a == 0) && (b == 0) && (c == 0)){
      _result = LINE;
    }
    else {
      _result = TWOPOINTS;
      _intersection_point = 
      Point_T2<R>(a,b,c);
    }
    return _result;
  }

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

  bool                intersection(Point_T2<R> &result) const;
  bool                intersection(Line_T2<R> &result) const;
protected:
  Line_T2<R> const*   _line1;
  Line_T2<R> const *  _line2;
  mutable bool                    _known;
  mutable Intersection_results    _result;
  mutable Point_T2<R>         _intersection_point; 
  //the other intersection point is the antipode of this one
};

template <class R>
inline bool do_intersect(
  const Line_T2<R> &p1,
  const Line_T2<R> &p2)
{
  return true;
}

CGAL_END_NAMESPACE

#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Line_T2<R> &line1, const Line_T2<R> &line2)
{
  typedef Line_T2_Line_T2_pair<R> is_t;
  is_t linepair(&line1, &line2);
  switch (linepair.intersection_type()) {
    case is_t::TWOPOINTS: {
      Point_T2<R> pt;
      linepair.intersection(pt);
      return make_object(pt);
    }
    case is_t::LINE:
      return make_object(line1);
  }
}

CGAL_END_NAMESPACE


CGAL_BEGIN_NAMESPACE

template <class R>
Line_T2_Line_T2_pair<R>::Line_T2_Line_T2_pair()
{
  _line1 = 0;
  _line2 = 0;
  _known = false;
}

template <class R>
Line_T2_Line_T2_pair<R>::Line_T2_Line_T2_pair(
    Line_T2<R> const *line1, Line_T2<R> const *line2)
{
  _line1 = line1;
  _line2 = line2;
  _known = false;
}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Line_T2_Line_T2_pair<R>::Intersection_results
Line_T2_Line_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  const typename R::FT line1_X =_line1->X();
  const typename R::FT line2_X =_line2->X();
  const typename R::FT line1_Y =_line1->Y();
  const typename R::FT line2_Y =_line2->Y();
  const typename R::FT a = line1_X*line2_Y-line2_X*line1_Y;
  
  const typename R::FT b = -_line1->W()*_line2->Y()+_line2->W()*_line1->Y();
  
  const typename R::FT c = _line1->W()*_line2->X()-_line2->W()*_line1->X();
  if ((a == 0) && (b == 0) && (c == 0)){
      _result = LINE;
  }
  else {
    _result = TWOPOINTS;
    _intersection_point = 
    Point_T2<R>(a,b,c);
  }
  return _result;
 
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

template <class R>
bool
Line_T2_Line_T2_pair<R>::intersection(Point_T2<R> &pt) const
{
  if (!_known)
    intersection_type();
  if (_result != TWOPOINTS)
    return false;
  pt = _intersection_point;
  return true;
}

template <class R>
bool
Line_T2_Line_T2_pair<R>::intersection(Line_T2<R> &l) const
{
  if (!_known)
    intersection_type();
  if (_result != LINE)
    return false;
  l = *_line1;
  return true;
}

CGAL_END_NAMESPACE



#endif
