// Copyright (c) 1997  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Planar_map/include/CGAL/Planar_map_2.h,v $
// $Revision: 1.58 $ $Date: 2003/09/18 10:24:29 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Iddo Hanniel      <hanniel@math.tau.ac.il>
//                 Eyal Flato        <flato@post.tau.ac.il>
//                 Oren Nechushtan   <theoren@math.tau.ac.il>
//                 Eti Ezra          <estere@post.tau.ac.il>
//                 Shai Hirsch       <shaihi@post.tau.ac.il>
//                 Eugene Lipovetsky <eug@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_PLANAR_MAP_T2_H
#define CGAL_PLANAR_MAP_T2_H

#include <vector>

CGAL_BEGIN_NAMESPACE

template <class Traits>
class Planar_map_T2 {

public:
  typedef Planar_map_T2<Traits> Planar_map;
  typedef typename Traits::Point_T2 Point;
  typedef typename Traits::Segment_T2 Segment;
  typedef typename Traits::Antipode_T2 Antipode;
  typedef typename Traits::Compare_distance_to_point_T2 Compare_distance_to_point;
  typedef typename Traits::Has_on_segment_T2  Has_on_segment;
  typedef typename Traits::Orientation_T2 Orientation_T2;


  template <class Point>
  class _Vertex {
    private:
      Point p;
      int number_of_incident_faces;
      std::vector<int> indexes_incident_faces;

    public:

       _Vertex() {}

      _Vertex (const Point & _p) {
        p = _p;
	number_of_incident_faces=0;
      }

      template <typename Index_iterator>
      _Vertex (const Point & _p,Index_iterator indexes_begin,
               Index_iterator indexes_end
	       ) {
	number_of_incident_faces=0;
        for (Index_iterator it = indexes_begin; it != indexes_end; it++)
        {
          indexes_incident_faces.push_back(*it);
          number_of_incident_faces++;
        }
        p = _p;
      }

      template <typename Index_iterator>
      Point set_incident_faces(Index_iterator indexes_begin,
                               Index_iterator indexes_end) {
        for (Index_iterator it = indexes_begin; it != indexes_end; it++)
        {
          indexes_incident_faces.push_back(*it);
          number_of_incident_faces++;
        }
      }

      Point get_point() const {
        return p;
      }

      std::vector<int> get_indexes_incident_faces() const{
        return indexes_incident_faces;
      }

      void Print() const {
        std::cout << "Vertex = " << p << std::endl;
      }

      int get_number_of_incident_faces() const {
        return number_of_incident_faces;
      }

      bool operator == (const _Vertex & v) const {
        return (p == v.get_point());
      }

      void set_index_incident_face(const int index_vertex) {
        indexes_incident_faces.push_back(index_vertex);
        number_of_incident_faces++;
      }
      void set_indexes_incident_faces(std::vector<int> v) {
        indexes_incident_faces=v;
	 //std::cout << "vertex set\n";
        number_of_incident_faces=v.size();
      }
  };


  template <class Vertex>
  class _Face {
    private:
      std::vector<Vertex> vertices_face;
      std::vector<int> indexes_neighbors;
      int index_current_edge;

    public :

      _Face() {}

      template <typename Vertex_iterator>
      _Face(Vertex_iterator vertices_begin,
           Vertex_iterator vertices_end)
      {
        int j=0;
        for (Vertex_iterator vi = vertices_begin; vi != vertices_end; vi++)
        {
          vertices_face.push_back(*vi);
          j++;
        }
        for (int k=0; k < j; k++)
        {
          indexes_neighbors.push_back(0);
        }
      }

      void set_incident_face(const int index, const int t)
      {
        vertices_face[index].set_index_incident_face(t);
      }

      void set_index_current_edge(const int new_index)
      {
        i = new_index;
      }

      void set_vertex(const int index, const Vertex v)
      {
      
        vertices_face[index]=v;
	
      }
      void set_index_neighbor(const int index, const int index_neighbor)
      {
        indexes_neighbors[index] = index_neighbor;
      }

      int get_index_neighbor(const int index) const
      {
        return indexes_neighbors[index];
      }

      int get_index_edge(const Vertex & s,
                         const Vertex & t) const
      {
        CGAL_precondition(!(s==t));
        unsigned int i=0;
	bool found = false;
	do {
          if (s == vertices_face[i])
	    found = true;
	  if (!found)
	    i++;
	} while(!found && i < vertices_face.size());
        if (!found) return -1;
	return i;
      }

      int get_index_vertex(const Vertex & s) const
      {
        unsigned int i=0;
	bool found = false;
	do {
	  if (s == vertices_face[i])
            found=true;
          else
	    i++;
	} while (!found && i < vertices_face.size());
	if (!found) return -1;
	return i;
      }

      int index_cw(const int i) const
      {
        return ((i - 1 + get_number_of_vertices())%get_number_of_vertices());
      }

      int index_ccw(const int i) const
      {
        return ((i + 1)%get_number_of_vertices());
      }

      Vertex cw(const int i) const {
        return vertices_face[index_cw(i)];
      }

      Vertex ccw(const int i) const {
        return vertices_face[index_ccw(i)];
      }

      bool is_vertex(const Vertex & v) const
      {
        bool belongs_face = false;
	unsigned int i = 0;
	do {
	  if (v == vertices_face[i])
	    belongs_face = true;
	  i++;
	} while (!belongs_face && i < vertices_face.size());
	return belongs_face;
      }

      unsigned int get_number_of_vertices() const
      {
        return vertices_face.size();
      }

      int get_index_current_edge() const
      {
        return index_current_edge;
      }

      Vertex get_vertex(int index) const
      {
        return vertices_face[index];
      }

      void Print() const
      {
        for (unsigned int k=0; k < vertices_face.size(); k++)
          vertices_face[k].Print();
      }
  };


public:
  typedef _Vertex<Point> Vertex;
  typedef _Face<Vertex> Face;
  typedef std::vector<Face> all_the_faces;

private:
   bool has_unbounded_face;
   all_the_faces pm;

public:

   typedef enum {
    VERTEX = 1,
    EDGE,
    FACE ,
    UNBOUNDED_FACE
  } Locate_type;

   Planar_map_T2() {}

   
   template <typename all_faces_iterator>
   Planar_map_T2(bool _has_unbounded_face,
                 const all_faces_iterator begin,
                 const all_faces_iterator end) {
     initialize(_has_unbounded_face,begin,end,(*begin).begin());
   }

   template <typename all_faces_iterator, typename all_vertices_iterator>
   void initialize(bool _has_unbounded_face,
                 const all_faces_iterator begin,
                 const all_faces_iterator end,
		 const all_vertices_iterator begin2) {

     has_unbounded_face = _has_unbounded_face;
     std::vector<int> v_incident_faces;
     std::vector<Vertex> v_vertex_face;
     bool equal = false;
     //set the faces;
     all_faces_iterator afi = begin,afi2;
     all_vertices_iterator avi,avi2;
     int index_ifc;
     for (; afi != end; afi++) {
       avi = (*afi).begin();
       for(; avi!=(*afi).end(); avi++){
         index_ifc = 0;
         for (afi2 = begin; afi2 != end; afi2++) {
	   avi2 = (*afi2).begin();
	   equal = false;
           while (!equal && avi2 != (*afi2).end())
	   {
	     if (avi2->get_point() == avi->get_point()) {
	       equal = true;
	       v_incident_faces.push_back(index_ifc);
	     }
	     avi2++;
	   }
	   index_ifc++;
	 }
	 v_vertex_face.push_back(
	           Vertex(avi->get_point(),
		          v_incident_faces.begin(),
			  v_incident_faces.end()));
         v_incident_faces.erase(v_incident_faces.begin(),v_incident_faces.end());
       }
       Face f(v_vertex_face.begin(),v_vertex_face.end());
       pm.push_back(f);
       v_vertex_face.erase(v_vertex_face.begin(),v_vertex_face.end());
     }

     //set the neighbors
     for (unsigned int k=0; k < pm.size(); k++) {

       for (unsigned int z = 0; z < pm[k].get_number_of_vertices(); z++) {

         for (unsigned int t = 0; t < pm.size(); t++) {

           if (t != k) {

             if (pm[t].is_vertex(pm[k].get_vertex(z)) &&
	         pm[t].is_vertex(
		     pm[k].get_vertex((z+1)%pm[k].get_number_of_vertices()))) {
		       pm[k].set_index_neighbor(z,t);

             }
	   }
	 }
       }
     }


   }



   bool is_closer_on_line(const Point p, const Point q, const Point r) const
   {
     Orientation_T2 orientation = Traits().orientation_T2_object();

     CGAL_precondition(orientation(p,q,r)==COLLINEAR);

     Has_on_segment has_on_segment = Traits().has_on_segment_T2_object();
     return (has_on_segment(Segment(q,r),p) && p!=q);

   }

   Locate_type locate(const Point p, int & index1, int & index2, Face & f) {
     Has_on_segment has_on_segment = Traits().has_on_segment_T2_object();
     Orientation_T2 orientation = Traits().orientation_T2_object();

     int index_face = 0;
     int index_s;
     
     //pick up a vertex, from it we draw a segment till p
     Vertex s = pm[0].get_vertex(0);
     Vertex vt;

     //if p and s are antipodal, the segment through them is not
     //defined, so we pick up another vertex to be s
     if (antipodal(s.get_point(),p))
       s = pm[0].get_vertex(1);
     if (p == s.get_point()) {
       index1 = pm[0].get_index_vertex(s);
       index2 = index1;
       f = pm[0];
       return VERTEX;
     }
     //the next, we get the indexes of vertex s incident faces
     std::vector<int> v_incident_faces = s.get_indexes_incident_faces();

     //auxiliar variables
     unsigned int j = 0;
     bool is_in_interior_overlap_face = false;
     Orientation o,b,t;

     //the index of each edge of a face
     int k;

     //in this loop, we get the index of a face f that
     //segment sp comes in or overlaps a edge of f
    // std::cout << "000\n";
     //std::cout << v_incident_faces.size() << " - 000\n";
     Locate_type lt1 = VERTEX,_lt1;
     Locate_type lt2 = VERTEX,_lt2;
     while (j < v_incident_faces.size() &&
            !is_in_interior_overlap_face) {
     
       index_s = pm[v_incident_faces[j]].get_index_vertex(s);
       pm[v_incident_faces[j]].Print();
       o = orientation(p,s.get_point(),
                       pm[v_incident_faces[j]].ccw(index_s).get_point());
       b = orientation(p,
                       pm[v_incident_faces[j]].cw(index_s).get_point(),
                         s.get_point());
       t = orientation(pm[v_incident_faces[j]].cw(index_s).get_point(),
                       s.get_point(),
                       pm[v_incident_faces[j]].ccw(index_s).get_point());
       if (o == 1 && b == 1) {
     
         is_in_interior_overlap_face = true;
         index_face = v_incident_faces[j];
         k = (index_s+1)%pm[index_face].get_number_of_vertices();
       }
       else if ((o * b == -1) && t == -1) {
     
         is_in_interior_overlap_face = true;
         index_face = v_incident_faces[j];
         k = (index_s+1)%pm[index_face].get_number_of_vertices();
       }
       //overlaps one edge of the face at index j
       else if (o==0 && (has_on_segment(Segment(s.get_point(),p),
                        pm[v_incident_faces[j]].ccw(index_s).get_point()) &&
			has_on_segment(Segment(s.get_point(),p),
                        pm[v_incident_faces[j]].ccw(index_s).get_point()))) {
        
         is_in_interior_overlap_face = true;
	 index_face = v_incident_faces[j];
	 k = index_s;
	 lt2 = EDGE;
       }
       //overlaps one edge of the face at index j
       else if (b==0 && (has_on_segment(Segment(s.get_point(),p),
                        pm[v_incident_faces[j]].cw(index_s).get_point()) ||
			has_on_segment(Segment(s.get_point(),
			 pm[v_incident_faces[j]].cw(index_s).get_point()),p))) {
        
         is_in_interior_overlap_face = true;
	 index_face = v_incident_faces[j];
	 k = (index_s-1 + pm[index_face].get_number_of_vertices())%
	        pm[index_face].get_number_of_vertices();
	 lt2 = EDGE;
       }
       
       j++;
     }
     if (has_unbounded_face && !is_in_interior_overlap_face) {
       
       is_in_interior_overlap_face = true;
       index_face = 0;
       index_s = pm[v_incident_faces[j]].get_index_vertex(s);
       k = (index_s+1)%pm[index_face].get_number_of_vertices();
     }
     bool found = false;
     bool intercept,_intercept;
     Object obj;
     Point q = s.get_point();
     Point pt;
     unsigned int i;
     Segment seg;
     int index_edge;
     
     while (!found)
     {
     
       _intercept = false;
       i = 0;
       Locate_type lt_aux = lt2;
       while ((lt_aux != VERTEX &&
                (i < pm[index_face].get_number_of_vertices()-1)) ||
              (lt_aux == VERTEX &&
                (i < pm[index_face].get_number_of_vertices()-2)) && !found)
       {
     
         intercept = false;
         lt1 = lt2;
	 Segment s1(s.get_point(),p);
	 Point pt1 = pm[index_face].get_vertex(k).get_point();
	
	 Point pt2 = pm[index_face].get_vertex((k+1)%
                     pm[index_face].get_number_of_vertices()).get_point();
        
	 Segment s2(pt1,pt2);
	 
	 obj = intersection(s1,s2);
         
         //treat all the case when the intersection between sp e a edge of the
	 //planar map is a point
         if (assign(pt,obj)) {
	 
	   intercept = true;
	   lt2 = EDGE;
	   if (is_closer_on_line(pt,q,p)) {
	   if (!(int(orientation(pm[index_face].get_vertex(k).get_point(),
	                         pm[index_face].get_vertex(
                         (k+1)%pm[index_face].get_number_of_vertices()).get_point()
			         ,s.get_point())) *
                 int(orientation(pm[index_face].get_vertex(k).get_point(),
	                         pm[index_face].get_vertex((k+1)%
			       pm[index_face].get_number_of_vertices()).get_point()
			         ,p))==-1)) {
             found = true;
	 
	     index1 = k;
	     index2 = (k+1)%pm[index_face].get_number_of_vertices();
           }
	   if (!(int(orientation(p,s.get_point(),
	                         pm[index_face].get_vertex(k).get_point())) *
                 int(orientation(p,s.get_point(),pm[index_face].get_vertex(
                                 (k+1)%
		   	       pm[index_face].get_number_of_vertices()).get_point()
		                ))==-1)) {
             lt2 = VERTEX;
	 
	     if (pt == p) {
	 
               found = true;
	     }
	   }
	   }
	 }
         //treat all the case when the intersection between sp e a edge of the
	 //planar map is a segment
	 else if (assign(seg,obj))
         {
	 
	   if (seg.has_on(p)) {
	     pt = p;
	 
             found = true;
	     intercept = true;
	     f=pm[index_face];
	     if (p == pm[index_face].get_vertex(k).get_point()) {
	       lt2 = VERTEX;
	       index1=k;
	       index2=k;
	 
	     }
	     else if (p == pm[index_face].get_vertex(
	                     (k+1)%
			     pm[index_face].get_number_of_vertices()).get_point()) {
               lt2 = VERTEX;
	       index1=k;
	       index2=k;
	 
	     }
	     else {
	       lt2 = EDGE;
	       index1=k;
	       index2=(k+1)%pm[index_face].get_number_of_vertices();
	 
	     }
	   }
	   else {
	     if (is_closer_on_line(pm[index_face].get_vertex(k).get_point(),
                        pm[index_face].get_vertex((k+1)%
			pm[index_face].get_number_of_vertices()).get_point(),p)) {
	 
	       pt = pm[index_face].get_vertex(k).get_point();
	     }
	     else {
	 
	       pt = pm[index_face].get_vertex((k+1)%
	            pm[index_face].get_number_of_vertices()).get_point();
	     }
	 
	     intercept = true;
	     lt2 = VERTEX;
	     vt = Vertex(pt);
	   }
	 }
	 if (intercept)
         {
	 
	   if (is_closer_on_line(pt,q,p)) {
	     _intercept = intercept;
	     q = pt;
	     index_edge = k;
	 
	     _lt1 = lt1;
	     _lt2 = lt2;
	 
	   }
         }
	 i++;
	 if (((lt_aux != VERTEX &&
                i < pm[index_face].get_number_of_vertices()-1) ||
              (lt_aux == VERTEX &&
                i < pm[index_face].get_number_of_vertices()-2))) {
	   k=(k+1)%pm[index_face].get_number_of_vertices();
	 
         }
      }
      if (!_intercept) {
        
        index1 = -1;
        index2 = -1;
	found = true;
        f=pm[index_face];
	
        if (index_face == 0 && has_unbounded_face)
	  return UNBOUNDED_FACE;
        return FACE;
      }
      else if (_lt2 == VERTEX && !found) {
        
        v_incident_faces = vt.get_indexes_incident_faces();
	unsigned int v = 0;
	is_in_interior_overlap_face = false;
	while (v < v_incident_faces.size() && !is_in_interior_overlap_face)
	{
	  k = pm[v_incident_faces[v]].get_index_vertex(vt);
	  o = orientation(p,vt.get_point(),
	                  pm[v_incident_faces[v]].ccw(k).get_point());
          b = orientation(p,
	                  pm[v_incident_faces[v]].cw(k).get_point(),
                          vt.get_point());
          t = orientation(pm[v_incident_faces[v]].cw(k).get_point(),
                          vt.get_point(),
                          pm[v_incident_faces[v]].ccw(k).get_point());
          if (o == 1 && b == 1) {
	    is_in_interior_overlap_face = true;
	    index_face = v_incident_faces[v];
	    k = (k+1)%pm[index_face].get_number_of_vertices();
	  }
	  //overlaps one edge of the face at index j
          else if (o==0 && has_on_segment(Segment(vt.get_point(),p),
                             pm[v_incident_faces[v]].ccw(k).get_point())) {
            is_in_interior_overlap_face = true;
	    index_face = v_incident_faces[v];
	    k = (k+1)%pm[index_face].get_number_of_vertices();
          }
          //overlaps one edge of the face at index j
          else if (b==0 && has_on_segment(Segment(vt.get_point(),p),
                             pm[v_incident_faces[v]].cw(k).get_point())) {
            is_in_interior_overlap_face = true;
	    index_face = v_incident_faces[v];
	    k = (k-1 + pm[index_face].get_number_of_vertices())%
	        pm[index_face].get_number_of_vertices();
          }
	  v++;
        }
	if (has_unbounded_face && !is_in_interior_overlap_face) {
          is_in_interior_overlap_face = true;
          index_face = 0;
	  k = pm[0].get_index_vertex(vt);
          k = (k+1)%pm[index_face].get_number_of_vertices();
        }
      }
      else if (!found){
        
        int _index_face = index_face;
	
        index_face = pm[index_face].get_index_neighbor(index_edge);
	
	if (has_unbounded_face && (index_face == 0 ||
	                           _index_face == 0))
          k = pm[index_face].get_index_edge(pm[index_face].get_vertex(index_edge),
                           pm[index_face].get_vertex(
	      (index_edge+1)%pm[index_face].get_number_of_vertices()));
        else
	  k = pm[index_face].get_index_edge(
                           pm[_index_face].get_vertex(
	      (index_edge+1)%pm[_index_face].get_number_of_vertices()),
	                   pm[_index_face].get_vertex(index_edge));
	k = (k+1)%pm[index_face].get_number_of_vertices();
      }
      lt1 = _lt1;
      lt2 = _lt2;
    }
    return lt2;
  }
};



CGAL_END_NAMESPACE

#endif
