// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/Point_2.h,v $
// $Revision: 1.16 $ $Date: 2003/10/21 12:18:10 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_POINT_T1_H
#define CGAL_SIGNED_HOMOGENEOUS_POINT_T1_H

//#include <CGAL/Origin.h>
//#include <CGAL/Bbox_2.h>

CGAL_BEGIN_NAMESPACE

template < class R_ >
class Point_T1
{
  typedef typename R_::FT                   FT;
  typedef typename R_::Point_T1              TPoint_T1;
  typedef typename R_::rep  		    rep;

private:
	Point_2<rep> *Point;

public:
  typedef R_                                     R;

  Point_T1(){}

 
  Point_T1(const FT &w, const FT &x)
  {	
    CGAL_kernel_precondition_msg((w!=0 || x!=0),
		 "Invalid pair in T^2");
    Point = new Point_2<rep>(w,x);	
  }

 
  const FT& w() const
  {
      return Point->x();
  }
  const FT& x() const
  {
      return Point->y();
  }

  int dimension() const
  {
      return 1;
  }

  bool operator==(const Point_T1 &p) const
  {
      if (this->x()*p.w() != this->w()*p.x())
         return false;
      return ! ((sign(this->x()) != sign(p.x())) || (sign(this->w()) != sign(p.w())));
  }
  
  bool operator!=(const Point_T1 &p) const
  {
      return !(*this == p);
  }

  TPoint_T1 operator-() const
  {
    return TPoint_T1(this->w(),-this->x());
  }


};


//#ifndef CGAL_NO_OSTREAM_INSERT_POINTC2
template < class R >
std::ostream &
operator<<(std::ostream &os, const Point_T1<R> &p)
{
    switch(os.iword(IO::mode)) {
    case IO::ASCII :
        return os << p.w() << ' ' << p.x();
    case IO::BINARY :
        write(os, p.w());
        write(os, p.x());
        return os;
    default:
        return os << "Point_T1(" << p.w() << ", " << p.x() << ')';
    }
}
//#endif // CGAL_NO_OSTREAM_INSERT_POINTC2

//#ifndef CGAL_NO_ISTREAM_EXTRACT_POINTC2
template < class R >
std::istream &
operator>>(std::istream &is, Point_T1<R> &p)
{
    typename R::FT w, x;
    switch(is.iword(IO::mode)) {
    case IO::ASCII :
        is >> w >> x;
        break;
    case IO::BINARY :
        read(is, w);
        read(is, x);
        break;
    default:
        std::cerr << "" << std::endl;
        std::cerr << "Stream must be in ascii or binary mode" << std::endl;
        break;
    }
    if (is)
	p = Point_T1<R>(w, x);
    return is;
}

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_POINT_T1_H
