// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Point_2_Ray_2_intersection.h,v $
// $Revision: 1.7 $ $Date: 2003/10/21 12:16:49 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_POINT_T2_RAY_T2_INTERSECTION_H
#define CGAL_POINT_T2_RAY_T2_INTERSECTION_H

#include <CGAL/Ray_T2.h>
#include <CGAL/Point_T2.h>

CGAL_BEGIN_NAMESPACE


template <class R>
inline bool
do_intersect(const Point_T2<R> &pt, const Ray_T2<R> &ray)
{
    return ray.has_on(pt);
}

CGAL_END_NAMESPACE

#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Point_T2<R> &pt, const Ray_T2<R> &ray)
{
    if (do_intersect(pt,ray)) {
        return make_object(pt);
    }
    return Object();
}


template <class R>
inline bool
do_intersect(const Ray_T2<R> &ray, const Point_T2<R> &pt)
{
    return ray.has_on(pt);
}


template <class R>
inline Object
intersection(const Ray_T2<R> &ray, const Point_T2<R> &pt)
{
    if (do_intersect(pt,ray)) {
        return make_object(pt);
    }
    return Object();
}

CGAL_END_NAMESPACE

#endif
