// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Point_2_Triangle_2_intersection.h,v $
// $Revision: 1.8 $ $Date: 2003/10/21 12:16:50 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_POINT_T2_TRIANGLE_T2_INTERSECTION_H
#define CGAL_POINT_T2_TRIANGLE_T2_INTERSECTION_H

#include <CGAL/Point_T2.h>
#include <CGAL/Triangle_T2.h>
#include <CGAL/Point_T2.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Point_T2_Triangle_T2_pair {
public:
    enum Intersection_results {NO, POINT};
    Point_T2_Triangle_T2_pair() ;
    Point_T2_Triangle_T2_pair(Point_T2<R> const *pt,
                              Triangle_T2<R> const *trian);
    ~Point_T2_Triangle_T2_pair() {}
#ifdef CGAL_CFG_RETURN_TYPE_BUG_2
    Intersection_results intersection_type() const
    {
        if (_known)
            return _result;
    // The non const this pointer is used to cast away const.
        _known = true;
        if (_trian->has_on_unbounded_side(*_pt)) {
            _result = NO;
        } else {
            _result = POINT;
        }
        return _result;
    }
    
#else
    Intersection_results intersection_type() const;
#endif // CGAL_CFG_RETURN_TYPE_BUG_2
    bool                intersection(Point_T2<R> &result) const;
protected:
    Point_T2<R> const *    _pt;
    Triangle_T2<R> const * _trian;
    mutable bool                       _known;
    mutable Intersection_results       _result;
    mutable Point_T2<R>            _intersection_point;
    mutable Point_T2<R>            _other_point;
};

template <class R>
inline bool do_intersect(
    const Point_T2<R> &p1,
    const Triangle_T2<R> &p2)
{
    typedef Point_T2_Triangle_T2_pair<R> pair_t;
    pair_t pair(&p1, &p2);
    return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE



#include <CGAL/Line_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>
#include <CGAL/Straight_2.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Point_T2_Triangle_T2_pair<R>::
Point_T2_Triangle_T2_pair()
{
    _known = false;
    _pt = 0;
    _trian = 0;
}

template <class R>
Point_T2_Triangle_T2_pair<R>::
Point_T2_Triangle_T2_pair(Point_T2<R> const *pt,
                            Triangle_T2<R> const *trian)
{
    _known = false;
    _pt = pt;
    _trian = trian;
}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Point_T2_Triangle_T2_pair<R>::Intersection_results
Point_T2_Triangle_T2_pair<R>::intersection_type() const
{
    if (_known)
        return _result;
// The non const this pointer is used to cast away const.
    _known = true;
    if (_trian->has_on_unbounded_side(*_pt)) {
        _result = NO;
    } else {
        _result = POINT;
    }
    return _result;
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2



template <class R>
bool
Point_T2_Triangle_T2_pair<R>::
intersection(Point_T2<R> &result) const
{
    if (!_known)
        intersection_type();
    if (_result != POINT)
        return false;
    result = *_pt;
    return true;
}

CGAL_END_NAMESPACE



#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Point_T2<R> &pt, const Triangle_T2<R>&tr)
{
    typedef Point_T2_Triangle_T2_pair<R> is_t;
    is_t ispair(&pt, &tr);
    switch (ispair.intersection_type()) {
    case is_t::NO:
    default:
        return Object();
    case is_t::POINT: {
        return make_object(pt);
    }
    }
}

template <class R>
class Triangle_T2_Point_T2_pair
: public Point_T2_Triangle_T2_pair<R> {
public:
    Triangle_T2_Point_T2_pair(
            Triangle_T2<R> const *trian,
            Point_T2<R> const *pt) :
                        Point_T2_Triangle_T2_pair<R>(pt, trian) {}
};

template <class R>
inline bool do_intersect(
    const Triangle_T2<R> &p1,
    const Point_T2<R> &p2)
{
    typedef Triangle_T2_Point_T2_pair<R> pair_t;
    pair_t pair(&p1, &p2);
    return pair.intersection_type() != pair_t::NO;
}

template <class R>
inline Object
intersection(const Triangle_T2<R> &tr, const Point_T2<R> &pt)
{
    return intersection(pt, tr);
}

CGAL_END_NAMESPACE

#endif
