// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Ray_2_Ray_2_intersection.h,v $
// $Revision: 1.11 $ $Date: 2003/10/21 12:16:52 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_RAY_T2_RAY_T2_INTERSECTION_H
#define CGAL_RAY_T2_RAY_T2_INTERSECTION_H

#include <CGAL/Ray_T2.h>
#include <CGAL/Segment_T2.h>
#include <CGAL/Point_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Ray_T2_Ray_T2_pair {
public:
  enum Intersection_results {NO, POINT, SEGMENT, RAY};
  Ray_T2_Ray_T2_pair() ;
  Ray_T2_Ray_T2_pair(Ray_T2<R> const *ray1,
                     Ray_T2<R> const *ray2);
  ~Ray_T2_Ray_T2_pair() {}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
  Intersection_results intersection_type() const;
#else
  Intersection_results intersection_type() const
  {
    if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  const Line_T2<R> &l1 = _ray1->supporting_line();
  const Line_T2<R> &l2 = _ray2->supporting_line();
  Line_T2_Line_T2_pair<R> linepair(&l1, &l2);
  switch ( linepair.intersection_type()) {
  case Line_T2_Line_T2_pair<R>::TWOPOINTS:
    linepair.intersection(_intersection_point);
    if (_ray1->collinear_has_on(_intersection_point)
        && _ray2->collinear_has_on(_intersection_point)) {
      _result = POINT;
      return _result;
    }
    if (_ray1->collinear_has_on(antipode(_intersection_point))
           && _ray2->collinear_has_on(antipode(_intersection_point)) ) {
      _result = POINT;
      _intersection_point = antipode(_intersection_point);
      return _result;
    }
    _result = NO;
    return _result;
  case Line_T2_Line_T2_pair<R>::LINE:
  {
    Point_T2<R> *sray1 = new Point_T2<R> (_ray1->start());
    Point_T2<R> *sray2 = new Point_T2<R> (_ray2->start());
    Point_T2<R> *spray1 = new Point_T2<R> (_ray1->second_point());
    Point_T2<R> *spray2 = new Point_T2<R> (_ray2->second_point());
    if (_ray1->direction() == _ray2->direction()) {
      if (*spray1 == *sray2) {
        _intersection_point = *spray1;
        _result = POINT;
        return _result;
      }
      if (*sray1 == *spray2) {
        _intersection_point = *sray1;
        _result = POINT;
        return _result;
      }
      if (*spray1 == *spray2) {
        _intersection_point = *spray1;
        if (_ray1->collinear_has_on(*sray2))
          _other_point = *sray2;
        else
          _other_point = *sray1;
        _result = RAY;
        return _result;
      }
      if (*sray2 == *sray1) {
        _intersection_point = *sray1;
        if (_ray1->collinear_has_on(*spray2))
          _other_point = *spray2;
        else
          _other_point = *spray1;
        _result = RAY;
        return _result;
      }
      if (_ray1->collinear_has_on(*sray2)) {
        _intersection_point = *sray2;
        _other_point = *spray1;
        if (is_at_infinity(*sray1))
          _result = SEGMENT;
        else
          _result = RAY;
        return _result;
      }
      if (_ray2->collinear_has_on(*sray1)) {
        _intersection_point = *sray1;
        _other_point = *spray2;
        if (is_at_infinity(*sray2))
          _result = SEGMENT;
        else
          _result = RAY;
        return _result; 
      }
      _result = NO;
      return _result;
    }
    else {
      if (*sray1 == *sray2) {
        _intersection_point = *sray1;
        _result = POINT;
        return _result;
      }
      if (*spray2 == *spray1) {
        _intersection_point = *spray1;
        _result = POINT;
        return _result;
      }
      if (_ray2->collinear_has_on(*sray1)) {
        _intersection_point = *sray1;
        _other_point = *sray2;
        _result = SEGMENT;
        return _result;
      }
      _result = NO;
      return _result;
    }

  }
  default:
        CGAL_kernel_assertion(false); // should not be reached:
        return _result;
    }
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

    bool                intersection(Point_T2<R> &result) const;
    bool                intersection(Segment_T2<R> &result) const;
    bool                intersection(Ray_T2<R> &result) const;
protected:
    Ray_T2<R> const*   _ray1;
    Ray_T2<R> const*   _ray2;
    mutable bool                    _known;
    mutable Intersection_results    _result;
    mutable Point_T2<R>         _intersection_point, _other_point;
};

template <class R>
inline bool do_intersect(
    const Ray_T2<R> &p1,
    const Ray_T2<R> &p2)
{
    typedef Ray_T2_Ray_T2_pair<R> pair_t;
    pair_t pair(&p1, &p2);
    return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE

#include <CGAL/Line_T2.h>
#include <CGAL/Line_T2_Line_T2_intersection.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Ray_T2_Ray_T2_pair<R>::Ray_T2_Ray_T2_pair()
{
    _ray1 = 0;
    _ray2 = 0;
    _known = false;
}

template <class R>
Ray_T2_Ray_T2_pair<R>::Ray_T2_Ray_T2_pair(
    Ray_T2<R> const *ray1, Ray_T2<R> const *ray2)
{
    _ray1 = ray1;
    _ray2 = ray2;
    _known = false;
}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Ray_T2_Ray_T2_pair<R>::Intersection_results
Ray_T2_Ray_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  const Line_T2<R> &l1 = _ray1->supporting_line();
  const Line_T2<R> &l2 = _ray2->supporting_line();
  Line_T2_Line_T2_pair<R> linepair(&l1, &l2);
  switch ( linepair.intersection_type()) {
  case Line_T2_Line_T2_pair<R>::TWOPOINTS:
    linepair.intersection(_intersection_point);
    if (_ray1->collinear_has_on(_intersection_point)
        && _ray2->collinear_has_on(_intersection_point)) {
      _result = POINT;
      return _result;
    }
    if (_ray1->collinear_has_on(antipode(_intersection_point))
           && _ray2->collinear_has_on(antipode(_intersection_point)) ) {
      _result = POINT;
      _intersection_point = antipode(_intersection_point);
      return _result;
    }
    _result = NO;
    return _result;
  case Line_T2_Line_T2_pair<R>::LINE:
  {
    Point_T2<R> *sray1 = new Point_T2<R> (_ray1->start());
    Point_T2<R> *sray2 = new Point_T2<R> (_ray2->start());
    Point_T2<R> *spray1 = new Point_T2<R> (_ray1->second_point());
    Point_T2<R> *spray2 = new Point_T2<R> (_ray2->second_point());
    if (_ray1->direction() == _ray2->direction()) {
      if (*spray1 == *sray2) {
        _intersection_point = *spray1;
        _result = POINT;
        return _result;
      }
      if (*sray1 == *spray2) {
        _intersection_point = *sray1;
        _result = POINT;
        return _result;
      }
      if (*spray1 == *spray2) {
        _intersection_point = *spray1;
        if (_ray1->collinear_has_on(*sray2))
          _other_point = *sray2;
        else
          _other_point = *sray1;
        _result = RAY;
        return _result;
      }
      if (*sray2 == *sray1) {
        _intersection_point = *sray1;
        if (_ray1->collinear_has_on(*spray2))
          _other_point = *spray2;
        else
          _other_point = *spray1;
        _result = RAY;
        return _result;
      }
      if (_ray1->collinear_has_on(*sray2)) {
        _intersection_point = *sray2;
        _other_point = *spray1;
        if (is_at_infinity(*sray1))
          _result = SEGMENT;
        else
          _result = RAY;
        return _result;
      }
      if (_ray2->collinear_has_on(*sray1)) {
        _intersection_point = *sray1;
        _other_point = *spray2;
        if (is_at_infinity(*sray2))
          _result = SEGMENT;
        else
          _result = RAY;
        return _result; 
      }
      _result = NO;
      return _result;
    }
    else {
      if (*sray1 == *sray2) {
        _intersection_point = *sray1;
        _result = POINT;
        return _result;
      }
      if (*spray2 == *spray1) {
        _intersection_point = *spray1;
        _result = POINT;
        return _result;
      }
      if (_ray2->collinear_has_on(*sray1)) {
        _intersection_point = *sray1;
        _other_point = *sray2;
        _result = SEGMENT;
        return _result;
      }
      _result = NO;
      return _result;
    }

  }
  default:
        CGAL_kernel_assertion(false); // should not be reached:
        return _result;
    }
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

template <class R>
bool
Ray_T2_Ray_T2_pair<R>::intersection(Point_T2<R> &result) const
{
    if (!_known)
        intersection_type();
    if (_result != POINT)
        return false;
    result = _intersection_point;
    return true;
}

template <class R>
bool
Ray_T2_Ray_T2_pair<R>::intersection(Segment_T2<R> &result) const
{
    if (!_known)
        intersection_type();
    if (_result != SEGMENT)
        return false;
    result = Segment_T2<R>(_ray1->start(), _ray2->start());
    return true;
}

template <class R>
bool
Ray_T2_Ray_T2_pair<R>::intersection(Ray_T2<R> &result) const
{
    if (!_known)
        intersection_type();
    if (_result != RAY)
        return false;
    result = Ray_T2<R>(_intersection_point, _ray1->direction());
    return true;
}

CGAL_END_NAMESPACE


#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Ray_T2<R> &ray1, const Ray_T2<R>&ray2)
{
    typedef Ray_T2_Ray_T2_pair<R> is_t;
    is_t ispair(&ray1, &ray2);
    switch (ispair.intersection_type()) {
    case is_t::NO:
    default:
        return Object();
    case is_t::POINT: {
        Point_T2<R> pt;
        ispair.intersection(pt);
        return make_object(pt);
    }
    case is_t::SEGMENT: {
        Segment_T2<R> iseg;
        ispair.intersection(iseg);
        return make_object(iseg);
    }
    case is_t::RAY: {
        Ray_T2<R> iray;
        ispair.intersection(iray);
        return make_object(iray);
    }
    }
}

CGAL_END_NAMESPACE

#endif
