// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Ray_2_Segment_2_intersection.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:16:53 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_RAY_T2_SEGMENT_T2_INTERSECTION_H
#define CGAL_RAY_T2_SEGMENT_T2_INTERSECTION_H

#include <CGAL/Ray_T2.h>
#include <CGAL/Segment_T2.h>
#include <CGAL/Point_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Ray_T2_Segment_T2_pair {
public:
    enum Intersection_results {NO, POINT, SEGMENT, RAY};
    Ray_T2_Segment_T2_pair() ;
    Ray_T2_Segment_T2_pair(Ray_T2<R> const *ray,
                           Segment_T2<R> const *seg);
    ~Ray_T2_Segment_T2_pair() {}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
    Intersection_results intersection_type() const;
#else
    Intersection_results intersection_type() const
{
  if (_known)
      return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  Segment_T2<R> seg_result;
  Point_T2<R> rs = _ray->start();
  Point_T2<R> rsp = _ray->second_point();
  Segment_T2<R> segr(rs,rsp);
  Segment_T2_Segment_T2_pair<R> sspair(_seg, &segr);
  switch (sspair.intersection_type()) {
  case Segment_T2_Segment_T2_pair<R>::SEGMENT:
    sspair.intersection(seg_result);
    _intersection_point = seg_result.source();
    _other_point = seg_result.target();
    if (is_at_infinity(_intersection_point))
      _result = RAY;
    else if (is_at_infinity(_other_point))
      _result = RAY;
    else 
      _result = SEGMENT;
    return _result;
  case Segment_T2_Segment_T2_pair<R>::POINT:
    sspair.intersection(_intersection_point);
    _result = POINT;
    return _result;
  case Segment_T2_Segment_T2_pair<R>::NO:
    _result = NO;
    return _result; 
  default:
    CGAL_kernel_assertion(false); // should not be reached:
    _result = NO;
    return _result;
  }
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

  bool                intersection(Point_T2<R> &result) const;
  bool                intersection(Segment_T2<R> &result) const;
  bool                intersection(Ray_T2<R> &result) const;
protected:
  Ray_T2<R> const *      _ray;
  Segment_T2<R> const *  _seg;
  mutable bool                    _known;
  mutable Intersection_results    _result;
  mutable Point_T2<R>         _intersection_point, _other_point;
};

template <class R>
inline bool do_intersect(
    const Ray_T2<R> &p1,
    const Segment_T2<R> &p2)
{
  typedef Ray_T2_Segment_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE


#include <CGAL/Line_T2.h>
#include <CGAL/Line_T2_Line_T2_intersection.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Ray_T2_Segment_T2_pair<R>::Ray_T2_Segment_T2_pair()
{
  _ray = 0;
  _seg = 0;
  _known = false;
}

template <class R>
Ray_T2_Segment_T2_pair<R>::Ray_T2_Segment_T2_pair(
    Ray_T2<R> const *ray, Segment_T2<R> const *seg)
{
  _ray = ray;
  _seg = seg;
  _known = false;
}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Ray_T2_Segment_T2_pair<R>::Intersection_results
Ray_T2_Segment_T2_pair<R>::intersection_type() const
{
  if (_known)
      return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  Segment_T2<R> seg_result;
  Point_T2<R> rs = _ray->start();
  Point_T2<R> rsp = _ray->second_point();
  Segment_T2<R> segr(rs,rsp);
  Segment_T2_Segment_T2_pair<R> sspair(_seg, &segr);
  switch (sspair.intersection_type()) {
  case Segment_T2_Segment_T2_pair<R>::SEGMENT:
    sspair.intersection(seg_result);
    _intersection_point = seg_result.source();
    _other_point = seg_result.target();
    if (is_at_infinity(_intersection_point))
      _result = RAY;
    else if (is_at_infinity(_other_point))
      _result = RAY;
    else
      _result = SEGMENT;
    return _result;
  case Segment_T2_Segment_T2_pair<R>::POINT:
    sspair.intersection(_intersection_point);
    _result = POINT;
    return _result;
  case Segment_T2_Segment_T2_pair<R>::NO:
    _result = NO;
    return _result; 
  default:
    CGAL_kernel_assertion(false); // should not be reached:
    _result = NO;
    return _result;
  }
}
#endif // CGAL_CFG_RETURN_TYPE_BUG_2

template <class R>
bool
Ray_T2_Segment_T2_pair<R>::intersection(Point_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != POINT)
    return false;
  result = _intersection_point;
  return true;
}

template <class R>
bool
Ray_T2_Segment_T2_pair<R>::intersection(Segment_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != SEGMENT)
    return false;
  result = Segment_T2<R>(_intersection_point, _other_point);
  return true;
}

template <class R>
bool
Ray_T2_Segment_T2_pair<R>::intersection(Ray_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != RAY)
    return false;
  Ray_T2<R> r(_intersection_point,_other_point);
  result = r;
  return true;
}

CGAL_END_NAMESPACE



#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Ray_T2<R> &ray, const Segment_T2<R>&seg)
{
  typedef Ray_T2_Segment_T2_pair<R> is_t;
  is_t ispair(&ray, &seg);
  switch (ispair.intersection_type()) {
  case is_t::NO:
  default:
    return Object();
  case is_t::POINT: {
    Point_T2<R> pt;
    ispair.intersection(pt);
    return make_object(pt);
  }
  case is_t::SEGMENT: {
    Segment_T2<R> iseg;
    ispair.intersection(iseg);
    return make_object(iseg);
  }
}
}

template <class R>
class Segment_T2_Ray_T2_pair: public Ray_T2_Segment_T2_pair<R> {
public:
  Segment_T2_Ray_T2_pair(
        Segment_T2<R> const *seg,
        Ray_T2<R> const *ray) :
                          Ray_T2_Segment_T2_pair<R>(ray, seg) {}
};

template <class R>
inline bool do_intersect(
    const Segment_T2<R> &p1,
    const Ray_T2<R> &p2)
{
  typedef Segment_T2_Ray_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

template <class R>
inline Object
intersection(const Segment_T2<R> &seg, const Ray_T2<R> &ray)
{
  return intersection(ray, seg);
}

CGAL_END_NAMESPACE
#endif
