// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/Cartesian/Rotation_rep_2.h,v $
// $Revision: 1.7 $ $Date: 2003/10/21 12:14:21 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Herve Bronnimann
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_ROTATION_REP_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_ROTATION_REP_T2_H

#include <CGAL/rational_rotation.h>

CGAL_BEGIN_NAMESPACE

template < class R >
class Rotation_repT2: public Aff_transformation_rep_baseT2<R>
{
friend class Aff_transformation_repT2<R>;
friend class Translation_repT2<R>;
friend class Scaling_repT2<R>;

public:
  typedef Aff_transformation_rep_baseT2<R> Aff_t_base;
  typedef typename Aff_t_base::FT                        FT;
  typedef typename Aff_t_base::TPoint_T2                 TPoint_T2;
  typedef typename Aff_t_base::TVector_T2                TVector_T2;
  typedef typename Aff_t_base::TDirection_T2             TDirection_T2;
  typedef typename Aff_t_base::TAff_transformation_T2    TAff_transformation_T2;
  typedef Aff_transformation_repT2<R>                    Transformation;
  typedef Translation_repT2<R>                           Translation;
  typedef Rotation_repT2<R>                              Rotation;
  typedef Scaling_repT2<R>                               Scaling;

  Rotation_repT2() {}

  Rotation_repT2(const FT &sinus, const FT &cosinus)
    : sinus_(sinus), cosinus_(cosinus) {}

  Rotation_repT2(const TDirection_T2 &d,
                 const FT &eps_num,
                 const FT &eps_den = FT(1))
  { // FIXME : construction
    FT sin_num;
    FT cos_num;
    FT denom;

    rational_rotation_approximation(d.to_vector().x(),
                                    d.to_vector().y(),
                                    sin_num,
                                    cos_num,
                                    denom,
                                    eps_num,
                                    eps_den);
    sinus_   = sin_num/denom;
    cosinus_ = cos_num/denom;
  }

  TPoint_T2 transform(const TPoint_T2 &p) const
  { // FIXME : construction
    return TPoint_T2(p.w(), cosinus_ * p.x() - sinus_ * p.y(),
                     sinus_ * p.x() + cosinus_ * p.y());
  }

  TVector_T2 transform(const TVector_T2 &v) const
  { // FIXME : construction
    return TVector_T2(v.w(), cosinus_ * v.x() - sinus_ * v.y(),
                    sinus_ * v.x() + cosinus_ * v.y());
  }

  TDirection_T2  transform(const TDirection_T2 &d) const
  { // FIXME : construction
    TVector_T2  v = d.to_vector();
    return TDirection_T2(v.w(),cosinus_ * v.x() - sinus_ * v.y(),
                        sinus_ * v.x() + cosinus_ * v.y());
  }

  TAff_transformation_T2 inverse() const
  { // FIXME : construction
    return TAff_transformation_T2(ROTATION, - sinus_, cosinus_);
  }

  TAff_transformation_T2 operator*(const Aff_t_base &t) const
  {
    return t.compose(*this);
  }

  TAff_transformation_T2 compose(const Translation &t) const
  { // FIXME : construction
    return TAff_transformation_T2(t.translationvector_.w(),
                                 t.translationvector_.x()*cosinus_-
                                 t.translationvector_.y()*sinus_,     
                                 t.translationvector_.x()*sinus_+
                                 t.translationvector_.y()*cosinus_,
                                 cosinus_*t.translationvector_.w(),
                                 sinus_*t.translationvector_.w(),
                                 -sinus_*t.translationvector_.w(),
                                 cosinus_*t.translationvector_.w());
  }

  TAff_transformation_T2 compose(const Rotation &t) const
  { // FIXME : construction
    return TAff_transformation_T2(ROTATION,
                                t.sinus_*cosinus_ + t.cosinus_*sinus_,
                                t.cosinus_*cosinus_-t.sinus_*sinus_ );
  }

  TAff_transformation_T2 compose(const Scaling &t) const
  { // FIXME : construction
    return TAff_transformation_T2(t.scalefactor_*cosinus_,
                                 t.scalefactor_*sinus_,
                                 -t.scalefactor_*sinus_,
                                 t.scalefactor_*cosinus_);
  }

  TAff_transformation_T2 compose(const Transformation &t) const
  { // FIXME : construction
    return TAff_transformation_T2(t.t11,t.t12*cosinus_-t.t13*sinus_,
                                  t.t12*sinus_+t.t13*cosinus_,
                                  cosinus_*t.t22 - sinus_*t.t23,
                                  sinus_*t.t22 + cosinus_*t.t23,
                                  cosinus_*t.t32 + sinus_*t.t33,
                                  sinus_*t.t32 + cosinus_*t.t33);
  }

  bool is_even() const
  {
    return true;
  }

  FT cartesian(int i, int j) const
  { // FIXME : construction
    switch (i)
    {
    case 0: switch (j)
            {
              case 0: return FT(1);
              default: return FT(0);
            }
    case 1: switch (j)
            {
              case 0: return FT(0);
              case 1: return cosinus_;
              case 2: return -sinus_;
            }
    case 2: switch (j)
            {
              case 0: return FT(0);
              case 1: return sinus_;
              case 2: return cosinus_;
            }
    }
    return FT(0);
  }

  std::ostream &print(std::ostream &os) const
  {
    os << "Aff_transformationT2(" << sinus_ << ", " << cosinus_ <<  ")";
    return os;
  }

private:
  FT sinus_, cosinus_;
};

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_ROTATION_REP_T2_H
