// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Segment_2_Iso_rectangle_2_intersection.h,v $
// $Revision: 1.11 $ $Date: 2003/10/21 12:16:53 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SEGMENT_T2_ISO_RECTANGLE_T2_INTERSECTION_H
#define CGAL_SEGMENT_T2_ISO_RECTANGLE_T2_INTERSECTION_H

#include <CGAL/Iso_rectangle_T2.h>
#include <CGAL/Segment_T2.h>
#include <CGAL/Segment_T2_Segment_T2_intersection.h>
#include <CGAL/Point_T2.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Segment_T2_Iso_rectangle_T2_pair {
public:
  enum Intersection_results {NO, POINT, SEGMENT};
  Segment_T2_Iso_rectangle_T2_pair() ;
  Segment_T2_Iso_rectangle_T2_pair(Segment_T2<R> const *seg,
                        Iso_rectangle_T2<R> const *rect) ;

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
  Intersection_results intersection_type() const;

#else
  Intersection_results intersection_type() const
  {
 if (_known)
    return _result;
  _known = true;
  Segment_T2<R> s(_iso->vertex(0),_iso->vertex(1));
  Segment_T2_Segment_T2_pair<R> ss(_seg, &s); 
  Intersection_results ir = intersection_by_side(ss);
  if (ir != N0) {
    _result = ir;
    return _result;
  }
  s = Segment_T2<R>(_iso->vertex(1),_iso->vertex(2));
  ss = Segment_T2_Segment_T2_pair<R>(_seg,&s);
  ir = intersection_by_side(ss);
  if (ir != N0) {
    _result = ir;
    return _result;
  }
  s = Segment_T2<R>(_iso->vertex(2),_iso->vertex(3));
  ss = Segment_T2_Segment_T2_pair<R>(_seg,&s);
  ir = intersection_by_side(ss);
  if (ir != N0) {
    _result = ss.intersection_type();
    return _result;
  }
  s = Segment_T2<R>(_iso->vertex(3),_iso->vertex(0));
  ss = Segment_T2_Segment_T2_pair<R>(_seg,&s);
  _result = ir;
  return _result;
  }
 
#endif // CGAL_CFG_RETURN_TYPE_BUG_2

  bool intersection(Point_T2<R> &result) const;
  bool intersection(Segment_T2<R> &result) const;
  Intersection_results intersection_by_side(
                  Segment_T2_Segment_T2_pair<R> *ss) const;
protected:
  mutable bool                       _known;
  mutable Intersection_results       _result;
  const Segment_T2<R>                     *_seg;
  const Iso_rectangle_T2<R>               *_iso;
  mutable Point_T2<R>                _intersection_point;
  mutable Point_T2<R>                _other_point;
};

template <class R>
inline bool do_intersect(
    const Segment_T2<R> &p1,
    const Iso_rectangle_T2<R> &p2)
{
  typedef Segment_T2_Iso_rectangle_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE

#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(
    const Segment_T2<R> &seg,
    const Iso_rectangle_T2<R> &iso)
{
    typedef Segment_T2_Iso_rectangle_T2_pair<R> is_t;
    is_t ispair(&seg, &iso);
    switch (ispair.intersection_type()) {
    case is_t::NO:
    default:
        return Object();
    case is_t::POINT: {
        Point_T2<R> ipt;
        ispair.intersection(ipt);
        return make_object(ipt);
    }
    case is_t::SEGMENT: {
        Segment_T2<R> iseg;
        ispair.intersection(iseg);
        return make_object(iseg);
    }
    }
}

CGAL_END_NAMESPACE




CGAL_BEGIN_NAMESPACE

template <class R>
Segment_T2_Iso_rectangle_T2_pair<R>::Segment_T2_Iso_rectangle_T2_pair()
{
  _known = false;
}

template <class R>
Segment_T2_Iso_rectangle_T2_pair<R>::
Segment_T2_Iso_rectangle_T2_pair(
        Segment_T2<R> const *seg,
        Iso_rectangle_T2<R> const *iso)
{
  _known = false;
  _iso = iso;
  _seg = seg;
}

template <class R>
typename Segment_T2_Iso_rectangle_T2_pair<R>::Intersection_results
Segment_T2_Iso_rectangle_T2_pair<R>::
intersection_by_side(Segment_T2_Segment_T2_pair<R> *ss) const
{
  Segment_T2<R> seg;
  switch(ss->intersection_type())
  {
    case POINT:
      ss->intersection(_intersection_point);
      _result = POINT;
      return _result;
    case SEGMENT:
      ss->intersection(seg);
      _intersection_point = seg.source();
      _other_point = seg.target();
      _result = SEGMENT;
      return _result;
    case NO:
      _result = NO;
      return NO;
     default: { // should not happen. 
        CGAL_kernel_assertion_msg(false, "Internal CGAL error.");
        _result = NO;
        return _result;
      }
  }
}


#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Segment_T2_Iso_rectangle_T2_pair<R>::Intersection_results
Segment_T2_Iso_rectangle_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  _known = true;
  Segment_T2<R> s(_iso->vertex(0),_iso->vertex(1));
  Segment_T2_Segment_T2_pair<R> ss(_seg, &s); 
  Intersection_results ir = intersection_by_side(&ss);
  if (intersection_by_side(&ss) != 0) {
    _result = ir;
    return _result;
  }
  s = Segment_T2<R>(_iso->vertex(1),_iso->vertex(2));
  ss = Segment_T2_Segment_T2_pair<R>(_seg,&s);
  ir = intersection_by_side(&ss);
  if (ir != 0) {
    _result = ir;
    return _result;
  }
  s = Segment_T2<R>(_iso->vertex(2),_iso->vertex(3));
  ss = Segment_T2_Segment_T2_pair<R>(_seg,&s);
  ir = intersection_by_side(&ss);
  if (ir != 0) {
    _result = ir;
    return _result;
  }
  s = Segment_T2<R>(_iso->vertex(3),_iso->vertex(0));
  ss = Segment_T2_Segment_T2_pair<R>(_seg,&s);
  _result = intersection_by_side(&ss);
  return _result;
  
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

template <class R>
bool Segment_T2_Iso_rectangle_T2_pair<R>::
intersection(Segment_T2<R> &seg) const
{
  if (!_known)
      intersection_type();
  if (_result != SEGMENT)
      return false;
  seg = Segment_T2<R>(_intersection_point, _other_point);
  return true;
}

template <class R> bool Segment_T2_Iso_rectangle_T2_pair<R>::
intersection(Point_T2<R> &pt) const
{
  if (!_known)
    intersection_type();
  if (_result != POINT)
    return false;
  pt = _intersection_point;
  return true;
}

CGAL_END_NAMESPACE



CGAL_BEGIN_NAMESPACE

template <class R>
class Iso_rectangle_T2_Segment_T2_pair:
          public Segment_T2_Iso_rectangle_T2_pair<R> {
public:
  Iso_rectangle_T2_Segment_T2_pair() {}
  Iso_rectangle_T2_Segment_T2_pair(Iso_rectangle_T2<R> const *rect,
                             Segment_T2<R> const *seg)
              :Segment_T2_Iso_rectangle_T2_pair<R> (seg, rect){}
};

template <class R>
inline bool do_intersect(
  const Iso_rectangle_T2<R> &p1,
  const Segment_T2<R> &p2)
{
  typedef Iso_rectangle_T2_Segment_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

template <class R>
inline Object
intersection(
    const Iso_rectangle_T2<R>&iso,
    const Segment_T2<R>&seg)
{
  return intersection(seg, iso);
}

CGAL_END_NAMESPACE

#endif // CGAL_SEGMENT_T2_ISO_RECTANGLE_T2_INTERSECTION_H
