// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Segment_2_Line_2_intersection.h,v $
// $Revision: 1.8 $ $Date: 2003/10/21 12:16:54 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SEGMENT_T2_LINE_T2_INTERSECTION_H
#define CGAL_SEGMENT_T2_LINE_T2_INTERSECTION_H

#include <CGAL/Line_T2.h>
#include <CGAL/Segment_T2.h>
#include <CGAL/Point_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Segment_T2_Line_T2_pair {
public:
  enum Intersection_results {NO, POINT, SEGMENT};
  Segment_T2_Line_T2_pair() ;
  Segment_T2_Line_T2_pair(Segment_T2<R> const *seg,
                          Line_T2<R> const *line);
  ~Segment_T2_Line_T2_pair() {}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
  Intersection_results intersection_type() const;
#else
  Intersection_results intersection_type() const
  {
    if (_known)
      return _result;
    // The non const this pointer is used to cast away const.
    _known = true;
    if (_seg->source() == _seg->target()) {
      if (_line->has_on(_seg->source())) {
        _intersection_point = _seg->source();
        _result = POINT;
      }
      else {
        _intersection_point = _seg->source();
        _result = NO;
      }
      return _result;
    }
    const Line_T2<R> l1 = _seg->supporting_line();
    Line_T2_Line_T2_pair<R> linepair(&l1, _line);
    switch ( linepair.intersection_type()) {
      case Line_T2_Line_T2_pair<R>::TWOPOINTS:
        linepair.intersection(_intersection_point);
        if (_seg->collinear_has_on(_intersection_point) )
          _result = POINT;
        else if (_seg->collinear_has_on(antipode(_intersection_point)) ){
          _result = POINT;
          _intersection_point = antipode(_intersection_point);
        } 
        else 
          _result = NO;
        break;
      case Line_T2_Line_T2_pair<R>::LINE:
        _result = SEGMENT;
        break;
    }
    return _result;
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

  bool                intersection(Point_T2<R> &result) const;
  bool                intersection(Segment_T2<R> &result) const;
protected:
  const Segment_T2<R> *_seg;
  const Line_T2<R>   *_line;
  mutable bool                    _known;
  mutable Intersection_results     _result;
  mutable Point_T2<R>         _intersection_point;
};

template <class R>
inline bool do_intersect(
  const Segment_T2<R> &p1,
  const Line_T2<R> &p2)
{
  typedef Segment_T2_Line_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE

#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Segment_T2<R> &seg, const Line_T2<R> &line)
{
  typedef Segment_T2_Line_T2_pair<R> is_t;
  is_t ispair(&seg, &line);
  switch (ispair.intersection_type()) {
    case is_t::NO:
    default:
      return Object();
    case is_t::POINT: {
      Point_T2<R> pt;
        ispair.intersection(pt);
        return make_object(pt);
    }
    case is_t::SEGMENT:
      return make_object(seg);
    }
}

template <class R>
class Line_T2_Segment_T2_pair: public Segment_T2_Line_T2_pair<R> {
public:
  Line_T2_Segment_T2_pair(
          Line_T2<R> const *line,
          Segment_T2<R> const *seg) :
                              Segment_T2_Line_T2_pair<R>(seg, line) {}
};

template <class R>
inline bool do_intersect(
  const Line_T2<R> &p1,
  const Segment_T2<R> &p2)
{
  typedef Line_T2_Segment_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

template <class R>
inline Object
intersection(const Line_T2<R> &line, const Segment_T2<R> &seg)
{
  return intersection(seg, line);
}

CGAL_END_NAMESPACE


#include <CGAL/Line_T2_Line_T2_intersection.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Segment_T2_Line_T2_pair<R>::Segment_T2_Line_T2_pair()
{
  _seg = 0;
  _line = 0;
  _known = false;
}

template <class R>
Segment_T2_Line_T2_pair<R>::Segment_T2_Line_T2_pair(
  Segment_T2<R> const *seg, Line_T2<R> const *line)
{
  _seg = seg;
  _line = line;
  _known = false;
}

#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Segment_T2_Line_T2_pair<R>::Intersection_results
Segment_T2_Line_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  if (_seg->source() == _seg->target()) {
    if (_line->has_on(_seg->source())) {
      _intersection_point = _seg->source();   
      _result = POINT;
    } 
    else {
      _intersection_point = _seg->source();
      _result = NO;
    }
    return _result;
  }
  const Line_T2<R> &l1 = _seg->supporting_line();
  Line_T2_Line_T2_pair<R> linepair(&l1,_line);
  switch ( linepair.intersection_type()) {
    case Line_T2_Line_T2_pair<R>::TWOPOINTS:
      linepair.intersection(_intersection_point);
      if (_seg->collinear_has_on(_intersection_point) )
        _result = POINT;
      else if (_seg->collinear_has_on(antipode(_intersection_point)) ){
        _result = POINT;
        _intersection_point = antipode(_intersection_point);
      }
      else
        _result = NO;
      break;
    case Line_T2_Line_T2_pair<R>::LINE:
      _result = SEGMENT;
      break;
  }
  return _result;
}

#endif // CGAL_CFG_RETURN_TYPE_BUG_2

template <class R>
bool
Segment_T2_Line_T2_pair<R>::intersection(Point_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != POINT)
    return false;
  result = _intersection_point;
  return true;
}

template <class R>
bool
Segment_T2_Line_T2_pair<R>::intersection(Segment_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != SEGMENT)
    return false;
  result = *_seg;
  return true;
}

CGAL_END_NAMESPACE



#endif
