// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Segment_2_Triangle_2_intersection.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:16:57 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SEGMENT_T2_TRIANGLE_T2_INTERSECTION_H
#define CGAL_SEGMENT_T2_TRIANGLE_T2_INTERSECTION_H

#include <CGAL/Segment_T2.h>
#include <CGAL/Triangle_T2.h>
#include <CGAL/Point_T2.h>
#include <CGAL/Segment_T2_Segment_T2_intersection.h>

CGAL_BEGIN_NAMESPACE

template <class R>
class Segment_T2_Triangle_T2_pair {
public:
  enum Intersection_results {NO, POINT, SEGMENT};
  Segment_T2_Triangle_T2_pair() ;
  Segment_T2_Triangle_T2_pair(Segment_T2<R> const *seg,
                              Triangle_T2<R> const *trian);
  ~Segment_T2_Triangle_T2_pair() {}
#ifdef CGAL_CFG_RETURN_TYPE_BUG_2

  Intersection_results intersection_by_side_triangle
          (Segment_T2_Segment_T2_pair<R> *lspair) const
  {
    Segment_T2<R> seg;
    switch(lspair->intersection_type()){
        case POINT:
        lspair->intersection(_intersection_point);
        _result = POINT;
        return _result;
      case SEGMENT:
        lspair->intersection(seg);
        _intersection_point = seg.source();
        _other_point = seg.target();
        _result = SEGMENT;
        return _result;
      case NO:
        _result = NO;
        return _result;
      default: { // should not happen.
        CGAL_kernel_assertion_msg(false, "Internal CGAL error.");
        _result = NO;
        return _result;
      }
    }
  }

  Intersection_results intersection_type() const
  {
    if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
    _known = true;
    Point_T2<R> v0 = _trian->vertex(0);
    Point_T2<R> v1 = _trian->vertex(1);
    Point_T2<R> v2 = _trian->vertex(2);
    Segment_T2<R> seg0(v0,v1);
    Segment_T2_Segment_T2_pair<R> sspair0(&seg0,_seg);

    if (intersection_by_side_triangle(&sspair0)!=NO){
      if (_seg->has_on(v2))
        _result = SEGMENT;
      return _result;
    }
    Segment_T2<R> seg1(v1,v2);
    Segment_T2_Segment_T2_pair<R> sspair1(&seg1,_seg);

    if (intersection_by_side_triangle(&sspair1)!=NO)
      return _result;
    Segment_T2<R> seg2(v2,v0);
    Segment_T2_Segment_T2_pair<R> sspair2(&seg2,_seg);
    return intersection_by_side_triangle(&sspair2);
}

#else
    Intersection_results intersection_type() const;
#endif // CGAL_CFG_RETURN_TYPE_BUG_2
    Intersection_results intersection_by_side_triangle
                        (Segment_T2_Segment_T2_pair<R> *lspair) const;
    bool                intersection(Point_T2<R> &result) const;
    bool                intersection(Segment_T2<R> &result) const;
protected:
    Segment_T2<R> const *  _seg;
    Triangle_T2<R> const * _trian;
    mutable bool                       _known;
    mutable Intersection_results       _result;
    mutable Point_T2<R>            _intersection_point;
    mutable Point_T2<R>            _other_point;
};

template <class R>
inline bool do_intersect(
  const Segment_T2<R> &p1,
  const Triangle_T2<R> &p2)
{
  typedef Segment_T2_Triangle_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE



#include <CGAL/Line_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>


CGAL_BEGIN_NAMESPACE

template <class R>
Segment_T2_Triangle_T2_pair<R>::
Segment_T2_Triangle_T2_pair()
{
  _known = false;
  _seg = 0;
  _trian = 0;
}

template <class R>
Segment_T2_Triangle_T2_pair<R>::
Segment_T2_Triangle_T2_pair(Segment_T2<R> const *seg,
                          Triangle_T2<R> const *trian)
{
  _known = false;
  _seg = seg;
  _trian = trian;
}

template <class R>
typename Segment_T2_Triangle_T2_pair<R>::Intersection_results
Segment_T2_Triangle_T2_pair<R>::
intersection_by_side_triangle(Segment_T2_Segment_T2_pair<R> *lspair) const
{
  Segment_T2<R> seg;
  switch(lspair->intersection_type()){
    case POINT:
      lspair->intersection(_intersection_point);
      _result = POINT;
      return _result;
    case SEGMENT:
      lspair->intersection(seg);
      _intersection_point = seg.source();
      _other_point = seg.target();
      _result = SEGMENT;
      return _result;
    case NO:
      _result = NO;
      return _result;
    default: { // should not happen.
      CGAL_kernel_assertion_msg(false, "Internal CGAL error.");
      _result = NO;
      return _result;
    }
  }
}



#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Segment_T2_Triangle_T2_pair<R>::Intersection_results
Segment_T2_Triangle_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  Point_T2<R> v0 = _trian->vertex(0);
  Point_T2<R> v1 = _trian->vertex(1);
  Point_T2<R> v2 = _trian->vertex(2);
  Segment_T2<R> seg0(v0,v1);
  Segment_T2_Segment_T2_pair<R> sspair0(&seg0,_seg);

  if (intersection_by_side_triangle(&sspair0)!=NO){
    if (_seg->has_on(v2))
      _result = SEGMENT;
    return _result;
  }
  Segment_T2<R> seg1(v1,v2);
  Segment_T2_Segment_T2_pair<R> sspair1(&seg1,_seg);

  if (intersection_by_side_triangle(&sspair1)!=NO)
    return _result;
  Segment_T2<R> seg2(v2,v0);
  Segment_T2_Segment_T2_pair<R> sspair2(&seg2,_seg);
  return intersection_by_side_triangle(&sspair2);
}
#endif // CGAL_CFG_RETURN_TYPE_BUG_2

template <class R>
bool
Segment_T2_Triangle_T2_pair<R>::
intersection(Point_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != POINT)
    return false;
  result = _intersection_point;
  return true;
}

template <class R>
bool
Segment_T2_Triangle_T2_pair<R>::
intersection(Segment_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != SEGMENT)
    return false;
  result = Segment_T2<R>(_intersection_point, _other_point);
  return true;
}

CGAL_END_NAMESPACE




#include <CGAL/Object.h>

CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Segment_T2<R> &seg, const Triangle_T2<R>&tr)
{
  typedef Segment_T2_Triangle_T2_pair<R> is_t;
  is_t ispair(&seg, &tr);
  switch (ispair.intersection_type()) {
  case is_t::NO:
  default:
      return Object();
  case is_t::POINT: {
      Point_T2<R> pt;
      ispair.intersection(pt);
      return make_object(pt);
  }
  case is_t::SEGMENT: {
      Segment_T2<R> iseg;
      ispair.intersection(iseg);
      return make_object(iseg);
  }
  }
}

template <class R>
class Triangle_T2_Segment_T2_pair
: public Segment_T2_Triangle_T2_pair<R> {
public:
  Triangle_T2_Segment_T2_pair(
          Triangle_T2<R> const *trian,
          Segment_T2<R> const *seg) :
                      Segment_T2_Triangle_T2_pair<R>(seg, trian) {}
};

template <class R>
inline bool do_intersect(
  const Triangle_T2<R> &p1,
  const Segment_T2<R> &p2)
{
  typedef Triangle_T2_Segment_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

template <class R>
inline Object
intersection(const Triangle_T2<R> &tr, const Segment_T2<R> &seg)
{
  return intersection(seg, tr);
}

CGAL_END_NAMESPACE

#endif
