// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/H2/include/CGAL/Homogeneous.h,v $
// $Revision: 1.15 $ $Date: 2003/10/21 12:16:06 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_H
#define CGAL_SIGNED_HOMOGENEOUS_H


#include <CGAL/basic.h>
#include <CGAL/Handle_for_virtual.h>
#include <CGAL/basic_classes.h>
#include <CGAL/user_classes.h>

#include <CGAL/Twotuple.h>
#include <CGAL/Threetuple.h>
#include <CGAL/utility.h>

#include <CGAL/Cartesian.h>


#include <CGAL/definitions_T2.h>
#include <CGAL/utility_T2.h>

#include <CGAL/Point_T1.h>

#include <CGAL/Point_T2.h>
#include <CGAL/Direction_T2.h>
#include <CGAL/Segment_T2.h>
#include <CGAL/Vector_T2.h>
#include <CGAL/Line_T2.h>
#include <CGAL/Ray_T2.h>
#include <CGAL/Triangle_T2.h>
#include <CGAL/Circle_T2.h>
#include <CGAL/Bbox_T2.h>
#include <CGAL/Iso_rectangle_T2.h>
#include <CGAL/Aff_transformation_T2.h>

#include <CGAL/predicates_on_points_T2.h>
#include <CGAL/predicates_on_vectors_T2.h>
#include <CGAL/global_operators_T2.h>
#include <CGAL/global_predicates_T2.h>
#include <CGAL/predicates_on_lines_T2.h>
#include <CGAL/predicates_on_segments_T2.h>
#include <CGAL/predicates_on_directions_T2.h>
#include <CGAL/global_operators_T1.h>

#include <CGAL/distance_predicates_T2.h>
#include<CGAL/intersection_T2.h>

#include <CGAL/basic_constructions_T2.h>
#include <CGAL/basic_constructions_T1.h>

#include <CGAL/Signed_homogeneous/function_objects_T2.h>
#include <CGAL/representation_tags.h>

CGAL_BEGIN_NAMESPACE

template< class R, class FT_ >
struct Signed_homogeneous_base
{
    typedef FT_                                         RT;
    typedef FT_                                         FT;
    typedef Cartesian_tag                               Rep_tag;
    typedef Cartesian_tag                               Kernel_tag;

    typedef CGAL::Object                                Object_T2;


    typedef Point_T2<R>                                 Point_T2;
    typedef Direction_T2<R>			                        Direction_T2;	
    typedef Segment_T2<R>				                        Segment_T2;
    typedef Vector_T2<R>				                        Vector_T2;
    typedef Ray_T2<R>					                          Ray_T2;
    typedef Line_T2<R>					                        Line_T2;
    typedef Point_T1<R>					                        Point_T1;
    typedef Triangle_T2<R>			                        Triangle_T2;
    typedef Circle_T2<R>  			                        Circle_T2;
    typedef Iso_rectangle_T2<R>                         Iso_rectangle_T2;
    typedef Aff_transformation_T2<R>                    Aff_transformation_T2;
    typedef Bbox_T2<R>                                  Bbox_T2;

	// Functors types and access functions.
#define CGAL_Kernel_pred(Y,Z) typedef Signed_homogeneousKernelFunctors::Y<R> Y; \
                              Y Z() const { return Y(); }
#define CGAL_Kernel_cons(Y,Z) CGAL_Kernel_pred(Y,Z)
#include <CGAL/Signed_homogeneous/interface_macros_T2.h>

};


template< class FT_ >
struct Signed_homogeneous
  : public Signed_homogeneous_base< Signed_homogeneous<FT_>, FT_ >
{
    // Number types and representation tag (to avoid ambiguity)
    typedef FT_                                           RT;
    typedef FT_                                           FT;
    typedef Cartesian_tag                                 Rep_tag;
    typedef Cartesian_tag                                 Kernel_tag;

    typedef Signed_homogeneous<FT>                        Self;
    typedef Signed_homogeneous_base<Self,FT>              Kernel_base;

    // Now CGAL::Point_2<R> is only a wrapper around CGAL::PointC2<R>
    // It is necessary to redefine here the classes to ensure that
    // Cartesian<FT>::Point_2 is exactly CGAL::Point_2< Cartesian<FT> >

    typedef typename Kernel_base::Object_T2                Object_T2;
   

    // Note: necessary to qualify Point_T2 by CGAL:: to disambiguate between
    // Point_2 in the current scope (nested within CGAL) and
    // CGAL::Point_T2< R > (which is in the CGAL namespace)

    typedef CGAL::Point_T2<Self>                           Point_T2;
    typedef CGAL::Direction_T2<Self>                       Direction_T2;
    typedef CGAL::Segment_T2<Self>			                   Segment_T2;
    typedef CGAL::Vector_T2<Self>			                     Vector_T2;
    typedef CGAL::Ray_T2<Self>				                     Ray_T2;
    typedef CGAL::Line_T2<Self>			                       Line_T2;
    typedef CGAL::Triangle_T2<Self>			                   Triangle_T2;
    typedef CGAL::Circle_T2<Self>                          Circle_T2;
    typedef CGAL::Iso_rectangle_T2<Self>                   Iso_rectangle_T2;
    typedef CGAL::Point_T1<Self>			                     Point_T1;
    typedef CGAL::Aff_transformation_T2<Self>              Aff_transformation_T2;
    typedef CGAL::Bbox_T2<Self>			                       Bbox_T2;

    typedef CGAL::Handle_for<Triple<Point_T2, Point_T2,
                                       Orientation> >                                   
                                            Circle_handle_T2;
    typedef CGAL::Handle_for<Triple<Point_T2, Point_T1, Point_T1> >  
                                                  Iso_rectangle_handle_T2;
    typedef CGAL::Handle_for_virtual<Aff_transformation_rep_baseT2<Self> >
			                           Aff_transformation_handle_T2;
    typedef CGAL::Handle_for<Triple< Point_T2, Point_T1, Point_T1> > Bbox_handle_T2;

    typedef CGAL::Handle_for<std::pair<Point_T2, Point_T2> >                                 
                                            Line_handle_T2;
    
    typedef CGAL::Cartesian<FT>				                     rep;

  
    // TODO: cleanup
    static   FT make_FT(const RT & num, const RT& denom) { return num/denom;}
    static   FT make_FT(const RT & num)                  { return num;}
    static   RT FT_numerator(const FT &r)                { return r;}
    static   RT FT_denominator(const FT &)               { return RT(1);}


};


CGAL_END_NAMESPACE

//#include <CGAL/iterator_traits_pointer_specs_for_cartesian_kernel.h>

#endif // CGAL_SIGNED_HOMOGENEOUS_H
