// Copyright (c) 1999,2000,2001,2002,2003  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/H2/include/CGAL/Homogeneous/Homogeneous_base.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:16:09 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra, Sylvain Pion
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_BASE_H
#define CGAL_SIGNED_HOMOGENEOUS_BASE_H

#define CGAL_REP_CLASS_DEFINED
#include <CGAL/basic.h>
#include <CGAL/Handle_for_virtual.h>
#include <CGAL/basic_classes.h>
#include <CGAL/user_classes.h>

#include <CGAL/Twotuple.h>
#include <CGAL/Threetuple.h>
#include <CGAL/utility.h>

#include <CGAL/Cartesian.h>
//#include <CGAL/Origin.h>

#include <CGAL/definitions_T2.h>
#include <CGAL/utility_T2.h>

#include <CGAL/Point_T1.h>

#include <CGAL/Point_T2.h>
#include <CGAL/Direction_T2.h>
#include <CGAL/Segment_T2.h>
#include <CGAL/Vector_T2.h>
#include <CGAL/Line_T2.h>
#include <CGAL/Ray_T2.h>
#include <CGAL/Triangle_T2.h>
#include <CGAL/Circle_T2.h>
#include <CGAL/Bbox_T2.h>
#include <CGAL/Iso_rectangle_T2.h>
#include <CGAL/Aff_transformation_T2.h>

#include <CGAL/predicates_on_points_T2.h>
#include <CGAL/predicates_on_vectors_T2.h>
#include <CGAL/global_operators_T2.h>
#include <CGAL/global_predicates_T2.h>
#include <CGAL/predicates_on_lines_T2.h>
#include <CGAL/predicates_on_segments_T2.h>
#include <CGAL/global_operators_T1.h>

#include <CGAL/distance_predicates_T2.h>
#include<CGAL/intersection_T2.h>

#include <CGAL/basic_constructions_T2.h>
#include <CGAL/basic_constructions_T1.h>

#include <CGAL/Signed_homogeneous/function_objects_T2.h>
#include <CGAL/representation_tags.h>

CGAL_BEGIN_NAMESPACE

template < typename K_ >
struct Signed_homogeneous_base
{
    typedef K_                                          Kernel;

    typedef Cartesian_tag                               Rep_tag;
    typedef Cartesian_tag                               Kernel_tag;

  //  typedef CGAL::Object                                Object_2;
   // typedef CGAL::Object                                Object_3;

    typedef Point_T2<Kernel>                             Point_T2;
    typedef Vector_T2<Kernel>                            Vector_T2;
    typedef Direction_T2<Kernel>                         Direction_T2;
    typedef Segment_T2<Kernel>                           Segment_T2;
    typedef Line_T2<Kernel>                              Line_T2;
    typedef Ray_T2<Kernel>                               Ray_T2;
    typedef Triangle_T2<Kernel>                          Triangle_T2;
    typedef Circle_T2<Kernel>                            Circle_T2;
    typedef Iso_rectangle_T2<Kernel>                     Iso_rectangle_T2;
    typedef Aff_transformation_T2<Kernel>                Aff_transformation_T2;

     // Functors types and access functions.
#define CGAL_Kernel_pred(Y,Z) typedef Signed_homogeneousKernelFunctors::Y<Kernel> Y; \
                              Y Z() const { return Y(); }
#define CGAL_Kernel_cons(Y,Z) CGAL_Kernel_pred(Y,Z)

#include <CGAL/Signed_homogeneous/interface_macros_T2.h>

};

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_BASE_H
