// Copyright (c) 2003  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/STL_Extension/include/CGAL/function_objects.h,v $
// $Revision: 1.54 $ $Date: 2003/10/21 12:23:41 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Michael Hoffmann <hoffmann@inf.ethz.ch>
//                 Lutz Kettner <kettner@mpi-sb.mpg.de>
//                 Sylvain Pion <Sylvain.Pion@mpi-sb.mpg.de>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_FUNCTION_OBJECTS_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_FUNCTION_OBJECTS_T2_H

#include <CGAL/functional_base.h>
#include <CGAL/intersection_T2.h>

CGAL_BEGIN_NAMESPACE

namespace Signed_homogeneousKernelFunctors {

// VC++ 7.0 doesn't have a using namespace that works, and for better
// testing, I prefer to use its workaround on all compilers for now.
//#if 0 // !defined CGAL_CFG_USING_NAMESPACE_BUG
  //using namespace CommonKernelFunctors;
//#else

  template <typename K>
  class Compare_angle_with_x_axis_T2
  {
    typedef typename K::Direction_T2  Direction_T2;
    typedef typename K::Point_T2      Point_T2;
    typedef typename K::FT	      FT;
  public:
    typedef Comparison_result        result_type;
    typedef Arity_tag< 2 >           Arity;

    Comparison_result
    operator()(const Direction_T2& d1, const Direction_T2& d2) const
    {
      return compare_angle(d1,d2);
    }
   
    Comparison_result
    operator()(const Point_T2& p, const Point_T2& q) const
    {
      CGAL_kernel_precondition(p.w() == 0 && q.w() == 0);
      Direction_T2 d1(p.w(),p.x(),p.y());
      Direction_T2 d2(q.w(),q.x(),q.y());
      return compare_angle(d1,d2);
    }
  };

  template <typename K>
  class Smaller_T1
  {
    typedef typename K::Point_T1 Point_T1;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Smaller_T1() {}
  
    bool
    operator()( const Point_T1& p, const Point_T1& q) const
    { return is_smaller_T1p(p,q); }
  };

  template <typename K>
  class Equal_T1
  {
    typedef typename K::Point_T1 Point_T1;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Equal_T1() {}
  
    bool
    operator()( const Point_T1& p, const Point_T1& q) const
    { return equal_T1(p,q); }
  };
  
  template <typename K>
  class Smaller_xy_one_range_infinity_T2
  {
    typedef typename K::Point_T2 Point_T2;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Smaller_xy_one_range_infinity_T2() {}
  
    bool
    operator()( const Point_T2& p, const Point_T2& q) const
    { return smaller_xy_one_range_infinity_T2(p, q); }
  };

  template <typename K>
  class Smaller_one_range_infinity_T1
  {
    typedef typename K::Point_T1 Point_T1;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Smaller_one_range_infinity_T1() {}
  
    bool
    operator()( const Point_T1& p, const Point_T1& q) const
    { return smaller_one_range_infinity_T1(p, q); }
  };

  template <typename K>
  class Equal_x_T2
  {
    typedef typename K::Point_T2 Point_T2;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Equal_x_T2() {}
  
    bool
    operator()(const Point_T2& p, const Point_T2& q) const
    { return equal_x(p,q); }
  };

  template <typename K>
  class Equal_y_T2
  {
    typedef typename K::Point_T2 Point_T2;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Equal_y_T2() {}
  
    bool
    operator()(const Point_T2& p, const Point_T2& q) const
    { return equal_y(p,q); }
  };



  template <typename K>
  class Smaller_yx_one_range_infinity_T2
  {
    typedef typename K::Point_T2 Point_T2;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 2 >   Arity;

    Smaller_yx_one_range_infinity_T2() {}
  
    bool
    operator()( const Point_T2& p, const Point_T2& q) const
    { return smaller_yx_one_range_infinity_T2(p, q); }
  };

  template <typename K>
  class Orientation_T2
  {
    typedef typename K::Point_T2        Point_T2;
    public:
      typedef Orientation             result_type;
      typedef Arity_tag< 3 >   Arity;
    //Orientation_T2() {}

      Orientation
      operator()(const Point_T2& r, const Point_T2& p, const Point_T2& q) const
      { return orientation(r,p,q);}
  };


  template <typename K>
  class Collinear_are_ordered_along_line_T2
  {
    typedef typename K::Point_T2        Point_T2;

  public:
    typedef bool             result_type;
    typedef Arity_tag< 3 >   Arity;
    Collinear_are_ordered_along_line_T2() {}

    bool
    operator()(const Point_T2& r, 
               const Point_T2& p,
               const Point_T2& q) const
    { return collinear_are_ordered_along_line(r,p,q);}


  };

  template <typename K>
  class Collinear_T2
  {
    typedef typename K::Point_T2        Point_T2;

  public:
    typedef bool             result_type;
    typedef Arity_tag< 3 >   Arity;
    
    bool
    operator()(const Point_T2& r, 
               const Point_T2& p,
               const Point_T2& q) const
    { return (orientation(r,p,q)==COLLINEAR);}


  };


  template <typename K>
  class Less_rotate_ccw_T2
  {
    typedef typename K::Point_T2        Point_T2;
    typedef typename K::Orientation_T2  Orientation_T2;
    typedef typename K::Collinear_are_ordered_along_line_T2 
    Collinear_are_ordered_along_line_T2;
    Orientation_T2 o;
    Collinear_are_ordered_along_line_T2 co;
  public:
    typedef bool             result_type;
    typedef Arity_tag< 3 >   Arity;

    Less_rotate_ccw_T2() {}
    Less_rotate_ccw_T2(const Orientation_T2& o_, 
                      const Collinear_are_ordered_along_line_T2& co_) 
      : o(o_), co(co_)
    {}

    bool
    operator()(const Point_T2& r, const Point_T2& p, const Point_T2& q) const
    {
      std::cout << "OLHAR \n";
      std::cout << "r = " << r << std::endl;
      std::cout << "p = " << p << std::endl;
      std::cout << "q = " << q << std::endl;
      Orientation ori = o(r, p, q);
      std::cout << "ori = " << ori << std::endl;
      
      if ( ori == LEFT_TURN )
        return true;
      else if ( ori == RIGHT_TURN )
        return false;
      else
      {
        if (p == r) return false;
        if (q == r) return true;
        if (p == q) return false;
        return co( r, q, p);
      }
    }
  };

  template <typename K>
  class Equal_T2
  {
  public:
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    // template to replace n different versions
    template <typename T>
    bool
    operator()(const T& p, const T& q) const
    { return p == q; }
  };

  template <typename K>
  class Less_xy_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Point_T2& p, const Point_T2& q) const
    { return less_xy_T2(p,q); }
  };

  template <typename K>
  class Left_turn_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 3 >   Arity;

    bool
    operator()(const Point_T2& p, 
               const Point_T2& q,
               const Point_T2& r) const
    { return left_turn(p,q,r); }
  };

  template <typename K>
  class Equal_relative_distance_to_omega_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 3 >   Arity;

    bool
    operator()(const Point_T2& p,  const Point_T2& q, const Point_T2& r) const
    { return has_equal_relative_distance_to_omega(p,q,r); }
  };

  template <typename K>
  class Smaller_relative_distance_to_omega_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 3 >   Arity;

    bool
    operator()(const Point_T2& p,  const Point_T2& q, const Point_T2& r) const
    { return has_smaller_relative_distance_to_omega(p,q,r); }
  };

  template <typename K>
  class Smaller_x_one_range_infinity_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Point_T2& p,  const Point_T2& q) const
    { return smaller_x_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Smaller_y_one_range_infinity_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Point_T2& p,  const Point_T2& q) const
    { return smaller_y_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Smaller_equal_x_one_range_infinity_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Point_T2& p, const Point_T2& q) const
    { return smaller_equal_x_one_range_infinity_T2(p,q); }
  };	

  template <typename K>
  class Greater_x_one_range_infinity_T2
  {
    public:
      typedef typename K::Point_T2        Point_T2;
      typedef bool                        result_type;
      typedef Arity_tag< 2 >              Arity;

      bool
      operator()(const Point_T2& p, const Point_T2& q) const
      { return greater_x_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Greater_y_one_range_infinity_T2
  {
    public:
      typedef typename K::Point_T2        Point_T2;
      typedef bool                        result_type;
      typedef Arity_tag< 2 >              Arity;

      bool
      operator()(const Point_T2& p, const Point_T2& q) const
      { return greater_y_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Back_range_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 1 >   Arity;

    bool
    operator()(const Point_T2& p) const
    { return is_at_back_range(p); }
  };

  template <typename K>
  class Antipode_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef Point_T2             result_type;
    typedef  Arity_tag< 1 >   Arity;

    Point_T2
    operator()(const Point_T2& p) const
    { return antipode(p); }
  };

  template <typename K>
  class Antipode_T1
  {
  public:
    typedef typename K::Point_T1        Point_T1;
    typedef Point_T1             result_type;
    typedef  Arity_tag< 1 >   Arity;

    Point_T1
    operator()(const Point_T1& p) const
    { return antipodeT1(p); }
  };

  template <typename K>
  class Front_range_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 1 >   Arity;

    bool
    operator()(const Point_T2& p) const
    { return is_at_front_range(p); }
  };	

  template <typename K>
  class Project_to_the_other_range_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef Point_T2                    result_type;
    typedef Arity_tag< 1 >              Arity;

    Point_T2
    operator()(const Point_T2& p) const
    { return project_to_the_other_range(p); }
  };	

  template <typename K>
  class Y_equal_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Point_T2& p, const Point_T2& q) const
    { return y_equal(p,q); }
  };

  template <typename K>
  class Compute_greater_y_one_range_infinity_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef Point_T2              result_type;
      typedef Arity_tag< 2 >        Arity;

      Point_T2
      operator()(const Point_T2& p, const Point_T2& q) const
      { return compute_greater_y_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Compute_greater_x_one_range_infinity_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef Point_T2              result_type;
      typedef Arity_tag< 2 >        Arity;

      Point_T2
      operator()(const Point_T2& p, const Point_T2& q) const
      { return compute_greater_x_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Compute_smaller_x_one_range_infinity_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef Point_T2              result_type;
      typedef Arity_tag< 2 >        Arity;

      Point_T2
      operator()(const Point_T2& p, const Point_T2& q) const
      { return compute_smaller_x_one_range_infinity_T2(p,q); }
  };

  template <typename K>
  class Compute_smaller_y_one_range_infinity_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef Point_T2              result_type;
      typedef Arity_tag< 2 >        Arity;

      Point_T2
      operator()(const Point_T2& p, const Point_T2& q) const
      { return compute_smaller_y_one_range_infinity_T2(p,q); }
  };


  template <typename K>
  class X_equal_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Point_T2& p, const Point_T2& q) const
    { return x_equal(p,q); }
  };

  template <typename K>
  class Positive_side_line_T2
  {
  public:
    typedef typename K::Point_T2  Point_T2;
    typedef typename K::Line_T2   Line_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Line_T2& l, const Point_T2& p) const
    { return has_on_positive_side_line(l,p); }
  };

  template <typename K>
  class Negative_side_line_T2
  {
  public:
    typedef typename K::Point_T2  Point_T2;
    typedef typename K::Line_T2   Line_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Line_T2& l, const Point_T2& p) const
    { return l.has_on_negative_side(p); }
  };
  
  
  template <typename K>
  class Intersect_segments_T2
  {
  public:
    typedef typename K::Segment_T2  Segment_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Segment_T2& s1, const Segment_T2& s2) const
    { return do_intersect(s1,s2); }
  };

  template <typename K>
  class Less_signed_distance_to_line_T2
  {
  public:
    typedef typename K::Point_T2  Point_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 4 >   Arity;

    bool
    operator()(const Point_T2& p, const Point_T2& q,
               const Point_T2& r, const Point_T2& s) const
    { return has_smaller_signed_distance_to_line(p,q,r,s); }
  };


  template <typename K>
  class Coincident_T2
  {
  public:
    typedef typename K::Line_T2  Line_T2;
    typedef bool             result_type;
    typedef  Arity_tag< 2 >   Arity;

    bool
    operator()(const Line_T2& l1, const Line_T2& l2) const
    { return coincident(l1,l2); }
  };

  template <typename K>
  class Not_at_infinity_T2
  {
  public:
    typedef typename K::Point_T2  Point_T2;
    typedef bool                  result_type;
    typedef  Arity_tag< 1 >       Arity;

    bool
    operator()(const Point_T2& p) const
    { return !is_at_infinity(p); }
  };

  template <typename K>
  class Point_smaller_relative_distance_to_omega_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef Point_T2                    result_type;
    typedef Arity_tag< 3 >              Arity;

    Point_T2
    operator()(const Point_T2& p, 
               const Point_T2& q,
               const Point_T2& r) const
    { 
      if (has_smaller_relative_distance_to_omega(p,q,r))
        return p;
      return q;
    }
  };

  template <typename K>
  class Has_on_segment_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef typename K::Segment_T2      Segment_T2;
    typedef bool                        result_type;
    typedef Arity_tag< 2 >              Arity;

    bool
    operator()(const Segment_T2& s, 
               const Point_T2& p) const
    { 
      return has_on_segment(p,s.source(),s.target());
    }
  };

  template <typename K>
  class Point_on_segment_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool                        result_type;
    typedef Arity_tag< 2 >              Arity;

    bool
    operator()(const Point_T2& p, 
               const Point_T2& q,
               const Point_T2& r) const
    { 
      return has_on_segment(p,q,r);
    }
  };


  template <typename K>
  class Back_range_or_minus_infinite_x_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool                        result_type;
    typedef Arity_tag< 1 >              Arity;

    bool
    operator()(const Point_T2& p) const
    { 
      if (is_at_front_range(p))
        return false;
      if (is_at_back_range(p))
        return true;
      if (sign(p.x()) == -1)
        return true;
      return (sign(p.y())==-1 && p.x()==0);
    }
  };

  template <typename K>
  class Collinear_has_on_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef bool                        result_type;
    typedef Arity_tag< 3 >              Arity;

    bool
    operator()(const Point_T2& p,
               const Point_T2& q,
               const Point_T2& r ) const
    { 
      return collinear_has_on(p,q,r);
    }
  };

  template <typename K>
  class At_infinity_T2
  {
  public:
    typedef typename K::Point_T2  Point_T2;
    typedef bool                  result_type;
    typedef  Arity_tag< 1 >       Arity;

    bool
    operator()(const Point_T2& p) const
    { return is_at_infinity(p); }
  };

  template <typename K>
  class Squared_distance_T2
  {
  public:
    typedef typename K::Point_T1  Point_T1;
    typedef typename K::Point_T2  Point_T2;
    typedef Point_T1              result_type;
    typedef  Arity_tag< 2 >       Arity;

    Point_T1
    operator()(const Point_T2& p, const Point_T2& q) const
    { return squared_distance(p,q); }
  };

 
  template <typename K>
  class Same_orientation_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef bool                  result_type;
      typedef  Arity_tag< 2 >       Arity;

    bool
    operator()(const Point_T2& o,
               const Point_T2& p,
               const Point_T2& q,
               const Point_T2& r) const
    { return same_orientation(o,p,q,r); }
  };

  template <typename K>
  class Ccw_around_point_or_collinear_and_closer_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef bool                  result_type;
      typedef  Arity_tag< 3 >       Arity;

    bool
    operator()(const Point_T2& p,
               const Point_T2& q,
               const Point_T2& r) const
    { return ccw_around_point_or_collinear_and_closer(p,q,r); }
  };

  template <typename K>
  class Compare_x_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef bool                  result_type;
      typedef  Arity_tag< 2 >       Arity;

    bool
    operator()(const Point_T2& p,
               const Point_T2& q) const

    { return compare_x(p,q); }
  };

  template <typename K>
  class Compare_y_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef bool                  result_type;
      typedef  Arity_tag< 2 >       Arity;

    bool
    operator()(const Point_T2& p,
               const Point_T2& q) const

    { return compare_y(p,q); }
  };

  template <typename K>
  class Collinear_between_T2
  {
    public:
      typedef typename K::Point_T2  Point_T2;
      typedef bool                  result_type;
      typedef  Arity_tag< 2 >       Arity;

    bool
    operator()(const Point_T2& p,
               const Point_T2& q,
               const Point_T2& r) const

    { return collinear_between(p,q); }
  };

  template <typename K>
  class Construct_opposite_segment_T2
  {
    typedef typename K::Segment_T2    Segment_T2;
    public:
      typedef Segment_T2       result_type;
      typedef Arity_tag< 1 >   Arity;

      Segment_T2
      operator()(const Segment_T2& s)	const
      { return Segment_T2(s.target(),s.source()); }
  };

  template <typename K>
  class Construct_perpendicular_direction_T2
  {
    public:
      typedef typename K::Point_T2          Point_T2;
      typedef typename K::Direction_T2      Direction_T2;
      typedef Direction_T2                  result_type;
      typedef Arity_tag< 2 >                Arity;

      Direction_T2
      operator()(const  Point_T2 &p, const Point_T2 &q) const
      { return perpendicular_direction(p,q); }
  };

  template <typename K>
  class Construct_squared_radius_T2
  {
    public:
      typedef typename K::Point_T2    Point_T2;
      typedef typename K::Point_T1    Point_T1;
      typedef Point_T1                result_type;
      typedef Arity_tag< 3 >          Arity;

      Point_T1
      operator()(const Point_T2& p, Point_T2& q, Point_T2& r)	const
      {
        Point_T2 c = circumcenter(p,q,r);
        return squared_distance_T2(p,c); 
      }
  };

  template <typename K>
  class Construct_circumcenter_T2
  {
    public:
      typedef typename K::Point_T2    Point_T2;
      typedef Point_T2                result_type;
      typedef Arity_tag< 3 >          Arity;

      Point_T2
      operator()(const Point_T2& p, Point_T2& q, Point_T2& r)const
      { return circumcenter(p,q,r); }
  };

  template <typename K>
  class Construct_triangle_T2
  {
    typedef typename K::Triangle_T2   Triangle_T2;
    typedef typename K::Point_T2      Point_T2;
  public:
    typedef Triangle_T2       result_type;
    typedef Arity_tag< 3 >   Arity;

    Triangle_T2
    operator()() const
    { return Triangle_T2(); }

    Triangle_T2
    operator()( const Point_T2& p, const Point_T2& q,
								const Point_T2& r) 			const
    { return Triangle_T2(p, q, r); }
  };

  template <typename K>
  class Construct_segment_T2
  {
    typedef typename K::Segment_T2    Segment_T2;
    typedef typename K::Point_T2      Point_T2;
  public:
    typedef Segment_T2       result_type;
    typedef Arity_tag< 2 >   Arity;

    Segment_T2
    operator()() const
    { return Segment_T2(); }

    Segment_T2
    operator()( const Point_T2& p, const Point_T2& q)	const
    { return Segment_T2(p,q); }
  };

  template <typename K>
  class Intersect_T2
  {
      typedef typename K::Object_T2    Object_T2;
    public:
      typedef Object_T2         result_type;
      typedef Arity_tag< 2 >   Arity;

    // 25 possibilities, so I keep the template.
    template <class T1, class T2>
    Object_T2
    operator()(const T1& t1, const T2& t2) const
    { return intersection(t1, t2); }
  };

  template <typename K>
  class Construct_line_T2
  {
    typedef typename K::RT                        RT;
    typedef typename K::FT                        FT;
    typedef typename K::Point_T2                  Point_T2;
    typedef typename K::Direction_T2              Direction_T2;
    typedef typename K::Vector_T2                 Vector_T2;
    typedef typename K::Segment_T2                Segment_T2;
    typedef typename K::Ray_T2                    Ray_T2;
    typedef typename K::Line_T2                   Line_T2;
   
  public:
    typedef Line_T2            result_type;
    typedef Arity_tag< 2 >    Arity;

    Construct_line_T2() {}
    
    Line_T2
    operator()() const
    { return Line_T2(); }

    Line_T2
    operator()(const Point_T2& p, const Point_T2& q) const
    { 
      return Line_T2(p,q);
    }

    Line_T2
    operator()(const Point_T2& p, const Direction_T2& d) const
    { 
      return Line_T2(p,d);
    }

    //Line_T2
    //operator()(const Point_T2& p, const Vector_T2& v) const
    //{ 
     // return Line_2(a, b, cc);
    //}

    Line_T2
    operator()(const Segment_T2& s) const
    { return Line_T2(s.source(), s.target()); }

    Line_T2
    operator()(const Ray_T2& r) const
    { return Line_T2(r.start(), r.second_point()); }
  };

  template <typename K>
  class Construct_vector_T2
  {
    typedef typename K::RT            RT;
    typedef typename K::FT            FT;
    typedef typename K::Segment_T2    Segment_T2;
    typedef typename K::Ray_T2        Ray_T2;
    typedef typename K::Line_T2       Line_T2;
    typedef typename K::Vector_T2     Vector_T2;
    typedef typename K::Point_T2      Point_T2;
  public:
    typedef Vector_T2         result_type;
    typedef Arity_tag< 2 >   Arity;

    Vector_T2
    operator()() const
    { return Vector_T2(); }

    Vector_T2
    operator()( const Point_T2& p, const Point_T2& q) const
    { return Vector_T2(p,q); }

    Vector_T2
    operator()( const Segment_T2& s) const
    { return s.to_vector(); }

    Vector_T2
    operator()( const Ray_T2& r) const
    { return r.to_vector(); }

    Vector_T2
    operator()( const Line_T2& l) const
    { return l.to_vector(); }

  };

  template <typename K>
  class Construct_direction_T2
  {
    typedef typename K::Direction_T2     Direction_T2;
    typedef typename K::Vector_T2        Vector_T2;
    typedef typename K::Line_T2          Line_T2;
    typedef typename K::Ray_T2           Ray_T2;
    typedef typename K::Segment_T2       Segment_T2;
    typedef typename K::RT              RT;
  public:
    typedef Direction_T2       result_type;
    typedef Arity_tag< 1 >    Arity;

    Direction_T2
    operator()() const
    { return Direction_T2(); }

    Direction_T2
    operator()(const Vector_T2& v) const
    { return Direction_T2(v); }

    Direction_T2
    operator()(const Line_T2& l) const
    { return Direction_T2(l); }

    Direction_T2
    operator()(const Ray_T2& r) const
    { return Direction_T2(r); }

    Direction_T2
    operator()(const Segment_T2& s) const
    { return Direction_T2(s); }
};

  template <typename K>
  class Construct_vertex_T2
  {
    typedef typename K::Point_T2          Point_T2;
    typedef typename K::Segment_T2        Segment_T2;
    typedef typename K::Iso_rectangle_T2  Iso_rectangle_T2;
    typedef typename K::Triangle_T2       Triangle_T2;
  public:
    typedef Point_T2          result_type;
    typedef Arity_tag< 2 >   Arity;

    const Point_T2 &
    operator()( const Segment_T2& s, int i) const
    { return s.vertex(i); }

    const Point_T2 &
    operator()( const Triangle_T2& t, int i) const
    { return t.vertex(i); }

    const Point_T2
    operator()( const Iso_rectangle_T2& r, int i) const
    { return r.vertex(i); }
  };


  template <typename K>
  class Origin_T2
  {
    typedef typename K::Point_T2          Point_T2;
    typedef typename K::FT                FT;
  public:
    typedef Point_T2          result_type;
    typedef Arity_tag< 0 >   Arity;

    const Point_T2 &
    operator()() const
    { return Point_T2(FT(1),FT(0),FT(0)); }
  };



  template <typename K>
  class Compare_distance_to_point_T2
  {
    typedef typename K::Point_T2          Point_T2;
    public:
      typedef Comparison_result          result_type;
      typedef Arity_tag< 3 >             Arity;

    Comparison_result
    operator()(const Point_T2& p,
               const Point_T2& q,
               const Point_T2& r) {
      return compare_distance_to_point(p,q,r);
    }
  };

  template <typename K>
  class Compare_around_origin_T2
  {
     typedef typename K::Point_T2          Point_T2;
    public:
     typedef bool                 result_type;
     typedef Arity_tag< 4 >       Arity;

     bool
     operator()(const Point_T2& p,
                const Point_T2& q) {
       return compare_around_origin(p,q);
     }
  };

  template <typename K>
  class Collinear_has_on_segment_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef typename K::Segment_T2      Segment_T2;
    typedef bool                        result_type;
    typedef Arity_tag< 2 >              Arity;

    bool
    operator()(const Point_T2& p, const Segment_T2& seg) const
    { 
      return collinear_has_on(p,seg.source(),seg.target());
    }
  };

  template <typename K>
  class Cmp_dist_collinear_origin_T2
  {
  public:
    typedef typename K::Point_T2        Point_T2;
    typedef Comparison_result           result_type;
    typedef Arity_tag< 2 >              Arity;

    Comparison_result
    operator()(const Point_T2& p, const Point_T2& q) const
    { 
      return cmp_dist_collinear_origin(p,q);
    }
  };

  template <typename K>
  class Compare_T1p
  {
  public:
    typedef typename K::Point_T1          Point_T1;
    typedef Comparison_result             result_type;
    typedef Arity_tag< 2 >                Arity;

    Comparison_result
    operator()(const Point_T1& p, const Point_T1& q) const
    { 
      return compare_T1p(p,q);
    }
  };


}
//#endif

CGAL_END_NAMESPACE
#endif
