// Copyright (c) 1997  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Arrangement/include/CGAL/Sweep_line_2/Pmwx_sweep_line_event.h,v $
// $Revision: 1.12 $ $Date: 2003/09/18 10:19:51 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Tali Zvi <talizvi@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_EVENT_INFO_H
#define CGAL_EVENT_INFO_H

CGAL_BEGIN_NAMESPACE

template < class Traits >
class Event_info
{
  typedef typename Traits::Point_T2 Point_T2;
	protected:
		Point_T2 current_event;
		Point_T2 previous_event;
	public:
		Event_info()
		{	}
		
		Event_info(const Point_T2& previous,const Point_T2 &cur)
		{
			current_event = cur;
			previous_event = previous;
		}

		void set_current(const Point_T2& cur)
		{
			current_event = cur;
		}		

		void set_previous(const Point_T2& previous)
		{
			previous_event = previous;
		}

		const Point_T2& get_previous() const
		{
			return previous_event;
		}

		const Point_T2& get_current() const
		{
			return current_event;
		}		
	
		bool operator=(Event_info& ei)
		{
			if (get_previous() != ei.get_previous())
				return false;
			return (get_current() == (ei.get_current()));
		}
};

CGAL_END_NAMESPACE

#endif
