// Copyright (c) 1997  Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Sweep_line_2/include/CGAL/Sweep_line_2/Sweep_line_functors.h,v $
// $Revision: 1.15 $ $Date: 2003/09/18 10:25:50 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Tali Zvi <talizvi@post.tau.ac.il>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SWEEP_LINE_FUNCTORS_H
#define CGAL_SWEEP_LINE_FUNCTORS_H

CGAL_BEGIN_NAMESPACE


template <class Point, class SweepLineTraits_T2>
class Point_less_functor 
{
public:
  typedef SweepLineTraits_T2 Traits;
  
  Point_less_functor(Traits *traits) : m_traits(traits) {}
  
  bool operator()(const Point* e1, const Point* e2) const 
  { 
    return (m_traits->compare_around_origin(*e1,*e2));
  }

private:

  /*! a pointer to a trits class */
  Traits *m_traits;
};


template <class SweepLineTraits_T2, class Subcurve>
class Status_line_curve_less_functor 
{
public:
  typedef SweepLineTraits_T2 Traits;
  typedef typename Traits::Point_T2 Point_T2;
  typedef typename Traits::Segment_T2 Segment_T2;
  typedef typename Traits::Object_T2 Object_T2;
  typedef typename Traits::X_monotone_curve_T2 X_monotone_curve_T2;
  typedef bool (Status_line_curve_less_functor::*func)
                      (const Subcurve*, const Subcurve*) const;


  struct Compare_param {
    Compare_param(Traits *t) : m_compare_func(1), m_traits(t)  {}
    int m_compare_func;
    Traits *m_traits;
  };

  Compare_param *m_compare_param;

  Status_line_curve_less_functor(Compare_param *p) : m_compare_param(p) {
    m_compare[0] = &Status_line_curve_less_functor::compare_at;
    m_compare[1] = &Status_line_curve_less_functor::compare_right;
  }

  bool operator()(const Subcurve* c1, const Subcurve* c2) const {
    return (this->*m_compare[m_compare_param->m_compare_func])(c1, c2);
  }


  bool compare_at(const Subcurve* c1, const Subcurve* c2)  const 
  {
    Point_T2 p = c2->get_last_point();
    if (!m_compare_param->m_traits->compare_around_origin(c1->get_last_point(),  
                                                c2->get_last_point()))
      p = c1->get_last_point();

    Comparison_result r =  m_compare_param->m_traits->curves_compare_dirSweepLine_at_perpendicularDir(
                                                  c1->get_curve(), 
                                                  c2->get_curve(), 
                                                  p);
    if ( r == SMALLER) {
      return true;
    } 
    return false;
  }

  bool compare_right(const Subcurve* c1, const Subcurve* c2)  const 
  {
    
    const X_monotone_curve_T2 &cv1 = c1->get_curve();
    const X_monotone_curve_T2 &cv2 = c2->get_curve();

    Traits *t = m_compare_param->m_traits;
    
    typedef typename Point_T2::FT FT;
    Point_T2 o(FT(1),FT(0),FT(0)), ao(FT(-1),FT(0),FT(0));
   
    //cv1 overlaps sweep line in the current position
    if (t->curve_overlapSweepLine(cv1))
    {
      Point_T2 p = c1->get_top_end();
      if (p == Point_T2(-1,0,0))
        p = c1->get_bottom_end();
      if (t->point_in_PerpendicularDirSweep_range(cv2,p)) {
        if (c1->get_bottom_end() == Point_T2(-1,0,0)) {
           return false;
        }
        if (t->curve_compare_dirSweepLine_at_perpendicularDir(c1->get_top_end(),cv2) 
                                                                        == SMALLER) {
             return true;
        } 
      }
      return false;
    }
  
    if ( t->curve_overlapSweepLine(cv2))
    {
      Point_T2 p = c2->get_top_end();
      if (p == Point_T2(-1,0,0))
        p = c2->get_bottom_end();
      if (t->point_in_PerpendicularDirSweep_range(cv1,p)) {
        if (c2->get_bottom_end() == Point_T2(-1,0,0)) {
          return (t->curve_compare_dirSweepLine_at_perpendicularDir(c2->get_top_end(),cv1)
                                                                       == LARGER);
        }
        if (t->curve_compare_dirSweepLine_at_perpendicularDir(c2->get_bottom_end(),cv1) 
                                                                      == LARGER) {
          return true;
        } 
      }
      return false;
    }
    Point_T2 p = c2->get_last_point();
    if (!m_compare_param->m_traits->compare_around_origin(c1->get_last_point(),  
                                                c2->get_last_point()))
      p = c1->get_last_point();

    // non of the curves is vertical...
    //std::cout << "ENTROU compare_right5\n"; 
    Comparison_result r =  t->curves_compare_dirSweepLine_at_perpendicularDir(
                                                  c1->get_curve(), 
                                                  c2->get_curve(), 
                                                  p);
    
    if (r == EQUAL) {
     if ((p == o) || (p == ao)) 
        p = c1->get_SweepLinePos();
     r = t->curves_compare_forward(c1->get_curve(), 
                                    c2->get_curve(), 
                                    p);
    }
    if ( r == SMALLER) {
      return true;
    } 
    if ( r == LARGER ) {
      return false;
    }

    // r = EQUAL
    return ( c1->getId() < c2->getId() );
  }
   

  
  
private:
  func m_compare[2];


};

CGAL_END_NAMESPACE

#endif // CGAL_SWEEP_LINE_FUNCTORS_H
