// Copyright (c) 1998  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Distance_2/include/CGAL/squared_distance_2_1.h,v $
// $Revision: 1.19 $ $Date: 2003/10/21 12:15:21 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
//                 Michel Hoffmann <hoffmann@inf.ethz.ch>
//                 Andreas Fabri <Andreas.Fabri@geometryfactory.com>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_T1_CONSTRUCTIONS_T2_H
#define CGAL_T1_CONSTRUCTIONS_T2_H

//#include<CGAL/predicates_on_lines_T2.h>

CGAL_BEGIN_NAMESPACE

template < class R >
inline
Point_T1<R>
squared_distance(const Point_T2<R> &p,
                 const Point_T2<R> &q)
{
	CGAL_kernel_precondition_msg( ((p.w() != 0) && (q.w() != 0)),
  "This function is defined only for proper points!");
	return Point_T1<R>(sign(p.w()*q.w()) * CGAL_NTS square(p.w()*q.w()),
                    (CGAL_NTS square(p.x()*q.w()-q.x()*p.w()) + 
                     CGAL_NTS square(p.y()*q.w()-q.y()*p.w())));
}

template < class R >
inline
Point_T1<R>
scaled_distance_to_line(const Point_T2<R> &p,
                        const Line_T2<R> &l)
{
  
  return scaled_distance_to_lineT2(l.W(),l.X(),l.Y(),p);
}

template < class R >
inline
Point_T1<R>
scaled_distance_to_lineT2( const typename R::FT &lw, 
                           const typename R::FT &lx, 
                           const typename R::FT &ly,
                           const Point_T2<R> &p)
{
	typedef typename R::FT	FT;
	FT d = lw*p.w() + lx*p.x() + ly*p.y();
	if (d == 0)
		return Point_T1<R>(FT(1),d);
	//the infinite line	
  if (lx == 0 && ly == 0) 
		return Point_T1<R>(FT(0),d);              
  return Point_T1<R>(abs(p.w()),d);
}




CGAL_END_NAMESPACE

#endif
