// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/Cartesian/Translation_rep_2.h,v $
// $Revision: 1.8 $ $Date: 2003/10/21 12:14:24 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Herve Bronnimann
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_TRANSLATION_REP_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_TRANSLATION_REP_T2_H

CGAL_BEGIN_NAMESPACE

template < class R >
class Translation_repT2 : public Aff_transformation_rep_baseT2<R>
{
friend class Aff_transformation_repT2<R>;
friend class Rotation_repT2<R>;
friend class Scaling_repT2<R>;

public:
  typedef typename R::FT                              FT;
  typedef Aff_transformation_rep_baseT2<R>            Aff_t_base;
  typedef Aff_transformation_repT2<R>                 Transformation;
  typedef Translation_repT2<R>                        Translation;
  typedef Rotation_repT2<R>                           Rotation;
  typedef Scaling_repT2<R>                            Scaling;
  typedef typename Aff_t_base::TPoint_T2              TPoint_T2;
  typedef typename Aff_t_base::TVector_T2             TVector_T2;
  typedef typename Aff_t_base::TDirection_T2          TDirection_T2;
  typedef typename Aff_t_base::TAff_transformation_T2 TAff_transformation_T2;

  Translation_repT2()
  {}

  Translation_repT2(const TVector_T2 &tv)
    : translationvector_(tv)
  {}

  TPoint_T2 transform(const TPoint_T2 &p) const
                   { return p + translationvector_; }
  TVector_T2 transform(const TVector_T2 &v) const { return v; }
  TDirection_T2 transform(const TDirection_T2 &d) const { return d; }

  TAff_transformation_T2 operator*(const Aff_t_base &t) const
  {
    return t.compose(*this);
  }

  TAff_transformation_T2 compose(const Translation &t) const
  {
    return TAff_transformation_T2(TRANSLATION,
                                translationvector_ + t.translationvector_);
  }

  TAff_transformation_T2 compose(const Rotation &t) const
  {
    return TAff_transformation_T2(translationvector_.w(),
                                  translationvector_.x(),
                                  translationvector_.y(),
                                  translationvector_.w() * t.cosinus_,
                                  translationvector_.w() * t.sinus_,
                                  -t.sinus_ * translationvector_.w(),
                                  t.cosinus_ * translationvector_.w());
  }

  TAff_transformation_T2 compose(const Scaling &t) const
  {
    return TAff_transformation_T2(translationvector_.w(),
                                 translationvector_.x(),
                                 translationvector_.y(),             
                                 t.scalefactor_*translationvector_.w(),
                                 FT(0),
                                 FT(0),       
                                 t.scalefactor_*translationvector_.w());
  }

  TAff_transformation_T2 compose(const Transformation &t) const
  {
    return TAff_transformation_T2(t.t11 * translationvector_.w(),
                                 t.t11 * translationvector_.x() +
                                 t.t12 * translationvector_.w(),
                                 t.t11 * translationvector_.y()+
                                 t.t13 * translationvector_.w(),
                                 t.t22 * translationvector_.w(),
                                 t.t23 * translationvector_.w(),   
                                 t.t32 * translationvector_.w(),
                                 t.t33 * translationvector_.w());
  }

  TAff_transformation_T2 inverse() const
  {
    return TAff_transformation_T2(TRANSLATION, -translationvector_);
  }

  bool is_even() const
  {
    return true;
  }

  FT cartesian(int i, int j) const
  {
    if (j==i) return translationvector_.w();
    if (i!=0) return FT(0);
    return translationvector_[i];
  }

  std::ostream &print(std::ostream &os) const
  {
    os << "Aff_transformationT2(VectorT2(" << translationvector_.w() << ", "   
       << translationvector_.x()  << ", "
       << translationvector_.y()  <<  "))";
    return os;
  }

private:
  TVector_T2   translationvector_;
};

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_TRANSLATION_REP_T2_H
