// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Intersections_2/include/CGAL/Triangle_2_Triangle_2_intersection.h,v $
// $Revision: 1.7 $ $Date: 2003/10/21 12:17:01 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Geert-Jan Giezeman
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGLE_T2_TRIANGLE_T2_INTERSECTION_H
#define CGAL_TRIANGLE_T2_TRIANGLE_T2_INTERSECTION_H

#include <CGAL/Segment_T2.h>
#include <CGAL/Triangle_T2.h>

CGAL_BEGIN_NAMESPACE

template <class R>
struct Pointlist_T2_rec_ {
  Pointlist_T2_rec_ *next;
  Point_T2<R> point;
  Oriented_side side;
};

template <class R>
struct Pointlist_T2_ {
  int size;
  Pointlist_T2_rec_<R> *first;
  Pointlist_T2_() ;
  void _finalize_list();
  ~Pointlist_T2_() ;
};

template <class R>
class Triangle_T2_Triangle_T2_pair {
public:
  enum Intersection_results {NO, POINT, SEGMENT, TRIANGLE, POLYGON};
  Triangle_T2_Triangle_T2_pair() ;
  Triangle_T2_Triangle_T2_pair(
                       Triangle_T2<R> const *trian1,
                       Triangle_T2<R> const *trian2) ;
  ~Triangle_T2_Triangle_T2_pair() {}
#ifdef CGAL_CFG_RETURN_TYPE_BUG_2
  Intersection_results intersection_type() const
  {
    if (_known)
    return _result;
    // The non const this pointer is used to cast away const.
    _known = true;
    _init_list(_pointlist, *_trian1);
    if (_trian2->is_degenerate()) {
      // _not_implemented();
      CGAL_kernel_assertion(false);
    }else {
      Point_T2<R> v0 =_trian2->vertex(0);
      Point_T2<R> v1 =_trian2->vertex(1);
      Point_T2<R> v2 =_trian2->vertex(2);

      if  (!_trian2->has_on_unbounded_side(_trian1->vertex(0)) &&
           !_trian2->has_on_unbounded_side(_trian1->vertex(1)) &&
           !_trian2->has_on_unbounded_side(_trian1->vertex(2))){
        _result = TRIANGLE;
        return _result;
      }
      else if (!_trian1->has_on_unbounded_side(v0) &&
               !_trian1->has_on_unbounded_side(v1) &&
               !_trian1->has_on_unbounded_side(v2)){
        _pointlist._finalize_list();
        _init_list(_pointlist, *_trian2);
        _result = TRIANGLE;
        return _result;
      }
      if (_trian2->orientation() == COUNTERCLOCKWISE) {
        // counterclockwise triangle
        _cut_off(_pointlist, v0, v1);
        _cut_off(_pointlist, v1, v2);
        _cut_off(_pointlist, v2, v0);
      } else {
        _cut_off(_pointlist, v1, v0);
        _cut_off(_pointlist, v0, v2);
        _cut_off(_pointlist, v2, v1);
      }
    }
    switch (_pointlist.size) {
      case 0:
        _result = NO;
        break;
      case 1:
        _result = POINT;
        break;
      case 2:
        _result = SEGMENT;
        break;
      case 3:
        _result = TRIANGLE;
        break;
      default:
        _result = POLYGON;
     }
     return _result;
   }

#else
    Intersection_results intersection_type() const;
#endif // CGAL_CFG_RETURN_TYPE_BUG_2
  bool                intersection(Point_T2<R> &result) const;
  bool                intersection(Segment_T2<R> &result) const;
  bool                intersection(Triangle_T2<R> &result) const;
  bool                intersection(/*Polygon_T2<R> &result*/) const;
  int                 vertex_count() const;
  Point_T2<R>     vertex(int i) const;
protected:
  Triangle_T2<R> const*   _trian1;
  Triangle_T2<R> const *  _trian2;
  mutable bool                    _known;
  mutable Intersection_results    _result;
  mutable Pointlist_T2_<R>    _pointlist;
};

template <class R>
inline bool do_intersect(
    const Triangle_T2<R> &p1,
    const Triangle_T2<R> &p2)
{
  typedef Triangle_T2_Triangle_T2_pair<R> pair_t;
  pair_t pair(&p1, &p2);
  return pair.intersection_type() != pair_t::NO;
}

CGAL_END_NAMESPACE



#include <CGAL/Line_T2.h>
#include <CGAL/utils.h>
#include <CGAL/number_utils.h>
#include <vector>

CGAL_BEGIN_NAMESPACE

template <class R>
Pointlist_T2_<R>::Pointlist_T2_()
{
  size = 0;
  first = 0;
}

template <class R>
void Pointlist_T2_<R>::_finalize_list()
{ 
  Pointlist_T2_rec_<R> *cur;
  for (int i=0; i<size; i++) {
    cur = first;
    first = cur->next;
    delete cur;
  }
}

template <class R>
Pointlist_T2_<R>::~Pointlist_T2_()
{
  _finalize_list();
}




template <class R>
void _init_list(Pointlist_T2_<R> &list,
                const Triangle_T2<R> &trian)
{
  // check on degeneracies of trian.

    list.size = 3;
    list.first = 0;
    for (int i=0; i<3; i++) {
      Pointlist_T2_rec_<R> *newrec =
                new Pointlist_T2_rec_<R>;
      newrec->next = list.first;
      list.first = newrec;
      newrec->point = trian[i];
    }

}

CGAL_END_NAMESPACE


CGAL_BEGIN_NAMESPACE

template <class R>
void _cut_off(Pointlist_T2_<R> &list,
              const Point_T2<R> &p1,
              const Point_T2<R> &p2)
{
  int i;
  int add = 0;
  Segment_T2<R> scutter(p1,p2);
  Pointlist_T2_rec_<R> *cur, *last=0, *newrec;
  for (i=0, cur = list.first; i<list.size; i++, cur = cur->next) {
      cur->side = Oriented_side(orientation(p1,p2,cur->point));
      last = cur;
  }
  for (cur = list.first, i=0; i<list.size; i++, cur = cur->next) {
    if ((cur->side == ON_POSITIVE_SIDE
         && last->side == ON_NEGATIVE_SIDE)
         || (cur->side == ON_NEGATIVE_SIDE
         && last->side == ON_POSITIVE_SIDE)) {
            // add a vertex after cur
      
      Segment_T2<R> s(cur->point, last->point);
      Segment_T2_Segment_T2_pair<R> segpair(&scutter,  &s);
      if (segpair.intersection_type() ==
                            Segment_T2_Segment_T2_pair<R>::POINT) {
        newrec = new Pointlist_T2_rec_<R>;
        newrec->next = last->next;
        last->next = newrec;
        newrec->side = ON_ORIENTED_BOUNDARY;
        segpair.intersection(newrec->point);
        add++;

      }
    }
    last = cur;
  }
  CGAL_kernel_assertion(add <= 2);
  Pointlist_T2_rec_<R> **curpt;
  curpt = &list.first;
  while (*curpt != 0) {
    cur = *curpt;
    if (cur->side == ON_NEGATIVE_SIDE) {
      add--;
      *curpt = cur->next;
      delete cur;
    } else {
      curpt = &cur->next;
    }
  }
  if (list.size == 2 && add == 1) {
    add = 0;
    cur = list.first;
    if (cur->side == ON_ORIENTED_BOUNDARY) {
      list.first = cur->next;
      delete cur;
    } else {
      last = cur;
      cur = cur->next;
      last->next = cur->next;
      delete cur;
    }
  }
  list.size += add;
}

template <class R>
Triangle_T2_Triangle_T2_pair<R>::
Triangle_T2_Triangle_T2_pair()
{
  _trian1 = 0;
  _trian2 = 0;
  _known = false;
}

template <class R>
Triangle_T2_Triangle_T2_pair<R>::
Triangle_T2_Triangle_T2_pair(Triangle_T2<R> const *trian1,
         Triangle_T2<R> const *trian2)
{
  _trian1 = trian1;
  _trian2 = trian2;
  _known = false;
}

//#ifndef CGAL_CFG_RETURN_TYPE_BUG_2
template <class R>
typename Triangle_T2_Triangle_T2_pair<R>::Intersection_results
Triangle_T2_Triangle_T2_pair<R>::intersection_type() const
{
  if (_known)
    return _result;
  // The non const this pointer is used to cast away const.
  _known = true;
  _init_list(_pointlist, *_trian1);
  if (_trian2->is_degenerate()) {
    // _not_implemented();
    CGAL_kernel_assertion(false);
  }else {
    Point_T2<R> v0 =_trian2->vertex(0);
    Point_T2<R> v1 =_trian2->vertex(1);
    Point_T2<R> v2 =_trian2->vertex(2);
    if  (!_trian2->has_on_unbounded_side(_trian1->vertex(0)) &&
         !_trian2->has_on_unbounded_side(_trian1->vertex(1)) &&
         !_trian2->has_on_unbounded_side(_trian1->vertex(2))){
      _result = TRIANGLE;
      return _result;
    }
    else if (!_trian1->has_on_unbounded_side(v0) &&
             !_trian1->has_on_unbounded_side(v1) &&
             !_trian1->has_on_unbounded_side(v2)){
      _pointlist._finalize_list();
      _init_list(_pointlist, *_trian2);
      _result = TRIANGLE;
      return _result;
    }
    if (_trian2->orientation() == COUNTERCLOCKWISE) {
      // counterclockwise triangle
      _cut_off(_pointlist, v0, v1);
      _cut_off(_pointlist, v1, v2);
      _cut_off(_pointlist, v2, v0);

    } else {
      _cut_off(_pointlist, v1, v0);
      _cut_off(_pointlist, v0, v2);
      _cut_off(_pointlist, v2, v1);
    }
  }
  switch (_pointlist.size) {
    case 0:
      _result = NO;
      break;
    case 1:
      _result = POINT;
      break;
    case 2:
      _result = SEGMENT;
      break;
    case 3:
      _result = TRIANGLE;
      break;
    default:
      _result = POLYGON;
   }
   return _result;
}


//#endif


template <class R>
bool
Triangle_T2_Triangle_T2_pair<R>::intersection(
        /* Polygon_2<R> &result */) const
{
  if (!_known)
    intersection_type();
  if (_result != TRIANGLE  &&  _result != POLYGON)
    return false;
  Pointlist_T2_rec_<R> *cur;
  int i;
  return true;
}

template <class R>
int
Triangle_T2_Triangle_T2_pair<R>::vertex_count() const
{
  CGAL_kernel_assertion(_known);
  return _pointlist.size;
}

template <class R>
Point_T2<R>
Triangle_T2_Triangle_T2_pair<R>::vertex(int n) const
{
  CGAL_kernel_assertion(_known);
  CGAL_kernel_assertion(n >= 0 && n < _pointlist.size);
  Pointlist_T2_rec_<R> *cur;
  int k;
  for (k=0, cur = _pointlist.first;
    k < n;
    k++, cur = cur->next) {
  }
  return cur->point;
}

template <class R>
bool
Triangle_T2_Triangle_T2_pair<R>::intersection(
        Triangle_T2<R> &result) const
{
  if (!_known)
    intersection_type();
  if (_result != TRIANGLE)
    return false;
  result = Triangle_T2<R>(_pointlist.first->point,
                          _pointlist.first->next->point,
                          _pointlist.first->next->next->point);
  return true;
}

template <class R>
bool
Triangle_T2_Triangle_T2_pair<R>::intersection(
        Segment_T2<R> &seg) const
{
  if (!_known)
    intersection_type();
  if (_result != SEGMENT)
    return false;
  seg = Segment_T2<R>(_pointlist.first->point,
                    _pointlist.first->next->point);
  return true;
}

template <class R>
bool
Triangle_T2_Triangle_T2_pair<R>::intersection(
        Point_T2<R> &pt) const
{
  if (!_known)
    intersection_type();
  if (_result != POINT)
    return false;
  pt = _pointlist.first->point;
  return true;
}

CGAL_END_NAMESPACE



CGAL_BEGIN_NAMESPACE

template <class R>
Object
intersection(const Triangle_T2<R> &tr1, const Triangle_T2<R >&tr2)
{
  typedef Triangle_T2_Triangle_T2_pair<R> is_t;
  is_t ispair(&tr1, &tr2);
  switch (ispair.intersection_type()) {
    case is_t::NO:
    default:
      return Object();
    case is_t::POINT: {
      Point_T2<R> pt;
      ispair.intersection(pt);
      return make_object(pt);
    }
    case is_t::SEGMENT: {
      Segment_T2<R> iseg;
      ispair.intersection(iseg);
      return make_object(iseg);
    }
    case is_t::TRIANGLE: {
      Triangle_T2<R> itr;
      ispair.intersection(itr);
      return make_object(itr);
    }
    case is_t::POLYGON: {
      typedef CGAL_STD::vector<Point_T2<R> > Container;
      Container points(ispair.vertex_count());
      for (int i =0; i < ispair.vertex_count(); i++) {
          points[i] = ispair.vertex(i);
      }
      return make_object(points);
    }
  }
}

CGAL_END_NAMESPACE

#endif
