// Copyright (c) 1997  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Triangulation_2.h,v $
// $Revision: 1.120 $ $Date: 2003/09/18 10:26:08 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Olivier Devillers, Mariette Yvinec
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGULATION_T2_H
#define CGAL_TRIANGULATION_T2_H

#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <utility>
#include <iostream>

#include <CGAL/iterator.h>
#include <CGAL/Iterator_project.h>
#include <CGAL/function_objects.h>

#include <CGAL/Triangulation_short_names_2.h>
#include <CGAL/triangulation_assertions.h>
#include <CGAL/Triangulation_utils_2.h>

#include <CGAL/Triangulation_data_structure_T2.h>
#include <CGAL/Triangulation_vertex_base_2.h>
#include <CGAL/Triangulation_ds_face_base_T2.h>
#include <CGAL/Triangulation_face_base_2.h>
#include <CGAL/Triangulation_line_face_circulator_T2.h>
#include <CGAL/Triangulation_traits_T2.h>

CGAL_BEGIN_NAMESPACE
template < class Gt, class Tds > class Triangulation_T2;
template < class Gt, class Tds > std::istream& operator>>
    (std::istream& is, Triangulation_T2<Gt,Tds> &tr);
template < class Gt, class Tds >  std::ostream& operator<<
  (std::ostream& os, const Triangulation_T2<Gt,Tds> &tr);


template < class Gt,
           class Tds = Triangulation_data_structure_T2 <
                             Triangulation_vertex_base_2<Gt>,
                             Triangulation_face_base_2<Gt, Triangulation_ds_face_base_T2<> > > >
class Triangulation_T2
  : public Triangulation_cw_ccw_2
{
  friend std::istream& operator>> CGAL_NULL_TMPL_ARGS
                (std::istream& is, Triangulation_T2 &tr);
  typedef Triangulation_T2<Gt,Tds>             Self;

public:
  typedef Tds                                        Triangulation_data_structure;
  typedef Gt                                         Geom_traits;
  typedef typename Gt::K                             K;
  typedef typename Geom_traits::Point_2              Point;
  typedef typename Geom_traits::Point_T1             Point_T1;
  typedef typename Geom_traits::Segment_2            Segment;
  typedef typename Geom_traits::Triangle_2           Triangle;
  typedef typename Geom_traits::Line_2               Line;
  typedef typename Geom_traits::Compare_T1p          Compare_T1p;

  typedef typename Tds::Vertex                       Vertex;
  typedef typename Tds::Face                         Face;
  typedef typename Tds::Edge                         Edge;
  typedef typename Tds::Vertex_handle                Vertex_handle;
  typedef typename Tds::Face_handle                  Face_handle;

  typedef typename Tds::Face_circulator              Face_circulator;
  typedef typename Tds::Vertex_circulator            Vertex_circulator;
  typedef typename Tds::Edge_circulator              Edge_circulator;

  typedef typename Tds::Face_iterator                All_faces_iterator;
  typedef typename Tds::Edge_iterator                All_edges_iterator;
  typedef typename Tds::Vertex_iterator              All_vertices_iterator;

  typedef std::vector<Segment>                       vector_Segment;


 // This class is used to generate the Finite_*_iterators.
 class Infinite_tester
 {
    const Triangulation_T2 *t;
  public:
    Infinite_tester() {}
   /* it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << it->vertex(j)->point() << inini*/

  Infinite_tester(const Triangulation_T2 *tr)	  : t(tr) {}

    bool operator()(const All_vertices_iterator & vit) const  {
      return t->is_infinite(vit);
    }
    bool operator()(const All_faces_iterator & fit ) const {
      return t->is_infinite(fit);
    }
    bool operator()(const All_edges_iterator & eit) const {
      return t->is_infinite(eit);
    }
 };

  //We derive in order to add a conversion to handle.
  class Finite_vertices_iterator :
    public Filter_iterator<All_vertices_iterator, Infinite_tester>
  {
    typedef Filter_iterator<All_vertices_iterator, Infinite_tester> Base;
    typedef Finite_vertices_iterator                          Self;
  public:
    Finite_vertices_iterator() : Base() {}
    Finite_vertices_iterator(const Base &b) : Base(b) {}
    Self & operator++() { Base::operator++(); return *this; }
    Self & operator--() { Base::operator--(); return *this; }
    Self operator++(int) { Self tmp(*this); ++(*this); return tmp; }
    Self operator--(int) { Self tmp(*this); --(*this); return tmp; }
    operator Vertex_handle() const { return Base::base(); }
  };

  class Finite_faces_iterator
    : public Filter_iterator<All_faces_iterator, Infinite_tester>
  {
    typedef Filter_iterator<All_faces_iterator, Infinite_tester> Base;
    typedef Finite_faces_iterator                           Self;
  public:
    Finite_faces_iterator() : Base() {}
    Finite_faces_iterator(const Base &b) : Base(b) {}
    Self & operator++() { Base::operator++(); return *this; }
    Self & operator--() { Base::operator--(); return *this; }
    Self operator++(int) { Self tmp(*this); ++(*this); return tmp; }
    Self operator--(int) { Self tmp(*this); --(*this); return tmp; }
    operator Face_handle() const { return Base::base(); }
  };


  typedef Filter_iterator<All_edges_iterator,
                           Infinite_tester>
                                               Finite_edges_iterator;

  //for backward compatibility
  typedef Finite_faces_iterator                Face_iterator;
  typedef Finite_edges_iterator                Edge_iterator;
  typedef Finite_vertices_iterator             Vertex_iterator;
  typedef typename std::vector<Face_handle>::iterator
                                               Degenerate_faces_iterator;

  typedef Triangulation_line_face_circulator_T2<Gt,Tds>  Line_face_circulator;

  // Auxiliary iterators for convenience
  // do not use default template argument to please VC++
  typedef Project_point<Vertex>                           Proj_point;
  typedef Iterator_project<Finite_vertices_iterator,
                           Proj_point,
                           const Point&,
                           const Point*,
                           CGAL_CLIB_STD::ptrdiff_t,
                           std::bidirectional_iterator_tag>  Point_iterator;

  typedef Point                value_type; // to have a back_inserter
  typedef const value_type&    const_reference;
  typedef value_type&          reference;


  enum Locate_type {VERTEX=0,
                    EDGE, //1
                    FACE, //2
                    OUTSIDE_CONVEX_HULL, //3
                    OUTSIDE_AFFINE_HULL}; //4

protected:
  Gt  _gt;
  Tds _tds;
  Vertex_handle _infinite_vertex;
  bool ch_is_defined;
  std::vector<Face_handle> vector_degenerate_faces;



public:

  // CONSTRUCTORS
  Triangulation_T2(const Geom_traits& geom_traits=Geom_traits());
  Triangulation_T2(const Triangulation_T2<Gt,Tds> &tr);

  //Assignement
  Triangulation_T2 &operator=(const Triangulation_T2 &tr);

  //Helping
  void copy_triangulation(const Triangulation_T2 &tr);
  void swap(Triangulation_T2 &tr);
  void clear();

  //ACCESS FUNCTIONs
  const Geom_traits& geom_traits() const { return _gt;}
  const Tds & tds() const  { return _tds;}
  Tds & tds()   { return _tds;}
  int dimension() const { return _tds.dimension();}
  int number_of_vertices() const {return _tds.number_of_vertices() - 1;}

  int number_of_faces() const;
  Vertex_handle infinite_vertex() const;
  Vertex_handle finite_vertex() const;
  Face_handle infinite_face() const;
  Infinite_tester infinite_tester() const;

  //SETTING
  void set_infinite_vertex(const Vertex_handle& v) {_infinite_vertex=v;}

  // CHECKING
  bool is_valid(bool verbose = false, int level = 0) const;

  // TEST INFINITE FEATURES AND OTHER FEATURES
  bool is_infinite(Face_handle f) const;
  bool is_infinite(Vertex_handle v) const;
  bool is_infinite(Face_handle f, int i) const;
  bool is_infinite(const Edge& e) const;
  bool is_infinite(const Edge_circulator& ec) const;
  bool is_infinite(const All_edges_iterator& ei) const;
  bool is_edge(Vertex_handle va, Vertex_handle vb) const;
  bool is_edge(Vertex_handle va, Vertex_handle vb, Face_handle& fr,
               int & i) const;
  bool includes_edge(Vertex_handle va, Vertex_handle vb,
                     Vertex_handle& vbb,
                     Face_handle& fr, int & i) const;
  bool is_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3) const;
  bool is_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3,
       Face_handle &fr) const;


  
 // GEOMETRIC FEATURES AND CONSTRUCTION
  Triangle triangle(Face_handle f) const;
  Segment segment(Face_handle f, int i) const;
  vector_Segment segment(const Edge& e) const;
  Segment segment(const Edge_circulator& ec) const;
  Segment segment(const All_edges_iterator& ei) const;
  Segment segment(const Finite_edges_iterator& ei) const;
  Point circumcenter(Face_handle  f) const;
  Point circumcenter(const Point& p0,
                     const Point& p1,
                     const Point& p2) const;


  //INSERTION - DELETION - Flip
  public:
    void   flip(Face_handle f, int i);

    Vertex_handle insert_first(const Point& p);
    Vertex_handle insert_second(const Point& p);
    Vertex_handle insert_in_edge(const Point& p, Face_handle f,int i);
    Vertex_handle insert_in_face(const Point& p, Face_handle f);
    Vertex_handle insert_outside_convex_hull(const Point& p, Face_handle f);
    Vertex_handle insert_outside_affine_hull(const Point& p);
    Vertex_handle insert(const Point &p, Face_handle start = Face_handle(NULL) );
    Vertex_handle insert(const Point& p,
                         Locate_type lt,
                         Face_handle loc, int li );
    //template < class InputIterator >
    //int insert(InputIterator first, InputIterator last);
    Vertex_handle push_back(const Point& a);

    // POINT LOCATION
    Face_handle locate_in_degenerate_faces(const Point& t,
                                           Locate_type& lt,
                                           int& li,
                                           bool& found = false) const;
    Face_handle
    march_locate_1D(const Point& t, Locate_type& lt, int& li) const ;
    Face_handle
    march_locate_2D(Face_handle start,
                    const Point& t,
                    Locate_type& lt,
                    int& li) const;
    Face_handle
    locate(const Point& p,
           Locate_type& lt,
           int& li,
           Face_handle start = Face_handle(NULL)) const;

    Face_handle
    locate(const Point &p,
           Face_handle start = Face_handle(NULL)) const;
    

  
    //TRAVERSING : ITERATORS AND CIRCULATORS
    Finite_faces_iterator finite_faces_begin() const;
    Finite_faces_iterator finite_faces_end() const;
    Finite_vertices_iterator finite_vertices_begin() const;
    Finite_vertices_iterator finite_vertices_end() const;
    Finite_edges_iterator finite_edges_begin() const;
    Finite_edges_iterator finite_edges_end() const;
    Point_iterator points_begin() const;
    Point_iterator points_end() const;

    All_faces_iterator all_faces_begin() const;
    All_faces_iterator all_faces_end() const;
    All_vertices_iterator all_vertices_begin() const;
    All_vertices_iterator all_vertices_end() const;
    All_edges_iterator all_edges_begin() const;
    All_edges_iterator all_edges_end() const; 

    //for compatibility with previous versions
    Face_iterator faces_begin() const {return finite_faces_begin();}
    Face_iterator faces_end() const {return finite_faces_end();}
    Edge_iterator edges_begin() const {return finite_edges_begin();}
    Edge_iterator edges_end() const {return finite_edges_end();}
    Vertex_iterator vertices_begin() const {return finite_vertices_begin();}
    Vertex_iterator vertices_end() const {return finite_vertices_end();}

    Face_circulator incident_faces( Vertex_handle v,
                                    Face_handle f = Face_handle(NULL)) const;
    Face_iterator incident_faces(Vertex_handle v0,  Vertex_handle v1,
                                 Face_iterator & result) const;

    Vertex_circulator incident_vertices(Vertex_handle v,
                                        Face_handle f = Face_handle(NULL)) const;
    Edge_circulator incident_edges(Vertex_handle v,
                                   Face_handle f = Face_handle(NULL)) const;

    Line_face_circulator line_walk(const Point& p,
                                   const Point& q,
                                   Face_handle f = Face_handle(NULL)) const;

    // TO DEBUG
    void show_all() const;
    void show_vertex(Vertex_handle vh) const;
    void show_face( Face_handle fh) const;


    //PREDICATES
    Oriented_side
    oriented_side(const Point &p0, const Point &p1,
                  const Point &p2, const Point &p) const;

    Bounded_side
    bounded_side(const Point &p0, const Point &p1,
                 const Point &p2, const Point &p) const;
    
    Oriented_side
    oriented_side(Face_handle f, const Point &p) const;

    Oriented_side
    side_of_oriented_circle(Face_handle f, const Point & p) const;

    bool equal_x(const Point& p, const Point& q) const;
    bool equal_y(const Point& p, const Point& q) const;
    bool  xy_equal(const Point& p, const Point& q) const;
    Orientation orientation(const Point& p, const Point& q, const Point& r) const;
    bool collinear_between(const Point& p, const Point& q, const Point& r, const Vertex_handle& dir = NULL) const;

    bool collinear_has_on_segment(const Point& p, const Point& q, const Point& r) const;
    bool antipodal(const Point& p, const Point& q) const;

  

  protected:
    Point compute_dir_antipodalEdge(Vertex_handle v0, Vertex_handle v2,
                                    bool infinite_antipodal);
    bool test_dim_down(Vertex_handle v);
    void fill_hole(Vertex_handle v, std::list<Edge> & hole);
    void fill_hole_delaunay(std::list<Edge> & hole);

  public:

    bool check_ch_defined() const;

    void make_hole(Vertex_handle v, std::list<Edge> & hole);

    Face_handle create_face(Face_handle f1, int i1,
                            Face_handle f2, int i2,
                            Face_handle f3, int i3);
    Face_handle create_face(Face_handle f1, int i1,
                            Face_handle f2, int i2);
    Face_handle create_face(Face_handle f, int i, Vertex_handle v);
    Face_handle create_face(Vertex_handle v1, Vertex_handle v2,Vertex_handle v3);
    Face_handle create_face(Vertex_handle v1, Vertex_handle v2,Vertex_handle v3,
                            Face_handle f1, Face_handle f2, Face_handle f3);
    Face_handle create_face();
    Face_handle create_face(Face_handle); //calls copy constructor of Face
    void delete_face(Face_handle f);
    void delete_vertex(Vertex_handle v);

    Vertex_handle file_input(std::istream& is);
    void file_output(std::ostream& os) const;

  private:
    Vertex_handle insert_outside_convex_hull_1(const Point& p, Face_handle f);
    Vertex_handle insert_outside_convex_hull_2(const Point& p, Face_handle f);

    // template members
  public:
    template < class Stream >
    Stream&  draw_triangulation(Stream& os) const 
    {
      Finite_edges_iterator it = finite_edges_begin();
      for( ;it != finite_edges_end() ; ++it) {
        os << segment(it);
      }
      return os;
    }

    template < class InputIterator >
    int insert(InputIterator first, InputIterator last)
    {
      int n = number_of_vertices();
      while(first != last){
        insert(*first);
        ++first;
      }
      return number_of_vertices() - n;
    }

  public:
    template<class EdgeIt>
    Vertex_handle star_hole(const Point& p, 
                            EdgeIt edge_begin,
                            EdgeIt edge_end) {
    std::list<Face_handle> empty_list;
    return star_hole(p, 
                     edge_begin, 
                     edge_end,
                     empty_list.begin(),
                     empty_list.end());
  }

  template<class EdgeIt, class FaceIt>
  Vertex_handle star_hole( const Point& p, 
                           EdgeIt edge_begin,
                           EdgeIt edge_end,
                           FaceIt face_begin,
                           FaceIt face_end) {
    typedef typename Triangulation_data_structure::Edge  Tds_Edge;
    typedef typename Triangulation_data_structure::Face  Tds_Face;
    Vertex_handle v = _tds.star_hole( edge_begin, edge_end,
                                    face_begin, face_end);
    v->set_point(p);
    return v;
  }

};

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
check_ch_defined() const {
  return ch_is_defined;
}

// CONSTRUCTORS
template <class Gt, class Tds >
Triangulation_T2<Gt, Tds>::
Triangulation_T2(const Geom_traits& geom_traits)
     : _gt(geom_traits), _tds()
{
  _infinite_vertex = _tds.insert_first();
  ch_is_defined = true;
}
  
// copy constructor duplicates vertices and faces
template <class Gt, class Tds >
Triangulation_T2<Gt, Tds>::
Triangulation_T2(const Triangulation_T2 &tr)
    : _gt(tr._gt)
{
  _infinite_vertex = _tds.copy_tds(tr._tds, tr.infinite_vertex());
} 
 

  //Assignement
template <class Gt, class Tds >
Triangulation_T2<Gt, Tds> &
Triangulation_T2<Gt, Tds>::
operator=(const Triangulation_T2 &tr)
{
  copy_triangulation(tr);
  return *this;
}

  // Helping functions

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::   
copy_triangulation(const Triangulation_T2 &tr)
{
  _tds.clear();
  _gt = tr._gt;
  _infinite_vertex = _tds.copy_tds(tr._tds, tr._infinite_vertex);
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
swap(Triangulation_T2 &tr)
{
  Vertex_handle v= _infinite_vertex;
  _infinite_vertex = tr._infinite_vertex;
  tr._infinite_vertex = v;
  
  _tds.swap(tr._tds);

  Geom_traits t = geom_traits();
  _gt = tr.geom_traits();
  tr._gt = t; 
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>:: 
clear()
{
  _tds.clear(); //detruit tous les sommets et toutes les faces
  _infinite_vertex = _tds.insert_first();
}

template <class Gt, class Tds >
int
Triangulation_T2<Gt, Tds>::
number_of_faces() const
{
  int count = _tds.number_of_faces();
  Face_circulator fc = incident_faces(infinite_vertex()),
    done(fc);
  if ( ! fc.is_empty() ) {
    do { 
      --count; ++fc;
    }  while (fc != done);
  }
  return count;
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
infinite_vertex() const
{
  return  _infinite_vertex;
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt,Tds>::Vertex_handle 
Triangulation_T2<Gt,Tds>::
finite_vertex() const
{
  CGAL_triangulation_precondition (number_of_vertices() >= 1);
  return (finite_vertices_begin());
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt,Tds>::Face_handle 
Triangulation_T2<Gt,Tds>::
infinite_face() const
{
  return infinite_vertex()->face();
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt,Tds>::Infinite_tester
Triangulation_T2<Gt,Tds>::
infinite_tester() const
{
  return Infinite_tester(this);
}


template <class Gt, class Tds >
bool
Triangulation_T2<Gt, Tds>::
is_valid(bool verbose, int level) const
{
  bool result = _tds.is_valid(verbose, level);
  if (dimension() <= 0 ||
      (dimension()==1 && number_of_vertices() == 2 ) ) return result;

  if (dimension() == 1) {
    Finite_vertices_iterator it1 = finite_vertices_begin(),
                             it2(it1), it3(it1);
    ++it2;
    ++it3; ++it3;
    while( it3 != finite_vertices_end()) {
     Orientation s = orientation(it1->point(),
                                 it2->point(),
                                 it3->point()); 
     result = result && s == COLLINEAR ;
     CGAL_triangulation_assertion(result);
     ++it1 ; ++it2; ++it3;
    }
  }    

  else { //dimension() == 2
    for(Finite_faces_iterator it=finite_faces_begin(); 
        it!=finite_faces_end(); it++) {
      CGAL_triangulation_assertion( ! is_infinite(it));
      Orientation s = orientation(it->vertex(0)->point(),
                                  it->vertex(1)->point(),
                                  it->vertex(2)->point());
      CGAL_triangulation_assertion( s == LEFT_TURN );
      result = result && ( s == LEFT_TURN );
    }

    Vertex_circulator start = incident_vertices(infinite_vertex());
    Vertex_circulator pc(start);
    Vertex_circulator qc(start); ++qc;
    Vertex_circulator rc(start); ++rc; ++rc;
    do{
      Orientation s = orientation(pc->point(),
                                  qc->point(),
                                  rc->point());
      CGAL_triangulation_assertion( s != LEFT_TURN );
      result = result && ( s != LEFT_TURN );
      ++pc ; ++qc ; ++rc;
    }while(pc != start);

    // check number of faces. This cannot be done by the Tds
    // which does not know the number of components nor the genus
    result = result && (number_of_faces() == 2*(number_of_vertices()+1) - 4
                                            - infinite_vertex()->degree());
    CGAL_triangulation_assertion( result);
  }
  return result;
}


template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_infinite(Face_handle f) const 
{
  return f->has_vertex(infinite_vertex());
}


template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_infinite(Vertex_handle v) const 
{
  return v == infinite_vertex();
}

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_infinite(Face_handle f, int i) const 
{
  return is_infinite(f->vertex(ccw(i))) ||
         is_infinite(f->vertex(cw(i)));
}

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_infinite(const Edge& e) const 
{
  return is_infinite(e.first,e.second);
}

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_infinite(const Edge_circulator& ec) const 
{
  return is_infinite(*ec);
}

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_infinite(const All_edges_iterator& ei) const 
{
  return is_infinite(*ei);
}

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_edge(Vertex_handle va, Vertex_handle vb) const
{
  return _tds.is_edge( va, vb);
}

template <class Gt, class Tds >
inline bool
Triangulation_T2<Gt, Tds>::
is_edge(Vertex_handle va, Vertex_handle vb, Face_handle& fr, int & i) const
{
  return _tds.is_edge(va, vb, fr, i);
}

template <class Gt, class Tds >
bool 
Triangulation_T2<Gt, Tds>::
includes_edge(Vertex_handle va, Vertex_handle vb,
              Vertex_handle & vbb,
              Face_handle& fr, int & i) const
  // returns true if the line segment ab contains an edge e of t
  // incident to a, false otherwise
  // if true, vbb becomes the vertex of e distinct from a
  // fr is the face incident to e and e=(fr,i)
  // fr is on the right side of a->b
{
  Vertex_handle v;
  Orientation orient;
  int indv;
  Edge_circulator ec = incident_edges(va), done(ec);
  if (ec != 0) {
    do { 
      //find the index of the other vertex of *ec
      indv = 3 - ((*ec).first)->index(va) - (*ec).second ; 
      v = ((*ec).first)->vertex(indv);
      if (!is_infinite(v)) {
        if (v==vb) {
          vbb = vb;
          fr=(*ec).first;
          i= (*ec).second;
          return true;
        }
        else {
          orient = orientation(va->point(),
                               vb->point(),
                               v->point()); 
          if ((orient==COLLINEAR) &&
             (collinear_between (va->point(),
                                 v->point(),
                                 vb->point()))) {
            vbb = v;
            fr=(*ec).first;
            i= (*ec).second;
            return true;
          }
        }
      }
    } while (++ec != done);
  }
  return false;
}

template <class Gt, class Tds >
inline bool 
Triangulation_T2<Gt, Tds>::
is_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3) const
{
  return _tds.is_face(v1, v2, v3);
}

template <class Gt, class Tds >
inline bool 
Triangulation_T2<Gt, Tds>::
is_face(Vertex_handle v1, 
        Vertex_handle v2, 
        Vertex_handle v3,
        Face_handle &fr) const
{
  return _tds.is_face(v1, v2, v3, fr);
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Triangle
Triangulation_T2<Gt, Tds>::
triangle(Face_handle f) const
{
  CGAL_triangulation_precondition( ! is_infinite(f) );
  typename Gt::Construct_triangle_T2 
     construct_triangle = geom_traits().construct_triangle_T2_object();
  return construct_triangle(f->vertex(0)->point(),
                            f->vertex(1)->point(),
                            f->vertex(2)->point());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Segment
Triangulation_T2<Gt, Tds>::
segment(Face_handle f, int i) const
{
  CGAL_triangulation_precondition( ! is_infinite(f,i));
  typename Gt::Construct_segment_T2
     construct_segment = geom_traits().construct_segment_T2_object();
  return construct_segment(f->vertex(ccw(i))->point(),
                           f->vertex(cw(i))->point());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::vector_Segment
Triangulation_T2<Gt, Tds>::
segment(const Edge& e) const
{
  CGAL_triangulation_precondition(! is_infinite(e)); 
  typename Gt::K kernel;
  typename Gt::K::Construct_segment_T2 
     construct_segment = kernel.construct_segment_T2_object();
  std::vector<Segment> vector;
  if (antipodal(e.first->vertex(ccw(e.second))->point(),
                e.first->vertex( cw(e.second))->point())) {
    vector.push_back(construct_segment(e.first->vertex(ccw(e.second))->point(),
                                       e.first->get_dir()->point()));
    vector.push_back(construct_segment(e.first->get_dir()->point(),
                                       e.first->vertex( cw(e.second))->point()));
  }
  else
    vector.push_back(construct_segment(e.first->vertex(ccw(e.second))->point(),
                                       e.first->vertex( cw(e.second))->point()));
  return vector;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Segment
Triangulation_T2<Gt, Tds>::
segment(const Edge_circulator& ec) const
{
  return segment(*ec);
}
    
template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Segment
Triangulation_T2<Gt, Tds>::
segment(const Finite_edges_iterator& ei) const
{
  return segment(*ei);
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Segment
Triangulation_T2<Gt, Tds>::
segment(const All_edges_iterator& ei) const
{
  return segment(*ei);
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
flip(Face_handle f, int i)
{
  CGAL_triangulation_precondition ( f != NULL );
  CGAL_triangulation_precondition (i == 0 || i == 1 || i == 2);
  CGAL_triangulation_precondition( dimension()==2); 
    
  CGAL_triangulation_precondition( !is_infinite(f) && 
                                   !is_infinite(f->neighbor(i)) );
  CGAL_triangulation_precondition(
                  orientation(f->vertex(i)->point(),
                              f->vertex(cw(i))->point(),
                              f->mirror_vertex(i)->point()) != LEFT_TURN &&
                  orientation(f->vertex(i)->point(),
                              f->vertex(ccw(i))->point(),
                              f->mirror_vertex(i)->point()) !=  RIGHT_TURN); 
  _tds.flip(f, i);
  return;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_first(const Point& p)
{
  CGAL_triangulation_precondition(number_of_vertices() == 0);
  Vertex_handle v = _tds.insert_second();
  v->set_point(p);
  return  v;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_second(const Point& p)
{
  CGAL_triangulation_precondition(number_of_vertices() == 1);
  Vertex_handle v = _tds.insert_dim_up(infinite_vertex(), true);
  v->set_point(p);
  return v;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_in_edge(const Point& p, Face_handle f,int i)
{

  CGAL_triangulation_exactness_precondition(orientation(f->vertex(cw(i))->point(),
      p,
      f->vertex(ccw(i))->point()) == COLLINEAR &&
      collinear_between(f->vertex(cw(i))->point(), p,f->vertex(ccw(i))->point() ) );
  Vertex_handle v;
  bool constructed_degenerate = false;
  if (dimension() == 1)
    v = _tds.insert_in_edge(f,i);
  else {
    //verify if f is degenerate
    if (orientation(f->vertex(0)->point(),
                    f->vertex(1)->point(),
                    f->vertex(2)->point()) == 0 &&
                    f->get_dir() == NULL ) {
    //if so, we must remove it from the vector degenerate_face and
      Degenerate_faces_iterator it = vector_degenerate_faces.begin();
      bool removed = false;
      while (it != vector_degenerate_faces.end() && !removed) {
         if (((*it)->vertex(0)->point() == f->vertex(0)->point()) &&
             ((*it)->vertex(1)->point() == f->vertex(1)->point()) &&
             ((*it)->vertex(2)->point() == f->vertex(2)->point())) {
           vector_degenerate_faces.erase(it);
           removed = true;
         }
         if (!removed)
           it++;
      }
      constructed_degenerate = true;
    }
    else if (orientation(f->vertex(0)->point(),
                    f->vertex(1)->point(),
                    f->vertex(2)->point()) == 0 &&
             (antipodal(f->vertex(0)->point(),
                    f->vertex(1)->point()) && (has_on_segment(p,f->vertex(1)->point(),
                    f->vertex(2)->point()) || (has_on_segment(p,f->vertex(0)->point(),
                    f->vertex(2)->point()))))  ||
                    (antipodal(f->vertex(1)->point(),f->vertex(2)->point()) &&
                     (has_on_segment(p,f->vertex(0)->point(),
                    f->vertex(1)->point()) || (has_on_segment(p,f->vertex(0)->point(),
                    f->vertex(2)->point()))) ) ||
                    (antipodal(f->vertex(0)->point(),
                    f->vertex(2)->point()) && (has_on_segment(p,f->vertex(0)->point(),
                    f->vertex(1)->point()) || (has_on_segment(p,f->vertex(1)->point(),
                    f->vertex(2)->point()))) ) )
    constructed_degenerate = true;
    Vertex_handle dir0 = f->get_dir();
    Face_handle fh = f->neighbor(i);
    int j = fh->index(f);
    Vertex_handle vh0 = f->vertex(i);
    Vertex_handle vh1 = fh->vertex(j);
    Vertex_handle dir1 = fh->get_dir();
    v = _tds.insert_in_edge(f,i);
    std::vector<Face_handle> f_v_vh0, f_v_vh1;
    //find incident face edge v-vh0
    Face_circulator fi0 = incident_faces(v), done(fi0);
    do {
      int k = fi0->index(v);
      if (!is_infinite(fi0->vertex(cw(k)))) {
        if (xy_equal(fi0->vertex(cw(k))->point(),vh0->point()))
          f_v_vh0.push_back(fi0);
        else if (xy_equal(fi0->vertex(cw(k))->point(),vh1->point()))
          f_v_vh1.push_back(fi0);
      }
      if (!is_infinite(fi0->vertex(ccw(k)))) {
        if (xy_equal(fi0->vertex(ccw(k))->point(),vh0->point()))
          f_v_vh0.push_back(fi0);
        else if (xy_equal(fi0->vertex(ccw(k))->point(),vh1->point()))
          f_v_vh1.push_back(fi0);
      }
      fi0++;
    } while (fi0 != done);
    int index = 0;
    bool found = false;
    if (f_v_vh0.size() > 0) {
      do {
        fh = f_v_vh0[index];
        int v_index = fh->index(v);
        if (antipodal(fh->vertex(cw(v_index))->point(),
                      fh->vertex(ccw(v_index))->point())) {
          fh->set_dir(dir0);
          found = true;
        }
        else {
          fh->set_dir(NULL);
        }
        fh++;
        index++;
      } while (index < 2 && !found);
    }
    index = 0;
    found = false;
    if (f_v_vh1.size() > 0) {
      do {
        fh = f_v_vh1[index];
        int v_index = fh->index(v);
        if (antipodal(fh->vertex(cw(v_index))->point(),
                      fh->vertex(ccw(v_index))->point())) {
          fh->set_dir(dir1);
          found = true;
        }
        else
          fh->set_dir(NULL);
        fh++;
        index++;
      } while (index < 2 && !found);
    }

  }
  v->set_point(p);
  if (constructed_degenerate) { // insert the new one's
    Face_circulator fc = v->incident_faces() ,done(fc);
    do {
      if (orientation(fc->vertex(0)->point(),
                      fc->vertex(1)->point(),
                      fc->vertex(2)->point()) == 0 &&
                      fc->get_dir() == NULL)
        vector_degenerate_faces.push_back(fc);
      fc++;
    } while(fc != done);
  }

  return v;
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_in_face(const Point& p, Face_handle f)
{
  // CGAL_triangulation_precondition(oriented_side(f,p) ==   //ON_POSITIVE_SIDE);
  Vertex_handle v= _tds.insert_in_face(f);
  Face_handle f1 = f->neighbor(1);
  if (antipodal(f1->vertex(0)->point(),f1->vertex(2)->point()))
    f1->set_dir(f->get_dir());
  else {
    Face_handle f2 = f->neighbor(2);
    if (antipodal(f2->vertex(0)->point(),
                  f2->vertex(1)->point()))
                  f2->set_dir(f->get_dir());
  }
  if (!antipodal(f->vertex(1)->point(),f->vertex(2)->point()))
    f->set_dir(NULL);
  v->set_point(p);
  return v;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_outside_convex_hull(const Point& p, Face_handle f)
{
  CGAL_triangulation_precondition(is_infinite(f) && dimension() >= 1);
  Vertex_handle v;
  if (dimension() == 1)  v=insert_outside_convex_hull_1(p, f);
  else   v=insert_outside_convex_hull_2(p, f);
  v->set_point(p);
  return v;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_outside_convex_hull_1(const Point& p, Face_handle f)
{
  CGAL_triangulation_precondition( is_infinite(f) && dimension()==1);
  CGAL_triangulation_precondition(
  orientation(f->mirror_vertex(f->index(infinite_vertex()))->point(),
              f->vertex(1- f->index(infinite_vertex()))->point(),
              p) == COLLINEAR);
  Vertex_handle v;
  if (check_ch_defined()) {
    int ivf = f->index(infinite_vertex());
    Vertex_handle a = f->vertex(1-ivf);
    Face_handle f_neighbor = f->neighbor(ivf);
    int index_a = f_neighbor->index(a);
    Vertex_handle na = f_neighbor->vertex(1-index_a);
    Face_handle ff = f->neighbor(1-ivf); //the other infinite face
    int ivff = ff->index(infinite_vertex());
    Vertex_handle b = ff->vertex(1-ivff);
    f_neighbor = ff->neighbor(ivff);
    int index_b = f_neighbor->index(b);
    Vertex_handle nb = f_neighbor->vertex(1-index_b);
    if (collinear_has_on_segment(a->point(),antipode(p),b->point())) {
      ch_is_defined = false;
      _tds.insert_in_infinite_vertex(f,ivf,v);
      if (antipodal(a->point(),p)) {
        Vertex_handle dir = _tds.create_vertex();
        const Point & pt = antipode(na->point());
        dir->set_point(pt);
        f->set_dir(dir);
        dir = f->get_dir();
      }
      else
        f->set_dir(NULL);
      if (antipodal(b->point(),p)) {
        Vertex_handle dir = _tds.create_vertex();
        const Point & pt = antipode(nb->point());
        dir->set_point(pt);
        ff->set_dir(dir);
        dir = ff->get_dir();
      }
      else
        ff->set_dir(NULL);
      v->set_point(p);
      Vertex_handle inf = _tds.create_vertex();
      set_infinite_vertex(inf);
      return v;
    }
  }
  v=_tds.insert_in_edge(f, 2);
  v->set_point(p);

  return v;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_outside_convex_hull_2(const Point& p, Face_handle f)
{
  CGAL_triangulation_precondition(is_infinite(f));
  Vertex_handle v;
  int li = f->index(infinite_vertex());
  //std::cout << "vai inserir out ch 1\n"; 
  CGAL_triangulation_precondition(
               orientation(p,f->vertex(ccw(li))->point(),
                           f->vertex(cw(li))->point()) == LEFT_TURN);
  //std::cout << "vai inserir out ch 2\n"; 
  std::list<Face_handle> ccwlist;
  std::list<Face_handle> cwlist;
  std::list<Face_handle> collinearlist;


  Face_circulator fc = incident_faces(infinite_vertex(), f);
  Face_circulator fc1 = fc;
  bool done = false;
  fc--;
  while(!done && fc!=fc1) {
    li = fc->index(infinite_vertex());
    const Point& q = fc->vertex(ccw(li))->point();
    const Point& r = fc->vertex(cw(li))->point();
  
    if (p == antipode(q) || p == antipode(r)) {
      collinearlist.push_back(fc);
      fc--;
    }
    else if (orientation(p,q,r) == LEFT_TURN) {
      ccwlist.push_back(fc);
      fc--;
    }
    else if (has_on_segment(antipode(q),p,antipode(r))) {
      ccwlist.push_back(fc);
      fc--;
    }
    else {done=true;}
  }
  if (fc != fc1) {
    fc = incident_faces(infinite_vertex(), f);
    fc1 = fc;
    done = false;
    fc++;
    while(!done && fc1 != fc){
      li = fc->index(infinite_vertex());
      const Point& q = fc->vertex(ccw(li))->point();
      const Point& r = fc->vertex(cw(li))->point();
      if(orientation(p,q,r) == LEFT_TURN) {
        cwlist.push_back(fc);
      }
      else {done=true;}
      fc++;
    }
  }
  else
    ch_is_defined = false;
  if (ch_is_defined) {
    Vertex_handle v = _tds.insert_in_face(f);
    v->set_point(p);
    Face_handle fh;
    while ( ! ccwlist.empty()) {
      fh = ccwlist.front();
      li = ccw(fh->index(infinite_vertex()));
      _tds.flip( fh, li);
      ccwlist.pop_front();
    }
    while ( ! cwlist.empty()) {
      fh = cwlist.front();
      li = cw(fh->index(infinite_vertex()));
      _tds.flip( fh, li);
      cwlist.pop_front();
    }
    //reset infinite_vertex()->face();
    fc = incident_faces(v);
    while( ! is_infinite(fc)) { fc++;}
    infinite_vertex()->set_face(fc);
    return v;
  }
  else {
   Vertex_handle dir = _tds.create_vertex();
   bool infinite_antipodals = false;
   Face_handle fc2;
   int i,j;
   All_faces_iterator afi;
   for(afi = all_faces_begin(); afi != all_faces_end(); afi++) {
     if (is_infinite(afi)) {
       i = afi->index(infinite_vertex());
       if (antipodal(p,afi->vertex(ccw(i))->point())) {
         fc2 = afi->neighbor(i);
         j = fc2->index(afi->vertex(ccw(i)));
         if (p.w() == 0) infinite_antipodals = true;
         const Point& pt = compute_dir_antipodalEdge(fc2->vertex(ccw(j)),
                                                     fc2->vertex(cw(j)),
                                                     infinite_antipodals);
         dir->set_point(pt);
         afi->set_dir(dir);
       }
       else if (antipodal(p,afi->vertex(cw(i))->point())) {
         fc2 = afi->neighbor(i);
         j = fc2->index(afi->vertex(cw(i)));
         if (p.w() == 0) infinite_antipodals = true;
         const Point& pt = compute_dir_antipodalEdge(fc2->vertex(ccw(j)),
                                                     fc2->vertex(cw(j)),
                                                     infinite_antipodals);
         dir->set_point(pt);
         afi->set_dir(dir);
       }
       else if (orientation(p,afi->vertex(cw(i))->point(),
                            afi->vertex(ccw(i))->point())==0){
         vector_degenerate_faces.push_back(afi);
       }
     }

  }
  
  int help = fc->index(infinite_vertex());
  fc1 = fc;
  std::vector<int> v_index_iv;
  _tds.insert_in_infinite_vertex(fc,help,v);
  v->set_point(p);
  Vertex_handle inf = _tds.create_vertex();
  set_infinite_vertex(inf);
  return v;
}
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Point
Triangulation_T2<Gt,Tds>::
compute_dir_antipodalEdge(Vertex_handle v0,Vertex_handle v2,bool infinite_antipodal)
{
  typedef typename Point::FT FT;
  Point p = v0->point();
  Point q = v2->point();
  FT x = -(abs(q.w())*p.x() + abs(p.w())*q.x());
  FT y = -(abs(q.w())*p.y() + abs(p.w())*q.y());
  if (!infinite_antipodal) {
    return Point(FT(0),x,y);
  }
  return Point(-(abs(q.w())*p.w() + abs(p.w())*q.w()),x,y);
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert_outside_affine_hull(const Point& p)
{
  CGAL_triangulation_precondition(dimension() == 1);
  Face_handle f = (*finite_edges_begin()).first;
  Orientation orient = orientation( f->vertex(0)->point(),
                                    f->vertex(1)->point(),p);
  CGAL_triangulation_precondition(orient != COLLINEAR);
  bool conform = ( orient == COUNTERCLOCKWISE);

  Vertex_handle v = _tds.insert_dim_up( infinite_vertex(), conform);
  v->set_point(p);
  return v;
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert(const Point &p, Face_handle start)
{
  Locate_type lt;
  int li;
  Face_handle loc = locate (p, lt, li, start);
  return insert(p, lt, loc, li);
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
insert(const Point& p, Locate_type lt, Face_handle loc, int li)
  // insert a point p, whose localisation is known (lt, f, i)
{
  if(number_of_vertices() == 0) {
     return(insert_first(p));
  }

  if(number_of_vertices() == 1) {
    if (lt == VERTEX) return finite_vertex();
    else return(insert_second(p));
  }
  switch(lt) {
  case FACE:
    return insert_in_face(p,loc);     
  case EDGE:
    return insert_in_edge(p,loc,li);
  case OUTSIDE_CONVEX_HULL:
    return  insert_outside_convex_hull(p,loc);
  case OUTSIDE_AFFINE_HULL:
    return  insert_outside_affine_hull(p);
  case VERTEX:
    return loc->vertex(li);
  }
  CGAL_triangulation_assertion(false);  // locate step failed
  return Vertex_handle(NULL);
}


template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt,Tds>::Vertex_handle
Triangulation_T2<Gt,Tds>::
push_back(const Point &p)
{
  return insert(p);
}

template <class Gt, class Tds >
bool
Triangulation_T2<Gt,Tds>::
test_dim_down(Vertex_handle v)
{
  //test the dimensionality of the resulting triangulation
  //upon removing of vertex v
  //it goes down to 1 iff
  // 1) any finite face is incident to v
  // 2) all vertices are colinear
  CGAL_triangulation_precondition(dimension() == 2);
  bool  dim1 = true; 
  Finite_faces_iterator fit = finite_faces_begin();
  while (dim1==true && fit != finite_faces_end()) {
    dim1 = dim1 && fit->has_vertex(v);
    fit++;
  }
  Face_circulator fic = incident_faces(v);
  while (is_infinite(fic)) {++fic;}
  Face_circulator done(fic);
  Face_handle start(fic); int iv = start->index(v);
  const Point& p = start->vertex(cw(iv))->point(); 
  const Point& q = start->vertex(ccw(iv))->point();
  while ( dim1 && ++fic != done) {
    iv = fic->index(v);
    if (fic->vertex(ccw(iv)) != infinite_vertex()) {
      dim1 = dim1 &&
      orientation(p, q, fic->vertex(ccw(iv))->point()) == COLLINEAR; 
    }
  }
  return dim1;
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
make_hole ( Vertex_handle v, std::list<Edge> & hole)
{
  std::list<Face_handle> to_delete;

  Face_handle  f, fn;
  int i, in ;
   Vertex_handle  vv;
      
  Face_circulator fc = incident_faces(v);
  Face_circulator done(fc);
  do {
    f = fc; fc++;
    i = f->index(v);
    fn = f->neighbor(i);
    in = fn->index(f);
    vv = f->vertex(cw(i));
    if( vv->face()==  f) vv->set_face(fn);
    vv = f->vertex(ccw(i));
    if( vv->face()== f) vv->set_face(fn);
    fn->set_neighbor(in, NULL);
    hole.push_back(Edge(fn,in));
    to_delete.push_back(f);
  }
  while(fc != done);

  while (! to_delete.empty()){
    delete_face(to_delete.front());
    to_delete.pop_front();
  }
  return;
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::						     
fill_hole ( Vertex_handle v, std::list< Edge > & hole )
{
  // uses the fact that the hole is starshaped
  // with repect to v->point()
  typedef std::list<Edge> Hole;

  Face_handle  ff, fn;
  int ii , in; 
  Vertex_handle v0, v1, v2;
  Bounded_side side;

  //stack algorithm to create faces
  // create face v0,v1,v2
  //if v0,v1,v2 are finite vertices
  // and form a left_turn
  // and triangle v0v1v2 does not contain v->point()
  if( hole.size() != 3) {
    typename Hole::iterator hit = hole.begin();
    typename Hole::iterator next= hit; 
    while( hit != hole.end() && hole.size() != 3) {
      ff = (*hit).first;  
      ii = (*hit).second;
      v0 = ff->vertex(cw(ii));
      v1 = ff->vertex(ccw(ii));
      if( !is_infinite(v0) && !is_infinite(v1)) {
        next=hit; next++;
        if(next == hole.end()) next=hole.begin();
        fn = (*next).first;
        in = (*next).second;
        v2 = fn->vertex(ccw(in));	
        if ( !is_infinite(v2) &&
          orientation(v0->point(), v1->point(), v2->point()) == LEFT_TURN ) {
          side =  bounded_side(v0->point(), 
                               v1->point(), 
                               v2->point(),
                               v->point());
          if (side == ON_UNBOUNDED_SIDE || 
             (side == ON_BOUNDARY && orientation(v0->point(),
                                                 v->point(),
                                                 v2->point()) == COLLINEAR &&
              collinear_between(v0->point(),v->point(),v2->point()) )) 
          {
              //create face
              Face_handle  newf = create_face(ff,ii,fn,in); 
              typename Hole::iterator tempo=hit;
              hit = hole.insert(hit,Edge(newf,1)); //push newf
              hole.erase(tempo); //erase ff
              hole.erase(next); //erase fn
              if (hit != hole.begin() ) --hit;
              continue;
          }
        }
      }
      ++hit; 
    }
  }

  // either the hole has only three edges
  // or all its finite vertices are reflex or flat
  // except may be one vertex whose corresponding ear
  // includes the vertex being removed

  // deal with the last left_turn if any
  if(hole.size() != 3) {
    typename Hole::iterator hit=hole.begin();
    while(hit != hole.end()) {
      ff = (*hit).first;  ii = (*hit).second;
      hit++;
      if(hit != hole.end()) { fn = (*hit).first; in = (*hit).second;}
      else { fn = ((hole.front()).first); in = (hole.front()).second;}
      if ( !is_infinite(ff->vertex(cw(ii))) &&
	   !is_infinite(fn->vertex(cw(in))) &&
	   !is_infinite(fn->vertex(ccw(in))) &&
	   orientation(ff->vertex(cw(ii))->point(),
		       fn->vertex(cw(in))->point(),
		       fn->vertex(ccw(in))->point()) == LEFT_TURN) {
	  create_face(ff,ii,fn,in);
	  break;
	}
    }
  }

  // deal with a reflex chain of convex hull edges
  if(hole.size() != 3) {
    // look for infinite vertex
    ff = (hole.front()).first;
    ii = (hole.front()).second;
    while ( ! is_infinite(ff->vertex(cw(ii)))){
      hole.push_back(hole.front());
      hole.pop_front();
      ff = (hole.front()).first;
      ii = (hole.front()).second;
    }
    //create faces
    while(hole.size() != 3){
      ff = (hole.front()).first;
      ii = (hole.front()).second;
      hole.pop_front();
      fn = (hole.front()).first;
      in = (hole.front()).second;
      hole.pop_front();
      Face_handle  newf = create_face(ff,ii,fn,in);
      hole.push_front(Edge(newf,1));
    }
  }

  // now hole has three edges
  typename Hole::iterator hit;
  hit = hole.begin();
  //  I don't know why the following yelds a segmentation fault
  //    create_face( (*hit).first, (*hit).second,
  // 	     (* ++hit).first, (*hit).second,
  // 	     (* ++hit).first, (*hit).second);
  ff = (*hit).first;      ii = (*hit).second;
  fn = (* ++hit).first;   in = (*hit).second;
  Face_handle f3 = (* ++hit).first;
  int i3 = (*hit).second;
  create_face(ff,ii,fn,in,f3,i3);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face(Face_handle f1, int i1,
            Face_handle f2, int i2,
            Face_handle f3, int i3)
{
  return _tds.create_face(f1, i1, f2, i2, f3, i3);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face(Face_handle f1, int i1,
            Face_handle f2, int i2)
{
  return _tds.create_face(f1, i1, f2, i2);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face(Face_handle f, int i, Vertex_handle v)
{
  return _tds.create_face(f, i, v);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3)
{
  return _tds.create_face(v1, v2, v3);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3,
	    Face_handle f1, Face_handle f2,  Face_handle f3)
{
  return _tds.create_face(v1, v2, v3, f1, f2, f3);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face(Face_handle fh)
{
  return _tds.create_face(fh);
}



template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
create_face()
{
  return _tds.create_face();
}

template <class Gt, class Tds >
inline
void
Triangulation_T2<Gt, Tds>::
delete_face(Face_handle f)
{
  _tds.delete_face(f);
}

template <class Gt, class Tds >
inline
void
Triangulation_T2<Gt, Tds>::
delete_vertex(Vertex_handle v)
{
  _tds.delete_vertex(v);
}

// POINT LOCATION
template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
march_locate_1D(const Point& t,
                Locate_type& lt,
                int& li) const
{
  if (check_ch_defined()) {
    Face_handle ff = infinite_face();
    int iv = ff->index(infinite_vertex());

    Face_handle f = ff->neighbor(iv);
    Vertex_handle vh0= ff->vertex(1-iv);
    ff = ff->neighbor(1-iv);
    iv = ff->index(infinite_vertex());
    Vertex_handle vh1= ff->vertex(1-iv);
    if (collinear_has_on_segment(vh0->point(),antipode(t),vh1->point())) {
      lt = OUTSIDE_CONVEX_HULL;
      li = iv;
      return ff;
    }

    f = ff->neighbor(iv);
    int i= f->index(ff);
    if (orientation(t,f->vertex(1-i)->point(), f->vertex(i)->point())==0) {
      std::cout << collinear_has_on_segment(t,f->vertex(1-i)->point(),
      f->vertex(i)->point()) << std::endl;
      if (collinear_has_on_segment(t,f->vertex(1-i)->point(),f->vertex(i)->point())) {
        lt = OUTSIDE_CONVEX_HULL;
        li = iv;
        return ff;
      }
    }

    if ( xy_equal(t,f->vertex(1-i)->point()) ){
      lt = VERTEX;
      li=1-i;
      return f;
    }

    ff = ff->neighbor(1-iv); //the other infinite face
    iv = ff->index(infinite_vertex());
    f = ff->neighbor(iv);
    i = f->index(ff);
    if (orientation(t,f->vertex(1-i)->point(),f->vertex(i)->point())==0) {
      std::cout << collinear_has_on_segment(t,f->vertex(1-i)->point(),
      f->vertex(i)->point()) << std::endl;
      if (collinear_has_on_segment(t,f->vertex(1-i)->point(),
                                     f->vertex(i)->point())) {
        lt = OUTSIDE_CONVEX_HULL;
        li = iv;
        return ff;
      }
    }
    if( xy_equal(t,f->vertex(1-i)->point()) ){
      lt = VERTEX;
      li=1-i;
      return f;
    }
  }

  Finite_edges_iterator eit= finite_edges_begin();

  Orientation pqt = orientation((*eit).first->vertex(0)->point(),
                                (*eit).first->vertex(1)->point(),t);
  if(pqt == RIGHT_TURN || pqt == LEFT_TURN) {
    lt = OUTSIDE_AFFINE_HULL;
    li = 4 ;// should not be used
    return Face_handle(NULL);
  }

  Vertex_handle u,v,vh_dir;
  Point dir;
  for ( ; eit != finite_edges_end() ; eit++) {
    u = (*eit).first->vertex(0);
    v = (*eit).first->vertex(1);
    if (antipodal(u->point(),v->point())) {
      Finite_edges_iterator _eit = eit;
      _eit++;
      if (_eit == finite_edges_end()) {
        _eit = finite_edges_begin();
        _eit++;
      }
      if ((*_eit).first->vertex(0)->point()==u->point() ||
          (*_eit).first->vertex(0)->point()==v->point()) {
          dir = antipode((*_eit).first->vertex(1)->point());
          vh_dir = (*_eit).first->vertex(1);
      }
      else {
        dir = antipode((*_eit).first->vertex(0)->point());
        vh_dir = (*_eit).first->vertex(0);
      }
      vh_dir->set_point(dir);
    }
    if(xy_equal(t,v->point())){
      lt = VERTEX;
      li = 1;
      return (*eit).first;
    }
    if(collinear_between(u->point(), t, v->point(),vh_dir)){
      lt = EDGE;
      li =  2;
      return (*eit).first;
    }
  }
  CGAL_triangulation_assertion(false);
  return Face_handle(NULL);
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
locate_in_degenerate_faces(const Point& t,
                           Locate_type& lt,
                           int& li,
                           bool &found) const
{
  unsigned int i = 0;
  Face_handle fh;
  Segment s;
  for (; i < vector_degenerate_faces.size(); i++) {
    if (vector_degenerate_faces[i]->vertex(0)->point() == t) {
      lt = VERTEX;
      li = 0;
      found = true;
      return vector_degenerate_faces[i];
    }
    else if (vector_degenerate_faces[i]->vertex(1)->point() == t) {
      lt = VERTEX;
      li = 1;
      found = true;
      return vector_degenerate_faces[i];
    }
    else if (vector_degenerate_faces[i]->vertex(2)->point() == t) {
      lt = VERTEX;
      li = 2;
      found = true;
      return vector_degenerate_faces[i];
    }
    else if (has_on_segment(t,vector_degenerate_faces[i]->vertex(0)->point(),
                              vector_degenerate_faces[i]->vertex(1)->point())) {
      if (!found) {
        s = Segment(vector_degenerate_faces[i]->vertex(0)->point(),
                    vector_degenerate_faces[i]->vertex(1)->point());
        fh = vector_degenerate_faces[i];
      }
      else if (_gt.compare_distance(vector_degenerate_faces[i]->vertex(0)->point(),
                                    vector_degenerate_faces[i]->vertex(1)->point(),
                                    s.source(),s.target()) == SMALLER) {
        s = Segment(vector_degenerate_faces[i]->vertex(0)->point(),
                    vector_degenerate_faces[i]->vertex(1)->point());
        fh = vector_degenerate_faces[i];
      }
      li = 2;
      lt = EDGE;
      found = true;
    }
    else if (has_on_segment(t,vector_degenerate_faces[i]->vertex(1)->point(),
                              vector_degenerate_faces[i]->vertex(2)->point())) {
      if (!found) {
        s = Segment(vector_degenerate_faces[i]->vertex(1)->point(),
                    vector_degenerate_faces[i]->vertex(2)->point());
        fh = vector_degenerate_faces[i];
      }
      else if (_gt.compare_distance(vector_degenerate_faces[i]->vertex(1)->point(),
                                    vector_degenerate_faces[i]->vertex(2)->point(),
                          s.source(),s.target()) == SMALLER) {
        s = Segment(vector_degenerate_faces[i]->vertex(1)->point(),
                    vector_degenerate_faces[i]->vertex(2)->point());
        fh = vector_degenerate_faces[i];
      }
      li = 0;
      lt = EDGE;
      found = true;
    }
    else if (has_on_segment(t,vector_degenerate_faces[i]->vertex(2)->point(),
                              vector_degenerate_faces[i]->vertex(0)->point())) {
      if (!found) {
        s = Segment(vector_degenerate_faces[i]->vertex(2)->point(),
                    vector_degenerate_faces[i]->vertex(0)->point());
        fh = vector_degenerate_faces[i];
      }
      else if (_gt.compare_distance(vector_degenerate_faces[i]->vertex(2)->point(),
                                    vector_degenerate_faces[i]->vertex(0)->point(),
                          s.source(),s.target()) == SMALLER) {
        s = Segment(vector_degenerate_faces[i]->vertex(2)->point(),
                    vector_degenerate_faces[i]->vertex(0)->point());
        fh = vector_degenerate_faces[i];
      }
      li = 1;
      lt = EDGE;
      found = true;
    }
  }
  return fh;
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt, Tds>::
march_locate_2D(Face_handle start,
                const Point& t,
                Locate_type& lt,
                int& li) const
{
  bool found = false;
  Face_handle result_fh;
  if (vector_degenerate_faces.begin() != vector_degenerate_faces.end()) {
    result_fh = locate_in_degenerate_faces(t,lt,li,found);
  }
  if (found)
    return result_fh;
  Point p = start->vertex(0)->point();
  Vertex_handle st = start->vertex(0);
  if (antipodal(p,t)) {
    if (!is_infinite(start->vertex(1))) {
      p = start->vertex(1)->point();
      st = start->vertex(1);
    }
    else {
      p = start->vertex(2)->point();
      st = start->vertex(2);
    }
  }
  if(xy_equal(t,p)) {
    lt = VERTEX;
    li = 0;
    return start;
  }
  Line_face_circulator lfc(st, this, t);

  if(lfc==0) {
    // point t lies outside or on the convex hull
    // we walk on the convex hull to find it out
    Face_circulator fc = incident_faces(infinite_vertex());
    Face_circulator done(fc);
    int ic = fc->index(infinite_vertex());
    if (xy_equal(t,fc->vertex(cw(ic))->point())){
      lt = VERTEX;
      li = cw(ic);
      return fc;
    }
    Orientation ori;
    do { // walking ccw around convex hull
      ic = fc->index(infinite_vertex());
      if (xy_equal(t,fc->vertex(ccw(ic))->point())){
        lt = VERTEX;
        li = ccw(ic);
        return fc;
      }
      ori = orientation( fc->vertex(cw(ic))->point(),
      fc->vertex(ccw(ic))->point(), t);
      if (ori == RIGHT_TURN) {
        lt = OUTSIDE_CONVEX_HULL;
        li = ic;
        return fc;
      }
      if (ori == COLLINEAR && collinear_between(fc->vertex(cw(ic))->point(),t,
                                                fc->vertex(ccw(ic))->point()) ) {
        lt = EDGE;
        li = ic;
        return fc;
      }
    } while (--fc != done);
    //should not arrive there;
    CGAL_triangulation_assertion(fc != done);
  }
  while(! lfc.locate(t, lt, li) ){
    ++lfc;
  }
  return lfc;
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Face_handle
Triangulation_T2<Gt,Tds>::
locate(const Point& p,
       Locate_type& lt,
       int& li,
       Face_handle start) const
{
  if( dimension() <= 0) {
    if (number_of_vertices() == 0) {
      lt = OUTSIDE_AFFINE_HULL;
      li = 4; // li should not be used in this case
    } else { // number_of_vertices() == 1
      if (xy_equal(p,finite_vertex()->point())){
        lt = VERTEX ;
      }
      else if (antipodal(p,finite_vertex()->point())) {
        std::cout << "The triangulation through antipodal points isn't defined\n";
        exit(0);
      }
      else{
        lt = OUTSIDE_AFFINE_HULL;
      }
      li = 4; // li should not be used in this case
    }
    return NULL;
  }
  if(dimension() == 1){
    if (number_of_vertices() == 2) {
      Face_handle ff = infinite_face();
      int iv = ff->index(infinite_vertex());
      Face_handle f = ff->neighbor(iv);
      if (antipodal(f->vertex(0)->point(),p) || antipodal(f->vertex(1)->point(),p)) {
        std::cout << "Triangulation through three points isn't defined when two of them are antipodal \n";
        exit(0);
      }
    }
    return march_locate_1D(p, lt, li);
  }
  if(start == NULL){
    Vertex_handle vstart = finite_vertex();
    start = vstart->incident_faces(); 
  }else if(is_infinite(start)){
    start = start->neighbor(start->index(infinite_vertex()));
  }
  return march_locate_2D(start, p, lt, li);
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>:: Face_handle
Triangulation_T2<Gt, Tds>::
locate(const Point &p,
       Face_handle start) const
{
  Locate_type lt;
  int li;
  return locate(p, lt, li, start);
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Finite_faces_iterator
Triangulation_T2<Gt, Tds>::
finite_faces_begin() const
{
  if ( dimension() < 2 )
    return finite_faces_end();
  return filter_iterator( all_faces_begin(),
                          all_faces_end(),
                          Infinite_tester(this));
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Finite_faces_iterator
Triangulation_T2<Gt, Tds>::
finite_faces_end() const
{
  return filter_iterator( all_faces_begin(),
                          all_faces_end(),
                          Infinite_tester(this),
                          all_faces_end() );
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Finite_vertices_iterator
Triangulation_T2<Gt, Tds>::
finite_vertices_begin() const
{
  if ( number_of_vertices() <= 0 )
    return finite_vertices_end();
  return filter_iterator(all_vertices_begin(),
                         all_vertices_end(),
                         Infinite_tester(this));
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Finite_vertices_iterator
Triangulation_T2<Gt, Tds>::
finite_vertices_end() const
{
  return filter_iterator(all_vertices_begin(),
                         all_vertices_end(),
                         Infinite_tester(this),
                         all_vertices_end());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Finite_edges_iterator
Triangulation_T2<Gt, Tds>::
finite_edges_begin() const
{
  if ( dimension() < 1 )
    return finite_edges_end();
  return filter_iterator(all_edges_begin(),
                         all_edges_end(),
                         infinite_tester());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Finite_edges_iterator
Triangulation_T2<Gt, Tds>::
finite_edges_end() const
{
    return filter_iterator(all_edges_begin(),
                           all_edges_end(),
                           infinite_tester(),
                           all_edges_end());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Point_iterator
Triangulation_T2<Gt, Tds>::
points_begin() const
{
  return Point_iterator(finite_vertices_begin());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Point_iterator
Triangulation_T2<Gt, Tds>::
points_end() const
{
  return Point_iterator(finite_vertices_end());
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::All_faces_iterator
Triangulation_T2<Gt, Tds>::
all_faces_begin() const
{
  return _tds.faces_begin();
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::All_faces_iterator
Triangulation_T2<Gt, Tds>::
all_faces_end() const
{
  return _tds.faces_end();
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::All_vertices_iterator
Triangulation_T2<Gt, Tds>::
all_vertices_begin() const
{
  return _tds.vertices_begin();
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::All_vertices_iterator
Triangulation_T2<Gt, Tds>::
all_vertices_end() const
{
  return _tds.vertices_end();
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::All_edges_iterator
Triangulation_T2<Gt, Tds>::
all_edges_begin() const
{
  return _tds.edges_begin();
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::All_edges_iterator
Triangulation_T2<Gt, Tds>::
all_edges_end() const
{
  return _tds.edges_end();
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_circulator
Triangulation_T2<Gt, Tds>::
incident_faces(Vertex_handle v, Face_handle f) const
{
  return _tds.incident_faces(v,f);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Face_iterator
Triangulation_T2<Gt, Tds>::
incident_faces(Vertex_handle v0, Vertex_handle v1,
               Face_iterator & result) const
{
  Face_circulator fc0 = incident_faces(v0), done(fc0);
  Face_iterator res;
  do {
    int index = fc0->index(v0);
    if (xy_equal(v1->point(),fc0->vertex(cw(index))->point())) {
      if (fc0 == done)
        result = res;
      res = fc0;
    }
    else if (xy_equal(v1->point(),fc0->vertex(ccw(index))->point())) {
      if (fc0 == done)
        result = res;
      res = fc0;
    }
    fc0++;
    res++;
  }while (fc0!=done);

  fc0 = incident_faces(v1);
  done = fc0;
  do {
    int index = fc0->index(v1);
    if (xy_equal(v0->point(),fc0->vertex(cw(index))->point()))
      res = fc0;
    else if (xy_equal(v0->point(),fc0->vertex(ccw(index))->point()))
      res = fc0;
    fc0++;
    res++;
  }while (fc0!=done);
  return res;
}




template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Vertex_circulator
Triangulation_T2<Gt, Tds>::
incident_vertices(Vertex_handle v, Face_handle f) const
{
  return _tds.incident_vertices(v,f);
}

template <class Gt, class Tds >
inline
typename Triangulation_T2<Gt, Tds>::Edge_circulator
Triangulation_T2<Gt, Tds>::
incident_edges(Vertex_handle v, Face_handle f) const
{
  return _tds.incident_edges(v,f);
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Line_face_circulator
Triangulation_T2<Gt, Tds>::
line_walk(const Point& p, const Point& q,  Face_handle f) const
{
  CGAL_triangulation_precondition( (dimension() == 2) &&
				! xy_equal(p,q));
  Line_face_circulator lfc = (f == NULL)
    ? Line_face_circulator(p, q, this)
    : Line_face_circulator(p, q, f, this);

  // the following lines may be useless :
  //  Line_face_circulator(p,q...) returns either a null circulator
  //  or a pointer to a finite face (to be checked)
  if( (!lfc.is_empty()) && is_infinite( lfc )){
    do {      ++lfc ;}
    while (is_infinite(lfc));
  }
  return lfc;
}

template <class Gt, class Tds >
Oriented_side
Triangulation_T2<Gt, Tds>::
oriented_side(const Point &p0, const Point &p1,
	      const Point &p2, const Point &p) const
{
  // return position of point p with respect to the oriented triangle p0p1p2
  // depends on the orientation of the vertices
  Bounded_side bs=bounded_side(p0,p1,p2,p);
  if (bs == ON_BOUNDARY) return ON_ORIENTED_BOUNDARY;
  Orientation      ot = orientation(p0, p1, p2);
  if (bs == ON_BOUNDED_SIDE)
    return (ot == LEFT_TURN) ? ON_POSITIVE_SIDE : ON_NEGATIVE_SIDE;
  // bs == ON_UNBOUNDED_SIDE
  return (ot == LEFT_TURN) ? ON_NEGATIVE_SIDE : ON_POSITIVE_SIDE;
}



template <class Gt, class Tds >
Bounded_side
Triangulation_T2<Gt, Tds>::
bounded_side(const Point &p0, const Point &p1,
	     const Point &p2, const Point &p) const
{
  // return position of point p with respect to triangle p0p1p2
  CGAL_triangulation_precondition( orientation(p0, p1, p2) != COLLINEAR);
  Orientation o1 = orientation(p0, p1, p),
              o2 = orientation(p1, p2, p),
              o3 = orientation(p2, p0, p);

  if (o1 == COLLINEAR){
    if (o2 == COLLINEAR ||  o3 == COLLINEAR) return ON_BOUNDARY;
    if (collinear_between(p0, p, p1))        return ON_BOUNDARY;
    return ON_UNBOUNDED_SIDE;
  }

  if (o2 == COLLINEAR){
    if (o3 == COLLINEAR)                     return ON_BOUNDARY;
    if (collinear_between(p1, p, p2))        return ON_BOUNDARY;
    return ON_UNBOUNDED_SIDE;
  }

  if (o3 == COLLINEAR){
    if (collinear_between(p2, p, p0))        return ON_BOUNDARY;
    return ON_UNBOUNDED_SIDE;
  }

  // from here none ot, o1, o2 and o3 are known to be non null
    if (o1 == o2 && o2 == o3)  return ON_BOUNDED_SIDE;
    return ON_UNBOUNDED_SIDE;
}


template <class Gt, class Tds >
Oriented_side
Triangulation_T2<Gt, Tds>::
oriented_side(Face_handle f, const Point &p) const
{
  CGAL_triangulation_precondition ( dimension()==2);
  return oriented_side(f->vertex(0)->point(),
		       f->vertex(1)->point(),
		       f->vertex(2)->point(),
		       p);
}


template <class Gt, class Tds >
bool
Triangulation_T2<Gt, Tds>::
antipodal(const Point& p, const Point& q) const
{
  return (p.antipodal(q));
}

template <class Gt, class Tds >
bool
Triangulation_T2<Gt, Tds>::
collinear_between(const Point& p, const Point& q,
                  const Point& r, const Vertex_handle& dir) const
{

  if (dir == NULL || !antipodal(p,r)) {
     return (has_on_segment(q,p,r) && q!=p && q!=r);
  }
  else {
    if (q == dir->point())
      return true;
    if (collinear_between(p,q,dir->point(),NULL))
      return true;
    return collinear_between(dir->point(),q,r,NULL);
  }
}

template <class Gt, class Tds >
bool
Triangulation_T2<Gt, Tds>::
collinear_has_on_segment(const Point& p,const Point& q,const Point& r) const
{
  return has_on_segment(q,p,r);
}

template <class Gt, class Tds >
inline
bool
Triangulation_T2<Gt, Tds>::
equal_x(const Point& p, const Point& q) const
{
  return (p.w()*q.x() == q.w()*p.x() && sign(p.w())==sign(q.w())
                      && sign(p.x())==sign(q.x()) );
}

template <class Gt, class Tds >
inline
bool
Triangulation_T2<Gt, Tds>::
equal_y(const Point& p, const Point& q) const
{
  return (p.w()*q.y() == q.w()*p.y() && sign(p.w())==sign(q.w())
             && sign(p.y())==sign(q.y()) );
}

template <class Gt, class Tds >
inline
bool
Triangulation_T2<Gt, Tds>::
xy_equal(const Point& p, const Point& q) const
{
  if ((p.w() != 0) || (q.w()!=0)) {
    if (!((p.w() * q.x()) == (p.x() * q.w())))
      return false;
    if (!((p.w() * q.y()) == (p.y() * q.w())))
      return false;
    if ( (sign(p.w()) != sign(q.w())) &&
         ( (sign(p.x()) != sign(q.x())) ||
         (p.x() == 0 && q.x() == 0)) &&
         ((sign(p.y()) != sign(q.y())) ||
         (p.y() == 0 && q.y() == 0)) )
      return false;
    return true;
  }
  else
    return (((p.y() * q.x()) == (p.x() * q.y()))) &&
             (sign(p.x()) == sign(q.x())) &&
             (sign(p.y()) == sign(q.y()));
}

template <class Gt, class Tds >
inline
Orientation
Triangulation_T2<Gt, Tds>::
orientation(const Point& p, const Point& q,const Point& r ) const
{
  if (sign_of_determinant3x3(p.w(),p.x(),p.y(),
                            q.w(),q.x(),q.y(),
                            r.w(),r.x(),r.y()) == 0)
    return COLLINEAR;
  if (sign_of_determinant3x3(p.w(),p.x(),p.y(),
                            q.w(),q.x(),q.y(),
                            r.w(),r.x(),r.y()) > 0)
    return LEFT_TURN;
  return RIGHT_TURN;
}

template<class Gt, class Tds>
inline
typename Triangulation_T2<Gt,Tds>::Point
Triangulation_T2<Gt,Tds>::
circumcenter (const Point& p0, const Point& p1, const Point&  p2) const
{
  return _gt.construct_circumcenter(p0,p1,p2);
}


template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Point
Triangulation_T2<Gt, Tds>::
circumcenter(Face_handle  f) const
{
  CGAL_triangulation_precondition (dimension()==2);
  return circumcenter((f->vertex(0))->point(),
                      (f->vertex(1))->point(),
                      (f->vertex(2))->point());
}


template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
show_all() const
{
  std::cerr<< "AFFICHE TOUTE LA TRIANGULATION :"<<std::endl;
  std::cerr << std::endl<<"====> "<< this;
  std::cerr <<  " dimension " <<  dimension() << std::endl;
  std::cerr << "nb of vertices " << number_of_vertices() << std::endl;

  if (dimension() < 1) return;
  if(dimension() == 1) {
    std::cerr<<" all edges "<<std::endl;
    All_edges_iterator aeit;
    for(aeit = all_edges_begin(); aeit != all_edges_end(); aeit++){
      show_face(aeit->first);
    }
    return;
  }

  std::cerr<<" faces finies "<<std::endl;
  Finite_faces_iterator fi;
  for(fi = finite_faces_begin(); fi != finite_faces_end(); fi++) {
    show_face(fi);
  }

  std::cerr <<" faces infinies "<<std::endl;
  All_faces_iterator afi;
  for(afi = all_faces_begin(); afi != all_faces_end(); afi++) {
    if(is_infinite(afi)) show_face(afi);
  }

  if (number_of_vertices()>1) {
    std::cerr << "affichage des sommets de la triangulation reguliere"
	      <<std::endl;
    //All_vertices_iterator vi;
    Finite_vertices_iterator vi;
    for( vi = finite_vertices_begin(); vi != finite_vertices_end(); vi++){
      if ((void*)(&(*(vi->face()))) != NULL) {
        show_vertex(vi);
        std::cerr << "  / face associee : "
                  << (void*)(&(*(vi->face())))<< std::endl; }
      }
      std::cerr<<std::endl;
  }
  return;
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
show_vertex(Vertex_handle vh) const
{
  if(is_infinite(vh)) std::cerr << "inf \t";
  else std::cerr << vh->point() << "\t";
  return;
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
show_face(Face_handle fh) const
{
  std::cerr << "face : "<<(void*)&(*fh)<<" => "<<std::endl;
  int i = fh->dimension();
  switch(i){
  case 0:
    std::cerr <<"point :" ; show_vertex(fh->vertex(0));
    std::cerr <<" / voisin " << &(*(fh->neighbor(0)));
    std::cerr <<"[" ; show_vertex(fh->neighbor(0)->vertex(0));
    std::cerr <<"]"  << std::endl;
    break;
  case 1:
     std::cerr <<"point :" ; show_vertex(fh->vertex(0));
     std::cerr <<" / voisin " << &(*(fh->neighbor(0)));
     std::cerr <<"[" ; show_vertex(fh->neighbor(0)->vertex(0));
     std::cerr <<"/" ; show_vertex(fh->neighbor(0)->vertex(1));
     std::cerr <<"]" <<std::endl;

     std::cerr <<"point :" ; show_vertex(fh->vertex(1));
     std::cerr <<" / voisin " << &(*(fh->neighbor(1)));
     std::cerr <<"[" ; show_vertex(fh->neighbor(1)->vertex(0));
     std::cerr <<"/" ; show_vertex(fh->neighbor(1)->vertex(1));
     std::cerr <<"]" <<std::endl;
     break;
  case 2:
    std::cerr <<"point :" ; show_vertex(fh->vertex(0));
    std::cerr <<" / voisin " << &(*(fh->neighbor(0)));
    std::cerr <<"[" ; show_vertex(fh->neighbor(0)->vertex(0));
    std::cerr <<"/" ; show_vertex(fh->neighbor(0)->vertex(1));
    std::cerr <<"/" ; show_vertex(fh->neighbor(0)->vertex(2));
    std::cerr <<"]" <<std::endl;

    std::cerr <<"point :" ; show_vertex(fh->vertex(1));
    std::cerr <<" / voisin " << &(*(fh->neighbor(1)));
    std::cerr <<"[" ; show_vertex(fh->neighbor(1)->vertex(0));
    std::cerr <<"/" ; show_vertex(fh->neighbor(1)->vertex(1));
    std::cerr <<"/" ; show_vertex(fh->neighbor(1)->vertex(2));
    std::cerr <<"]" <<std::endl;

    std::cerr <<"point :" ; show_vertex(fh->vertex(2));
    std::cerr <<" / voisin " << &(*(fh->neighbor(2)));
    std::cerr <<"[" ; show_vertex(fh->neighbor(2)->vertex(0));
    std::cerr <<"/" ; show_vertex(fh->neighbor(2)->vertex(1));
    std::cerr <<"/" ; show_vertex(fh->neighbor(2)->vertex(2));
    std::cerr <<"]" <<std::endl;
    break;
  }
  return;
}

template <class Gt, class Tds >
void
Triangulation_T2<Gt, Tds>::
file_output(std::ostream& os) const
{
  _tds.file_output(os, infinite_vertex(), true);
}

template <class Gt, class Tds >
typename Triangulation_T2<Gt, Tds>::Vertex_handle
Triangulation_T2<Gt, Tds>::
file_input(std::istream& is)
{
  clear();
  Vertex_handle v= _tds.file_input(is, true);
  set_infinite_vertex(v);
  return v;
}

template <class Gt, class Tds >
std::ostream&
operator<<(std::ostream& os, const Triangulation_T2<Gt, Tds> &tr)
{
  tr.file_output(os);
  return os ;
}



template < class Gt, class Tds >
std::istream&
operator>>(std::istream& is, Triangulation_T2<Gt, Tds> &tr)
{
  tr.file_input(is);
  CGAL_triangulation_assertion(tr.is_valid());
  return is;
}


CGAL_END_NAMESPACE

#endif // CGAL_TRIANGULATION_T2_H
