// Copyright (c) 1997  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Triangulation_data_structure_2.h,v $
// $Revision: 1.33 $ $Date: 2003/10/30 15:33:53 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Mariette Yvinec
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGULATION_DATA_STRUCTURE_T2_H
#define CGAL_TRIANGULATION_DATA_STRUCTURE_T2_H

#include <CGAL/Triangulation_data_structure_2.h>

CGAL_BEGIN_NAMESPACE

template < class Vb = Triangulation_ds_vertex_base_2<>,
           class Fb = Triangulation_ds_face_base_2<> >
class Triangulation_data_structure_T2
{
public:
  typedef Triangulation_data_structure_2<Vb,Fb>                tds_2;
  typedef Triangulation_data_structure_T2                      Tds_T2;
  typedef typename tds_2::Face_handle                          Face_handle;
  typedef typename tds_2::Vertex_handle                        Vertex_handle;
  typedef typename tds_2::Vertex                               Vertex;
  typedef typename tds_2::Face                                 Face;
  typedef typename tds_2::Edge                                 Edge;
  typedef typename tds_2::Face_iterator                        Face_iterator;
  typedef typename tds_2::Vertex_iterator                      Vertex_iterator;
  typedef typename tds_2::Edge_iterator                        Edge_iterator;
  typedef typename tds_2::Face_circulator                      Face_circulator;
  typedef typename tds_2::Vertex_circulator                    Vertex_circulator;
  typedef typename tds_2::Edge_circulator                      Edge_circulator;

  Triangulation_data_structure_T2();

  Triangulation_data_structure_T2(const Tds_T2 &tds1);

  Vertex_handle insert_in_edge(Face_handle f, int i);

  Vertex_handle insert_in_face(Face_handle f);
 
  void insert_in_infinite_vertex(Face_handle f, int i, Vertex_handle & v);

  Vertex_handle insert_in_infinite_vertex(Face_circulator f, 
                                 Face_circulator done,std::vector<int> v); //int i, Vertex_handle &v);

  void flip(Face_handle f, int i);

  Vertex_handle create_vertex(const Vertex &v);

  Vertex_handle insert_dim_up(Vertex_handle w,  bool orient);

  Vertex_handle insert_first();

  Vertex_handle	create_vertex();

  Vertex_handle insert_second();

  int number_of_vertices() const;

  int dimension() const;

  Face_iterator faces_begin() const;

  Face_iterator faces_end() const;

  Vertex_iterator vertices_begin() const;

  Vertex_iterator vertices_end() const;

  Edge_iterator edges_begin() const;

  Edge_iterator edges_end() const;

  Face_circulator incident_faces (Vertex_handle v,
                                  Face_handle f) const;

  Vertex_circulator incident_vertices(Vertex_handle v,
                                      Face_handle f = Face_handle()) const;

  Edge_circulator incident_edges(Vertex_handle v,
                                 Face_handle f = Face_handle()) const;

  bool operator= (const Tds_T2 &tds1);

  bool is_vertex(Vertex_handle v) const;

  ~Triangulation_data_structure_T2();

  Face_handle create_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3);

protected:
  tds_2 *tds;
};

template < class Vb, class Fb>
Triangulation_data_structure_T2<Vb,Fb> ::
Triangulation_data_structure_T2()
{
  tds = new tds_2();
}

template < class Vb, class Fb>
Triangulation_data_structure_T2<Vb,Fb> ::
Triangulation_data_structure_T2(const Tds_T2 &tds1)
{
  tds = new tds_2(tds1);
}

template < class Vb, class Fb>
Triangulation_data_structure_T2<Vb,Fb> ::
~Triangulation_data_structure_T2()
{
  delete tds;
}

template < class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb, Fb>::Face_handle
Triangulation_data_structure_T2<Vb, Fb>::
create_face(Vertex_handle v1, Vertex_handle v2, Vertex_handle v3)
{
  return tds->create_face(v1, v2, v3);
}

template <class Vb, class Fb >
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::create_vertex()
{
  return tds->create_vertex();
}

template <class Vb, class Fb >
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::insert_first() 
{
  return tds->insert_first();
}

template <class Vb, class Fb >
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::insert_second() 
{
  return tds->insert_second();
}

template <class Vb, class Fb >
int
Triangulation_data_structure_T2<Vb,Fb>::
number_of_vertices() const
{
  return tds->number_of_vertices();
}

template <class Vb, class Fb >
int
Triangulation_data_structure_T2<Vb,Fb>::
dimension() const
{
  return tds->dimension();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::
//insert_in_infinite_vertex(Face_handle f, int i, Vertex_handle & v)
insert_in_infinite_vertex(Face_circulator f, Face_circulator done,
std::vector<int> v)
{

  Vertex_handle vh = tds->create_vertex();
  int i = 0;
  do {
    f->set_vertex(v[i],vh);
    f++;
    i++;
  }while (f != done);
  return vh;

}

template <  class Vb, class Fb>
void
Triangulation_data_structure_T2<Vb,Fb>::
insert_in_infinite_vertex(Face_handle f, int i, Vertex_handle & v)
{
  v = f->vertex(i);
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::
insert_in_edge(Face_handle f, int i)
{
  CGAL_triangulation_precondition(f != Face_handle() && dimension() >= 1); 
  return tds->insert_in_edge(f,i);
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::
insert_in_face(Face_handle f)
{
  return tds->insert_in_face(f);	
}

template <class Vb, class Fb>
void
Triangulation_data_structure_T2<Vb,Fb>::
flip(Face_handle f, int i)
{
  tds->flip(f,i);
}

template <class Vb, class Fb>
inline
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::
create_vertex(const Vertex &v)
{
  return tds->create_vertex(v);
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_handle
Triangulation_data_structure_T2<Vb,Fb>::
 insert_dim_up(Vertex_handle w,  bool orient)
{
  return tds->insert_dim_up(w,orient);
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Face_iterator 
Triangulation_data_structure_T2<Vb,Fb>::faces_begin() const
{
  return tds->faces_begin();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Face_iterator 
Triangulation_data_structure_T2<Vb,Fb>::faces_end() const
{
  return tds->faces_end();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_iterator 
Triangulation_data_structure_T2<Vb,Fb>::vertices_begin() const
{
  return tds->vertices_begin();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_iterator 
Triangulation_data_structure_T2<Vb,Fb>::vertices_end() const
{
  return tds->vertices_end();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Edge_iterator 
Triangulation_data_structure_T2<Vb,Fb>::edges_begin() const
{
  return tds->edges_begin();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Edge_iterator 
Triangulation_data_structure_T2<Vb,Fb>::edges_end() const
{
  return tds->edges_end();
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Face_circulator Triangulation_data_structure_T2<Vb,Fb>::incident_faces
                                   (Vertex_handle v, 
                                    Face_handle f =  Face_handle()) const 
{
  return tds->incident_faces(v,f);
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Vertex_circulator Triangulation_data_structure_T2<Vb,Fb>::incident_vertices(
                                      Vertex_handle v, 
                                      Face_handle f) const
{    
  return tds->incident_vertices(v,f);  
}

template <  class Vb, class Fb>
typename Triangulation_data_structure_T2<Vb,Fb>::Edge_circulator
Triangulation_data_structure_T2<Vb,Fb>::incident_edges(
                                 Vertex_handle v, 
                                 Face_handle f) const
{
  return tds->incident_edges(v,f);
}

template <class Vb, class Fb>
inline bool
Triangulation_data_structure_T2<Vb,Fb>::
is_vertex(Vertex_handle v) const
{
  return tds->is_vertex(v);
}


CGAL_END_NAMESPACE

#endif
