// Copyright (c) 1997  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Triangulation_face_base_2.h,v $
// $Revision: 1.23 $ $Date: 2003/09/18 10:26:15 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Mariette Yvinec
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGULATION_FACE_BASE_T2_H
#define CGAL_TRIANGULATION_FACE_BASE_T2_H

#include <CGAL/basic.h>
#include <CGAL/triangulation_assertions.h>
#include <CGAL/Triangulation_short_names_2.h>
#include <CGAL/Triangulation_ds_face_T2.h>

CGAL_BEGIN_NAMESPACE 

template < typename Gt, typename Fb = Triangulation_ds_face_T2 >
class Triangulation_face_base_T2 
  : public Fb
{
public:
  typedef Gt                                           Geom_traits;
  typedef typename Fb::Vertex_handle                   Vertex_handle;
  typedef typename Fb::Face_handle                     Face_handle;

  template < typename TDS2 >
  struct Rebind_TDS {
    typedef typename Fb::template Rebind_TDS<TDS2>::Other  FbT2;
    typedef Triangulation_face_base_T2<Gt, FbT2>             Other;
  };

public:
  Triangulation_face_base_T2()
       : Fb() {}

  Triangulation_face_base_T2(Vertex_handle v0, 
			    Vertex_handle v1, 
			    Vertex_handle v2, Vertex_handle vdir = NULL)
    : Fb(v0,v1,v2,vdir) {}

  Triangulation_face_base_T2(Vertex_handle v0, 
			    Vertex_handle v1, 
			    Vertex_handle v2,
			    Face_handle n0, 
			    Face_handle n1, 
			    Face_handle n2, Vertex_handle vdir = NULL)
    : Fb(v0,v1,v2,n0,n1,n2,vdir) {}
};


CGAL_END_NAMESPACE 

#endif //CGAL_TRIANGULATION_FACE_BASE_T2_H
