// Copyright (c) 1997  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Triangulation_2/include/CGAL/Regular_triangulation_2.h,v $
// $Revision: 1.69 $ $Date: 2003/09/18 10:26:06 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Frederic Fichel, Mariette Yvinec, Julia Floetotto
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_TRIANGULATION_NOT_DELAUNAY_T2_H
#define CGAL_TRIANGULATION_NOT_DELAUNAY_T2_H

#include <CGAL/Triangulation_T2.h>
#include <CGAL/iterator.h>

CGAL_BEGIN_NAMESPACE

template < class Gt,
           class Tds = Triangulation_data_structure_T2 <
                           Triangulation_vertex_base_2<Gt> ,
Triangulation_face_base_2<Gt, Triangulation_ds_face_base_T2<> > > >
class Triangulation_not_Delaunay_T2 : public Triangulation_T2<Gt,Tds>
{
public:
  typedef Gt                                                   Geom_traits;
  typedef typename Geom_traits::K                           K;
  typedef typename K::FT                                    FT;
  typedef typename Geom_traits::Point_2                     Point;
  typedef typename Geom_traits::Point_T1                    Point_T1;
  typedef typename Geom_traits::Segment_2                   Segment;
  typedef typename Geom_traits::Triangle_2                  Triangle;
  typedef typename Geom_traits::Line_2                      Line;
  typedef typename Geom_traits::Direction_T2                Direction;
  typedef typename Geom_traits::Orientation_2               Orientation_2;


  typedef Triangulation_T2<Gt,Tds>                          Triangulation;
  typedef typename Triangulation::Locate_type               Locate_type;
  typedef typename Triangulation::Face_handle               Face_handle;
  typedef typename Triangulation::Face_circulator           Face_circulator;
  typedef typename Triangulation::Vertex_handle             Vertex_handle;
  typedef typename Triangulation::Edge                      Edge;
  typedef typename Triangulation::Edge_circulator           Edge_circulator;
  typedef typename Triangulation::Finite_edges_iterator
                                                            Finite_edges_iterator;
  typedef typename Triangulation::Finite_faces_iterator     Finite_faces_iterator;
  typedef typename Triangulation::Finite_vertices_iterator  Finite_vertices_iterator;
  typedef typename std::vector<Segment>::iterator           MDST_iterator;


  Triangulation_not_Delaunay_T2(const Gt& gt = Gt())
  : Triangulation_T2<Gt,Tds>(gt) {}

  bool is_not_delaunay(Face_handle& f, Face_handle& n, int index);

  bool is_delaunay(Face_handle& f, Face_handle& n, int index);

  //INSERTION
  Vertex_handle insert(const Point  &p,
                       Face_handle start = Face_handle(NULL) );
  Vertex_handle insert(const Point& p,
                       Locate_type lt,
                       Face_handle loc, int li );

  std::vector<Face_handle> incident_faces_not_Delaunay(const Vertex_handle v0,
                                                       const Vertex_handle v1)
  {
    Face_circulator fc0 = incident_faces(v0), done(fc0);
    std::vector<Face_handle> v;
    do {
      int index = fc0->index(v0);
      if (!is_infinite(fc0->vertex(cw(index)))) {
        if (xy_equal(v1->point(),fc0->vertex(cw(index))->point())) {
          v.push_back(create_face(fc0->vertex(0),fc0->vertex(1),fc0->vertex(2)));
        }
      }
      if (!is_infinite(fc0->vertex(ccw(index)))) {
        if (xy_equal(v1->point(),fc0->vertex(ccw(index))->point())) {
          v.push_back(create_face(fc0->vertex(0),fc0->vertex(1),fc0->vertex(2)));
        }
      }
      fc0++;
    }while (fc0!=done);


    fc0 = incident_faces(v1);
    done = fc0;
    do {
      int index = fc0->index(v1);
      if (!is_infinite(fc0->vertex(cw(index)))) {
        if (xy_equal(v0->point(),fc0->vertex(cw(index))->point()))
          v.push_back(create_face(fc0->vertex(0),fc0->vertex(1),fc0->vertex(2)));
      }
      if (!is_infinite(fc0->vertex(ccw(index)))) {
        if (xy_equal(v0->point(),fc0->vertex(ccw(index))->point()))
          v.push_back(create_face(fc0->vertex(0),fc0->vertex(1),fc0->vertex(2)));
      }
      fc0++;
    }while (fc0!=done);

    return v;
  }


private:
  void restore_not_Delaunay(Vertex_handle v);
  void propagating_flip(Face_handle& f,int i);

public:

  template < class InputIterator >
  int
  insert(InputIterator first, InputIterator last)
  {
    int n = number_of_vertices();
    while(first != last){
      insert(*first);
      ++first;
    }
    return number_of_vertices() - n;
  }
};

template < class Gt, class Tds >
inline
typename Triangulation_not_Delaunay_T2<Gt,Tds>::Vertex_handle
Triangulation_not_Delaunay_T2<Gt,Tds>::
insert(const Point  &p,  Face_handle start)
{
  Locate_type lt;
  int li;
  Face_handle loc = locate (p, lt, li, start);
  return insert(p, lt, loc, li);
}

template < class Gt, class Tds >
inline
typename Triangulation_not_Delaunay_T2<Gt,Tds>::Vertex_handle
Triangulation_not_Delaunay_T2<Gt,Tds>::
insert(const Point  &p, Locate_type lt, Face_handle loc, int li)
{
  Vertex_handle v = Triangulation_T2<Gt,Tds>::insert(p,lt,loc,li);
  restore_not_Delaunay(v);
  return(v);
}


template < class Gt, class Tds >
void
Triangulation_not_Delaunay_T2<Gt,Tds>::
restore_not_Delaunay(Vertex_handle v)
{
  if(dimension() <= 1) return;
  Face_handle f=v->face();
  Face_handle next;
  int i;
  Face_handle start(f);
  do {
    i = f->index(v);
    next = f->neighbor(ccw(i));

    propagating_flip(f,i);
    f=next;
  } while(next != start);
  return;
}


template < class Gt, class Tds >
bool
Triangulation_not_Delaunay_T2<Gt,Tds>::
is_delaunay(Face_handle& f, Face_handle& n, int index)
{
  int i = f->index(n);
  int in = n->index(f);

  Point a = f->vertex(i)->point();
  Point b = f->vertex(cw(i))->point();
  Point d = f->vertex(ccw(i))->point();
  Point c = n->vertex(in)->point();
  if (is_infinite(f))
    return true;

  if (Triangulation().antipodal(a,b))
    return true;
  if (Triangulation().antipodal(a,c))
    return true;
  if (Triangulation().antipodal(a,d))
    return true;
  if (Triangulation().antipodal(b,c))
    return true;
  if (Triangulation().antipodal(b,d))
    return true;
  if (Triangulation().antipodal(c,d))
    return true;
  if (orientation(a,b,c)==COUNTERCLOCKWISE || orientation(c,d,a)==COUNTERCLOCKWISE) {

    return true;
  }
  if (orientation(d,a,b)==COUNTERCLOCKWISE || orientation(b,c,d)==COUNTERCLOCKWISE) {

    return false;
  }
  if (orientation(a,b,c)==COLLINEAR || orientation(a,d,c)==COLLINEAR) {

    return true;
  }
  if (orientation(b,a,d)==COLLINEAR || orientation(b,c,d)==COLLINEAR) {

    return false;
  }
  if (sign(a.w()) == sign(b.w()) && sign(b.w()) == sign(d.w()) && sign(d.w()) != sign(c.w()))
      return true;
  if (sign(b.w()) == sign(d.w()) && sign(d.w()) == sign(c.w()) && sign(b.w()) != sign(a.w()))
      return true;
  if (sign(a.w()) == sign(b.w()) && sign(b.w()) == sign(c.w()) && sign(d.w()) != sign(c.w()))
      return false;
  if (sign(d.w()) == sign(a.w()) && sign(a.w()) == sign(c.w()) && sign(b.w()) != sign(a.w()))
      return false;

  if (sign(a.w()) == sign(b.w()) && sign(b.w()) == sign(c.w())  && sign(c.w()) == sign(d.w())
      && a.w()!=0) {

    Point c1 = circumcenter(a,b,d);
    if (sign(c1.w()) != sign(a.w()))
      c1 = Point(-c1.w(),-c1.x(),-c1.y());
    if (geom_traits().compare_T1(squared_distance(c1,a),squared_distance(c1,c)) == LARGER)
      return false;
    Point c2 = circumcenter(c,b,d);
    if (sign(c2.w()) != sign(c.w()))
      c2 = Point(-c2.w(),-c2.x(),-c2.y());
    if (geom_traits().compare_T1(squared_distance(c2,c),squared_distance(c2,a)) == LARGER)
      return false;
     return true;
  }
  //if the four points are proper
  else if (a.w()*b.w()*c.w()*d.w() != 0) {
    bool tr1 = false;
    bool tr2 = false;
    Point min_angle_tr1, min_angle_tr2;
    Point min_angle_t1 = Gt().find_minimum_angle(a,d,b);
    Point min_angle_t2 = Gt().find_minimum_angle(c,b,d);
    if  (K().compare_angle_with_x_axis_T2_object()(min_angle_t1,min_angle_t2) == LARGER)
      min_angle_tr1 = min_angle_t2;
    else {
      tr1 = true;
      min_angle_tr1 = min_angle_t1;
    }
    Point min_angle_t3 = Gt().find_minimum_angle(b,a,c);
    Point min_angle_t4 = Gt().find_minimum_angle(d,a,c);
    if (K().compare_angle_with_x_axis_T2_object()(min_angle_t3,min_angle_t4) == LARGER)
      min_angle_tr2 = min_angle_t4;
    else {
      tr2 = true;
      min_angle_tr2 = min_angle_t3;
    }
    return (K().compare_angle_with_x_axis_T2_object()(min_angle_tr2,min_angle_tr1) == SMALLER);
  }

  //three points are improper
  else if (a.w()!=0 && (b.w()==0 && c.w()==0 && d.w()==0))
    return false;
  else if (c.w()!=0 && (a.w()==0 && b.w()==0 && d.w()==0))
    return false;
  else if (b.w()!=0 && (a.w()==0 && c.w()==0 && d.w()==0))
    return true;
  else if (d.w()!=0 && (b.w()==0 && c.w()==0 && d.w()==0))
    return true;
  //two points are improper
  else if (a.w() == 0 &&  b.w() == 0) {
      return (geom_traits().compare_distance(b,d,a,c)!=LARGER);
  }
  else if (a.w() == 0 &&  c.w() == 0) { //b and d must be in the same hemisphere
    Direction dir_bd = Direction(b,d);
    Point p_inf = Point(0,dir_bd.dx(),dir_bd.dy());
    if (has_on_segment(p_inf,b,d))
      p_inf = antipode(p_inf);
    Segment smaller1,smaller2;
    Point a1,c1;
    if (geom_traits().compare_distance(a,p_inf,c,p_inf)!=LARGER)
      smaller1 = Segment(a,p_inf);
    else
      smaller1 = Segment(c,p_inf);
    a1 = Point(1,a.x(),a.y());
    FT norm = sqrt(a1.w()*a1.w() + a1.x()*a1.x() + a1.y()*a1.y());
    a1 = Point(norm,a1.x()*norm,a1.y()*norm);
    c1 = Point(1,c.x(),c.y());
    norm = sqrt(c1.w()*c1.w() + c1.x()*c1.x() + c1.y()*c1.y());;
    c1 = Point(norm,c1.x()*norm,c1.y()*norm);
    Direction dir_ac = perpendicular_direction(a1,c1);
    p_inf = Point(0,dir_ac.dx(),dir_ac.dy());
    if (has_on_segment(p_inf,a,c))
      p_inf = antipode(p_inf);
    if (geom_traits().compare_distance(d,p_inf,b,p_inf)!=LARGER)
      smaller2 = Segment(d,p_inf);
    else
      smaller2 = Segment(b,p_inf);
    return (geom_traits().compare_distance(smaller1.source(),
                                           smaller1.target(),
                                           smaller2.source(),
                                           smaller2.target())!=SMALLER);
  }
  else if (a.w() == 0 &&  d.w() == 0) //b and c must be in the same hemisphere
    return (geom_traits().compare_distance(b,d,a,c)!=LARGER);
  else if (b.w() == 0 &&  c.w() == 0) {
    //a and d must be in the same hemisphere
    return (geom_traits().compare_distance(b,d,a,c)!=LARGER);
  }
  else if (b.w() == 0 &&  d.w() == 0) {
    Direction dir_ac = Direction(a,c);
    Point p_inf = Point(0,dir_ac.dx(),dir_ac.dy());
    if (has_on_segment(p_inf,a,c))
      p_inf = antipode(p_inf);
    Segment smaller1,smaller2;
    Point b1,d1;
    if (geom_traits().compare_distance(b,p_inf,d,p_inf)!=LARGER)
      smaller1 = Segment(b,p_inf);
    else
      smaller1 = Segment(d,p_inf);
    b1 = Point(1,b.x(),b.y());
    FT norm = sqrt(b1.w()*b1.w() + b1.x()*b1.x() + b1.y()*b1.y());
    b1 = Point(norm,b1.x()*norm,b1.y()*norm);
    d1 = Point(1,d.x(),d.y());
    norm = sqrt(d1.w()*d1.w() + d1.x()*d1.x() + d1.y()*d1.y());
    d1 = Point(norm,d1.x()*norm,d1.y()*norm);
    Direction dir_bd = perpendicular_direction(b1,d1);
    p_inf = Point(0,dir_bd.dx(),dir_bd.dy());
    if (has_on_segment(p_inf,b,d))
      p_inf = antipode(p_inf);
    if (geom_traits().compare_distance(a,p_inf,c,p_inf)!=LARGER)
      smaller2 = Segment(a,p_inf);
    else
      smaller2 = Segment(c,p_inf);
    return (geom_traits().compare_distance(smaller1.source(),
                                           smaller1.target(),
                                           smaller2.source(),
                                           smaller2.target())!=SMALLER);
  }
  else if (c.w() == 0 &&  d.w() == 0) //a and b must be in the same hemisphere
    return (geom_traits().compare_distance(b,d,a,c)!=LARGER);

    else if (a.w() == 0 && ((sign(b.w()) == sign(c.w()) && sign(d.w()) != sign(c.w()))
             ||  (sign(c.w()) == sign(d.w()) && sign(d.w()) != sign(b.w())))) {
      Direction dir_bd = perpendicular_direction(b,d);
      Point pdir_bd = Point(0,dir_bd.dx(),dir_bd.dy());
      Direction dir = Direction(b,d);
      Point p_inf = Point(0,dir.dx(),dir.dy());
      Object obj = intersection(Segment(c,pdir_bd),Line(d,p_inf));
      Point p1,p2;
      Segment smaller1,smaller2;
      if (!assign(p1,obj)) {
        obj = intersection(Segment(c,antipode(pdir_bd)),Line(d,p_inf));
        assign(p1,obj);
      }
      obj = intersection(Segment(b,d),Line(Point(0,1,0),Point(0,0,1)));
      assign(p2,obj);
      if (geom_traits().compare_distance(c,p1,a,p2)!=LARGER)
        smaller1 = Segment(c,p1);
      else
        smaller1 = Segment(a,p2);
      Direction dir_ac;
      dir_ac = perpendicular_direction(a,c);

      Direction dir_ca = Direction(c,a);

      p_inf = Point(0,dir_ac.dx(),dir_ac.dy());
      Point pdir_ca =  Point(0,dir_ca.dx(),dir_ca.dy());

      obj = intersection(Segment(d,p_inf),Line(c,pdir_ca));
      if (!assign(p1,obj)){
        obj = intersection(Segment(d,antipode(p_inf)),Line(c,pdir_ca));
      }
      obj = intersection(Segment(b,p_inf),Line(c,pdir_ca));
      if (!assign(p2,obj)){
        obj = intersection(Segment(b,antipode(p_inf)),Line(c,pdir_ca));
        assign(p2,obj);
      }
      if (geom_traits().compare_distance(d,p1,b,p2)!=LARGER)
        smaller2 = Segment(d,p1);
      else
        smaller2 = Segment(b,p2);

      return (geom_traits().compare_distance(smaller1.source(),
                                           smaller1.target(),
                                           smaller2.source(),
                                           smaller2.target())!=SMALLER);
    }
    else if (b.w() == 0 && ((sign(d.w()) == sign(c.w()) && sign(c.w()) != sign(a.w()))
      || (sign(d.w()) == sign(a.w()) && sign(a.w()) != sign(c.w())))) {
      Direction dir_bd = perpendicular_direction(b,d);
      Point pdir_bd = Point(0,dir_bd.dx(),dir_bd.dy());
      Direction dir = Direction(b,d);
      Point p_inf = Point(0,dir.dx(),dir.dy());
      Object obj = intersection(Segment(c,pdir_bd),Line(d,p_inf));
      Point p1,p2;
      Segment smaller1,smaller2;
      if (!assign(p1,obj)) {
        obj = intersection(Segment(c,antipode(pdir_bd)),Line(d,p_inf));
        assign(p1,obj);
      }

      obj = intersection(Segment(a,pdir_bd),Line(d,p_inf));
      if (!assign(p2,obj)) {
        obj = intersection(Segment(a,antipode(pdir_bd)),Line(d,p_inf));
        assign(p2,obj);
      }

      if (geom_traits().compare_distance(c,p1,a,p2)!=LARGER)
        smaller1 = Segment(c,p1);
      else
        smaller1 = Segment(a,p2);

      obj = intersection(Segment(a,c),Line(Point(0,1,0),Point(0,0,1)));
      assign(p1,obj);

      Direction dir_ac;
      dir_ac = perpendicular_direction(a,c);

      Direction dir_ca = Direction(c,a);

      p_inf = Point(0,dir_ac.dx(),dir_ac.dy());
      Point pdir_ca =  Point(0,dir_ca.dx(),dir_ca.dy());

      obj = intersection(Segment(d,p_inf),Line(a,pdir_ca));
      if (!assign(p2,obj)){
        obj = intersection(Segment(d,antipode(p_inf)),Line(a,pdir_ca));
        assign(p2,obj);
      }

      if (geom_traits().compare_distance(b,p1,d,p2)!=LARGER)
        smaller2 = Segment(b,p1);
      else
        smaller2 = Segment(d,p2);
      return (geom_traits().compare_distance(smaller1.source(),
                                           smaller1.target(),
                                           smaller2.source(),
                                           smaller2.target())!=SMALLER);
    }
    else if (c.w() == 0 && ((sign(d.w()) == sign(a.w()) && sign(a.w()) != sign(b.w()))
      || (sign(a.w()) == sign(b.w()) && sign(a.w()) != sign(d.w()))))  {
      Direction dir_bd = perpendicular_direction(b,d);
      Point pdir_bd = Point(0,dir_bd.dx(),dir_bd.dy());
      Direction dir = Direction(b,d);
      Point p_inf = Point(0,dir.dx(),dir.dy());
      Object obj = intersection(Segment(a,pdir_bd),Line(d,p_inf));
      Point p1,p2;
      Segment smaller1,smaller2;
      if (!assign(p1,obj)) {
        obj = intersection(Segment(a,antipode(pdir_bd)),Line(d,p_inf));
        assign(p1,obj);
      }
      obj = intersection(Segment(b,d),Line(Point(0,0,1),Point(0,1,0)));
      assign(p2,obj);
      if (geom_traits().compare_distance(a,p1,c,p2)!=LARGER)
        smaller1 = Segment(a,p1);
      else
        smaller1 = Segment(c,p2);
      Direction dir_ac = perpendicular_direction(a,c);
      Point pdir_ac = Point(0,dir_ac.dx(),dir_ac.dy());
      Direction dir_ca = Direction(c,a);
      p_inf = Point(0,dir_ca.dx(),dir_ca.dy());
      obj = intersection(Segment(d,pdir_ac),Line(a,p_inf));
      if (!assign(p1,obj)) {
        obj = intersection(Segment(d,antipode(pdir_ac)),Line(a,p_inf));
        assign(p1,obj);
      }
      obj = intersection(Segment(b,pdir_ac),Line(a,p_inf));
      if (!assign(p2,obj)){
        obj = intersection(Segment(b,pdir_ac),Line(a,p_inf));
        assign(p2,obj);
      }
      if (geom_traits().compare_distance(d,p1,b,p2)!=LARGER)
        smaller2 = Segment(d,p1);
      else
        smaller2 = Segment(b,p2);

      return (geom_traits().compare_distance(smaller1.source(),
                                           smaller1.target(),
                                           smaller2.source(),
                                           smaller2.target())!=SMALLER);
    }

    Direction dir_bd = perpendicular_direction(b,d);

    Point pdir_bd = Point(0,dir_bd.dx(),dir_bd.dy());

    Direction dir1 = Direction(b,d);

    Point p_inf = Point(0,dir1.dx(),dir1.dy());

    Object obj = intersection(Segment(c,pdir_bd),Line(b,p_inf));

    Point p1,p2;

    Segment smaller1,smaller2;

    if (!assign(p1,obj)) {

      obj = intersection(Segment(c,antipode(pdir_bd)),Line(b,p_inf));

      assign(p1,obj);

    }
    obj = intersection(Segment(a,pdir_bd),Line(b,p_inf));

    if (!assign(p2,obj)) {

      obj = intersection(Segment(a,antipode(pdir_bd)),Line(b,p_inf));


      assign(p2,obj);

    }
    if (geom_traits().compare_distance(c,p1,a,p2)!=LARGER)  {

      smaller1 = Segment(c,p1);

    }
    else {

      smaller1 = Segment(a,p2);

    }
    obj = intersection(Segment(a,c),Line(Point(0,1,0),Point(0,0,1)));

    assign(p1,obj);

    Direction dir_ac;

    dir_ac = perpendicular_direction(a,c);
    Direction dir_ca = Direction(c,a);
    p_inf = Point(0,dir_ac.dx(),dir_ac.dy());

    Point pdir_ca =  Point(0,dir_ca.dx(),dir_ca.dy());
    obj = intersection(Segment(b,p_inf),Line(a,pdir_ca));

    if (!assign(p2,obj)){

      obj = intersection(Segment(b,antipode(p_inf)),Line(a,pdir_ca));

      assign(p2,obj);

    }
    if (geom_traits().compare_distance(d,p1,b,p2)!=LARGER) {

      smaller2 = Segment(d,p1);

    }
    else {

      smaller2 = Segment(b,p2);

    }

    return (geom_traits().compare_distance(smaller1.source(),
                                           smaller1.target(),
                                           smaller2.source(),
                                           smaller2.target())!=SMALLER);



}

template < class Gt, class Tds >
bool
Triangulation_not_Delaunay_T2<Gt,Tds>::
is_not_delaunay(Face_handle& f, Face_handle& n, int index)
{
  if (is_infinite(f))
    return true;
  return !is_delaunay(f,n,index);
}

template < class Gt, class Tds >
void
Triangulation_not_Delaunay_T2<Gt,Tds>::
propagating_flip(Face_handle& f,int i)
{
  Face_handle n = f->neighbor(i);

  if (is_not_delaunay(n, f, i))
    return;
  if (!(orientation(f->vertex(i)->point(),f->vertex(cw(i))->point(),
                    f->mirror_vertex(i)->point()) == RIGHT_TURN &&
        orientation(f->vertex(i)->point(),f->vertex(ccw(i))->point(),
                    f->mirror_vertex(i)->point()) == LEFT_TURN)) {
    if ((!orientation(f->vertex(i)->point(),f->vertex(cw(i))->point(),
                    f->mirror_vertex(i)->point()) == COLLINEAR) &&
        (!orientation(f->vertex(i)->point(),f->vertex(ccw(i))->point(),
                    f->mirror_vertex(i)->point()) == COLLINEAR))
      return;
  }
  flip(f, i);
  propagating_flip(f,i);
  i = n->index(f->vertex(i));
  propagating_flip(n,i);

}



CGAL_END_NAMESPACE

#endif // CGAL_DELAUNAY_TRIANGULATION_T2_H
