// Copyright (c) 1999  Max-Planck-Institute Saarbruecken (Germany).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Convex_hull_2/include/CGAL/ch_bykat.C,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:15:01 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_BYKAT_IMPLEMENTATION_T2_C
#define CGAL_CH_BYKAT_IMPLEMENTATION_T2_C

#ifndef CGAL_CH_BYKAT_IMPLEMENTATION_T2_H
#include <CGAL/ch_bykat_implementation_T2.h>
#endif // CGAL_CH_BYKAT_IMPLEMENTATION_T2_H


#include <CGAL/ch_assertions.h>
#include <CGAL/ch_selected_extreme_points_T2.h>
#include <CGAL/ch_graham_andrew.h>
#include <CGAL/stl_extensions.h>
#include <CGAL/functional.h>
#include <list>
#include <algorithm>

CGAL_BEGIN_NAMESPACE
template <class InputIterator, class OutputIterator, class Traits>
OutputIterator
ch_bykat_implementation_T2(InputIterator first, InputIterator last,
              OutputIterator  result,
              const Traits& ch_traits)
{
  typedef typename Traits::Point_T2                   Point_T2;
  typedef typename Traits::Left_turn_T2               Left_turn_T2;
  typedef typename Traits::Less_signed_distance_to_line_T2  Less_dist;
  typedef typename Traits::Equal_T2                         Equal_T2;

  Left_turn_T2 left_turn    = ch_traits.left_turn_T2_object();
  Less_dist   less_dist    = ch_traits.less_signed_distance_to_line_T2_object();
  Equal_T2     equal_points = ch_traits.equal_T2_object();

  if (first == last) return result;

  std::vector< Point_T2 >       P;    // Points in subsets
  std::vector< Point_T2 >       H;   // right endpoints of subproblems
  P.reserve(16);
  H.reserve(16);

  typedef typename std::vector< Point_T2 >::iterator   PointIterator;
  std::vector< PointIterator > L;      // start of subset range
  std::vector< PointIterator > R;      // end of subset range
  L.reserve(16);
  R.reserve(16);
  PointIterator           l;
  PointIterator           r;
  Point_T2                 a,b,c;

  std::copy(first,last,std::back_inserter(P));

  ch_we_point_T2(P.begin(), P.end(), l, r, ch_traits);
  
  a = *l;
  b = *r;

  if (equal_points(a,b))
  {
      *result = a;  ++result;
      return result;
  }

  OutputIterator  res(result);

  H.push_back( a );
  L.push_back( P.begin() );
  R.push_back( l = std::partition( P.begin(), P.end(),
                                   bind_1(bind_1(left_turn,a),b) ) );

  r = std::partition( l, P.end(), bind_1(bind_1(left_turn,b),a) );

  for (;;)
  {
    if ( l != r)
    {
        c = *std::min_element( l, r, bind_1(bind_1(less_dist, a), b));
        H.push_back( b );
        L.push_back( l );
        R.push_back( l = std::partition(l, r,
                                        bind_1(bind_1(left_turn,b),c)));

        r = std::partition(l, r, bind_1(bind_1(left_turn,c),a));
        b = c;
    }
    else
    {
      *res = a;  ++res;
      if ( L.empty() ) break;
      a = b;
      b = H.back(); H.pop_back();
      l = L.back(); L.pop_back();
      r = R.back(); R.pop_back();
    }
  }
  return res;

}



CGAL_END_NAMESPACE

#endif // CGAL_CH_BYKAT_IMPLEMENTATION_T2_C


