//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_C
#define CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_C

#include <CGAL/Cartesian.h>
#include <CGAL/Quotient.h>
#include <CGAL/Arr_segment_cached_traits_2.h>
#include <CGAL/Sweep_line_2.h>
#include <CGAL/polygon_operations_T2.h>
#include <CGAL/Segment_T2_Segment_T2_intersection.h>
#include <CGAL/functional.h>
#include <CGAL/algorithm.h>
#include <algorithm>
#include <vector>

#ifndef CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_H
#include <CGAL/ch_do_intersect_front_range_T2.h>
#endif //CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_H

CGAL_BEGIN_NAMESPACE



template <class ForwardIterator, class Traits>
inline
bool
ch_do_intersect_front_range_T2(ForwardIterator first1, 
															 ForwardIterator last1,
															 ForwardIterator first2,
															 ForwardIterator last2,
															 const Traits &traits)	
{
  typedef std::iterator_traits<ForwardIterator> 			  ITraits;
  typedef typename ITraits::value_type          			  value_type;
  typedef Kernel_traits<value_type>       			        KTraits;
  typedef typename KTraits::Kernel              			  Kernel;
	typedef typename Kernel::FT    											  FT;
	typedef Quotient<FT>                           			  NT;
	typedef Cartesian<NT>																	New_kernel;
	typedef Arr_segment_cached_traits_2<New_kernel >		  Traits_2;

	typedef typename Traits_2::Point_2                    Point_2;
	typedef typename Traits_2::Segment_2                  Segment_2;
	typedef typename Traits_2::Curve_2                    Curve_2;

	typedef typename std::list<Curve_2>                   CurveList;
	typedef typename CurveList::iterator                  CurveListIter;
	typedef CGAL::Sweep_line_2<CurveListIter, Traits_2>   Sweep_line;
	
	typedef typename Traits::Segment_T2 									Segment_T2;
	typedef typename Traits::Point_T2 										Point_T2;
	typedef typename Traits::Not_at_infinity_T2						
																							Not_at_infinity_T2;	
	typedef typename Traits::Compute_greater_x_one_range_infinity_T2
													 Compute_greater_x_one_range_infinity_T2;	
	typedef typename Traits::Compute_greater_y_one_range_infinity_T2
													 Compute_greater_y_one_range_infinity_T2;	
	typedef typename Traits::Compute_smaller_x_one_range_infinity_T2
													 Compute_smaller_x_one_range_infinity_T2;	
	typedef typename Traits::Compute_smaller_y_one_range_infinity_T2
													 Compute_smaller_y_one_range_infinity_T2;	
	typedef typename New_kernel::Has_on_2 	Has_on_2;
	typedef typename Traits::Point_smaller_relative_distance_to_omega_T2
													Point_smaller_relative_distance_to_omega_T2;

	Not_at_infinity_T2 not_at_infinity = 
										traits.not_at_infinity_T2_object();   
  Compute_greater_x_one_range_infinity_T2 
						compute_greater_x_one_range_infinity =
						traits.compute_greater_x_one_range_infinity_T2_object();   
  Compute_greater_y_one_range_infinity_T2 
						compute_greater_y_one_range_infinity =
						traits.compute_greater_y_one_range_infinity_T2_object();   
  Compute_smaller_x_one_range_infinity_T2 
						compute_smaller_x_one_range_infinity =
						traits.compute_smaller_x_one_range_infinity_T2_object();   
  Compute_smaller_y_one_range_infinity_T2 
						compute_smaller_y_one_range_infinity =
						traits.compute_smaller_y_one_range_infinity_T2_object(); 
	Point_smaller_relative_distance_to_omega_T2	
					  point_less_distance_infinity =
				traits.point_smaller_relative_distance_to_omega_T2_object();
	 New_kernel new_traits;
	 Has_on_2 has_on = new_traits.has_on_2_object();  						
	
	CGAL_precondition(first1 != last1 && first2 != last2);
	bool intersect = false;
	ForwardIterator it,it1,it2,it3;
	ForwardIterator f1 = first1;
	ForwardIterator l1 = last1;
	ForwardIterator f2 = first2;
	ForwardIterator l2 = last2;
	std::list<Curve_2> segments,segments1,segments2;
	std::vector<Segment_T2> v,v2,v3;
	Curve_2 c1;
	Segment_T2 seg,seg0,seg1;
	std::vector<Point_2> v_ch1,v_ch2;
	

	it1 = std::partition(f1,l1,not_at_infinity);
	it3 = std::partition(f2,l2,not_at_infinity);
	
	Point_T2 bbox1_p1,bbox1_p2,bbox2_p1,bbox2_p2,bbox1_p1x,
				bbox1_p1y,bbox1_p2x,bbox1_p2y;
	
	if (it1 == f1) {
		it1++;
		if (it1 != l1)
			seg0 = Segment_T2(*f1,*it1);
		else
			seg0 = Segment_T2(*f1,*f1);
		if (it3 == l2)
			return false;
		it = it3;
		it3++;
		if (it3 != l2)
			seg1 = Segment_T2(*it,*it3);
		else
			seg1 = Segment_T2(*it,*it);
		return do_intersect(seg0,seg1);
	}
	if (it3 == f2) {	
		
		it3++;
		if (it3 != l2)
			seg0 = Segment_T2(*f2,*it3);
		else
			seg0 = Segment_T2(*f2,*f2);
		if (it1 == l1)
			return false;
		it = it1;
		it1++;
		if (it1 != l1)
			seg1 = Segment_T2(*it,*it1);
		else
			seg1 = Segment_T2(*it,*it);
		return do_intersect(seg0,seg1);
	}


	it = f1;
	it2 = it;
	it2++;
	for (; it2!=it1; it2++) {
		v_ch1.push_back(Point_2(NT(it->x()*sign(it->w()),
												 abs(it->w())),
												 NT(it->y()*sign(it->w()),
												 abs(it->w()))));
		c1 = Curve_2(Point_2(NT(it->x()*sign(it->w()),
												 abs(it->w())),
												 NT(it->y()*sign(it->w()),
												 abs(it->w()))),
								 Point_2(NT(it2->x()*sign(it2->w()),
												 abs(it2->w())),	
												 NT(it2->y()*sign(it2->w()),
												 abs(it2->w()))));
		it = it2;
		segments.push_back(c1);
		segments1.push_back(c1);
	}
	v_ch1.push_back(Point_2(NT(it->x()*sign(it->w()),
												 abs(it->w())),
												 NT(it->y()*sign(it->w()),
												 abs(it->w()))));
	c1 = Curve_2(Point_2(NT(it->x()*sign(it->w()),
											 abs(it->w())),
											 NT(it->y()*sign(it->w()),
											 abs(it->w()))),
							 Point_2(NT(f1->x()*abs(f1->w()),abs(f1->w())),	
											 NT(f1->y()*abs(f1->w()),abs(f1->w()))));
	segments.push_back(c1);	
	segments1.push_back(c1);
	it = first2;
	it2 = it;
	it2++;
	
	for (; it2!=it3; it2++) {
		v_ch2.push_back(Point_2(NT(it->x()*sign(it->w()),
												 abs(it->w())),	
												 NT(it->y()*sign(it->w()),
												 abs(it->w()))));
		c1 = Curve_2(Point_2(NT(it2->x()*sign(it2->w()),
												 abs(it2->w())),
												 NT(it2->y()*sign(it2->w()),abs(it2->w()))),
								 Point_2(NT(it->x()*sign(it->w()),
												 abs(it->w())),	
												 NT(it->y()*sign(it->w()),
												 abs(it->w()))));
		it = it2;
		segments.push_back(c1);
		segments2.push_back(c1);
	}
	v_ch2.push_back(Point_2(NT(it->x()*sign(it->w()),
												 abs(it->w())),	
												 NT(it->y()*sign(it->w()),
												 abs(it->w()))));
	c1 = Curve_2(Point_2(NT(it->x()*sign(it->w()),abs(it->w())),
											 NT(it->y()*sign(it->w()),abs(it->w()))),
							 Point_2(NT(first2->x()*sign(first2->w()),
											 abs(first2->w())),	
											 NT(first2->y()*sign(first2->w()),
											 abs(first2->w()))));

	ch_bbox_T2(f1,it1,bbox1_p1,bbox1_p2,traits);
	ch_bbox_T2(f2,it2,bbox2_p1,bbox2_p2,traits);
	bbox1_p1x =
		compute_smaller_x_one_range_infinity(bbox1_p1,bbox2_p1);
	bbox1_p1y = 
		compute_smaller_y_one_range_infinity(bbox1_p1,bbox2_p1);
	bbox1_p2x =
		compute_greater_x_one_range_infinity(bbox1_p2,bbox2_p2);
	bbox1_p2y = 
		compute_greater_y_one_range_infinity(bbox1_p2,bbox2_p2);
	Segment_T2 s1(Point_T2(bbox1_p1x.w()*bbox1_p1y.w(),
												 bbox1_p1x.x()*bbox1_p1y.w(),
												 bbox1_p1y.y()*bbox1_p1x.w()),
								Point_T2(bbox1_p1y.w()*bbox1_p2x.w(),
												 bbox1_p1y.w()*bbox1_p2x.x(),
												 bbox1_p1y.y()*bbox1_p2y.w()));
	Segment_T2 s2(Point_T2(bbox1_p1y.w()*bbox1_p2x.w(),
												 bbox1_p1y.w()*bbox1_p2x.x(),
												 bbox1_p1y.y()*bbox1_p2y.w()),
								Point_T2(bbox1_p2x.w()*bbox1_p2y.w(),
												 bbox1_p2x.x()*bbox1_p2y.w(),
												 bbox1_p2x.w()*bbox1_p2y.y()));
	Segment_T2 s3(Point_T2(bbox1_p2x.w()*bbox1_p2y.w(),
												 bbox1_p2x.x()*bbox1_p2y.w(),
												 bbox1_p2x.w()*bbox1_p2y.y()),
								Point_T2(bbox1_p1x.w()*bbox1_p2y.w(),
												 bbox1_p1x.x()*bbox1_p2y.w(),
												 bbox1_p1x.w()*bbox1_p2y.y()));
	Segment_T2 s4(Point_T2(bbox1_p1x.w()*bbox1_p2y.w(),
												 bbox1_p1x.x()*bbox1_p2y.w(),
												 bbox1_p1x.w()*bbox1_p2y.y()),
								Point_T2(bbox1_p1x.w()*bbox1_p1y.w(),
												 bbox1_p1x.x()*bbox1_p1y.w(),
												 bbox1_p1x.w()*bbox1_p1y.y()));
	v2.push_back(s1);
	v2.push_back(s2);
	v2.push_back(s3);
	v2.push_back(s4); 
	
	ForwardIterator cur,previous; 
	Point_T2 p;
	unsigned int i,j;
	
	if (it1!=l1) {
		cur = first1;
		previous = cur;
		cur++;
		while (cur != last1)
		{
			if (!not_at_infinity(*cur) && not_at_infinity(*previous)) {
				v.push_back(Segment_T2(*previous,*cur));
			}
			else if (not_at_infinity(*cur) && !not_at_infinity(*previous)) {
				v.push_back(Segment_T2(*cur,*previous));
			}
			previous = cur;
			cur++;
		} 
		for (i=0; i < v.size(); i++) {
			for (j=0; j < v2.size(); j++) {
				Segment_T2_Segment_T2_pair<Kernel> segment_pair(&v[i],&v2[j]);
				
				if (segment_pair.intersection(p)) {
					
					segments.push_back(Curve_2(
														Point_2(NT(v[i].source().x()*
																		sign(v[i].source().w()),
														      	abs(v[i].source().w())),
																		NT(v[i].source().y()*
																		sign(v[i].source().w()),
														      	abs(v[i].source().w()))),
														Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
																	 NT(p.y()*sign(p.w()),abs(p.w())))));
					v_ch2.push_back(Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
											NT(p.y()*sign(p.w()),abs(p.w()))));
				}
				else if (segment_pair.intersection(seg)) {
					
					 //intersection is a segment
					segment_pair.intersection(seg);
					p = point_less_distance_infinity(seg.source(),
																						 seg.target(),
																						 v[i].target());
					segments.push_back(Curve_2(
														Point_2(NT(v[i].source().x()*
																		sign(v[i].source().w()),
														      	abs(v[i].source().w())),
																		NT(v[i].source().y()*
																		sign(v[i].source().w()),
														      	abs(v[i].source().w()))),
														Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
																	 NT(p.y()*sign(p.w()),abs(p.w())))));
					
					v_ch2.push_back(Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
											NT(p.y()*sign(p.w()),abs(p.w()))));
				}
				
			}
		} 
	}
	if (it2!=l2) {
		cur = first2;
		previous = cur;
		cur++;
		while (cur != last2)
		{
			if (!not_at_infinity(*cur) && not_at_infinity(*previous)) {
				v3.push_back(Segment_T2(*previous,*cur));
			}
			else if (not_at_infinity(*cur) && !not_at_infinity(*previous)) {
				v3.push_back(Segment_T2(*cur,*previous));
			}
			previous = cur;
			cur++;
		} 
		for (i=0; i < v3.size(); i++) {
			for (j=0; j < v2.size(); j++) {
				Segment_T2_Segment_T2_pair<Kernel> segment_pair(&v3[i],&v2[j]);
	
				if (segment_pair.intersection(p)) {
					segments.push_back(Curve_2(
														Point_2(NT(v3[i].source().x()*
																		sign(v3[i].source().w()),
														      	abs(v3[i].source().w())),
																		NT(v3[i].source().y()*
																		sign(v3[i].source().w()),
														      	abs(v3[i].source().w()))),
														Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
																	 NT(p.y()*sign(p.w()),abs(p.w())))));
					v_ch2.push_back(Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
											NT(p.y()*sign(p.w()),abs(p.w()))));
				}
				else if (segment_pair.intersection(seg)) {
					 //intersection is a segment
					segment_pair.intersection(seg);
					p = point_less_distance_infinity(seg.source(),
																						 seg.target(),
																						 v3[i].target());
					segments.push_back(Curve_2(
														Point_2(NT(v3[i].source().x()*
																		sign(v3[i].source().w()),
														      	abs(v3[i].source().w())),
																		NT(v3[i].source().y()*
																		sign(v3[i].source().w()),
														      	abs(v3[i].source().w()))),
														Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
																	 NT(p.y()*sign(p.w()),abs(p.w())))));
					
					v_ch2.push_back(Point_2(NT(p.x()*sign(p.w()),abs(p.w())),
											NT(p.y()*sign(p.w()),abs(p.w()))));
				}
				
			}
		} 
	}
	
	if (it1 != l1 && it3 !=l2)
	{
		i = 0;
		j = 0;
		while (i < v.size() && !intersect) {
			while (j < v3.size() && !intersect) {
				//std::cout << "do_intersect(v[i],v3[j]) = " << 
				//							do_intersect(v[i],v3[j]) << std::endl;
				if (do_intersect(v[i],v3[j])) {
					intersect = true;
	
				}
				j++;
	
			}
			i++;
	
		}
	
		std::cout << intersect << std::endl;
		if (intersect)
			return true;
	
		it2 = it1;
		it2++;
		if (it2 == l1)
			seg0 = Segment_T2(*it1,*it1);
		else
			seg0 = Segment_T2(*it1,*it2);
		it2 = it3;
		it2++;
	
		if (it2 == l2)
			seg1 = Segment_T2(*it3,*it3);
		else
			seg1 = Segment_T2(*it3,*it2);
		if (do_intersect(seg0,seg1))
			return true;
	}		
	
	
	segments.push_back(c1);	
	segments2.push_back(c1);
	Traits_2 traits2;
	// Use a sweep to create the sub curves  
 	std::list<Point_2> points;
	Sweep_line sl(&traits2);
	
	sl.get_intersection_points(segments.begin(), 
		   segments.end(), 
		   std::back_inserter(points), true);
	
	typename std::list<Point_2>::iterator scv_iter = points.begin(); 

	while (scv_iter != points.end() && !intersect) {	

		if (find_if(segments1.begin(),segments1.end(),
								bind_2(has_on,*scv_iter)) != segments1.end() &&
				find_if(segments2.begin(),segments2.end(),
								bind_2(has_on,*scv_iter)) != segments2.end())
			intersect = true;
		++scv_iter;
	}

	if (!intersect) {

		//verify if one point of ch1=[f1,it1) is in ch2=[first2,last2)
		i = 0;
		it = first1;
		do {
			i++;
			it++;
		}	while (it != last1 && i < 3);
		if (i >= 3) {
			if (oriented_side_polygon_T2(first1,last1,*first2) !=
																							 ON_NEGATIVE_SIDE) {

				return true;
			}
		}
		//verify if one point of ch1=[f1,it1) is in ch2=[first2,last2)
		i = 0;
		it = first2;
		do {
			i++;
			it++;
		}	while (it != last2 && i < 3);
		if (i >= 3) {
			//return 
			intersect= (oriented_side_polygon_T2(first2,last2,*first1) !=
																						 ON_POSITIVE_SIDE);

			return intersect;
		}
	}

  return intersect;
}


template <class ForwardIterator,class Point, class Traits>
inline
void
ch_bbox_T2(ForwardIterator first1, 
					 ForwardIterator last1,
					 Point &p1, Point &p2,
					 const Traits &traits)	
{
	typedef typename Traits::Point_T2 Point_T2;
	typedef typename Traits::Greater_x_one_range_infinity_T2
																		Greater_x_one_range_infinity_T2;	
	typedef typename Traits::Greater_y_one_range_infinity_T2
																		Greater_y_one_range_infinity_T2;	
	typedef typename Traits::Smaller_x_one_range_infinity_T2
																		Smaller_x_one_range_infinity_T2;	
	typedef typename Traits::Smaller_y_one_range_infinity_T2
																		Smaller_y_one_range_infinity_T2;	

	Greater_x_one_range_infinity_T2 greater_x_one_range_infinity =
										traits.greater_x_one_range_infinity_T2_object();   
  Greater_y_one_range_infinity_T2 greater_y_one_range_infinity =
										traits.greater_y_one_range_infinity_T2_object();   
  Smaller_x_one_range_infinity_T2 smaller_x_one_range_infinity =
										traits.smaller_x_one_range_infinity_T2_object();   
  Smaller_y_one_range_infinity_T2 smaller_y_one_range_infinity =
										traits.smaller_y_one_range_infinity_T2_object(); 

	ForwardIterator bigger_x1,bigger_y1,smaller_x1,smaller_y1,it;
	bigger_x1 = first1;
	it = bigger_x1;
	it++;
	while (it != last1) {
		if (greater_x_one_range_infinity(*it,*bigger_x1))
			bigger_x1 = it;
		it++;
	}
	bigger_y1 = first1;
	it = bigger_y1;
	it++;
	while (it != last1) {
		if (greater_y_one_range_infinity(*it,*bigger_y1))
			bigger_y1 = it;
		it++;
	}
	smaller_x1 = first1;
	it = smaller_x1;
	it++;
	while (it != last1) {
		if (smaller_x_one_range_infinity(*it,*smaller_x1))
			smaller_x1 = it;
		it++;
	}
	smaller_y1 = first1;
	it = smaller_y1;
	it++;
	while (it != last1) {
		if (smaller_y_one_range_infinity(*it,*smaller_y1))
			smaller_y1 = it;
		it++;
	}
	
	p1 = Point_T2(smaller_x1->w()*smaller_y1->w(),
								smaller_x1->x()*smaller_y1->w(),
								smaller_y1->y()*smaller_x1->w());

	p2 = Point_T2(bigger_x1->w()*bigger_y1->w(),
								bigger_x1->x()*bigger_y1->w(),
								bigger_y1->y()*bigger_x1->w());
}

CGAL_END_NAMESPACE
#endif // CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_C
