//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_H
#define CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_H

#include <CGAL/basic.h>
#include <iterator>

CGAL_BEGIN_NAMESPACE

template <class ForwardIterator, class Traits>
inline
bool
ch_do_intersect_front_range_T2(ForwardIterator first1, 
															 ForwardIterator last1,
															 ForwardIterator first2,
															 ForwardIterator last2,
															 const Traits &traits);

template <class ForwardIterator>
inline
bool
ch_do_intersect_front_range_T2(ForwardIterator first1, 
															 ForwardIterator last1,
															 ForwardIterator first2,
															 ForwardIterator last2)
{
  typedef std::iterator_traits<ForwardIterator> 		  ITraits;
  typedef typename ITraits::value_type          		  value_type;
  typedef Kernel_traits<value_type>       			      KTraits;
  typedef typename KTraits::Kernel              			Kernel;
	return 
		ch_do_intersect_front_range_T2(first1,last1,first2,last2,Kernel());
}

CGAL_END_NAMESPACE

#ifdef CGAL_CFG_NO_AUTOMATIC_TEMPLATE_INCLUSION
#include <CGAL/ch_do_intersect_front_range_T2.C>
#endif // CGAL_CFG_NO_AUTOMATIC_TEMPLATE_INCLUSION


#endif // CGAL_CH_DO_INTERSECT_FRONT_RANGE_T2_H
