// Copyright (c) 1999  Max-Planck-Institute Saarbruecken (Germany).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Convex_hull_2/include/CGAL/ch_graham_andrew.h,v $
// $Revision: 1.9 $ $Date: 2003/10/21 12:15:03 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_GRAHAM_ANDREW_T2_H
#define CGAL_CH_GRAHAM_ANDREW_T2_H

#include <CGAL/ch_graham_andrew_traits_T2.h>
#include <CGAL/ch_processor_T2.h>

CGAL_BEGIN_NAMESPACE

template <class ForwardIterator, class OutputIterator>
OutputIterator
ch_graham_andrew_T2(ForwardIterator first, ForwardIterator last, 
            				OutputIterator result)
{
  typedef std::iterator_traits<ForwardIterator> ITraits;
  typedef typename ITraits::value_type          value_type;
  typedef CGAL::Kernel_traits<value_type>       KTraits;
  typedef typename KTraits::Kernel              Kernel;
	return ch_processor_T2(first,last,result,Kernel(),
												 ch_graham_andrew_traits_T2());
}


template <class ForwardIterator, class OutputIterator, class Traits>
OutputIterator
ch_graham_andrew_T2(ForwardIterator first, ForwardIterator last, 
             OutputIterator  result,
             const Traits& ch_traits)
{
	return ch_processor_T2(first,last,result,ch_traits,
												 ch_graham_andrew_traits_T2());
}

CGAL_END_NAMESPACE

#endif //CGAL_CH_GRAHAM_ANDREW_T2_H
