// Copyright (c) 1999  Max-Planck-Institute Saarbruecken (Germany).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Convex_hull_2/include/CGAL/ch_graham_andrew.C,v $
// $Revision: 1.7 $ $Date: 2003/10/21 12:15:02 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_GRAHAM_ANDREW_IMPLEMENTATION_T2_C
#define CGAL_CH_GRAHAM_ANDREW_IMPLEMENTATION_T2_C

#ifndef CGAL_CH_GRAHAM_ANDREW_IMPLEMENTATION_T2_H
#include <CGAL/ch_graham_andrew_T2.h>
#endif // CGAL_CH_GRAHAM_ANDREW_IMPLEMENTATION_T2_H


#include <CGAL/ch_assertions.h>
#include <CGAL/stl_extensions.h>
#include <CGAL/ch_selected_extreme_points_T2.h>
#include <vector>
#include <algorithm>

CGAL_BEGIN_NAMESPACE

template <class ForwardIterator, class OutputIterator, class Traits>
OutputIterator
ch_graham_andrew_implementation_T2( ForwardIterator  first,
                       ForwardIterator  last,
                       OutputIterator result,
                       const Traits&  ch_traits)
{

  typedef  typename Traits::Point_T2                          Point_T2;
  typedef  typename Traits::Left_turn_T2                      Left_turn_T2;
  typedef  typename Traits::Equal_T2                          Equal_T2;
  typedef  typename  Traits::Ccw_around_point_or_collinear_and_closer_T2
                                                Ccw_around_point_or_collinear_and_closer_T2;
  typedef  typename std::vector< Point_T2 >::iterator         BidirectionalIterator;

  typedef  typename Traits::Back_range_or_minus_infinite_x_T2 Back_range_or_minus_infinite_x_T2;

  Equal_T2 equal_points = ch_traits.equal_T2_object();
  Ccw_around_point_or_collinear_and_closer_T2 rotation_predicate =
                     ch_traits.ccw_around_point_or_collinear_and_closer_T2_object();
  Left_turn_T2 left_turn = ch_traits.left_turn_T2_object();

  Back_range_or_minus_infinite_x_T2 is_at_back_range_or_minus_infinite_x =
                     ch_traits.back_range_or_minus_infinite_x_T2_object();


  if (first == last) {   return result;}

  last = std::unique(first,last);

  BidirectionalIterator q;
  q = first;
  q++;
  OutputIterator  res(result);
  if (q == last) {
    
    *res = *first;
    res++;
    return res;
  }
  q++;
  if (q == last) {
    *res = *first;
    q = first;
    q++;
    res++;
    *res = *q;
    res++;
    return res;
  }


  std::vector< Point_T2 >  V;
  std::vector< BidirectionalIterator >  Vres;
  BidirectionalIterator start, suc_start;


  if (std::find_if(first,last,is_at_back_range_or_minus_infinite_x)==last)
    ch_extreme_point_T2(first,last,start,Point_T2(1,0,0));
  else
    ch_extreme_point_T2(first,last,start,Point_T2(-1,0,0));
  
  std::copy( first, last, std::back_inserter(V) );
  std::sort(V.begin(),V.end(),bind_1(rotation_predicate,*start));
  for (q = V.begin(); q != V.end(); q++)
  

  Vres.push_back(V.begin());
  suc_start = V.begin();
  suc_start++;

  Vres.push_back(suc_start);
  q=suc_start;
  q++;
  typedef typename  std::vector<BidirectionalIterator>::reverse_iterator
                                                        reverse_iterator;
  reverse_iterator rev_iterator = Vres.rbegin();
  BidirectionalIterator beta = *(rev_iterator++);

  bool condition1 = true;
  bool condition2 = true;
  while (q != V.end()) {
    if (left_turn(**rev_iterator,*beta,*q)) {
      Vres.push_back(q);
      rev_iterator = Vres.rbegin();
      beta = *(rev_iterator++);
      q++;
    }
    else {
      condition1 = true;
      condition2 = true;
      do {
        Vres.pop_back();
        if (Vres.size() >= 2) {
          rev_iterator = Vres.rbegin();
          beta = *(rev_iterator++);
          condition1 = (!left_turn(**rev_iterator,*beta,*q));
        }
        else {
          Vres.push_back(q);
          q++;
          condition2 = false;
        }
      } while (condition1 && condition2);
    }
  }
  typedef typename std::vector< BidirectionalIterator >::iterator std_iterator;
  std_iterator iterator;
  for (iterator = Vres.begin();  iterator != Vres.end(); ++iterator)
  { *res = **iterator;  ++res; }


  return res;
}

CGAL_END_NAMESPACE

#endif // CGAL_CH_GRAHAM_ANDREW_IMPLEMENTATION_T2_C
