// Copyright (c) 1999  Max-Planck-Institute Saarbruecken (Germany).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Convex_hull_2/include/CGAL/ch_jarvis.C,v $
// $Revision: 1.8 $ $Date: 2003/10/21 12:15:03 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_JARVIS_IMPLEMENTATION_T2_C
#define CGAL_CH_JARVIS_IMPLEMENTATION_T2_C

#ifndef CGAL_CH_JARVIS_IMPLEMENTATION_T2_H
#include <CGAL/ch_jarvis_implementation_T2.h>
#endif // CGAL_CH_JARVIS_IMPLEMENTATION_T2_H

#ifndef CH_NO_POSTCONDITIONS
#include <CGAL/convexity_check_T2.h>
#endif // CH_NO_POSTCONDITIONS

#include <CGAL/ch_assertions.h>
#include <CGAL/ch_selected_extreme_points_T2.h>
#include <CGAL/functional.h>
//#include <CGAL/ch_selected_extreme_points_2.h>
#include <algorithm>


CGAL_BEGIN_NAMESPACE
template <class ForwardIterator, class ForwardIterator2, 
          class Point, class Traits>
ForwardIterator2
ch_jarvis_march_T2(ForwardIterator first, ForwardIterator last,
                const Point& start_p, 
                const Point& stop_p, 
                ForwardIterator2  result,
                const Traits& ch_traits)
{

  if (first == last) return result;
  typedef   typename Traits::Less_rotate_ccw_T2     Less_rotate_ccw;
  typedef   typename Traits::Point_T2               Point_T2;
  typedef   typename Traits::Equal_T2               Equal_T2; 
  
  Equal_T2     equal_points = ch_traits.equal_T2_object();   

	ForwardIterator ite = first;
	while (ite != last)
	{
		std::cout << "ite = "<< *ite << std::endl;
		ite++;
  }
 /* #if defined(CGAL_CH_NO_POSTCONDITIONS) || defined(CGAL_NO_POSTCONDITIONS) \
    || defined(NDEBUG)*/
  ForwardIterator2  res(result);
/*  #else
  Tee_for_output_iterator<OutputIterator,Point_T2> res(result);
  #endif // no postconditions ...*/
  CGAL_ch_assertion_code( \
      int count_points = 0; )
  CGAL_ch_assertion_code( \
      for (ForwardIterator fit = first; fit!= last; ++fit) ++count_points; std::cout << "count_points = " << count_points; )
  Less_rotate_ccw  
      rotation_predicate = ch_traits.less_rotate_ccw_T2_object( );
  *res = start_p;
  ++res;
  CGAL_ch_assertion_code( \
      int constructed_points = 1; )
  CGAL_ch_exactness_assertion_code( \
      Point previous_point = start_p; ) 

  
	ForwardIterator it = std::min_element( first, last, 
                                         bind_1(rotation_predicate, start_p) );
	while (! equal_points(*it, stop_p) )
  {
      CGAL_ch_exactness_assertion( \
          *it != previous_point );
      CGAL_ch_exactness_assertion_code( \
          previous_point = *it; )

      *res = *it;  ++res;
  
      CGAL_ch_assertion_code( \
          ++constructed_points;
)
      CGAL_ch_assertion( \
          constructed_points <= count_points + 1 );

      it = std::min_element( first, last, 
                             bind_1(rotation_predicate, *it) );
  } 
  return res;

}

template <class ForwardIterator, class ForwardIterator2, class Traits>
ForwardIterator2
ch_jarvis_implementation_T2(ForwardIterator first, ForwardIterator last, 
               ForwardIterator2  result, const Traits& ch_traits)
{
  if (first == last) return result;
  ForwardIterator start;
  ch_w_point_T2(first, last, start, ch_traits);
  return ch_jarvis_march_T2( first, last, *start, *start, result, ch_traits);
}

CGAL_END_NAMESPACE

#endif // CGAL_CH_JARVIS_IMPLEMENTATION_T2_C
