//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_MERGE_T2_C
#define CGAL_CH_MERGE_T2_C

#ifndef CGAL_CH_MERGE_T2_H
#include <CGAL/ch_merge_T2.h>
#endif // CGAL_CH_MERGE_T2_H

#include <CGAL/convexity_check_T2.h>
#include <CGAL/circulator.h>

CGAL_BEGIN_NAMESPACE

template <class BidirectionalIterator,class OutputIterator,class Traits>
OutputIterator
ch_merge_T2(BidirectionalIterator first1,
            BidirectionalIterator last1,
            BidirectionalIterator first2,
            BidirectionalIterator last2,
            OutputIterator result,
            const Traits& traits)
{



  CGAL_precondition(first1!=last1 || first2!=last2);
  

  typedef typename Traits::Orientation_T2 Orientation_T2;
  typedef typename Traits::Collinear_has_on_T2 Collinear_has_on_T2;
  typedef Circulator_from_iterator<BidirectionalIterator>  BidirectionalCirculator;

  Orientation_T2 orientation = traits.orientation_T2_object();
  Collinear_has_on_T2 collinear_has_on_segment = traits.collinear_has_on_T2_object();
  OutputIterator res(result);
  BidirectionalIterator it1;

  it1 = first1;
  while (it1 != last1) {

    it1++;
  }
  it1 = first2;
  while (it1 != last2) {

    it1++;
  }

  if (first1 == last1) {
    it1 = first2;
    while(it1 != last2)
    {
      *res = *it1;
      it1++;
      res++;
    }
    return res;
  }
  if (first2 == last2) {
    it1 = first1;
    while(it1 != last1)
    {
      *res = *it1;
      it1++;
      res++;
    }
    return res;
  }
  BidirectionalIterator it2,it3,it4;
  it1 = first1;
  it1++;
  it2 = first2;
  it2++;

  if (it1==last1 && it2==last2) {
    *res = *first1;
    res++;
    *res = *first2;
    res++;
    return res;
  }
  if (it1!=last1) {
    it3 = it1;
    it3++;
  }
  // a segment in [first1,last1) and a point in [first2,last2)
  if (it3 == last1 && it2 == last2) {
    if (orientation(*first2,*first1,*it1) == 0)
    {
      *res = *first2;
      res++;
      if (has_on_segment(*it1,*first2,*first1))
        *res = *first1;
      else
        *res = *it1;
    }
    else {
      *res=*first1;
      if (orientation(*first1,*it1,*first2) > 0) {
        *res = *it1;
        *res = *first2;
      }
      else {
        *res = *first2;
        *res = *it1;
      }
    }
    return res;
  }
  if (it2!=last2) {
    it4 = it2;
    it4++;
  }
  // a point in [first1,last1) and a segment in [first2,last2)
  if (it1 == last1 && it4 == last2) {
    if (orientation(*first1,*first2,*it2) == 0)
    {
      *res = *first2;
      res++;
      if (has_on_segment(*it2,*first2,*first1))
        *res = *first2;
      else
        *res = *it2;
    }
    else {
      *res=*first1;
      if (orientation(*first1,*it2,*first2) > 0) {
        *res = *it2;
        *res = *first2;
      }
      else {
        *res = *first2;
        *res = *it2;
      }
    }
    return res;
  }

  BidirectionalCirculator suc_p,pred_p,a1,a2,b1,b2,pred_q,suc_q;
  BidirectionalCirculator p(first1,last1);
  BidirectionalCirculator q(first2,last2);
  if (it3 == last1 && it4 == last2) {
    suc_p = p;
    suc_p++;
    suc_q = q;
    suc_q++;
    while (!(orientation(*p,*q,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q)) ||
      !(orientation(*p,*q,*suc_q) > 0 ||  has_on_segment(*suc_q,*p,*q)) ) {
      while (!(orientation(*p,*q,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q))) {
        p = suc_p;
        suc_p++;
      }
      while (!(orientation(*p,*q,*suc_q) > 0 || has_on_segment(*suc_q,*p,*q))) {
        q = suc_q;
        suc_q++;
      }
    }  
    a1 = p;
    b1 = q;
    while (!(orientation(*q,*p,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q)) ||
           !(orientation(*q,*p,*suc_q) > 0 || has_on_segment(*suc_q,*p,*q)) ) {
      while (!(orientation(*q,*p,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q))) {
        p = suc_p;
        suc_p++;
      }
      while (!(orientation(*q,*p,*suc_q) > 0 || has_on_segment(*suc_q,*p,*q))) {
        q = suc_q;
        suc_q++;
      }
    }  
    a2 = p;
    b2 = q;
  }
  else {

    a2 = p;
    b2 = q;
    pred_q = q;
    pred_q--;
    suc_q = q;
    suc_q++;

    if (it1 == last1)
    {
      while (!(orientation(*p,*q,*pred_q) > 0 && orientation(*p,*q,*suc_q) > 0) &&
             !(int(has_on_segment(*pred_q,*p,*q)) * int(orientation(*p,*q,*suc_q)) == 1) && 
             !(int(has_on_segment(*suc_q,*p,*q)) * int(orientation(*p,*q,*pred_q)) == 1)) {
        pred_q = q;
        q = suc_q;
        suc_q++;
      }
      a1 = p;
      b1 = q;
      pred_q = b2;
      pred_q--;
      suc_q = b2;
      suc_q++;
      while((!(orientation(*b2,*a2,*pred_q) > 0 && orientation(*b2,*a2,*suc_q) > 0) &&
             !(int(has_on_segment(*pred_q,*a2,*b2)) * int(orientation(*b2,*a2,*suc_q)) == 1) && 
             !(int(has_on_segment(*suc_q,*a2,*b2)) * int(orientation(*b2,*a2,*pred_q)) == 1))) {
        suc_q = b2;
        b2 = pred_q;
        pred_q--;
      }
    }
    else if (it2 == last2)
    {

      pred_p = p;
      pred_p--;
      suc_p = p;
      suc_p++;
      while (!(orientation(*p,*q,*pred_p) > 0 && orientation(*p,*q,*suc_p) > 0) &&
             !(int(has_on_segment(*pred_p,*p,*q)) * int(orientation(*p,*q,*suc_p)) == 1) && 
             !(int(has_on_segment(*suc_p,*p,*q)) * int(orientation(*p,*q,*pred_p)) == 1)) {
        pred_p = p;
        p = suc_p;
        suc_p++;
      }
      a1 = p;
      b1 = q;
      
      pred_p = a2;
      pred_p--;
      suc_p = a2;
      suc_p++;
      while((!(orientation(*b2,*a2,*pred_p) > 0 && orientation(*b2,*a2,*suc_p) > 0) &&
             !(int(has_on_segment(*pred_p,*a2,*b2)) * int(orientation(*b2,*a2,*suc_p)) == 1) && 
             !(int(has_on_segment(*suc_p,*a2,*b2)) * int(orientation(*b2,*a2,*pred_p)) == 1))) {
        suc_p = a2;
        a2 = pred_p;
        pred_p--;
      }
    }
    else if (it4 == last2) 
    {

      b2 = BidirectionalCirculator(first2,last2);
      b2++;
      pred_p = p;
      pred_p--;
      suc_p = p;
      suc_p++;
      suc_q = q;
      suc_q++;
      while (!(orientation(*p,*q,*suc_q) > 0 || has_on_segment(*suc_q,*p,*q)) ||
            (!(orientation(*p,*q,*pred_p) > 0 && orientation(*p,*q,*suc_p) > 0) &&
             !(int(has_on_segment(*pred_p,*p,*q)) * int(orientation(*p,*q,*suc_p)) == 1) && 
             !(int(has_on_segment(*suc_p,*p,*q)) * int(orientation(*p,*q,*pred_p)) == 1))) {
        while (!(orientation(*p,*q,*suc_q) > 0 || has_on_segment(*suc_q,*p,*q))) {
          q = suc_q;
          suc_q++;
        }
        while (!(orientation(*p,*q,*pred_p) > 0 && orientation(*p,*q,*suc_p) > 0) &&
               !(int(has_on_segment(*pred_p,*p,*q)) * int(orientation(*p,*q,*suc_p)) == 1) && 
               !(int(has_on_segment(*suc_p,*p,*q)) * int(orientation(*p,*q,*pred_p)) == 1))
        {

          pred_p = p;
          p = suc_p;
          suc_p++;
        }
      }

      a1 = p;
      b1 = q;
      pred_p = a2;
      pred_p--;
      suc_p = a2;
      suc_p++;
      suc_q= b2;
      suc_q++;
      while(!(orientation(*b2,*a2,*suc_q) > 0 || has_on_segment(*suc_q,*b2,*a2)) ||
           (!(orientation(*b2,*a2,*pred_p) > 0 && orientation(*b2,*a2,*suc_p) > 0) &&
            !(int(has_on_segment(*pred_p,*a2,*b2)) * int(orientation(*b2,*a2,*suc_p)) == 1) && 
            !(int(has_on_segment(*suc_p,*a2,*b2)) * int(orientation(*b2,*a2,*pred_p)) == 1)))  {
        while(!(orientation(*b2,*a2,*suc_q) > 0 || has_on_segment(*suc_q,*b2,*a2)))
        {
          b2 = suc_q;
          suc_q++;
        }
        while((!(orientation(*b2,*a2,*pred_p) > 0 && orientation(*b2,*a2,*suc_p) > 0) &&
          !(int(has_on_segment(*pred_p,*a2,*b2)) * int(orientation(*b2,*a2,*suc_p)) == 1) && 
          !(int(has_on_segment(*suc_p,*a2,*b2)) * int(orientation(*b2,*a2,*pred_p)) == 1)))  {
          suc_p = a2;
          a2 = pred_p;
          pred_p--;
        }
      }
    }
    else if (it3 == last1){

      a2 = BidirectionalCirculator(first1,last1);
      a2++;
      suc_p = p;
      suc_p++;
      suc_q = q;
      suc_q++;
      pred_q = q;
      pred_q--;
      std::cout << (!(orientation(*p,*q,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q))) << std::endl;
      while (!(orientation(*p,*q,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q)) ||
            (!(orientation(*p,*q,*pred_q) > 0 && orientation(*p,*q,*suc_q) > 0) &&
             !(int(has_on_segment(*pred_q,*p,*q)) * int(orientation(*p,*q,*suc_q)) == 1) && 
             !(int(has_on_segment(*suc_q,*p,*q)) * int(orientation(*p,*q,*pred_q)) == 1))) {
        std::cout << (!orientation(*p,*q,*suc_p) > 0 &&	!has_on_segment(*suc_p,*p,*q)) << std::endl;
        while (!(orientation(*p,*q,*suc_p) > 0 || has_on_segment(*suc_p,*p,*q))) {
          p = suc_p;
          suc_p++;
        }
        while (!(orientation(*p,*q,*pred_q) > 0 && orientation(*p,*q,*suc_q) > 0) &&
               !(int(has_on_segment(*pred_q,*p,*q)) * int(orientation(*p,*q,*suc_q)) == 1) && 
               !(int(has_on_segment(*suc_q,*p,*q)) * int(orientation(*p,*q,*pred_q)) == 1)) {
          pred_q = q;
          q = suc_q;
          suc_q++;
        }
      }
      a1 = p;
      b1 = q;
      pred_q = b2;
      pred_q--;
      suc_q = b2;
      suc_q++;
      suc_p = a2;
      suc_p++;

      while(!(orientation(*b2,*a2,*suc_p) > 0 || has_on_segment(*suc_p,*b2,*a2)) ||
           (!(orientation(*b2,*a2,*pred_q) > 0 && orientation(*b2,*a2,*suc_q) > 0) &&
            !(int(has_on_segment(*pred_q,*a2,*b2)) * int(orientation(*b2,*a2,*suc_q)) == 1) && 
            !(int(has_on_segment(*suc_q,*a2,*b2)) *  int(orientation(*b2,*a2,*pred_q)) == 1))) {
        while(!(orientation(*b2,*a2,*suc_p) > 0 || has_on_segment(*suc_p,*b2,*a2))) {
          a2 = suc_p;
          suc_p++;
        }
        while((!(orientation(*b2,*a2,*pred_q) > 0 &&  orientation(*b2,*a2,*suc_q) > 0) &&
               !(int(has_on_segment(*pred_q,*a2,*b2)) * int(orientation(*b2,*a2,*suc_q)) == 1) && 
               !(int(has_on_segment(*suc_q,*a2,*b2))* int(orientation(*b2,*a2,*pred_q)) == 1))) {
          suc_q = b2;
          b2 = pred_q;
          pred_q--;
       }
     }
   }
   else {
     pred_p = p;
     pred_p--;
     suc_p = p;
     suc_p++;
     while ((!(orientation(*p,*q,*pred_p) > 0 && orientation(*p,*q,*suc_p) > 0) &&
             !(int(has_on_segment(*pred_p,*p,*q)) * int(orientation(*p,*q,*suc_p)) == 1) && 
             !(int(has_on_segment(*suc_p,*p,*q))* int(orientation(*p,*q,*pred_p)) == 1)) || 
            (!(orientation(*p,*q,*pred_q) > 0 && orientation(*p,*q,*suc_q) > 0) &&
             !(int(has_on_segment(*pred_q,*p,*q))* int(orientation(*p,*q,*suc_q)) == 1) && 
             !(int(has_on_segment(*suc_q,*p,*q))* int(orientation(*p,*q,*pred_q)) == 1))) {
       while (!(orientation(*p,*q,*pred_p) > 0 && orientation(*p,*q,*suc_p) > 0) &&
              !(int(has_on_segment(*pred_p,*p,*q))* int(orientation(*p,*q,*suc_p)) == 1) && 
              !(int(has_on_segment(*suc_p,*p,*q))* int(orientation(*p,*q,*pred_p)) == 1)) {
         pred_p = p;
         p = suc_p;
         suc_p++;
       }
       while (!(orientation(*p,*q,*pred_q) > 0 && orientation(*p,*q,*suc_q) > 0) &&
              !(int(has_on_segment(*pred_q,*p,*q)) * int(orientation(*p,*q,*suc_q)) == 1) && 
              !(int(has_on_segment(*suc_q,*p,*q)) * int(orientation(*p,*q,*pred_q)) == 1)) {
         pred_q = q;
         q = suc_q;
         suc_q++;
       }
     }
     a1 = p;
     b1 = q;
     pred_p = a2;
     pred_p--;
     suc_p = a2;
     suc_p++;
     pred_q = b2;
     pred_q--;
     suc_q = b2;
     suc_q++;
     while (((!(orientation(*b2,*a2,*pred_p) > 0 && orientation(*b2,*a2,*suc_p) > 0) &&
              !(int(has_on_segment(*pred_p,*a2,*b2)) * int(orientation(*b2,*a2,*suc_p)) == 1) && 
              !(int(has_on_segment(*suc_p,*a2,*b2)) * int(orientation(*b2,*a2,*pred_p)) == 1) ) ) || 
            ((!(orientation(*b2,*a2,*pred_q) > 0 && orientation(*b2,*a2,*suc_q) > 0) &&
              !(int(has_on_segment(*pred_q,*a2,*b2))* int(orientation(*b2,*a2,*suc_q)) == 1) && 
              !(int(has_on_segment(*suc_q,*a2,*b2))* int(orientation(*b2,*a2,*pred_q)) == 1)))) {
       while((!(orientation(*b2,*a2,*pred_p) > 0 && orientation(*b2,*a2,*suc_p) > 0) &&
              !(int(has_on_segment(*pred_p,*a2,*b2)) * int(orientation(*b2,*a2,*suc_p)) == 1) && 
              !(int(has_on_segment(*suc_p,*a2,*b2))*int(orientation(*b2,*a2,*pred_p)) == 1) ))  {
         suc_p = a2;
         a2 = pred_p;
         pred_p--;
       }
       while((!(orientation(*b2,*a2,*pred_q) > 0 && orientation(*b2,*a2,*suc_q) > 0) &&
              !(int(has_on_segment(*pred_q,*a2,*b2))* int(orientation(*b2,*a2,*suc_q)) == 1) && 
              !(int(has_on_segment(*suc_q,*a2,*b2))* int(orientation(*b2,*a2,*pred_q)) == 1))) {
         suc_q = b2;
         b2 = pred_q;
         pred_q--;
       }
     }
   }
  }
  BidirectionalCirculator aux1(first1,last1);
  while (aux1 != a1) {
    aux1++;
  } 
  a1 = aux1;
  aux1= BidirectionalCirculator(first2,last2);
  while (aux1 != b2) {
    aux1++;
  } 
  b2 = aux1;
  suc_p=a1;
  suc_p++;
  pred_p = a1;
  pred_p--;
  suc_q=b2;
  suc_q++;
  pred_q=b2;
  pred_q--;
  bool sense1=false;
  bool sense2=false;

  if (orientation(*a1,*b1,*b2) < 0) {
    if (orientation(*a1,*pred_p,*b1)!=orientation(*a1,*pred_p,*suc_p))
      sense1 = true; 
    if (orientation(*b2,*pred_q,*a2)!=orientation(*b2,*pred_q,*suc_q))
      sense2 = true;
    *res = *a1;
    while (a1!=a2) {
      res++;
      if (sense1)
        a1++;
      else
        a1--;
      *res = *a1;
    } 
    *res = *b2;
    while (b2!=b1) {
      res++;
      if (sense2)
        b2++;
      else
        b2--;
      *res = *b2;
    } 
  }
  else {
    aux1= BidirectionalCirculator(first2,last2);
    while (aux1 != b1) {
      aux1++;
    } 
    b1 = aux1;
    aux1= BidirectionalCirculator(first1,last1);
    while (aux1 != a2) {
      aux1++;
    } 
    a2 = aux1;
    suc_q = b1;
    suc_q++;
    pred_q = b1;
    pred_q--;
    
    pred_p = a2;
    pred_p--;
    suc_p = a2;
    suc_p++;
    if (orientation(*a2,*pred_p,*b2)!=orientation(*a2,*pred_p,*suc_p))
      sense2 = true; 
    if (orientation(*b1,*pred_q,*a1)!=orientation(*b1,*pred_q,*suc_q))
      sense1 = true;

    *res = *b1;
    while (b1!=b2) {
      res++;
      if (sense1)
        b1++;
      else
        b1--;
      *res = *b1;
    }
    *res = *a2;
    while (a2!=a1) {
      res++;
      if (sense2)
        a2++;
      else
        a2--;
      *res = *a2;
    }
  }
  return res;
}
CGAL_END_NAMESPACE
#endif //CGAL_CH_MERGE_T2_C
