//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_POSITION_T2_C
#define CGAL_CH_POSITION_T2_C

#ifndef CGAL_CH_POSITION_T2_H
#include <CGAL/ch_position_T2.h>
#endif // CGAL_CH_POSITION_T2_H

#include <algorithm>

CGAL_BEGIN_NAMESPACE

/*
template <class ForwardIterator, class Traits>
bool
ch_is_left_T2(ForwardIterator first1, ForwardIterator last1,
	   ForwardIterator first2, ForwardIterator last2,
	   const Traits& traits)
{
	typedef typename Traits::Smaller_x_one_range_infinity_T2  
					Smaller_one_range_infinity_T2;
	typedef typename Traits::Back_range_T2  Back_range_T2;

	Smaller_one_range_infinity_T2 smaller_one_range_infinity =
		traits.smaller_x_one_range_infinity_T2_object();

	Back_range_T2 back_range = traits.back_range_T2_object();

	ForwardIterator it1,it2;
	bool changed1 = false;
	bool changed2 = false;
	bool result;

	if (at_least_one_element_at_container(first1,last1,back_range)) {
		changed1 = true;
		for(it1 = first1; it1!=last1; ++it1) {
			*it1 = antipode(*it1);
			std::cout << *it1 << std::endl;
		}
	}
	if (at_least_one_element_at_container(first2,last2,back_range)) {
		changed2 = true;
		for(it2 = first2; it2!=last2; ++it2) {
			*it2 = antipode(*it2);
			std::cout << *it2 << std::endl;
		}
	}

	it1 = std::min_element(first1, last1, smaller_one_range_infinity);
	it2 = std::min_element(first2, last2, smaller_one_range_infinity);

	result = smaller_one_range_infinity(*it1,*it2);
	std::cout << "pos 0 \n";
	if (changed1) {
		for(ForwardIterator it11 = first1; it11!=last1; ++it11)
				*it11 = antipode(*it11);
	}
	std::cout << "pos 1 \n";
	if (changed2) {
		for(ForwardIterator it22 = first2; it22!=last2; ++it22) 
				*it22 = antipode(*it22);
	}
		std::cout << "pos 2 \n";
	return result; 
}
*/

template <class ForwardIterator, class Traits>
bool
ch_is_left_T2(ForwardIterator first1, ForwardIterator last1,
	   ForwardIterator first2, ForwardIterator last2,
	   const Traits& traits)
{
	//PRECONDICAO: OS CASCOS NAO SE INTERCEPTAM

	typedef typename Traits::Smaller_equal_x_one_range_infinity_T2  
					Smaller_equal_one_range_infinity_T2;
	typedef typename Traits::Back_range_T2  Back_range_T2;
	typedef typename Traits::Antipode_T2 Antipode_T2;  

	Smaller_equal_one_range_infinity_T2 smaller_equal_one_range_infinity =
		traits.smaller_equal_x_one_range_infinity_T2_object();

	Back_range_T2 back_range = traits.back_range_T2_object();

	Antipode_T2 antipode = traits.antipode_T2_object();

	ForwardIterator it1,it2;
	bool changed1 = false;
	bool changed2 = false;
	bool result;

	if (find_if(first1,last1,back_range) != last1) {
		transform(first1,last1,first1,antipode);
		it1 = std::min_element(first1, last1,
											 smaller_equal_one_range_infinity);
		changed1 = true;
	}
	else 
		it1 = std::min_element(first1, last1, 	
											 smaller_equal_one_range_infinity);

	if (find_if(first2,last2,back_range) != last2) {
		transform(first2,last2,first2,antipode);
		it2 = std::min_element(first2, last2, 
											 smaller_equal_one_range_infinity);
		changed2 = true;
	}
	else 
		it2 = std::min_element(first2, last2, 
											 smaller_equal_one_range_infinity);
	
	result = smaller_equal_one_range_infinity(*it1,*it2);
	if (changed2)
		transform(first2,last2,first2,antipode);
	if (changed1)
		transform(first1,last1,first1,antipode);
	return result;
}

CGAL_END_NAMESPACE

#endif //CGAL_CH_POSITION_T2_C
