//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_PROCESSOR_T2_H
#define CGAL_CH_PROCESSOR_T2_H

#include <CGAL/ch_do_intersect_front_range_T2.h>
#include <CGAL/ch_merge_T2.h>
#include <algorithm>
#include <iterator>
#include <vector>


CGAL_BEGIN_NAMESPACE

template <class ForwardIterator, class OutputIterator, 
          class Traits, class Algorithm_Traits>
OutputIterator
ch_processor_T2(ForwardIterator first, ForwardIterator last, 
             OutputIterator  result,const Traits& kernel_traits, 
             const Algorithm_Traits& ch_algorithm_traits)
{
  typedef typename Algorithm_Traits::Ch_generate_T2          Ch_generate_T2;
  typedef std::iterator_traits<ForwardIterator>              ITraits;
  typedef typename ITraits::value_type                       value_type;
  typedef typename Traits::Point_T2                          Point_T2;
  typedef typename Traits::Segment_T2                        segment;
  typedef typename Traits::Antipode_T2                       Antipode_T2;
  typedef typename Traits::Back_range_or_minus_infinite_x_T2
                                                             Back_range_or_minus_infinite_x_T2;

  /*separate the points between two sets: one of points at 
    front range and infinity of T2 and the other of ones at 
    the back range*/

  Antipode_T2 antipode = kernel_traits.antipode_T2_object();
  Back_range_or_minus_infinite_x_T2 
                is_at_back_range_or_minus_infinite_x = 
                       kernel_traits.back_range_or_minus_infinite_x_T2_object();

  CGAL_precondition(first!=last);


  ForwardIterator ch1_first;
  ForwardIterator ch1_last;

  ForwardIterator ch2_first = first;
  ForwardIterator ch2_last = last;

  ch2_last = std::partition(ch2_first,last,is_at_back_range_or_minus_infinite_x);
  ch1_first = ch2_last;
  ch1_last = last;



  Ch_generate_T2 ch_generate = ch_algorithm_traits.ch_generate_T2_object();

  //construct the convex hull for each side of T2

  std::vector<value_type> v1,v1_aux;


  if (ch1_first != ch1_last)
    ch_generate(ch1_first,ch1_last, std::back_inserter(v1),kernel_traits);


  std::vector<value_type> v2,v2_aux;

  if (ch2_first != ch2_last)
    ch_generate(ch2_first,ch2_last,std::back_inserter(v2),kernel_traits);
   
  typename std::vector<value_type>::iterator  v2_it;

  std::copy(v1.begin(),v1.end(),std::back_inserter(v1_aux));
  std::copy(v2.begin(),v2.end(),std::back_inserter(v2_aux));
  //project the convex hull at the back range (ch2) to the front range
  transform(v2.begin(),v2.end(),v2.begin(),antipode);


  for (v2_it = v2.begin(); v2_it != v2.end(); ++v2_it)
  {
    std::cout << *v2_it << std::endl;
  }
  
  if (!(v1.begin()!=v1.end() && v2.begin()!=v2.end())) {
    transform(v2.begin(),v2.end(),v2.begin(),antipode);
    return ch_merge_T2(v1.begin(),v1.end(),v2.begin(),v2.end(),result,kernel_traits);
  }

  
  if (!ch_do_intersect_front_range_T2(v1.begin(),v1.end(),v2.begin(), v2.end(),kernel_traits))
  {

       
    //project back the convex hull to the front range
    //transform(v2.begin(),v2.end(),v2.begin(),antipode);
  for (v2_it = v2.begin(); v2_it != v2.end(); ++v2_it)
  {
    std::cout << *v2_it << std::endl;
  }
    //verify if both ch intersect
    //if so, merge them 
    return ch_merge_T2(v1_aux.begin(),v1_aux.end(),v2_aux.begin(),v2_aux.end(),result,kernel_traits);
  }

  return result;

}

CGAL_END_NAMESPACE

#endif //CGAL_CH_PROCESSOR_T2_H
