// Copyright (c) 1999  Max-Planck-Institute Saarbruecken (Germany).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Convex_hull_2/include/CGAL/ch_selected_extreme_points_2.C,v $
// $Revision: 1.4 $ $Date: 2003/10/21 12:15:04 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_SELECTED_EXTREME_POINTS_T2_C
#define CGAL_CH_SELECTED_EXTREME_POINTS_T2_C

#ifndef CGAL_CH_SELECTED_EXTREME_POINTS_T2_H
#include <CGAL/ch_selected_extreme_points_T2.h>
#endif // CGAL_CH_SELECTED_EXTREME_POINTS_T2_H

/*#ifndef CH_NO_POSTCONDITIONS
#include <CGAL/convexity_check_2.h>
#endif // CH_NO_POSTCONDITIONS*/

#include <algorithm>
#include <CGAL/functional.h>

CGAL_BEGIN_NAMESPACE

template <class ForwardIterator, class Traits>
void
ch_w_point_T2
( ForwardIterator first, ForwardIterator last,
                 ForwardIterator& w,
                 const Traits& ch_traits)
{

  typename Traits::Smaller_xy_one_range_infinity_T2
  lexicographically_xy_smaller = ch_traits.smaller_xy_one_range_infinity_T2_object();
  ForwardIterator ite = first;
  while (ite != last) {

    ite++;
  }

  ForwardIterator it = first;
  w = it;
  while ( it != last )
  {
    if ( lexicographically_xy_smaller( *it, *w ))  w = it;
      it++;
  }

  ite = first;
  while (ite != last) {

    ite++;
  }
}


template <class ForwardIterator, class Traits>
void
ch_we_point_T2( ForwardIterator first, ForwardIterator last,
                ForwardIterator& w,
                ForwardIterator& e,
                const Traits& ch_traits)
{
  typename Traits::Smaller_xy_one_range_infinity_T2
  lexicographically_xy_smaller = ch_traits.smaller_xy_one_range_infinity_T2_object();
  w = e = first;
  while ( first != last )
  {
    if ( lexicographically_xy_smaller( *first, *w ))  w = first;
    if ( lexicographically_xy_smaller( *e, *first ))  e = first;
    ++first;
  }
}

template <class ForwardIterator, class Point, class Traits>
void
ch_extreme_point_T2( ForwardIterator first, ForwardIterator last,
                  ForwardIterator& p,
                  const Point& o,
                  const Traits& ch_traits)
{
  typedef typename Traits::At_infinity_T2       At_infinity_T2;
  typedef typename Traits::Squared_distance_T2  Squared_distance_T2;
  typedef typename Traits::Less_rotate_ccw_T2   Less_rotate_ccw_T2;

  At_infinity_T2 is_at_infinity = ch_traits.at_infinity_T2_object();
  Squared_distance_T2
  squared_distance = ch_traits.squared_distance_T2_object();
  Less_rotate_ccw_T2 rotation_predicate = ch_traits.less_rotate_ccw_T2_object();
  p = first;
  ForwardIterator q;
  q++;
  p=std::partition(first,last,is_at_infinity);

  if (p == first) {
    q = p;
    while (q != last)
    {
      if (compare_T1p(squared_distance(*q,o), squared_distance(*p,o)) == LARGER)
        p = q;

      q++;
    }
  }
  else {
    p = (std::min_element(first,p,
                          bind_1(rotation_predicate, o)) );

  }

}


CGAL_END_NAMESPACE

#endif // CGAL_CH_SELECTED_EXTREME_POINTS_T2_C
