//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CH_SUPPORT_LINES_T2_C
#define CGAL_CH_SUPPORT_LINES_T2_C

#ifndef CGAL_CH_SUPPORT_LINES_T2_H
#include <CGAL/ch_support_lines_T2.h>
#endif // CGAL_CH_SUPPORT_LINES_T2_H

#include <algorithm>
#include <CGAL/circulator.h>
#include <CGAL/ch_position_T2.h>
#include <CGAL/convexity_check_T2.h>

CGAL_BEGIN_NAMESPACE



template <class BidirectionalIterator, class Point, class Traits>
void
ch_support_segment_extreme_points_T2(
												BidirectionalIterator first1,
												BidirectionalIterator last1,
                    		BidirectionalIterator first2,
												BidirectionalIterator last2,
                    		BidirectionalIterator e,
                    		BidirectionalIterator d,
												Point &p1,
                    		Point &q1,
												Point &p2,
                    		Point &q2,
												const Traits& ch_traits)
{


	CGAL_precondition(first1!=last1 && first2!=last2);

	typedef CGAL::Kernel_traits<Point>  KTraits;
  typedef typename KTraits::Kernel    Kernel;
	typedef typename Traits::Orientation_T2 Orientation_T2;
	typedef typename Traits::Segment_T2 Segment_T2;
	typedef typename Traits::Triangle_T2 Triangle_T2;
	typedef typename Traits::Back_range_T2 Back_range_T2;
	typedef typename Traits::Front_range_T2 Front_range_T2;
	typedef typename Traits::Y_equal_T2 Y_equal_T2;
	typedef typename Traits::X_equal_T2 X_equal_T2;
	typedef typename Traits::Equal_T2 Equal_T2;
	typedef typename Traits::Compute_greater_y_one_range_infinity_T2 
													 Compute_greater_y_one_range_infinity_T2;
	typedef typename Traits::Compute_greater_x_one_range_infinity_T2 
													 Compute_greater_x_one_range_infinity_T2;
	typedef typename Traits::Compute_smaller_x_one_range_infinity_T2 
													 Compute_smaller_x_one_range_infinity_T2;
	typedef typename Traits::Compute_smaller_y_one_range_infinity_T2 
													 Compute_smaller_y_one_range_infinity_T2;
	typedef typename Traits::Greater_y_one_range_infinity_T2 	
													 Greater_y_one_range_infinity_T2;

	typedef typename Traits::Smaller_y_one_range_infinity_T2 	
													 Smaller_y_one_range_infinity_T2;

	typedef Circulator_from_iterator<BidirectionalIterator> 
																	          BidirectionalCirculator;



	typedef typename Traits::Negative_side_line_T2
												Negative_side_line_T2;
	typedef typename Traits::Positive_side_line_T2
												Positive_side_line_T2;
		

	Negative_side_line_T2 has_on_negative_side = 
							ch_traits.negative_side_line_T2_object();	
	Positive_side_line_T2 has_on_positive_side = 
							ch_traits.positive_side_line_T2_object();	

  Back_range_T2 back_range = ch_traits.back_range_T2_object();
	Front_range_T2 front_range = ch_traits.front_range_T2_object();
	Orientation_T2 orientation = ch_traits.orientation_T2_object();
	Y_equal_T2 y_equal = ch_traits.y_equal_T2_object();
	X_equal_T2 x_equal = ch_traits.x_equal_T2_object();
	Equal_T2 equal_points = ch_traits.equal_T2_object();
	Compute_greater_y_one_range_infinity_T2 	
										compute_greater_y_one_range_infinity =
		ch_traits.compute_greater_y_one_range_infinity_T2_object(); 
	Compute_greater_x_one_range_infinity_T2 	
										compute_greater_x_one_range_infinity =
		ch_traits.compute_greater_x_one_range_infinity_T2_object(); 
	Compute_smaller_x_one_range_infinity_T2 	
										compute_smaller_x_one_range_infinity =
		ch_traits.compute_smaller_x_one_range_infinity_T2_object(); 
	Compute_smaller_y_one_range_infinity_T2 	
										compute_smaller_y_one_range_infinity =
		ch_traits.compute_smaller_y_one_range_infinity_T2_object(); 
	Greater_y_one_range_infinity_T2 	
										greater_y_one_range_infinity =
		ch_traits.greater_y_one_range_infinity_T2_object();
	Smaller_y_one_range_infinity_T2 	
										smaller_y_one_range_infinity =
		ch_traits.smaller_y_one_range_infinity_T2_object();

	
		
	if (find_if(first1,last1,back_range)==last1) {

		BidirectionalCirculator e1,d1,e2,d2;
		BidirectionalIterator a,b,c,f;
	
		for (a=first1; a!=last1; ++a)


		a = first1;
		b = first2;
		Point point;

		bool degenerate = false;

		BidirectionalCirculator suc_e, pred_e(first1,last1), 
														suc_e2, pred_e2;
		e1 = pred_e;
			
		while (*e1!=*e) 
			e1 = ++pred_e;
		
		pred_e = e1;
		suc_e = e1;
		pred_e--;
		suc_e++;
		e2 = e1;
		pred_e2 = pred_e;
		suc_e2 = suc_e;

		BidirectionalCirculator suc_d, pred_d(first2,last2),
														suc_d2, pred_d2;

		d1 = pred_d;
		while (*d1!=*d) 
			d1 = ++pred_d;
		pred_d = d1;
		suc_d = d1;
		pred_d--;
		suc_d++;
		
		d2 = d1;
		pred_d2 = pred_d;
		suc_d2 = suc_d;

		a++;
		b++;
		c = a;
		f = b;
		if (c != last1)
			c++;
		if (f != last2)
			f++;
		
		
		if (a == last1 && b == last2)
		{

			e1 = BidirectionalCirculator(first1,last1);
			d1 = BidirectionalCirculator(first2,last2);
			e2 = e1;
			d2 = d1;
		}
		else if (a == last1) {
			e1 = BidirectionalCirculator(first1,last1);	
			e2 = e1;

			if (f == last2) {

				Triangle_T2 t0(*first2,*e1,*e2);
				Triangle_T2 t1(*b,*e1,*e2);

				if (!t0.has_on_unbounded_side(*b)) {
					d1 = BidirectionalCirculator(first2,last2);
					d2 = d1;	
					degenerate = true;
				}
				else if (!t1.has_on_unbounded_side(*first2)) {
					d1 = BidirectionalCirculator(b,last2);
					d2 = d1;
					degenerate = true;
				}

			}
			if (!degenerate) {
				d1=ch_get_support_line_upper_point_back_range_T2(first1,last1,
																										 first2,last2,
																									 e1,d1,ch_traits);
				d2=ch_get_support_line_lower_point_back_range_T2(first1,last1,
																										 first2,last2,
																									 e2,d2,ch_traits);
			}

		}
		else if (b == last2) {

			d1 = BidirectionalCirculator(first2,last2);
			d2 = d1;

			if (c == last1) {
				Triangle_T2 t0(*first1,*d1,*d2);
				Triangle_T2 t1(*a,*d1,*d2);
				if (!t0.has_on_unbounded_side(*a)) {
					e1 = BidirectionalCirculator(first1,last1);
					e2 = e1;
					degenerate = true;
				}
				else if (!t1.has_on_unbounded_side(*first1)) {
					e1 = BidirectionalCirculator(a,last1);
					e2 = e1;
					degenerate = true;
				}
			}
			if (!degenerate) {
				e1=ch_get_support_line_upper_point_front_range_T2(first1,last1,
																									  first2,last2,
																									  e1,d1,ch_traits);

				e2=ch_get_support_line_lower_point_front_range_T2(first1,last1,
																										 first2,last2,
																									 e2,d2,ch_traits);
			}		
		}
		else {

	 		// ch1 is left

			if (c == last1 && f == last2) {
			
				Triangle_T2 t0(*first1,*a,*b);
				Triangle_T2 t1(*first1,*a,*first2);
				Triangle_T2 t2(*first2,*b,*a);
				Triangle_T2 t3(*first2,*b,*first1);

				!t0.has_on_unbounded_side(*first2) << std::endl;

				!t1.has_on_unbounded_side(*b) << std::endl;

				!t2.has_on_unbounded_side(*first1) << std::endl;

				!t3.has_on_unbounded_side(*a) << std::endl;
				if (!t0.has_on_unbounded_side(*first2)) {
					d1 = BidirectionalCirculator(b,last2); 
					d2 = d1; 
					degenerate = true;
				}
			  else if (!t1.has_on_unbounded_side(*b)) {
					d1 = BidirectionalCirculator(first2,last2); 
					d2 = d1; 
					degenerate = true;
				}
				else if (!t2.has_on_unbounded_side(*first1)) {
					e1 = BidirectionalCirculator(a,last1); 
					e2 = e1; 
					degenerate = true;
				}
				else if (!t3.has_on_unbounded_side(*a)) {
					e1 = BidirectionalCirculator(first1,last1); 
					e2 = e1; 
					degenerate = true;
				}
			}
			if (!degenerate) {
				if (ch_is_left_T2(first1,last1,first2,last2,ch_traits)) {
	
					if (orientation(*pred_e,*e1,*suc_e) > 0 ||
							orientation(*pred_d,*d1,*suc_d) > 0) {

						while (!(((orientation(*d1,*pred_e,*e1) > 0 &&
	  		  				orientation(*d1,*e1,*suc_e) <= 0) ||	
									(orientation(*d1,*pred_e,*e1) == 0 &&
  			  				orientation(*d1,*e1,*suc_e) == 0 && c==last1)) && 
									((orientation(*d1,*pred_d,*e1) < 0 &&
			 	  				orientation(*d1,*e1,*suc_d) >= 0) ||	
									(orientation(*d1,*pred_d,*e1) == 0 &&
  			  				orientation(*d1,*e1,*suc_d) == 0 && f==last2))))	{
							while (!((orientation(*d1,*pred_e,*e1) > 0 &&
			  	  		orientation(*d1,*e1,*suc_e) <= 0) ||	
								(orientation(*d1,*pred_e,*e1) == 0 &&
  			  			orientation(*d1,*e1,*suc_e) == 0 && c==last1))) {
									suc_e = e1;
							  	e1 = pred_e;
									pred_e--;	

							}
							while (!((orientation(*d1,*pred_d,*e1) < 0 &&
	  		  			orientation(*d1,*e1,*suc_d) >= 0) ||
								(orientation(*d1,*pred_d,*e1) == 0 &&
 			  				orientation(*d1,*e1,*suc_d) == 0 && f==last2))) {
									suc_d = d1;
									d1 = pred_d;
									pred_d--;				

							}
						}
						while (!(((orientation(*d2,*pred_e2,*e2) <= 0 &&
  			  				orientation(*d2,*e2,*suc_e2) > 0) ||
									(orientation(*d2,*pred_e2,*e2) == 0 &&
  			  				orientation(*d2,*e2,*suc_e2) == 0 && c==last1)) &&
									((orientation(*d2,*pred_d2,*e2) >= 0 &&
  			  				orientation(*d2,*e2,*suc_d2) < 0) || 
									(orientation(*d2,*pred_d2,*e2) == 0 &&
  			  				orientation(*d2,*e2,*suc_d2) == 0 && f==last2)))) {

						
							while (!((orientation(*d2,*pred_e2,*e2) <= 0 &&
  			  				orientation(*d2,*e2,*suc_e2) > 0) ||
									(orientation(*d2,*pred_e2,*e2) == 0 &&
  			  				orientation(*d2,*e2,*suc_e2) == 0 && c==last1))) {
									pred_e2 = e2;
									e2 = suc_e2;
									suc_e2++;
									
							}
							while (!((orientation(*d2,*pred_d2,*e2) >= 0 &&
  			  				orientation(*d2,*e2,*suc_d2) < 0) || 
									(orientation(*d2,*pred_d2,*e2) == 0 &&
  			  				orientation(*d2,*e2,*suc_d2) == 0 && f==last2))) 
							{
								pred_d2 = d2;
								d2 = suc_d2;
								suc_d2++;					
							
							}
						}
						if (c == last1) {
							Triangle_T2 t0(*first1,*d1,*d2);
							Triangle_T2 t1(*a,*d1,*d2);
							if (!t0.has_on_unbounded_side(*a)) {
								e1 = BidirectionalCirculator(first1,last1);
								e2 = e1;
							}
							else if (!t1.has_on_unbounded_side(*first1)) {
								e1 = BidirectionalCirculator(a,last1);
								e2 = e1;
							}
							while (orientation(*e1,*d1,*suc_d)!=1) {
								d1 = suc_d;
								suc_d++;
							}
							while (orientation(*d2,*e2,*pred_d2)!=1) {
								d2 = pred_d2;
								pred_d2--;
							}
						}
						else if (f == last2) {
							
							Triangle_T2 t0(*first2,*e1,*e2);
							Triangle_T2 t1(*b,*e1,*e2);
							if (!t0.has_on_unbounded_side(*b)) {
								d1 = BidirectionalCirculator(first2,last2);
								d2 = d1;
							}
							else if (!t1.has_on_unbounded_side(*first2)) {
								d1 = BidirectionalCirculator(b,last2);
								d2 = d1;
							}
							while (orientation(*e1,*d1,*pred_e)!=1) {
								e1 = pred_e;
								pred_e--;
							}
							while (orientation(*d2,*e2,*suc_e2)!=1) {
								e2 = suc_e2;
								suc_e2++;
							}
						}
						Segment_T2 l1(*e1,*d1);
						
						Segment_T2 l2(*e2,*d2);
						
						Segment_T2_Segment_T2_pair<Kernel> spair(&l1,&l2);
						if (do_intersect(l1,l2) && spair.intersection(point))
						{
							suc_d = d1;
							d1=d2;
							d2=suc_d;
							suc_d = d1;
							suc_d++;
							pred_d2 = d2;
							pred_d2--;
							if ((!equal_points(point,*e1) &&
									!equal_points(point,*e2) || (*e1!=*e2)) &&
									(!equal_points(point,*d1) &&
									!equal_points(point,*d2) || (*d1!=*d2))) {	
								while (orientation(*e1,*d1,*suc_d)!=1 ||					
											 orientation(*e1,*d1,*pred_e)!=1) {
		
									if (orientation(*e1,*d1,*suc_d)!=1) {
										d1 = suc_d;
										suc_d++;
									}
									if (orientation(*e1,*d1,*pred_e)!=1) {
										e1 = pred_e;
										pred_e--;
									}
								}
								while (orientation(*d2,*e2,*pred_d2)!=1 ||			
											 orientation(*d2,*e2,*suc_e2)!=1) {
									if (orientation(*d2,*e2,*pred_d2)!=1) {
										d2 = pred_d2;
										pred_d2--;
									}
									if (orientation(*d2,*e2,*suc_e2)!=1) {
										e2 = suc_e2;
										suc_e2++;
									}
								}
							}
						}
					}
				}
				//ch1 is_at_right
				else  { 
					
					if (orientation(*pred_e,*e1,*suc_e) > 0 ||
			  			orientation(*pred_d,*d1,*suc_d) > 0) { 
						while (!(((orientation(*d1,*pred_e,*e1) <= 0 &&
	  	  	  			 orientation(*d1,*e1,*suc_e) > 0) || 
									(orientation(*d1,*pred_e,*e1) == 0 &&
	  	  	  			 orientation(*d1,*e1,*suc_e) == 0 && c==last1)) &&
									((orientation(*d1,*pred_d,*e1) >= 0 &&
	  		  			   orientation(*d1,*e1,*suc_d) < 0) ||
									(orientation(*d1,*pred_d,*e1) == 0 &&
	  		  			   orientation(*d1,*e1,*suc_d) == 0 && f==last2)))) {  
					
							while (!((orientation(*d1,*pred_e,*e1) <= 0 &&
			  	  				orientation(*d1,*e1,*suc_e) > 0) ||  							 											(orientation(*d1,*pred_e,*e1)==0 &&
									 	orientation(*d1,*e1,*suc_e) ==0 && c==last1))) {
								pred_e = e1;
								e1 = suc_e;
								suc_e++;
							}
					
							while (!((orientation(*d1,*pred_d,*e1) >= 0 &&
			  	 					orientation(*d1,*e1,*suc_d) < 0) ||
										(orientation(*d1,*pred_d,*e1) == 0 &&
			  	 					orientation(*d1,*e1,*suc_d) == 0 && f==last2))){
										//case ch1 and ch2 CCW					
					
									pred_d = d1;
									d1 = suc_d;
									suc_d++;
							}
						}
						while (!(((orientation(*d2,*pred_e2,*e2) > 0 &&
	  		  				orientation(*d2,*e2,*suc_e2) <= 0) ||
									(orientation(*d2,*pred_e2,*e2) == 0 &&
	  		  				orientation(*d2,*e2,*suc_e2) == 0 && c==last1)) &&
									((orientation(*d2,*pred_d2,*e2) < 0 &&
	  		  				orientation(*d2,*e2,*suc_d2) >= 0) ||
									(orientation(*d2,*pred_d2,*e2) == 0 &&
	  		  				orientation(*d2,*e2,*suc_d2) == 0 && f==last2)))) {
							while (!((orientation(*d2,*pred_e2,*e2) > 0 &&
	  		  				orientation(*d2,*e2,*suc_e2) <= 0) || 
									(orientation(*d2,*pred_e2,*e2) == 0 &&
	  		  				orientation(*d2,*e2,*suc_e2) == 0 && c==last1))) 
							{	
							
						 		suc_e2 = e2;
								e2 = pred_e2;
								pred_e2--;
							}
							while (!((orientation(*d2,*pred_d2,*e2) < 0 &&
	  		  				orientation(*d2,*e2,*suc_d2) >= 0) ||
									(orientation(*d2,*pred_d2,*e2) == 0 &&
	  		  				orientation(*d2,*e2,*suc_d2) == 0 && f==last2)))
							{
								suc_d2 = d2;
								d2 = pred_d2;
								pred_d2--;
							}
						}
						
						if (c == last1) {
							Triangle_T2 t0(*first1,*d1,*d2);
						
							Triangle_T2 t1(*a,*d1,*d2);
							if (!t0.has_on_unbounded_side(*a)) {
								e1 = BidirectionalCirculator(first1,last1);
								e2 = e1;
							}
							else if (!t1.has_on_unbounded_side(*first1)) {
								e1 = BidirectionalCirculator(a,last1);
								e2 = e1;
							}
							while (orientation(*e1,*d1,*pred_d)!=-1) {
								d1 = pred_d;
								pred_d--;
							}
							while (orientation(*d2,*e2,*suc_d2)!=-1) {
								d2 = suc_d2;
								suc_d2++;
							}
									
						}
						else if (f == last2) {
							Triangle_T2 t0(*first2,*e1,*e2);
							Triangle_T2 t1(*b,*e1,*e2);
							if (!t0.has_on_unbounded_side(*b)) {
								d1 = BidirectionalCirculator(first2,last2);
								d2 = d1;
							}
							else if (!t1.has_on_unbounded_side(*first2)) {
								d1 = BidirectionalCirculator(b,last2);
								d2 = d1;
							}
							while (orientation(*e1,*d1,*suc_e)!=-1) {
								e1 = suc_e;
								suc_e++;
							}
							while (orientation(*d2,*e2,*pred_e2)!=-1) {
								e2 = pred_e2;
								pred_e2--;
							}
						}
						Segment_T2 l1(*e1,*d1);
						Segment_T2 l2(*e2,*d2);
						Segment_T2_Segment_T2_pair<Kernel> spair(&l1,&l2);
						if (do_intersect(l1,l2) && spair.intersection(point))
						{
							suc_d = d1;
							d1=d2;
							d2=suc_d;
							pred_d = d1;
							pred_d--;
							suc_d2 = d2;
							suc_d2++;
							if ((!equal_points(point,*e1) &&
									 !equal_points(point,*e2) || (*e1!=*e2))  &&
									(!equal_points(point,*d1) &&
									!equal_points(point,*d2) || (*d1!=*d2))) {	
								
								

								while (orientation(*e1,*d1,*pred_d)!=-1 ||					
									 orientation(*e1,*d1,*suc_e)!=-1) {
									if (orientation(*e1,*d1,*pred_d)!=-1) {
										d1 = pred_d;
										pred_d--;
									}
									if (orientation(*e1,*d1,*suc_e)!=-1) {
										e1 = suc_e;
										suc_e++;
									}
								}
								while (orientation(*d2,*e2,*suc_d2)!=-1 ||					
											 orientation(*d2,*e2,*pred_e2)!=-1) {
									if (orientation(*d2,*e2,*suc_d2)!=-1) {
										d2 = suc_d2;
										suc_d2++;
									}
									if (orientation(*d2,*e2,*pred_e2)!=-1) {
										e2 = pred_e2;
										pred_e2--;
									}
								}
							}
						}
					}
				}
			}
		}
		p1 = *e1;
		p2 = *e2;
		q1 = *d1;
		q2 = *d2;	

		
		
		

		}
		else
			return ch_support_segment_extreme_points_T2(
																			first2,last2,
																			first1,last1,
																			d,e,q1,p1,q2,p2,
																			ch_traits);	
}

template <class BidirectionalIterator, class Traits>
typename Traits::Line_T2
ch_inferior_support_line_T2(BidirectionalIterator first1,
												BidirectionalIterator last1,
                    		BidirectionalIterator first2,
												BidirectionalIterator last2,
                    		BidirectionalIterator e,
                    		BidirectionalIterator d,
												const Traits& ch_traits)
{

//ESTA FALTANDO TRATAR CASOS DE COLINEARIDADE
//PRE-CONDICAO: STRONGLY CONVEX
	
	CGAL_precondition(is_ccw_strongly_convex_T2
											(first1,last1,ch_traits) ||
										is_cw_strongly_convex_T2
											(first1,last1,ch_traits));
	CGAL_precondition(is_ccw_strongly_convex_T2
											(first2,last2,ch_traits) ||
										is_cw_strongly_convex_T2
											(first2,last2,ch_traits));
	typedef typename Traits::Orientation_T2 Orientation_T2;
	typedef typename Traits::Line_T2 Line_T2;
	typedef typename Traits::Back_range_T2 Back_range_T2;

	typedef Circulator_from_iterator<BidirectionalIterator> 
																	          BidirectionalCirculator;


	Orientation_T2 orientation = ch_traits.orientation_T2_object();
	Back_range_T2 back_range = ch_traits.back_range_T2_object();

	if (find_if(first1,last1,back_range)==last1) {
		BidirectionalCirculator e1, suc_e, pred_e(first1,last1);
		e1 = pred_e;
		while (*e1!=*e) 
			e1 = ++pred_e;
	
		pred_e = e1;
		suc_e = e1;
		pred_e--;
		suc_e++;
		BidirectionalCirculator d1, suc_d, pred_d(first2,last2);
	
		d1 = pred_d;
		while (*d1!=*d) 
			d1 = ++pred_d;
		pred_d = d1;
		suc_d = d1;
		pred_d--;
		suc_d++;
	
		// ch1 is left
		if (ch_is_left_T2(first1,last1,first2,last2,ch_traits)) {

			if (orientation(*pred_e,*e1,*suc_e) > 0) {
				while (!(orientation(*d1,*pred_e,*e1) <= 0 &&
  	  				orientation(*d1,*e1,*suc_e) > 0 &&
							orientation(*d1,*pred_d,*e1) >= 0 &&
  	  				orientation(*d1,*e1,*suc_d) < 0))	{
					while (!(orientation(*d1,*pred_e,*e1) <= 0 &&
  	  			orientation(*d1,*e1,*suc_e) > 0)) {
						pred_e = e1;
						e1 = suc_e;
						suc_e++;
					}
					while (!(orientation(*d1,*pred_d,*e1) >= 0 &&
  	  			orientation(*d1,*e1,*suc_d) < 0)) {
						pred_d = d1;
						d1 = suc_d;
						suc_d++;					
					}
				}
			}
			else {
				while (!(orientation(*d1,*pred_e,*e1) < 0 &&
  	  				orientation(*d1,*e1,*suc_e) >= 0 &&
							orientation(*d1,*pred_d,*e1) > 0 &&
  	  				orientation(*d1,*e1,*suc_d) <= 0))	{
					while (!(orientation(*d1,*pred_e,*e1) < 0 &&
  	  			orientation(*d1,*e1,*suc_e) >= 0)) {
						suc_e = e1;
						e1 = pred_e;
						pred_e--;				
					}
					while (!(orientation(*d1,*pred_d,*e1) > 0 &&
  	  			orientation(*d1,*e1,*suc_d) <= 0)) {
						suc_d = d1;
						d1 = pred_d;
						pred_d--;
					}
				}
			}
		}
  	//ch1 is_at_right
		else { //if (ch_is_left_T2(first2,last2,first1,last2,ch_traits)) {
			if (orientation(*pred_e,*e1,*suc_e) > 0) {
				while (!(orientation(*d1,*pred_e,*e1) > 0 &&
  	  			orientation(*d1,*e1,*suc_e) <= 0 &&
						orientation(*d1,*pred_d,*e1) < 0 &&
  	  			orientation(*d1,*e1,*suc_d) >= 0))	{
					while (!(orientation(*d1,*pred_e,*e1) > 0 &&
  	  			orientation(*d1,*e1,*suc_e) <= 0)) {
				 		suc_e = e1;
						e1 = pred_e;
						pred_e--;
					}
					while (!(orientation(*d1,*pred_d,*e1) < 0 &&
  	  			orientation(*d1,*e1,*suc_d) >= 0)) {
						suc_d = d1;
						d1 = pred_d;
						pred_d--;
					}
				}
			}
			else {
				while (!(orientation(*d1,*pred_e,*e1) >= 0 &&
  	  				orientation(*d1,*e1,*suc_e) < 0 &&
							orientation(*d1,*pred_d,*e1) <= 0 &&
  	  				orientation(*d1,*e1,*suc_d) > 0))	{
					while (!(orientation(*d1,*pred_e,*e1) >= 0 &&
  	  			orientation(*d1,*e1,*suc_e) < 0)) {
						pred_e = e1;
						e1 = suc_e;
						suc_e++;
					}
					while (!(orientation(*d1,*pred_d,*e1) <= 0 &&
  	  			orientation(*d1,*e1,*suc_d) > 0)) {
						pred_d = d1;
						d1 = suc_d;
						suc_d++;
					}
				}
			}
		}

		return Line_T2(*e1,*d1);
	}
	else
		return ch_inferior_support_line_T2(first2,last2,
																			 first1,last1,
																			 d,e,ch_traits);	
}	


template <class BidirectionalCirculator, class BidirectionalIterator,
					class Traits>
BidirectionalCirculator
ch_get_support_line_upper_point_back_range_T2(
												BidirectionalIterator first1,
												BidirectionalIterator last1,
												BidirectionalIterator first2,
												BidirectionalIterator last2,
												BidirectionalCirculator e1,
                    		BidirectionalCirculator d1,
												const Traits& ch_traits) {
	typedef typename Traits::Orientation_T2 Orientation_T2;
	Orientation_T2 orientation = ch_traits.orientation_T2_object();

	BidirectionalIterator f = first2;
	f++;
	BidirectionalCirculator suc_d = d1;
	BidirectionalCirculator pred_d = d1;
	suc_d++;
	pred_d--;
	

	if (ch_is_left_T2(first1,last1,first2,last2,ch_traits)) {
		if (orientation(*pred_d,*d1,*suc_d) > 0) {
			while (!((orientation(*d1,*pred_d,*e1) < 0 &&
  		  			 orientation(*d1,*e1,*suc_d) >= 0) ||
							(orientation(*d1,*pred_d,*e1) == 0 &&
  		  			 orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
			//case ch1 and ch2 CCW
				pred_d = d1;
				d1 = suc_d;
				suc_d++;	

			}
		}
		else {
			while (!((orientation(*d1,*pred_d,*e1) <= 0 &&
  	  				 orientation(*d1,*e1,*suc_d) > 0) ||
							(orientation(*d1,*pred_d,*e1) == 0 &&
  	  				 orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
				suc_d = d1;
				d1 = pred_d;
				pred_d--;			

			}
		}
	}					
	else {
		//ch1 is_at_right
		if (orientation(*pred_d,*d1,*suc_d) > 0) { 
			while (!((orientation(*d1,*pred_d,*e1) >= 0 &&
			   			orientation(*d1,*e1,*suc_d) < 0) ||
						 (orientation(*d1,*pred_d,*e1) == 0 &&
			   			orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
				pred_d = d1;
				d1 = suc_d;
				suc_d++;

			}
		}
		else {
			while (!((orientation(*d1,*pred_d,*e1) > 0 &&
	  	    		orientation(*d1,*e1,*suc_d) <= 0) ||
							(orientation(*d1,*pred_d,*e1) == 0 &&
	  	    		orientation(*d1,*e1,*suc_d) == 0 && f==last2))) {
				//case ch1 and ch2 CW
				suc_d = d1;
				d1 = pred_d;
				pred_d--;

			}
		}	
	}
	return d1;
}



template <class BidirectionalCirculator, class BidirectionalIterator,
					class Traits>
BidirectionalCirculator
ch_get_support_line_upper_point_front_range_T2(
												BidirectionalIterator first1,
												BidirectionalIterator last1,
												BidirectionalIterator first2,
												BidirectionalIterator last2,
												BidirectionalCirculator e1,
                    		BidirectionalCirculator d1,
												const Traits& ch_traits) {

	typedef typename Traits::Orientation_T2 Orientation_T2;
	Orientation_T2 orientation = ch_traits.orientation_T2_object();

	BidirectionalIterator c = first1;
	c++;
	c++;
	BidirectionalCirculator suc_e = e1;
	BidirectionalCirculator pred_e = e1;
	suc_e++;
	pred_e--;

	if (ch_is_left_T2(first1,last1,first2,last2,ch_traits)) {

		if (orientation(*pred_e,*e1,*suc_e) > 0) {

			while (!((orientation(*d1,*pred_e,*e1) > 0 &&
 		  				orientation(*d1,*e1,*suc_e) <= 0) ||
							(orientation(*d1,*pred_e,*e1) == 0 &&
 		  				orientation(*d1,*e1,*suc_e) == 0 && c==last1)))	{

				suc_e = e1;
				e1 = pred_e;
				pred_e--;	
			}
		}
		else {

			while (!((orientation(*d1,*pred_e,*e1) >= 0 &&
		  				orientation(*d1,*e1,*suc_e) < 0) ||
						  (orientation(*d1,*pred_e,*e1) == 0 &&
		  				orientation(*d1,*e1,*suc_e) == 0 && c==last1)))	{

				pred_e = e1;
				e1 = suc_e;
				suc_e++;				
			}
		}
	}
	//ch1 is_at_right
	else  {
		if (orientation(*pred_e,*e1,*suc_e) > 0) { 
			while (!((orientation(*d1,*pred_e,*e1) <= 0 &&
  						orientation(*d1,*e1,*suc_e) > 0) ||
							(orientation(*d1,*pred_e,*e1) == 0 &&
  						orientation(*d1,*e1,*suc_e) == 0 && c==last1)))	{
				pred_e = e1;
				e1 = suc_e;
				suc_e++;
 			}
		}
		else {
			while (!((orientation(*d1,*pred_e,*e1) < 0 &&
	 					orientation(*d1,*e1,*suc_e) >= 0) ||
						(orientation(*d1,*pred_e,*e1) == 0 &&
	 					orientation(*d1,*e1,*suc_e) == 0 && c==last1)))	{
				suc_e = e1;
				e1 = pred_e;
				pred_e--;
			}
		}
	}
	return e1;
}


template <class BidirectionalCirculator, class BidirectionalIterator,
					class Traits>
BidirectionalCirculator
ch_get_support_line_lower_point_front_range_T2(
												BidirectionalIterator first1,
												BidirectionalIterator last1,
												BidirectionalIterator first2,
												BidirectionalIterator last2,
												BidirectionalCirculator e1,
                    		BidirectionalCirculator d1,
												const Traits& ch_traits) {
	typedef typename Traits::Orientation_T2 Orientation_T2;
	Orientation_T2 orientation = ch_traits.orientation_T2_object();
	BidirectionalIterator c=first1;
	c++;
	c++;
	BidirectionalCirculator suc_e = e1;
	BidirectionalCirculator pred_e = e1;
	suc_e++;
	pred_e--;


	if (ch_is_left_T2(first1,last1,first2,last2,ch_traits)) {
		if (orientation(*pred_e,*e1,*suc_e) > 0) {
			while (!((orientation(*d1,*pred_e,*e1) <= 0 &&
  	  				orientation(*d1,*e1,*suc_e) > 0) ||
							(orientation(*d1,*pred_e,*e1) == 0 &&
  	  				orientation(*d1,*e1,*suc_e) == 0 && c==last1)))	{
				pred_e = e1;
				e1 = suc_e;
				suc_e++;
			}
		}
		else {
			while (!((orientation(*d1,*pred_e,*e1) < 0 &&
  	  				orientation(*d1,*e1,*suc_e) >= 0) ||
							(orientation(*d1,*pred_e,*e1) == 0 &&
  	  				orientation(*d1,*e1,*suc_e) == 0 && c==last1)))	{
				suc_e = e1;
				e1 = pred_e;
				pred_e--;				
			}
		}
	}
  //ch1 is_at_right
	else {

		if (orientation(*pred_e,*e1,*suc_e) > 0) {

			while (!((orientation(*d1,*pred_e,*e1) > 0 &&
  	  			orientation(*d1,*e1,*suc_e) <= 0) ||
						(orientation(*d1,*pred_e,*e1) == 0 &&
  	  			orientation(*d1,*e1,*suc_e) == 0 && c==last1))) {

		 		suc_e = e1;
				e1 = pred_e;
				pred_e--;
			}
		}
		else {

			while (!((orientation(*d1,*pred_e,*e1) >= 0 &&
  	  				orientation(*d1,*e1,*suc_e) < 0) ||
							(orientation(*d1,*pred_e,*e1) == 0 &&
  	  				orientation(*d1,*e1,*suc_e) == 0 && c==first1))) {

				pred_e = e1;
				e1 = suc_e;
				suc_e++;
			}
		}
	}

	return e1;
}


template <class BidirectionalCirculator, class BidirectionalIterator,
					class Traits>
BidirectionalCirculator
ch_get_support_line_lower_point_back_range_T2(	
												BidirectionalIterator first1,
												BidirectionalIterator last1,
												BidirectionalIterator first2,
												BidirectionalIterator last2,
												BidirectionalCirculator e1,
                    		BidirectionalCirculator d1,
												const Traits& ch_traits) {
	
	typedef typename Traits::Orientation_T2 Orientation_T2;
	Orientation_T2 orientation = ch_traits.orientation_T2_object();

	BidirectionalIterator f = last2;
	f++;
	f++;
	BidirectionalCirculator suc_d = d1;
	BidirectionalCirculator pred_d = d1;
	suc_d++;
	pred_d--;


	if (ch_is_left_T2(first1,last1,first2,last2,ch_traits)) {
		if (orientation(*pred_d,*d1,*suc_d) > 0) {
			while (!((orientation(*d1,*pred_d,*e1) >= 0 &&
  	  				orientation(*d1,*e1,*suc_d) < 0) ||
							(orientation(*d1,*pred_d,*e1) == 0 &&
  	  				orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
				pred_d = d1;
				d1 = suc_d;
				suc_d++;					
			}
		}
		else {
			while (!((orientation(*d1,*pred_d,*e1) > 0 &&
  	  				orientation(*d1,*e1,*suc_d) <= 0) ||
						 (orientation(*d1,*pred_d,*e1) == 0 &&
  	  			 orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
				suc_d = d1;
				d1 = pred_d;
				pred_d--;
			}
		}
	}
 	//ch1 is_at_right
	else { 
		if (orientation(*pred_d,*d1,*suc_d) > 0) {
			while (!((orientation(*d1,*pred_d,*e1) < 0 &&
  	  			   orientation(*d1,*e1,*suc_d) >= 0) ||
						  (orientation(*d1,*pred_d,*e1) == 0 &&
  	  			   orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
				suc_d = d1;
				d1 = pred_d;
				pred_d--;
			}
		}
		else {
			while (!((orientation(*d1,*pred_d,*e1) <= 0 &&
  	  				orientation(*d1,*e1,*suc_d) > 0) ||
						 (orientation(*d1,*pred_d,*e1) == 0 &&
  	  				orientation(*d1,*e1,*suc_d) == 0 && f==last2)))	{
				pred_d = d1;
				d1 = suc_d;
				suc_d++;
			}
		}
	}
	return d1;
}


template <class BidirectionalIterator, class Line, class Traits>
bool ch_is_superior_support_line (BidirectionalIterator first1, 
																	BidirectionalIterator last1,
																	BidirectionalIterator first2,
																  BidirectionalIterator last2,  
																	Line l,
																	const Traits& traits)
{
	typedef typename Traits::Negative_side_line_T2 Negative_side_line_T2;
	typedef typename Traits::Positive_side_line_T2 Positive_side_line_T2;
	Negative_side_line_T2 
				has_on_negative_side = traits.negative_side_line_T2_object();

	Positive_side_line_T2 
				has_on_positive_side = traits.positive_side_line_T2_object();

	if (find_if(first2,last2,
						bind_1(has_on_positive_side,l))!=last2)
		return false;
	if (find_if(first1,last1,
						bind_1(has_on_negative_side,l))!=last1)
		return false;
	return true;
}

template <class BidirectionalIterator, class Line, class Traits>
bool ch_is_inferior_support_line (BidirectionalIterator first1, 
																	BidirectionalIterator last1,
																	BidirectionalIterator first2,
																  BidirectionalIterator last2,  
																	Line l,
																	const Traits& traits)
{
	typedef typename Traits::Negative_side_line_T2 Negative_side_line_T2;
	typedef typename Traits::Positive_side_line_T2 Positive_side_line_T2;
	Negative_side_line_T2 
				has_on_negative_side = traits.negative_side_line_T2_object();

	Positive_side_line_T2 
				has_on_positive_side = traits.positive_side_line_T2_object();

	if (find_if(first2,last2,
						bind_1(has_on_negative_side,l))!=last2)
		return false;	
	if (find_if(first1,last1,
						bind_1(has_on_positive_side,l))!=last1)
		return false;
	return true;	

}
CGAL_END_NAMESPACE

#endif //CGAL_CH_SUPPORT_LINES_T2_C
