// Copyright (c) 1999  Max-Planck-Institute Saarbruecken (Germany).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Convex_hull_2/include/CGAL/convexity_check_2.C,v $
// $Revision: 1.9 $ $Date: 2003/10/21 12:15:07 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CONVEXITY_CHECK_T2_C
#define CGAL_CONVEXITY_CHECK_T2_C

#include <CGAL/convexity_check_T2.h>
#include <CGAL/stl_extensions.h>
#include <CGAL/functional.h>
#include <algorithm>

CGAL_BEGIN_NAMESPACE
template <class ForwardIterator, class Traits>
bool
is_ccw_strongly_convex_T2( ForwardIterator first, ForwardIterator last, 
                          const Traits& ch_traits)
{
  typedef  typename Traits::Less_xy_T2     Less_xy;
  typedef  typename Traits::Left_turn_T2    Left_turn;
  // added 
  typedef  typename Traits::Equal_T2        Equal_T2;
  
  Less_xy  smaller_xy    = ch_traits.less_xy_T2_object();
  Left_turn left_turn    = ch_traits.left_turn_T2_object();
  Equal_T2  equal_points  = ch_traits.equal_T2_object();   

  ForwardIterator iter1;
  ForwardIterator iter2;
  ForwardIterator iter3;

  if ( first == last) return true;

  iter2 = first;
  iter3 = ++iter2;

  if (iter3 == last ) return true;

  ++iter3;

  if (iter3 == last ) return (! equal_points(*first,*iter2) );

  iter1 = first;
  short int f = 0;

  while (iter3 != last) 
  {
		/*	std::cout << "left_turn(" << "[" << *iter1 << "]," << "[" 
								<< *iter2 << "]," << "[" << *iter3 << "])" <<
				left_turn( *iter1, *iter2, *iter3 ) << std::endl;
			std::cout << "smaller_xy("<< "[" << *iter2 << "]," << "[" <<
									*iter1 << "])"  << 
							smaller_xy( *iter2, *iter3 )  << std::endl;
			std::cout << "smaller_xy("<< "[" << *iter2 << "]," << "[" <<
									*iter3 << "])"  << 
							smaller_xy( *iter2, *iter3 )  << std::endl;
      if ( !left_turn( *iter1, *iter2, *iter3 ) ) return false; 
      if ( smaller_xy( *iter2, *iter1 ) && smaller_xy( *iter2, *iter3 )) ++f;*/

      ++iter1;
      ++iter2;
      ++iter3;
  }

	/*	std::cout << "left_turn(" << "[" << *iter1 << "]," << "[" 
								<< *iter2 << "]," << "[" << *iter3 << "])" <<
				left_turn( *iter1, *iter2, *iter3 ) << std::endl;
			std::cout << "smaller_xy("<< "[" << *iter2 << "]," << "[" <<
									*iter1 << "])"  << 
							smaller_xy( *iter2, *iter3 )  << std::endl;
			std::cout << "smaller_xy("<< "[" << *iter2 << "]," << "[" <<
									*iter3 << "])"  << 
							smaller_xy( *iter2, *iter3 )  << std::endl;*/

  iter3 = first;
  if ( !left_turn( *iter1, *iter2, *iter3 ) ) return false; 
  if ( smaller_xy( *iter2, *iter1 ) && smaller_xy( *iter2, *iter3 )) ++f;


/*		std::cout << "left_turn(" << "[" << *iter1 << "]," << "[" 
								<< *iter2 << "]," << "[" << *iter3 << "])" <<
				left_turn( *iter1, *iter2, *iter3 ) << std::endl;
			std::cout << "smaller_xy("<< "[" << *iter2 << "]," << "[" <<
									*iter1 << "])"  << 
							smaller_xy( *iter2, *iter3 )  << std::endl;
			std::cout << "smaller_xy("<< "[" << *iter2 << "]," << "[" <<
									*iter3 << "])"  << 
							smaller_xy( *iter2, *iter3 )  << std::endl;*/

  iter1 = iter2;
  iter2 = first;
  ++iter3;
  if ( !left_turn( *iter1, *iter2, *iter3 ) ) return false; 
  if ( smaller_xy( *iter2, *iter1 ) && smaller_xy( *iter2, *iter3 )) ++f;


  return ( f > 1 ) ? false : true;
}

template <class ForwardIterator, class Traits>
bool
is_cw_strongly_convex_T2( ForwardIterator first, ForwardIterator last, 
                         const Traits& ch_traits)
{
  typedef  typename Traits::Less_xy_T2      Less_xy;
  typedef  typename Traits::Left_turn_T2    Left_turn;
  typedef  typename Traits::Equal_T2        Equal_T2;  

  Less_xy  smaller_xy    = ch_traits.less_xy_T2_object();
  Equal_T2  equal_points  = ch_traits.equal_T2_object();  

  ForwardIterator iter1;
  ForwardIterator iter2;
  ForwardIterator iter3;

  if ( first == last) return true;

  iter2 = first;
  iter3 = ++iter2;

  if (iter3 == last ) return true;

  ++iter3;

  if (iter3 == last ) return (! equal_points(*first,*iter2) );

  iter1 = first;
  short int f = 0;

  while (iter3 != last) 
  {
      if ( !left_turn( *iter2, *iter1, *iter3 ) ) return false;
      if ( smaller_xy( *iter2, *iter1 ) && smaller_xy( *iter2, *iter3 )) ++f;

      ++iter1;
      ++iter2;
      ++iter3;
  }

  iter3 = first;
  if ( !left_turn( *iter2, *iter1, *iter3 ) ) return false;
  if ( smaller_xy( *iter2, *iter1 ) && smaller_xy( *iter2, *iter3 )) ++f;


  iter1 = iter2;
  iter2 = first;
  ++iter3;
  if ( !left_turn( *iter2, *iter1, *iter3 ) ) return false;
  if ( smaller_xy( *iter2, *iter1 ) && smaller_xy( *iter2, *iter3 )) ++f;


  return ( f > 1 ) ? false : true;
}



CGAL_END_NAMESPACE

#endif // CGAL_CONVEXITY_CHECK_T2_C
