// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/constructions/kernel_ftC2.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:14:37 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Sven Schoenherr, Herv Brnnimann, Sylvain Pion
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_CONSTRUCTIONS_KERNEL_FTT2_H
#define CGAL_CONSTRUCTIONS_KERNEL_FTT2_H

CGAL_BEGIN_NAMESPACE

template <class FT>
CGAL_KERNEL_MEDIUM_INLINE
void
point_on_lineT2(const FT &lpw, const FT &lpx, const FT &lpy,
                const FT &ldw, const FT &ldx, const FT &ldy,
	             	int i,
                FT &w, FT &x, FT &y)
{
 
  if (ldw == 0 && i==0) {
    w = lpw;
    x = lpx;
    y = lpy;
  }
  else if ((lpw < 0) && (ldw < 0))
  {
    w = -lpw*ldw;
    x = -lpx*ldw - ldx*lpw*FT(i);
    y = -lpy*ldw - ldy*lpw*FT(i);
  }
  else
  {
    w = lpw*ldw;
    x = lpx*ldw + ldx*lpw*FT(i);
    y = lpy*ldw + ldy*lpw*FT(i);
  }
}

template <class FT>
CGAL_KERNEL_MEDIUM_INLINE
void
projection_lineT2(const FT &pw, const FT &px, const FT &py,
                  const FT &lpw, const FT &lpx, const FT &lpy,
                  const FT &ldw, const FT &ldx, const FT &ldy,
                  FT &w, FT &x, FT &y)
{
  // projects p on the line l
  


  FT dpw = pw*lpw;
  FT dpx = px*lpw-pw*lpx;
  FT dpy = py*lpw-pw*lpy;

  FT lambdax = (ldx*dpx+ldy*dpy)*ldw*ldw;
  FT lambdaw = ldw*dpw*(ldx*ldx+ldy*ldy);
  
  if ((lpw < 0) && ((lambdaw * ldw) < 0))
  {
    w = -lpw * lambdaw * ldw;
    x = -lpx * lambdaw * ldw - lambdax * ldx * lpw;
    y = -lpy * lambdaw * ldw - lambdax * ldy * lpw;
  }
  else
  {
    w = lpw * lambdaw * ldw;
    x = lpx * lambdaw * ldw + lambdax * ldx * lpw;
    y = lpy * lambdaw * ldw + lambdax * ldy * lpw;  
  }
}

template <class FT>
inline
Oriented_side
side_of_oriented_lineT2(const FT &aw, const FT &ax, const FT &ay,
                        const FT &bw, const FT &bx, const FT &by,
                        const FT &pw, const FT &px, const FT &py)
{
  std::cout << aw*(bx*py-by*px) - ax*(bw*py-by*pw) + ay*(bw*px-bx*pw) << std::endl;
  return Oriented_side(CGAL_NTS sign(aw*(bx*py-by*px) - ax*(bw*py-by*pw) + ay*(bw*px-bx*pw)));
}

template < class FT >
inline
Orientation
orientationT2(const FT &pw, const FT &px, const FT &py,
              const FT &qw, const FT &qx, const FT &qy,
              const FT &rw, const FT &rx, const FT &ry)
{
  return Orientation (sign_of_determinant3x3(pw, px, py,
                                             qw, qx, qy,
                                             rw, rx, ry));
}


CGAL_END_NAMESPACE

#endif //CGAL_CONSTRUCTIONS_KERNEL_FTT2_H
