// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/Cartesian/point_constructions_2.h,v $
// $Revision: 1.9 $ $Date: 2003/10/21 12:14:30 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Herve Bronnimann
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_POINT_CONSTRUCTIONS_T1_H
#define CGAL_SIGNED_HOMOGENEOUS_POINT_CONSTRUCTIONS_T1_H



CGAL_BEGIN_NAMESPACE

template < class R>
inline
Point_T1<R> absolute_valueT1(const Point_T1<R> &p)
{
  if (is_at_front_rangeT1(p))
    return Point_T1<R>(p.w(), p.x());
  return Point_T1<R>(-p.w(), -p.x());
}

template < class R>
inline
Point_T1<R> antipode(const Point_T1<R> &p)
{
  return Point_T1<R>(-p.w(), -p.x());
}


template < class R >
inline
Point_T1<R> smaller_cartesian(const Point_T1<R> &p,
                              const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.w() > 0 && q.w() > 0) ;
  if (p.x()*q.w() < p.w()*q.x())
    return p;
  return q;
}

template < class R >
inline
Point_T1<R> smaller_cartesian(const Point_T1<R> &p,
                              const Point_T1<R> &q,
                              const Point_T1<R> &r)
{
  CGAL_kernel_precondition(p.w() > 0 && q.w() > 0 && r.w() > 0) ;
  Point_T1<R> smaller = p;
  if (p.w()*q.x() < p.x()*q.w())
    smaller = q;
  if (r.x()*smaller.w() < r.w()*smaller.x())
    smaller = r;
  return smaller;
}


template < class R >
inline
Point_T1<R> smaller_cartesian(const Point_T1<R> &p,
                              const Point_T1<R> &q,
                              const Point_T1<R> &r,
                              const Point_T1<R> &s)
{
  CGAL_kernel_precondition(p.w() > 0 && q.w() > 0 && r.w() > 0 && s.w() > 0) ;
  Point_T1<R> smaller = p;
  if (p.w()*q.x() < p.x()*q.w())
    smaller = q;
  if (r.x()*smaller.w() < r.w()*smaller.x())
    smaller = r;
  if (s.x()*smaller.w() < s.w()*smaller.x())
    smaller = s;
  return smaller;
}

template < class R >
inline
Point_T1<R> greater_cartesian(const Point_T1<R> &p, const Point_T1<R> &q)
{
  CGAL_kernel_precondition(p.w() > 0 && q.w() > 0);
  if (p.x()*q.w() > p.w()*q.x())
    return p;
  return q;
}


template < class R >
inline
Point_T1<R> greater_cartesian(const Point_T1<R> &p,
                              const Point_T1<R> &q,
                              const Point_T1<R> &r)

{
  CGAL_kernel_precondition(p.w() > 0 && q.w() > 0 && r.w() > 0) ;
  Point_T1<R> greater = p;
  if (p.w()*q.x() > p.x()*q.w())
    greater = q;
  if (r.x()*greater.w() > r.w()*greater.x())
    greater = r;
  return greater;
}


template < class R >
inline
Point_T1<R> greater_cartesian(const Point_T1<R> &p,
                              const Point_T1<R> &q,
                              const Point_T1<R> &r,
                              const Point_T1<R> &s)

{
  CGAL_kernel_precondition(p.w() > 0 && q.w() > 0 && r.w() > 0 && s.w() > 0) ;
  Point_T1<R> greater = p;
  if (p.w()*q.x() > p.x()*q.w())
    greater = q;
  if (r.x()*greater.w() > r.w()*greater.x())
    greater = r;
  if (s.x()*greater.w() > s.w()*greater.x())
    greater = s;
  return greater;
}

CGAL_END_NAMESPACE

#endif //CGAL_SIGNED_HOMOGENEOUS_POINT_CONSTRUCTIONS_T1_H
