//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_POLYGON_OPERATIONS_T2_C
#define CGAL_POLYGON_OPERATIONS_T2_C

CGAL_BEGIN_NAMESPACE

template < class ForwardIterator, class Point, class Traits >
inline
Oriented_side
oriented_side_polygon_T2(ForwardIterator first, ForwardIterator last, 
                         const Point &p, const Traits &traits)
{

  typedef typename Traits::Orientation_T2            Orientation_T2;
  typedef typename Traits::Segment_T2                Segment_T2;
  typedef typename Traits::Collinear_has_on_T2       Collinear_has_on_T2;

  Orientation_T2 orientation = traits.orientation_T2_object();
  Collinear_has_on_T2 collinear_has_on = traits.collinear_has_on_T2_object();

  CGAL_precondition(first != last);
  ForwardIterator it = first;
  it++;
  CGAL_precondition(it != last);
  it++;
  CGAL_precondition(it != last);
  Orientation s,b,t,z,k;
  
  ForwardIterator l = last;
  l--;
  ForwardIterator f = first;
  f++;
  t = orientation(*l,*first,*f);
  s = orientation(p,*first,*f);
  b = orientation(p,*l,*first);


  if (s*b == -1)
    s=Orientation(-t);
  else if (s*b == 0) {
    if (s == 0)
      s = Orientation((int(collinear_has_on(p,*first,*f))-1) * int(t));
    else {
      s = Orientation((int(collinear_has_on(p,*l,*first))-1) * int(t));
    }
  }
  int i=1;
  ForwardIterator previous = first;
  ForwardIterator next = f;
  next++;
  std::vector<Segment_T2 > vseg;
  vseg.push_back(Segment_T2(*first,p));
  while (s!=0 && next != last) {
    b = orientation(*f,*first,p);
    k = orientation(*next,*first,p);
    vseg.push_back(Segment_T2(*f,*next));
    if ( b == 0 && k == 0 )
       s = Orientation(int(!collinear_has_on(p,*f,*next))*int(s));
    if ( ( b != -1 && k == -1) ||
         ( b != 1 && k == 1) )  {

      z = orientation(*previous,*first,p);
      t = orientation(*l,*first,p);


      if (( ((1-b)*(1-z)*(k*t))!=1 && (1-b)*(z*k)!=1) || (*f == p))
        if (do_intersect(vseg[0],vseg[i]))
          s = Orientation(int(s)*int(orientation(*first,*f,*next))*
                          int(orientation(p,*f,*next)));
    }
    l = previous;
    previous = f;
    f = next;
    next++;
    i++;
  }
  return Oriented_side(s);
}

CGAL_END_NAMESPACE

#endif //CGAL_POLYGON_OPERATIONS_T2_C
