// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/Cartesian/predicates_on_directions_2.h,v $
// $Revision: 1.7 $ $Date: 2003/10/21 12:14:31 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Herve Bronnimann
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_PREDICATES_ON_DIRECTIONS_T2
#define CGAL_PREDICATES_ON_DIRECTIONS_T2

CGAL_BEGIN_NAMESPACE  

template <class R>
inline
Comparison_result compare_angle(const Direction_T2<R>& d1, const Direction_T2<R>& d2)
{
  typedef typename R::FT FT;

  Point_T2<R> _d1 = Point_T2<R>(FT(0),d1.dx(),d1.dy());
  Point_T2<R> _d2 = Point_T2<R>(FT(0),d2.dx(),d2.dy());
  Point_T2<R> _x_axis = Point_T2<R>(FT(0),FT(1),FT(0));
  Point_T2<R> _y_axis = Point_T2<R>(FT(0),FT(0),FT(1));
  Point_T2<R> o = Point_T2<R>(FT(1),FT(0),FT(0));
  if (int(orientation(o,_x_axis,_d1))*int(orientation(o,_x_axis,_d2)) != -1) {
    if (orientation(o,_d1,_d2) == 0) {
      if (orientation(o,_y_axis,_d1) == orientation(o,_y_axis,_d2))
        return EQUAL;
      return Comparison_result(orientation(o,_y_axis,_d1));
    }
    return Comparison_result(orientation(o,_d2,_d1)); 
  }
  return Comparison_result(orientation(o,_d1,_y_axis));
}


CGAL_END_NAMESPACE

#endif


