// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/predicates_on_lines_2.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:18:25 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_PREDICATES_ON_LINES_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_PREDICATES_ON_LINES_T2_H

CGAL_BEGIN_NAMESPACE

template < class R >
inline
bool
at_the_same_side(const Point_T2<R> &p, 
                 const Point_T2<R> &q,
                 const Line_T2<R> &l)
{
  return (l.has_on_positive_side(p) && l.has_on_positive_side(q)) ||
         (l.has_on_negative_side(p) && l.has_on_negative_side(q));
}

template < class R >
inline
bool
at_the_same_side(const Point_T2<R> &p, 
                 const Point_T2<R> &q,
                 const Point_T2<R> &r, 
                 const Line_T2<R> &l)
{
  bool positive0 = l.has_on_positive_side(p);
  bool positive1 = l.has_on_positive_side(q);
  bool positive2 = l.has_on_positive_side(r);
  bool positive3 = l.has_on_negative_side(p);
  
  bool positive4 = l.has_on_negative_side(q);
  bool positive5 = l.has_on_negative_side(r);
  std::cout << positive0 << std::endl;
  std::cout << positive1 << std::endl;
  std::cout << positive2 << std::endl;
  std::cout << positive3 << std::endl;
  std::cout << positive4 << std::endl;
 
  bool pos =(positive0 && 
            positive1 &&
            positive2);
  bool neg =(positive3 && 
            positive4 &&
            positive5);
   std::cout << positive5 << pos << neg << std::endl;
   return false;// pos || neg; 
 
}

template < class R >
inline
bool
at_the_same_side_or_on(const Point_T2<R> &p, 
                       const Point_T2<R> &q,
                       const Point_T2<R> &r, 
                       const Line_T2<R> &l)
{
 
  if (at_the_same_side(p,q,r,l)) {
 
    return true;
  }
  if (at_the_same_side(q,r,l) && l.has_on(p)) {
 
    return true;
  }
  if (at_the_same_side(p,r,l) && l.has_on(q)) {
 
    return true;
  }
  if (at_the_same_side(p,q,l) && l.has_on(r)) {
 
    return true;
  }
  if (l.has_on(p) && l.has_on(q)) {
 
    return true;
  }
  if (l.has_on(q) && l.has_on(r)) {
 
    return true;
  }
 
  return l.has_on(p) && l.has_on(r);
}
       
template < class R >
inline
bool
coincident(const Line_T2<R> &l0, 
           const Line_T2<R> &l1)
{
  return dual(l0) == dual(l1);
}

template < class R >
inline
bool
is_infinity_line(const Line_T2<R> &l)
{
  return is_at_infinity(l.point0()) &&
         is_at_infinity(l.point1());
}

template < class R >
inline
bool
has_on_positive_side_line(const Point_T2<R> &l0, 
                          const Point_T2<R> &l1,
                          const Point_T2<R> &p)
{
  return side_of_oriented_line(l0,l1,p) == ON_POSITIVE_SIDE;
}

template < class R >
inline
bool
has_on_positive_side_line(const Line_T2<R> &l, 
                          const Point_T2<R> &p)
{
  return l.has_on_positive_side(p);
}

template < class R >
inline
bool
has_on_negative_side_line(const Point_T2<R> &l0, 
                          const Point_T2<R> &l1,
                          const Point_T2<R> &p)
{
  return side_of_oriented_line(l0,l1,p) == ON_NEGATIVE_SIDE;
}

template < class R >
inline
bool
has_on_negative_side_line(const Line_T2<R> &l, 
                          const Point_T2<R> &p)
{
  return l.has_on_negative_side(p);
}


template < class R >
inline
bool
has_on_boundary_line(const Point_T2<R> &l0, 
                     const Point_T2<R> &l1,
                     const Point_T2<R> &p)
{
  return side_of_oriented_line(l0,l1,p) == ON_ORIENTED_BOUNDARY;
}

template < class R >
inline
Oriented_side
side_of_oriented_line(const Point_T2<R> &l0, 
                      const Point_T2<R> &l1,
                      const Point_T2<R> &p)
{
 
  return side_of_oriented_lineT2(l0.w(), l0.x(), l0.y(),
                                 l1.w(), l1.x(), l1.y(),
                                 p.w(), p.x(), p.y());
}

template < class R >
inline
Oriented_side
side_of_oriented_line(const Line_T2<R> &l, 
                      const Point_T2<R> &p)
{
  Point_T2<R> p0 = l.point0();
  Point_T2<R> p1 = l.point1();
  return side_of_oriented_line(p0,p1,p);
}

template < class R >
inline
bool
is_horizontal(const Line_T2<R> &l)
{
  Direction_T2<R> d= l.direction();
  return (d == Direction_T2<R>(1,1,0) || 
          d == Direction_T2<R>(1,-1,0) );
}

template < class R >
inline
bool
is_vertical(const Line_T2<R> &l)
{
  Direction_T2<R> d= l.direction();
  return ( d == Direction_T2<R>(1,0,1) || 
           d == Direction_T2<R>(1,0,-1) );
}

/*compares the slopes of the lines l1 and l2 */
template < class R >
inline
Comparison_result
compare_slopes (Line_T2<R> &l1, 
                Line_T2<R> &l2)
{
   return compare_slopesC2(l1.X(), l1.Y(), l2.X(), l2.Y());
}

/*RETURN true iff q is at the perpendicular line to the line pr,
  RETURN false otherwise*/
template <class R >
inline
bool
is_at_perpendicular_line(const Point_T2<R> &p, 
                         const Point_T2<R> &q,
                         const Point_T2<R> &r)
{
  Direction_T2<R> d(perpendicular_direction(p,r));
  Line_T2<R> l(p,d);
  return l.has_on(q);
}



CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_PREDICATES_ON_LINES_T2_H
