// Copyright (c) 1999  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Kernel_23/include/CGAL/predicates_on_points_3.h,v $
// $Revision: 1.11 $ $Date: 2003/10/21 12:18:26 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Andreas Fabri, Stefan Schirra
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_PREDICATES_ON_POINTS_T2
#define CGAL_PREDICATES_ON_POINTS_T2

#include <math.h>
#include <CGAL/predicates_on_points_T1.h>
#include <CGAL/Line_T2_Line_T2_intersection.h>
#include <CGAL/kernel_ftT2.h>


CGAL_BEGIN_NAMESPACE

template < class R >
inline
bool
equal_x(const Point_T2<R>& p, 
        const Point_T2<R>& q)
{
  Point_T1<R> px(p.w(),p.x());
  Point_T1<R> qx(q.w(),q.x());
  return equal_T1(px,qx);
}

template < class R >
inline
bool
equal_y(const Point_T2<R>& p, 
        const Point_T2<R>& q)
{
  Point_T1<R> py(p.w(),p.y());
  Point_T1<R> qy(q.w(),q.y());
  return equal_T1(py,qy);
}


template < class R >
inline
bool
compare_around_origin(const Point_T2<R>& current,
                      const Point_T2<R>& new_point) {
  if (current == new_point) return false;
  Point_T2<R> o(1,0,0);
  Point_T2<R> start(0,1,0);
  if (new_point == Point_T2<R>(-1,0,0)) {
    if (has_on_segment(current,o,start))
      return true;
    return (new_point != current && has_on_segment(current,antipode(o),start));
  }
  if (current == Point_T2<R>(-1,0,0)) {
    return (!has_on_segment(new_point,o,start) && 
            !has_on_segment(new_point,antipode(o),start));
  }
  if (current == o)
    return true;
  if (new_point == o)
    return false;
  Point_3<Cartesian<typename R::FT> > start_R3(start.w(),start.x(),start.y());
  Point_3<Cartesian<typename R::FT> > o_R3(typename R::FT(1),
                                        typename R::FT(0),
                                        typename R::FT(0)); 
  Point_3<Cartesian<typename R::FT> > oa_R3(typename R::FT(-1),
                                        typename R::FT(0),
                                        typename R::FT(0));
  Point_3<Cartesian<typename R::FT> > cur_R3(current.w(),current.x(),current.y());
  Point_3<Cartesian<typename R::FT> > new_R3(new_point.w(), new_point.x(), new_point.y());
 
  if (current == antipode(o)) {
    if (orientation(o,start,new_point)!=0)
      return true;
    return (int(coplanar_orientation(o_R3,oa_R3,new_R3)) * 
              int(coplanar_orientation(o_R3,oa_R3,start_R3)) == -1);
  }
  if (new_point == antipode(o))
    return  ( (orientation(o,start,new_point)==0) &&
              (int(coplanar_orientation(o_R3,oa_R3,cur_R3)) * 
               int(coplanar_orientation(o_R3,oa_R3,start_R3)) == 1) );

  Orientation o1 = orientation(o,start,current);
  Orientation o2 = orientation(o,current,new_point);
  if (o2 == 0) {
    if (coplanar_orientation(o_R3,oa_R3,cur_R3) == 
        coplanar_orientation(o_R3,oa_R3,new_R3)) {
      return (cmp_dist_collinear_origin(current,new_point)==SMALLER);
    }
    if (coplanar_orientation(o_R3,oa_R3,cur_R3) == 
            coplanar_orientation(o_R3,oa_R3,start_R3))
      return (orientation(o,start,current)!=-1);
    return (orientation(o,start,new_point)==-1);
  }
  if (o1 == 1) {
    if (o2 == 1) {
      return true;
    }
  //if (o2 == -1)	
  return (orientation(o,start,new_point) == -1);
  }
  if (o1 == -1) {
    if (orientation(o,start,new_point) == 1){
      return false;
    }
    return ((o2 == 1) && (orientation(o,start,new_point)==-1));
  }
  //o1 == 0
  if (has_on_segment(current,o,start) ||
      has_on_segment(current,antipode(o),start)){
    return true;
  }
  return (orientation(o,current,new_point) == 1);
}


template < class R >
inline
bool
is_at_same_range_proper_point(const Point_T2<R> &p,
                              const Point_T2<R> &q,
                              const Point_T2<R> &r)
{
  if (is_at_front_range(q) || is_at_front_range(r))
    return is_at_front_range(p);
  return is_at_back_range(p);  
}

/* Return TRUE iff p,q and r are at the front range */
template < class R >
inline
bool
is_at_front_range(const Point_T2<R> &p,
                  const Point_T2<R> &q,
                  const Point_T2<R> &r)
{
  return is_at_front_range(p) &&
         is_at_front_range(q) &&
         is_at_front_range(r);
}

/* Return TRUE iff one point is improper and the others
are at the front range */
template < class R >
inline
bool
is_one_improper_and_others_front_range
	(const std::vector<Point_T2<R> > &v)
{
  int i=0;
  int j=0;
  std::cout << v.size();
  while ( i < v.size() && (is_at_front_range(v[i]) || 
          is_at_infinity(v[i])) && j <= 1 ){ 
    if ( is_at_infinity(v[i]) )
      j++;
    i++;
  }
  return (j==1 && i==v.size());  
}

/* Return TRUE iff one point is improper and the others
are at the back range */
template < class R >
inline
bool
is_one_improper_and_others_back_range
(const std::vector<Point_T2<R> > &v)
{
  int i=0;
  int j=0;
  while ( i < v.size() && (is_at_back_range(v[i]) || 
          is_at_infinity(v[i])) && j <= 1 ){ 
    if ( is_at_infinity(v[i]) )
      j++;
    i++;
  }
  return (j==1 && i==v.size());
}

/* Return TRUE iff all the points are at the back range
or all are at the front range*/
template < class R>
inline
bool
is_all_at_same_range(const std::vector<Point_T2<R> > &v)
{
  int i=0;
  if (v.size()==1)
    return !(is_at_infinity(v[0]));
  while(i < v.size()-1 && is_at_same_range(v[i],v[i+1]))
    i++;
  return (i==v.size()-1);
}

/* Return TRUE iff p and q are at the 
same range */
template < class R>
inline
bool
is_at_same_range(const Point_T2<R> &p,
                 const Point_T2<R> &q)
{
  return (p.w() * q.w() > 0);
}

/* Return TRUE iff p and q aren't at the 
same range */
template < class R>
inline
bool
is_at_different_range(const Point_T2<R> &p,
                     const Point_T2<R> &q)
{
  return !is_at_same_range(p,q);
}

/* Return TRUE iff p, q and r are at the 
same range */
template < class R>
inline
bool
is_at_same_range(const Point_T2<R> &p,
                 const Point_T2<R> &q,
                 const Point_T2<R> &r)
{
  return (p.w() * q.w() > 0) && 
         (q.w() * r.w() > 0);
}

/* Return TRUE iff p,q and r aren't at the 
same range */
template < class R>
inline
bool
is_at_different_range(const Point_T2<R> &p,
                     const Point_T2<R> &q,
                     const Point_T2<R> &r)
{
  return !is_at_same_range(p, q, r);
}

/* Return TRUE iff p and q are at the 
same range  or p or q are improper*/
template < class R >
inline
bool
is_at_same_range_or_one_improper(const Point_T2<R> &p,
                                 const Point_T2<R> &q)
{
  if ((p.w() * q.w()) > 0)
    return true;
  return ((p.w() * q.w() == 0) && (p.w()!=0 || q.w()!=0));
}  

/* Return TRUE iff p and q are at the 
front range */
template < class R>
inline
bool
is_at_front_range(const Point_T2<R> &p, 
                  const Point_T2<R> &q)
{
  return (p.w() > 0) && (q.w() > 0);
}

/* Return TRUE iff p is at the
front range */
template < class R>
inline
bool
is_at_front_range(const Point_T2<R> &p)
{
  return (p.w() > 0);
}

/*Return TRUE iff p and q are at the
back range */
template < class R>
inline
bool
is_at_back_range(const Point_T2<R> &p,
                 const Point_T2<R> &q)
{
  return (p.w() < 0) && (q.w() < 0);
}

/*Return TRUE iff p is at the
back range */
template < class R>
inline
bool
is_at_back_range(const Point_T2<R> &p)
{
  return (p.w() < 0);
}

/*Return TRUE iff p is at the
infinity */
template < class R>
inline
bool
is_at_infinity(const Point_T2<R> &p)
{
  return (p.w() == 0);
}

/*Return LEFTTURN if r lies to the left of the oriented 
  line l defined by p and q, returns RIGHTTURN if r lies to 
   the right of l, and returns COLLINEAR if r lies on l.
*/
template < class R >
inline
Orientation
orientation(const Point_T2<R> &p,
            const Point_T2<R> &q,
            const Point_T2<R> &r)
{
  return orientationT2(p.w(), p.x(), p.y(), 
                       q.w(), q.x(), q.y(), 
                       r.w(), r.x(), r.y());
}

/*Return TRUE iff if r lies to the left of the oriented 
  line l defined by p and q */
template < class R >
inline
bool
left_turn(const Point_T2<R> &p,
          const Point_T2<R> &q,
          const Point_T2<R> &r)
{
  return orientation(p, q, r) == LEFT_TURN;
}

/*Return TRUE iff if r lies to the right of the oriented 
  line l defined by p and q  */
template < class R >
inline
bool
right_turn(const Point_T2<R> &p,
           const Point_T2<R> &q,
           const Point_T2<R> &r)
{
  return orientation(p, q, r) == RIGHT_TURN;
}

/*Return TRUE iff if p.x == q.x,
  otherwise return FALSE */
template < class R >
inline
bool
x_equal(const Point_T2<R> &p, 
        const Point_T2<R> &q)
{
  Point_T1<R> *p1 = new Point_T1<R>(p.w(),p.x());
  Point_T1<R> *q1 = new Point_T1<R>(q.w(),q.x());
  return equal_T1(*p1,*q1);
}

/*Return TRUE iff if p.y == q.y,
  otherwise return FALSE */
template < class R >
inline
bool
y_equal(const Point_T2<R> &p,
        const Point_T2<R> &q)
{
  Point_T1<R> *p1 = new Point_T1<R>(p.w(),p.y());
  Point_T1<R> *q1 = new Point_T1<R>(q.w(),q.y());
  return equal_T1(*p1,*q1);
}

/*Return TRUE iff if p.w == q.w,
  otherwise return FALSE */
template < class R >
inline
bool
w_equal(const Point_T2<R> &p,
        const Point_T2<R> &q)
{
  return p.w() == q.w();
}

/*precondition p and q are in T1+
  Return TRUE iff if p.x < q.x,
  otherwise return FALSE. */
template < class R >
inline
bool
is_smaller_x_T2p(const Point_T2<R> &p,
       const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.x() >= 0 && q.x() >= 0);
  Point_T1<R> *p1 = new Point_T1<R>(p.w(),p.x());
  Point_T1<R> *q1 = new Point_T1<R>(q.w(),q.x());
  return is_smaller_T1p(*p1,*q1);
}


template < class R >
inline
bool
smaller_x_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q))
    return (p.x() < 0  && q.x() >= 0) || (p.x() == 0 && q.x() > 0);
  if (is_at_infinity(p))
    return (p.x() < 0) || (p.x() == 0 && sign(q.w()*q.x()) > 0);
  if (is_at_infinity(q))
    return (q.x() == 0 && sign(p.w()*p.x()) < 0) || q.x() > 0;
  return (p.x()*q.w() < p.w()*q.x());
}

template < class R >
inline
bool
smaller_equal_x_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
	CGAL_kernel_precondition(p.w()*q.w()>=0);

	if (is_at_infinity(p) && is_at_infinity(q))
		return (p.x() < 0  && q.x() >= 0) ||
					 (p.x() == 0 && q.x() > 0) ||
					 (sign(p.x()) == sign(q.x()));
	if (is_at_infinity(p))
		return (p.x() < 0) || (p.x() == 0 && sign(q.w()*q.x()) > 0);
	if (is_at_infinity(q))
		return (q.x() == 0 && sign(p.w()*p.x()) < 0) || q.x() > 0;
	return (p.x()*abs(q.w())*sign(p.x()) <=
					abs(p.w())*q.x()*sign(q.w()));
}

template < class R >
inline
bool
greater_x_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q))
    return (p.x() > 0  && q.x() <= 0) || (p.x() == 0 && q.x() < 0);
  if (is_at_infinity(p))
    return (p.x() > 0) || (p.x() == 0 && sign(q.w()*q.x()) < 0);
  if (is_at_infinity(q))
    return (q.x() == 0 && sign(p.w()*p.x()) > 0) || q.x() < 0;
  return (p.x()*q.w() > p.w()*q.x());
}

template < class R >
inline
bool
greater_y_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q))
    return (p.y() > 0  && q.y() <= 0) || (p.y() == 0 && q.y() < 0);
  if (is_at_infinity(p))
    return (p.y() > 0) || (p.y() == 0 && sign(q.w()*q.y()) < 0);
  if (is_at_infinity(q))
    return (q.y() == 0 && sign(p.w()*p.y()) > 0) || q.y() < 0;
  return (p.y()*abs(q.w())*sign(p.w()) > abs(p.w())*q.y()*sign(q.w()));
}

template < class R >
inline
bool
smaller_y_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{

  CGAL_kernel_precondition(p.w()*q.w()>=0);

  if (is_at_infinity(p) && is_at_infinity(q))
    return (p.y() < 0  && q.y() >= 0) || (p.y() == 0 && q.y() > 0);
  if (is_at_infinity(p))
    return (p.y() < 0) || (p.y() == 0 && sign(q.w()*q.y()) < 0);
  if (is_at_infinity(q))
    return (q.y() == 0 && sign(p.w()*p.y()) < 0) || q.y() > 0;
  return (p.y()*q.w() < p.w()*q.y());
}

template < class R >
inline
bool
smaller_xy_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.w()*q.w()>=0);
  
  typedef typename Point_T2<R>::FT   FT;
  if (is_at_infinity(p) && !is_at_infinity(q))
    return false;
  if (!is_at_infinity(p) && is_at_infinity(q))
    return true;
  if (is_at_infinity(p) && is_at_infinity(q))
    return (orientation(Point_T2<R>(FT(1),FT(0),FT(0)),p,q)==CLOCKWISE);
  if (smaller_x_one_range_infinity_T2(p,q))
    return true;
  if (!greater_x_one_range_infinity_T2(p,q))
    return (smaller_y_one_range_infinity_T2(p,q));
  return false;
}

template < class R >
inline
bool
smaller_yx_one_range_infinity_T2(const Point_T2<R> &p,
             const Point_T2<R> &q)
{
	CGAL_kernel_precondition(p.w()*q.w()>=0);
	if (smaller_y_one_range_infinity_T2(p,q)) 
		return true;
	if (!greater_y_one_range_infinity_T2(p,q))
		return (smaller_x_one_range_infinity_T2(p,q)); 
	return false;
}

/*precondition p and q are in T1+
  Return TRUE iff if p.y < q.y,
  otherwise return FALSE. */
template < class R >
inline
bool
smaller_y(const Point_T2<R> &p,
       const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.y() >= 0 && q.y() >= 0);
  Point_T1<R> *p1 = new Point_T1<R>(p.w(),p.y());
  Point_T1<R> *q1 = new Point_T1<R>(q.w(),q.y());
  return smallerT1(*p1,*q1);
}

template < class R >
inline
bool
larger_x(const Point_T2<R> &p,
        const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.x() >= 0 && q.x() >= 0);
  return (!smaller_x(p,q) && !x_equal(p,q));
}

template < class R >
inline
bool
larger_y(const Point_T2<R> &p,
        const Point_T2<R> &q)
{
  CGAL_kernel_precondition(p.y() >= 0 && q.y() >= 0);
  return !smaller_y(p,q) && !y_equal(p,q);
}


/*Return TRUE iff if p.w < q.w,
  otherwise return FALSE. */
template < class R >
inline
bool
smaller_w(const Point_T2<R> &p,
          const Point_T2<R> &q)
{
  return p.w() < q.w();
}


/* this function is based on orientation tests
and require that p,q aren't antipodals*/
template < class R >
inline
bool
less_xy_T2(const Point_T2<R> &p,
           const Point_T2<R> &q)
{
	if (is_at_infinity(q))
		return is_at_front_range(p);
  if (orientation(p,q,Point_T2<R>(0,0,1)) > 0)
		return true;
	if (orientation(p,q,Point_T2<R>(0,0,1)) == 0)
		return (orientation(p,q,Point_T2<R>(0,1,0)) < 0);
	return false;
}


/*Return SMALLER iff if p.x < q.x,
  EQUAL iff p.x == q.x,
  otherwise return LARGER. */
template < class R >
inline
Comparison_result
compare_x(const Point_T2<R> &p, const Point_T2<R> &q)
{
  typename R::FT normp = sqrt(p.w()*p.w() + 
                              p.x()*p.x() +
                              p.y()*p.y());
  typename R::FT normq = sqrt(q.w()*q.w() +
                              q.x()*q.x() +
                              q.y()*q.y());
  typename R::FT p1=p.x()*normq;
  typename R::FT q1=q.x()*normp;
 
  if (p1 < q1)
    return SMALLER;
  if (p1 > q1)
    return LARGER;
  return EQUAL;
}

/*Return SMALLER iff if p.y < q.y,
  EQUAL iff p.y == q.y,
  otherwise return LARGER. */
template < class R >
inline
Comparison_result
compare_y(const Point_T2<R> &p,
          const Point_T2<R> &q)
{
  typename R::FT normp = sqrt(p.w()*p.w() + 
                              p.x()*p.x() +
                              p.y()*p.y());
  typename R::FT normq = sqrt(q.w()*q.w() +
                              q.x()*q.x() +
                              q.y()*q.y());
  typename R::FT p1=p.y()*normq;
  typename R::FT q1=q.y()*normp;
 
  if (p1 < q1)
    return SMALLER;
  if (p1 > q1)
    return LARGER;
  return EQUAL;
}

/*Return TRUE iff if p.x == q.x and
  p.y == q.y,
  otherwise return FALSE. */
template < class R >
inline
bool
equal_xy(const Point_T2<R> &p, 
         const Point_T2<R> &q)
{
  return compare_x(p,q) == EQUAL &&
         compare_y(p,q) == EQUAL;
}

/*Return TRUE iff if p == q,
  otherwise return FALSE. */
template < class R >
inline
bool
equal_wxy(const Point_T2<R> &p,
          const Point_T2<R> &q)
{
 return  p==q;
}

/*Return SMALLER iff if p < q,
  EQUAL iff p == q,
  otherwise return LARGER. */
template < class R >
inline
Comparison_result
compare_wxy(const Point_T2<R> &p, 
	          const Point_T2<R> &q)
{
  return compare_lexicographically_wxy(p,q); 
}

/*Return SMALLER iff if p < q,
  EQUAL iff p == q,
  otherwise return LARGER. */
template < class R >
inline
Comparison_result
compare_lexicographically_wxy(const Point_T2<R> &p, 
                     		     const Point_T2<R> &q)
{
  typename R::FT normp = sqrt(p.w()*p.w() + 
                              p.x()*p.x() + 
                              p.y()*p.y());
  typename R::FT normq = sqrt(q.w()*q.w() + 
                              q.x()*q.x() + 
                              q.y()*q.y());
  typename R::Point_T2 p1(p.w()*normq, 
                          p.x()*normq,
                          p.y()*normq);
  typename R::Point_T2 q1(q.w()*normp, 
                          q.x()*normp,
                          q.y()*normp);
 
  if (p1.w() < q1.w())
    return SMALLER;
  if (p1.w() == q1.w()) {
    if (p1.x() < q1.x())
      return SMALLER;
    if (p1.x() == q1.x()) {
      if (p1.y() < q1.y())
        return SMALLER;
      if (p1.y() == q1.y())
        return EQUAL;
      return LARGER;
    }
    return LARGER;
  }
  return LARGER;
}

/*Return TRUE iff if p <= q,
  otherwise return FALSE. */
template < class R >
inline
bool
lexicographically_wxy_smaller_or_equal(const Point_T2<R> &p, 
				                               const Point_T2<R> &q)
{ 
  return compare_lexicographically_wxy(p, q) != LARGER;
}

/*Return TRUE iff if p < q,
  otherwise return FALSE. */
template < class R >
inline
bool
lexicographically_wxy_smaller(const Point_T2<R> &p, 
			                        const Point_T2<R> &q)
{ 
  return compare_lexicographically_wxy(p, q) == SMALLER;
}

/*Return TRUE iff if p, q and r are collinear
  otherwise return FALSE. */
template < class R >
inline
bool
collinear(const Point_T2<R> &p,
          const Point_T2<R> &q,
          const Point_T2<R> &r)
{

  return (sign_of_determinant3x3(p.w(),p.x(),p.y(),
                                 q.w(),q.x(),q.y(),
                                 r.w(),r.x(),r.y()) == 0);
}

// return true if point q is strictly between p and r
// p,q and r are supposed to be collinear points
template < class R >
inline
bool
collinear_between(const Point_T2<R> &p,
          const Point_T2<R> &q,
          const Point_T2<R> &r)
{
	return (has_on_segment(q,p,r) && q!=p && q!=r);
}


/*Return TRUE iff if p and q are antipodal,
  otherwise return FALSE. */
template < class R >
inline bool
antipodal(const Point_T2<R> &p, 
          const Point_T2<R> &q)
{
  return antipode(p)==q;
}

/*verify if point pqr are collinear and p is at the segment qr*/
template < class R >
inline
bool
has_on_segment(const Point_T2<R> &p,
               const Point_T2<R> &q,
               const Point_T2<R> &r)
{
  CGAL_kernel_precondition(!antipodal(q,r));
  Point_3<Cartesian<typename R::FT> > a(q.w(),q.x(),q.y()); 
  Point_3<Cartesian<typename R::FT> > b(r.w(),r.x(),r.y());
  Point_3<Cartesian<typename R::FT> > c(p.w(),p.x(),p.y());
  Point_3<Cartesian<typename R::FT> > o(typename R::FT(0),
                                        typename R::FT(0),
                                        typename R::FT(0));
  Point_3<Cartesian<typename R::FT> > a1(-q.w(),-q.x(),-q.y()); 
  if (collinear(p,q,r)) {
    if (coplanar_orientation(c,o,a)!=coplanar_orientation(c,o,b)) {
      if (coplanar_orientation(a,a1,c) == 0)
        return (angle(a,o,c) == ACUTE);
      return (coplanar_orientation(a,a1,b) ==
              coplanar_orientation(a,a1,c)); 
     }
    return (p==q && q==r);
  }
  return false;
}

/*verify if point p is at the segment qr*/
template < class R > 
bool
collinear_has_on(const Point_T2<R> &p, 
                 const Point_T2<R> &q,
                 const Point_T2<R> &r)
{
   CGAL_kernel_precondition(!antipodal(q,r));
  Point_3<Cartesian<typename R::FT> > a(q.w(),q.x(),q.y()); 
  Point_3<Cartesian<typename R::FT> > b(r.w(),r.x(),r.y());
  Point_3<Cartesian<typename R::FT> > c(p.w(),p.x(),p.y());
  Point_3<Cartesian<typename R::FT> > o(typename R::FT(0),
                                        typename R::FT(0),
                                        typename R::FT(0));
  Point_3<Cartesian<typename R::FT> > a1(-q.w(),-q.x(),-q.y()); 
  if (coplanar_orientation(c,o,a)!=coplanar_orientation(c,o,b)) {
    if (coplanar_orientation(a,a1,c) == 0)
      return (angle(a,o,c) == ACUTE);
    return (coplanar_orientation(a,a1,b) ==
            coplanar_orientation(a,a1,c)); 
  }
  return (p==q && q==r);
}



/*Return TRUE iff p,q,r are collinear and
  ordered along line, otherwise return FALSE */
template < class R >
inline
bool
collinear_are_ordered_along_line(const Point_T2<R> &p,
                                 const Point_T2<R> &q,
                                 const Point_T2<R> &r)
{
  if (!has_on_segment(r,p,q))
    return true;
  if (r==p)
    return true;
  return r==q;
} 

/*Return TRUE iff p,q,r are collinear and strictly
  ordered along line, otherwise return FALSE */
template < class R >
inline
bool
collinear_are_strictly_ordered_along_line(const Point_T2<R> &p,
                                          const Point_T2<R> &q,
                                          const Point_T2<R> &r)
{
  return !has_on_segment(r,p,q);
} 

/*Return TRUE iff p,q,r are 
  ordered along line, otherwise return FALSE */
template < class R >
inline
bool
are_ordered_along_line(const Point_T2<R> &p,
                       const Point_T2<R> &q,
                       const Point_T2<R> &r)
{
    return collinear(p, q, r) && 
           collinear_are_ordered_along_line(p, q, r);
}


/*Return TRUE iff p,q,r are strictly
  ordered along line, otherwise return FALSE */
template < class R >
inline
bool
are_strictly_ordered_along_line(const Point_T2<R> &p,
                                const Point_T2<R> &q,
                                const Point_T2<R> &r)
{
    return collinear(p, q, r) && 
           collinear_are_strictly_ordered_along_line(p, q, r);
}

/* compares the x-coordinates of p and the horizontal 
projection of p on h;
*/
template < class R >
inline
Comparison_result 
compare_x_at_y (const Point_T2<R> &p, 
                      Line_T2<R> &h)
{
	if (is_horizontal(h))
    CGAL_kernel_precondition(h.has_on(p));
  Point_T2<R> q(p.w(),(-h.W()*p.w()-h.Y()*p.y())/h.X(),p.y()); 
  return compare_x(p,q); 
}

template < class R >
inline
Comparison_result
compare_x_at_y (const Point_T2<R> &p,
                 Line_T2<R> &h1,
                 Line_T2<R> &h2)
{
  if (is_horizontal(h1))
  	CGAL_kernel_precondition(h1.has_on(p));
  Point_T2<R> q1(p.w(),(-h1.W()*p.w()-h1.Y()*p.y())/h1.X(),p.y());
  if (is_horizontal(h2))
  	CGAL_kernel_precondition(h2.has_on(p));
  Point_T2<R> q2(p.w(),(-h2.W()*p.w()-h2.Y()*p.y())/h2.X(),p.y()); 
  return compare_x(q1,q2);             
}


/*Let p be the intersection of lines l1 and l2. This function compares
 the x-coordinates of p and the horizontal projection of p on h; 
 Precondition: l1 and l2 intersect and are not horizontal; 
 h is not horizontal.
*/
template < class R >
inline
Comparison_result
compare_x_at_y ( Line_T2<R> &l1,
                 Line_T2<R> &l2,
                 Line_T2<R> &h)
{
  CGAL_kernel_precondition(!coincident(l1,l2));
  Point_T2<R> p(l1.X()*l2.Y()-l1.Y()*l2.X(),
               -l1.W()*l2.Y()+l1.Y()*l2.W(),
                l1.W()*l2.X()-l1.X()*l2.W());
  if (is_horizontal(h))
		CGAL_kernel_precondition(h.has_on(p));
  Point_T2<R> q(p.w(),(-h.W()*p.w()-h.Y()*p.y())/h.X());
  return compare_x(p,q);
                 
}


/* Let p be the intersection of lines l1 and l2. This function 
   compares the x-coordinates of the horizontal projection of p 
   on h1 and on h2; see (d) in the figure below.
   Precondition: l1 and l2 intersect and are not horizontal;
    h1 and h2 are not horizontal.
*/
template < class R >
inline
Comparison_result
compare_x_at_y ( Line_T2<R> &l1,
                 Line_T2<R> &l2,
                 Line_T2<R> &h1,
                 Line_T2<R> &h2)
{
  CGAL_kernel_precondition(!coincident(l1,l2));
  Point_T2<R> p(l1.X()*l2.Y()-l1.Y()*l2.X(),
               -l1.W()*l2.Y()+l1.Y()*l2.W(),
                l1.W()*l2.X()-l1.X()*l2.W());
  if (is_horizontal(l1)) {
    CGAL_kernel_precondition(l1.has_on(p));
  }
  Point_T2<R> q1(p.w(), (-l1.W()*p.w()-l1.Y()*p.y())/l1.X(), p.y());

  if (is_horizontal(l2)) {
    CGAL_kernel_precondition(l2.has_on(p));  }
  Point_T2<R> q2(p.w(), (-l2.W()*p.w()-l2.Y()*p.y())/l2.X(), p.y());

	return compare_x(q1,q2);

}

/*compares the y-coordinates of p and the intersection of lines l1 and l2.*/
template < class R >
inline
Comparison_result 
compare_y (Point_T2<R> &p, 
           Line_T2<R> &l1, 
           Line_T2<R> &l2)
{
  CGAL_kernel_precondition(!coincident(l1,l2));
  return compare_y(p,Point_T2<R>(l1.X()*l2.Y()-l1.Y()*l2.X(),
                                -l1.W()*l2.X()+l2.W()*l1.X(),
                                 l1.W()*l2.Y()-l2.W()*l1.Y()));
}

/*compares the y-coordinates of the intersection of line l with 
line h1 and with line h2.*/
template < class R >
inline
Comparison_result
compare_y(Line_T2<R> &l, 
          Line_T2<R> &h1, 
          Line_T2<R> &h2)
{
  CGAL_kernel_precondition(!coincident(l,h1) && !coincident(l,h2));
  return compare_y(Point_T2<R>(l.X()*h1.Y()-l.Y()*h1.X(),
                            -l.W()*h1.X()+l.X()*h1.W(),
                             l.W()*h1.X()-h1.W()*l.X()),
                 Point_T2<R>(l.X()*h2.Y()-l.Y()*h2.X(),
                            -l.W()*h2.X()+l.X()*h2.W(),
                             l.W()*h2.X()-h2.W()*l.X()));

}

/*compares the y-coordinates of the intersection of lines l1 and l2 
and the intersection of lines h1 and h2 */
template < class R >
inline
Comparison_result
compare_y ( Line_T2<R> &l1,
            Line_T2<R> &l2,
            Line_T2<R> &h1,
            Line_T2<R> &h2)
{
  CGAL_kernel_precondition(!coincident(l1,l2) && !coincident(h1,h2));
  return compare_y(Point_T2<R>(l1.X()*l2.Y()-l1.Y()*l2.X(),
                              -l1.W()*l2.Y()+l2.W()*l1.Y(),
                               l1.W()*l2.X()-l2.W()*l1.X()),
                   Point_T2<R>(h1.X()*h2.Y()-h1.Y()*h2.X(),
                              -h1.W()*h2.Y()+h2.W()*h1.Y(),        
                               h1.W()*h2.X()-h2.W()*h1.X()));
}


/*compares the y-coordinates of p and the vertical projection of p on h
Precondition: h is not vertical. */
template < class R >
inline
Comparison_result     
compare_y_at_x (Point_T2<R> &p, Line_T2<R> &h)
{
  if (is_vertical(h))
		CGAL_kernel_precondition(h.has_on(p));
  Point_T2<R> q(p.w(),p.x(),
               (-h.W()*p.w() - h.X()*p.x())/h.Y());
  return compare_y(p,q);

}

/*compares the y-coordinates of the vertical projection of p
 on h1 and on h2
Precondition: h1 and h2 are not vertical.*/
template < class R >
inline
Comparison_result
compare_y_at_x ( Point_T2<R> &p,
                 Line_T2<R> &h1,
                 Line_T2<R> &h2)
{
	if (is_vertical(h1))
  	CGAL_kernel_precondition(h1.has_on(p));
  if (is_vertical(h2))
		CGAL_kernel_precondition(h2.has_on(p));
  Point_T2<R> q1(p.w(),p.x(),
                (-h1.W()*p.w() - h1.X()*p.x())/h1.Y());
  Point_T2<R> q2(p.w(),p.x(),
                (-h2.W()*p.w() - h2.X()*p.x())/h2.Y());

  return compare_y(q1,q2);

}

/* Let p be the intersection of lines l1 and l2. This function compares
 the y-coordinates of p and the vertical projection of p on h
Precondition: l1 and l2 not coincident and h is not vertical.*/
template < class R >
inline
Comparison_result
compare_y_at_x ( Line_T2<R> &l1,
                 Line_T2<R> &l2,
                 Line_T2<R> &h)
{
  CGAL_kernel_precondition(!coincident(l1,l2));
  Point_T2<R> p(l1.X()*l2.Y()-l1.Y()*l2.X(),
               -l1.W()*l2.Y()+l1.Y()*l2.W(),
                l1.W()*l2.X()-l1.X()*l2.W());
	if (is_vertical(h))
		CGAL_kernel_precondition(h.has_on(p));
	Point_T2<R> q(p.w(),p.x(),(-h.W()*p.w()-h.X()*p.x())/h.Y());
  return compare_y(p,q);
}


/*Let p be the intersection of lines l1 and l2. This function compares 
the y-coordinates of the vertical projection of p on h1 and on h2.
Precondition: l1 and l2 are not coincident; h1 and h2 are not vertical.*/
template < class R >
inline
Comparison_result
compare_y_at_x ( Line_T2<R> &l1,
                 Line_T2<R> &l2,
                 Line_T2<R> &h1,
                 Line_T2<R> &h2)
{
  CGAL_kernel_precondition(!coincident(l1,l2));
  Point_T2<R> p(l1.X()*l2.Y()-l1.Y()*l2.X(),
               -l1.W()*l2.Y()+l1.Y()*l2.W(),
                l1.W()*l2.X()-l1.X()*l2.W());
  if (is_vertical(h1)) 
		CGAL_kernel_precodition(h1.has_on(p));
  Point_T2<R> q1(p.w(),p.x(),(-h1.W()*p.w()-h1.X()*p.x())/h1.Y());
  if (is_vertical(h2)) 
		CGAL_kernel_precodition(h2.has_on(p));
  Point_T2<R> q2(p.w(),p.x(),(-h2.W()*p.w()-h2.X()*p.x())/h2.Y());
  return compare(q1,q2);
}

/*compares the y-coordinates of p and the vertical projection of p on s.
 If s is vertical, then return EQUAL when p lies on s, SMALLER when p
 lies under s, and LARGER otherwise.
Precondition: p is within the x range of s.*/
template < class R >
inline
Comparison_result 
compare_y_at_x ( Point_T2<R> &p, Segment_T2<R> &s)
{
  if (is_vertical(s)){
	  CGAL_kernel_precondition(orientation(s.source(), s.target(),p)
						== COLLINEAR);                                  
    if (s.has_on(p))
      return EQUAL;
    else if (orientation(s.source(),
						 Point_T2<R>(l.W(),l.X(),l.Y()),p) == 1)
      return LARGER;
    return SMALLER;
  }
  Line_T2<R> l=s.supporting_line();
  Point_T2<R> q(p.w(),p.x(),(-l.W()*p.w()-l.X()*p.x())/l.Y());
  CGAL_kernel_precondition(s.has_on(q));
  return compare_y(p,q);
}


/*compares the y-coordinates of the vertical projection of p 
on s1 and on s2. If s1 or s2 is vertical, then return EQUAL if
they intersect, otherwise return SMALLER if s1 lies below s2, 
and return LARGER otherwise.
Precondition: p is within the x range of s1 and s2.
*/
template < class R >
inline
Comparison_result
compare_y_at_x ( Point_T2<R> &p,
                 Segment_T2<R> &s1,
                 Segment_T2<R> &s2)
{
  Line_T2<R> l1=s1.supporting_line();
  Point_T2<R> q1(p.w(),p.x(),(-l1.W()*p.w()-l1.X()*p.x())/l1.Y());
  Line_T2<R> l2=s2.supporting_line();
  Point_T2<R> q2(p.w(),p.x(),(-l2.W()*p.w()-l2.X()*p.x())/l2.Y());
  CGAL_kernel_precondition(s1.has_on(q1) && s2.has_on(q2));
 
 /* if (vertical(s1) || vertical(s2)) {
    if (do_intersect(s1,s2))  
      return EQUAL;
    else if (lies_below(s1,s2))
      return SMALLER;
    return LARGER;
  }
 */
  return compare_y(q1,q2);
} 

/*Compares the coordinates of points p and q lexicographically
 in yx order: first y-coordinates are compared, if they are equal,
 x-coordinates are compared. */
template < class R >
inline
Comparison_result 
compare_wyx ( const Point_T2<R> &p, 
             const Point_T2<R> &q)
{
  return compare_lexicographically_wxy(p,q);
}
                                        

/* returns true iff p is lexicographically larger than 
   q with respect to xy order. */
template < class R >
inline
bool
lexicographically_wxy_larger( const Point_T2<R> &p,
                             const Point_T2<R> &q)
{
  return compare_lexicographically_wxy(p,q)
                          == LARGER; 
}

/*returns true iff p is lexicographically not smaller 
than q with respect to xy order.*/
template < class R >
inline
bool
lexicographically_wxy_larger_or_equal( const Point_T2<R> &p,
                                       const Point_T2<R> &q)
{
  return compare_lexicographically_wxy(p,q)
                          != SMALLER; 
}

/* returns OBTUSE, RIGHT or ACUTE depending on the angle formed 
by the three points p, q, r (q being the vertex of the angle).*/
template < class R >
inline
Angle angle( const Point_T2<R> &p, 
             const Point_T2<R> &q, 
             const Point_T2<R> &r)
{
	CGAL_kernel_precondition(p!=q && q!=r);
	CGAL_kernel_precondition(p!=antipode(q) && 
													 q!=antipode(r));
	typedef typename Point_T2<R>::FT FT;
	Point_T2<R> o(FT(1),FT(0),FT(0)); 
	if (is_at_infinity(q)) {
		if (is_at_infinity(p)) {
			if (is_at_infinity(r)) {
				if (orientation(o,q,p)!=
						orientation(o,q,r))
					return OBTUSE;
				return ACUTE;
			}
			return RIGHT;
		}	
		if (is_at_infinity(r))
			return RIGHT;
		return OBTUSE;
	}

	Line_T2<R> l1(q,p);
	Line_T2<R> l2(q,r);
	Point_T2<R> angle(FT(0),l2.Y()*l1.Y() + l2.X()*l1.X(),
											abs(l2.X()*l1.Y() - l2.Y()*l1.X()));
	Point_T2<R>	q1(FT(0),FT(1),FT(0)),
							p1(FT(0),FT(0),FT(1));
	return Angle(-orientation(angle,o,p1) * orientation(angle,o,q1));
}

/*template < class R >
inline
Comparison_result
compare_deltax_deltay(const PointC2<R>& p,
                      const PointC2<R>& q,
                      const PointC2<R>& r,
                      const PointC2<R>& s)
{
    return compare_deltax_deltayC2(p.x(), q.x(), r.y(), s.y());
}*/


template < class R >
inline
bool
ccw_around_point_or_collinear_and_closer
											(const Point_T2<R>& p,
                      const Point_T2<R>& q,
                      const Point_T2<R>& r)
{
	if (orientation(p,q,r) == COUNTERCLOCKWISE)
		return true;
	if (orientation(p,q,r) == COLLINEAR) {
		Line_T2<R> l;
		if (is_at_infinity(q) && is_at_infinity(r))
			 l = Line_T2<R>(Point_T2<R>(1,0,0),q);
		else {
			 if (q != r)
				 l = perpendicular_line(q,r);
			 else
				 return true;
		}
		return (!at_the_same_side(p,r,l));	
	}
	return false;
}

template < class R >
inline
bool
same_orientation(const Point_T2<R>& o,
								 const Point_T2<R>& p,
                 const Point_T2<R>& q,
                 const Point_T2<R>& r)
{
	return orientation(o,p,q) == orientation(o,p,r);
};

template < class R >
inline
bool
is_proper_point(const Point_T2<R>& p)
{
	return (p.w()!=0);
}

template < class R >
inline
bool
is_proper_points(const Point_T2<R>& p,
								const Point_T2<R>& q)
{
	return (p.w()*q.w()!=0);
}

template < class R >
inline
Comparison_result
cmp_dist_collinear_origin(const Point_T2<R>& p,
                          const Point_T2<R>& q) 
{
  Point_T2<R> o(1,0,0);
  CGAL_kernel_precondition(orientation(p,q,o)==COLLINEAR);
  if (is_proper_points(p,q)) {
    //std::cout << "squared_distance(p,o) - " << squared_distance(p,o) << std::endl;
    //std::cout << "squared_distance(q,o) - " << squared_distance(q,o) << std::endl;
    return (compare_T1p(squared_distance(p,o),
                        squared_distance(q,o)));
  }
  if (is_at_infinity(p)) {
    if (is_at_back_range(q))
      return SMALLER;
    else if (is_at_front_range(q))
      return LARGER;
    return EQUAL;
  }
  if (is_at_front_range(p))
    return SMALLER;
  else if (is_at_back_range(p))
    return LARGER;
  return EQUAL;
}

CGAL_END_NAMESPACE


#endif

