//  T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
//  Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//  All rights reserved.
//  --
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
//  --
//  T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
//  Architecture and Implementation: Alessandra G. Oliveira
//
//  The current version of T2 contained in this distribution is: 1.20050628
//  which has been tested under and is compatible with CGAL 3.0.1
//
//  Source code and other informations can be found at:
//  http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_PREDICATES_ON_VECTORS_T2
#define CGAL_PREDICATES_ON_VECTORS_T2

CGAL_BEGIN_NAMESPACE

template < class R >
inline
bool
is_LD (const Vector_T2<R> & u, const Vector_T2<R> & v)
{
   return (u.x() * v.y() - v.x() * u.y()) == 0;
}

template < class R >
inline
bool
is_opposite (const Vector_T2<R> & u, const Vector_T2<R> & v)
{
   return (is_LD(u,v) && ( (sign(u.w()) == sign(v.w())) && (sign(u.x()) != 
           sign(v.x())) && (sign(u.y()) != sign(v.y())) ) );
}

CGAL_END_NAMESPACE

#endif //CGAL_PREDICATES_ON_VECTORS_T2
