// Copyright (c) 2003  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/STL_Extension/include/CGAL/utility.h,v $
// $Revision: 1.22 $ $Date: 2003/10/21 12:23:43 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Michael Hoffmann <hoffmann@inf.ethz.ch>
//                 Lutz Kettner <kettner@mpi-sb.mpg.de>
//                 Sylvain Pion <Sylvain.Pion@mpi-sb.mpg.de>
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_UTILITY_T2_H 
#define CGAL_UTILITY_T2_H 

CGAL_BEGIN_NAMESPACE

//+---------------------------------------------------------------------+
//| Quintuple class                                                     |
//+---------------------------------------------------------------------+

template <class T1, class T2, class T3, class T4, class T5>
struct Quintuple
{
  typedef T1 first_type;
  typedef T2 second_type;
  typedef T3 third_type;
  typedef T4 fourth_type;
  typedef T5 fifth_type;

  T1 first;
  T2 second;
  T3 third;
  T4 fourth;
  T5 fifth;

  Quintuple() {}

  Quintuple(const T1& a, const T2& b, const T3& c, const T4& d, const T5& e)
  : first(a), second(b), third(c), fourth(d), fifth(e)
  {}
};

template <class T1, class T2, class T3, class T4, class T5>
inline
Quintuple<T1, T2, T3, T4, T5>
make_quintuple(const T1& x, const T2& y, const T3& z, const T4& zz, const T5& zzz)
{
  return Quintuple<T1, T2, T3, T4, T5>(x, y, z, zz, zzz);
}

template <class T1, class T2, class T3, class T4, class T5>
inline
bool
operator==(const Quintuple<T1, T2, T3, T4, T5>& x,
           const Quintuple<T1, T2, T3, T4, T5>& y)
{
  return ( (x.first == y.first) &&
           (x.second == y.second) &&
           (x.third == y.third) &&
           (x.fourth == y.fourth) && 
           (x.fifth == y.fifth) );
}

/*template <class T1, class T2, class T3, class T4, class T5>
inline
bool
operator<(const Quintuple<T1, T2, T3, T4, T5>& x,
          const Quintuple<T1, T2, T3, T4, T5>& y)
{
  return ( x.first < y.first ||
           ( (x.first == y.first) && (x.second < y.second) ) ||
           ( (x.first == y.first) && (x.second == y.second) &&
             (x.third < y.third) ) ||
           ( (x.first == y.first) && (x.second == y.second) &&
             (x.third == y.third) ) &&
           (x.fourth < y.fourth) ) ||
           ( (x.first == y.first) && (x.second == y.second) &&
             (x.third == y.third) ) && (x.fourth == y.fourth) &&
             (x.fifth < y.fifth) );
}*/


CGAL_END_NAMESPACE

#endif // CGAL_UTILITY_T2_H //
// EOF //
