// Copyright (c) 2000  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland), Freie Universitaet Berlin (Germany),
// INRIA Sophia-Antipolis (France), Martin-Luther-University Halle-Wittenberg
// (Germany), Max-Planck-Institute Saarbruecken (Germany), RISC Linz (Austria),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Source: /CVSROOT/CGAL/Packages/Cartesian_kernel/include/CGAL/Cartesian/vector_constructions_3.h,v $
// $Revision: 1.10 $ $Date: 2003/10/21 12:14:36 $
// $Name: CGAL_3_0_1  $
//
// Author(s)     : Herve Bronnimann
// ==============================================================================
// T2 is an extension of CGAL (www.cgal.org) to the Oriented Projective Plane T2.
// --
// This file is based on and contains modifications to the $Source file
// mentioned above. These modifications are
//      Copyright (c) 2005 Alessandra G. Oliveira & Pedro J. de Rezende.
//      All rights reserved.
//
// The Licensing terms of the $Source file remain in force for this file.
//
// The original header from the $Source file, together with copyright, licensing
// and authoring information, is preserved above.
// --
// T2 System Concept and Design: Alessandra G. Oliveira & Pedro J. de Rezende
// Architecture and Implementation: Alessandra G. Oliveira
//
// The current version of T2 contained in this distribution is: 1.20050628
// which has been tested under and is compatible with CGAL 3.0.1
//
// Source code and other information can be found at:
// http://www.ic.unicamp.br/~rezende/T2inCGALandT2Viewer.htm


#ifndef CGAL_SIGNED_HOMOGENEOUS_VECTOR_CONSTRUCTIONS_T2_H
#define CGAL_SIGNED_HOMOGENEOUS_VECTOR_CONSTRUCTIONS_T2_H

#include <CGAL/Vector_T2.h>
#include <CGAL/kernel_ftT2.h>

CGAL_BEGIN_NAMESPACE

template < class R >
typename R::Vector_T2
cross_product(const Vector_T2<R> &v, const Vector_T2<R> &w)
{
    return Vector_T2<R>(v.y() * w.w() - v.w() * w.y(),
                       v.w() * w.x() - v.x() * w.w(),
                       v.x() * w.y() - v.y() * w.x());
}

CGAL_END_NAMESPACE

#endif // CGAL_SIGNED_HOMOGENEOUS_VECTOR_CONSTRUCTIONS_T2_H
